//
//      Sound/Timer engine wrappers for Allegro
//

#ifndef _WIN32
#include <unistd.h>     // This is for sleep, in the timer check routine
#endif
#include <stdlib.h>     // atexit(), random
#include <string.h>     // String splicing routines

#include <allegro.h>
#include "ithelib.h"
#include "console.h"
#include "media.h"
#include "loadfile.h"
#include "sound.h"
#include "audlib.h"

// defines

// variables

extern char mus_chan;                      // Number of music channels
extern char sfx_chan;                      // Number of effects channels
extern int  driftlevel;                    // Amount of frequency variation

int sf_volume = 63;
int mu_volume = 63;

int Songs=0;
int Waves=0;
static int pausedata=0;   // Speed of song before pausing, 0 if not paused

struct SMTab *wavtab;       // Tables of data for each sound and song.
struct SMTab *mustab;       // Filled by script.cc

// functions

void S_Load();              // Load the sound and music
void S_SoundVolume();       // Set the Sound volume
void S_MusicVolume();       // Set the Music volume
void S_PlaySample(char *s,int v);   // Play a sound sample
void S_PlayMusic(char *s);  // Play a song
void S_StopMusic();         // Stop the music

static void LoadMusic();
static void LoadWavs();


// Public code

/*
 *      S_PollMusic - Keep the music playing
 */

void S_PollMusic()
{
#ifndef NO_SOUND
poll_aud();
#endif
}

/*
 *      S_Load - Load in the music and samples.
 */

void S_Load()
{
LoadMusic();
LoadWavs();
}


/*
 *      S_MusicVolume - Set the music level
 */

void S_MusicVolume(int vol)
{
mu_volume = vol;
#ifndef NO_SOUND
set_aud_volume(vol); // 0-255
#endif
return;
}


/*
 *      S_SoundVolume - Set the effects level
 */

void S_SoundVolume(int vol)
{
sf_volume = vol;
return;
}


/*
 *   S_PlayModule - play a music module from the array song[]
 *                  may disrupt the sound effects slightly
 */

void S_PlayMusic(char *name)
{
IFILE *ifp;
char filename[1024];
int offset,ctr;

#ifdef NO_SOUND
return;
#endif

for(ctr=0;ctr<Songs;ctr++)
	if(!istricmp(mustab[ctr].name,name))
		{
		// If the music is marked Not Present, ignore it (music is optional)
		if(mustab[ctr].fname[0] == 0)
			return;
		// Otherwise, no excuses
		if(!loadfile(mustab[ctr].fname,filename))
			{
			Bug("S_PlayMusic - could play song %s, could not find file '%s'\n",name,mustab[ctr].fname);
			return;
			}

		ifp = iopen(filename);
		offset = ifp->origin;
		strcpy(filename,ifp->truename);
		iclose(ifp);
#ifndef NO_SOUND
        load_aud_stream_offset(filename,mustab[ctr].loopsong,offset);
#endif
        return;
        }

Bug("S_PlayMusic - could not find song '%s'\n",name);
}

/*
 *   S_PauseModule - Stop music, not SFX
 */

void S_PauseMusic()
{
return;
}

/*
 *   S_UnPauseModule - Restart music
 */

void S_UnPauseMusic()
{
return;
}

/*
 *   S_PlaySample - play a sound sample.
 */

void S_PlaySample(char *name, int volume)
{
int freq,num,ctr;

#ifdef NO_SOUND
return;
#endif

for(ctr=0;ctr<Waves;ctr++)
    if(!istricmp(wavtab[ctr].name,name))
        {
        freq=1000;
        if(!wavtab[ctr].nodrift)
            {
            num = 10 - (rand()%20);
            freq+=(num*driftlevel)/10;
            }
		play_sample(wavtab[ctr].sample,volume,128,freq,0); // vol,pan,freq,loop

//        if(!wavtab[ctr].handle)
//                    MIDASerror();
        return;
        }

Bug("S_PlaySample- could not find sound '%s'\n",name);
}


//
//   Private code hereon
//

/*
 *   LoadMusic - Make sure the music files are there, else mark as Not Present
 */

void LoadMusic()
{
char filename[1024];
int ctr;

ilog_printf("  Checking music.. ");     // This line is not terminated!

for(ctr=0;ctr<Songs;ctr++)
	if(!loadfile(mustab[ctr].fname,filename))
		{
		ilog_quiet("Warning: Could not load music file '%s'\n",mustab[ctr].fname);
		mustab[ctr].fname[0]=0;	// Erase the filename to mark N/A
		}

ilog_printf("done.\n");  // This terminates the line above

}


/*
 *   LoadWavs - Load in the sounds
 */

void LoadWavs()
{
char filename[1024];
int ctr;

#ifdef NO_SOUND
return;
#endif

// load in each sound

ilog_printf("  Loading sounds");     // This line is not terminated, for the dots

Plot(Waves);    // Work out how many dots to print

for(ctr=0;ctr<Waves;ctr++)
	{
        if(!loadfile(wavtab[ctr].fname,filename))
            ithe_panic("LoadWavs: Cannot open WAV file",wavtab[ctr].fname);
        wavtab[ctr].sample=iload_wav(filename);
        if(!wavtab[ctr].sample)
            ithe_panic("LoadWavs: Invalid WAV file",filename);
        Plot(0);                                        // Print a dot
	}

ilog_printf("\n");  // End the line of dots

}



