//
//	Do a dependency check and build the VRM files
//	=============================================
//
//	Kind of like 'MAKE', but simpler, and infinitely more easy to use
//

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>

#define STRING_LEN 256
#define SLASH "/"

// Seer headers
#include "seer/seer.h"
//#include "seer/internal.h"
//#include "seer/export.h"
#define SCError(i,c...) {printf("SeeRC:");printf(c);exit(i);}

void rebuild(char *name);
void removevrm(char *name);
void buildit(char *name);
void scan(char *dname,char *ext,void (*callfunc)(char *fname));

#ifdef __BEOS__
long telldir(DIR *d);
void seekdir(DIR *d, long loc);
#endif

main(int argc,char *argv[])
{
long ctr;
char *ptr;
FILE *fp;


puts("Automatic VRM builder 2.10 - JP Morris, 1999");
puts("============================================\n");
puts("Use 'BUILD -a' to rebuild everything.");
puts("This program is in the public domain.\n\n");

fp=fopen("system.hpp","r");
if(!fp)
	{
	printf("Could not find include file 'system.hpp'.\n");
	printf("Make sure you are running BUILD from the right directory.\n");
	printf("For example flat\\code or tfm\\code or mygame\\code\n");
	return 1;
	}
fclose(fp);

if(argv[1])
    if(!strcmp(argv[1],"-a"))
        {
        printf("Deleting compiled VRMs..\n");
        scan(".",".vrm",removevrm);
        printf("\n\n");
        }

printf("Looking for out-of-date VRMs...\n");

scan(".",".cpp",buildit);

printf("  Finished.                                                     \r");
printf("\n\nDone.\n");
return 0;
}

void scan(char *dname,char *ext,void (*callfunc)(char *fname))
{
DIR *d;
struct dirent *de;
struct stat statbuf;
char name[STRING_LEN];
long loc;

d = opendir(dname);
while (de = readdir(d))
    {
    // Make relative path
    strcpy(name,dname);
    strcat(name,SLASH);
    strcat(name,de->d_name);

    // Is it a directory
    if(stat(name,&statbuf))
        printf("CAN NOT STAT '%s'\n",name);

    if(statbuf.st_mode&S_IFDIR)
        {
        if(strcmp(de->d_name,"."))              // Check raw unprocessed name
            if(strcmp(de->d_name,".."))
                {
                loc = telldir(d);
                scan(name,ext,callfunc);        // recurse using processed one
                seekdir(d,loc);
                }
        }
    else
        {
//        strlwr(name);
        if(strstr(name,ext))            // If the extension is correct
            callfunc(name);             // then call the function
        }
    }
closedir(d);
}

// Remove existing VRM file

void removevrm(char *name)
{
remove(name);
printf("                                                                \r");
printf("  %s\r",name);
}

// Build the file if necessary

void buildit(char *name)
{
struct stat vrm,cpp;
char fname[STRING_LEN];
char *ptr;

printf("                                                                \r");
printf("  %s\r",name);

strcpy(fname,name);
ptr=strstr(fname,".cpp");
strcpy(ptr,".vrm");

stat(name,&cpp);

// If it doesn't exist, definitely build it

if(stat(fname,&vrm))
    {
    rebuild(name);
    return;
    }

// If it does exist and is out of date rebuild it

if(vrm.st_mtime<cpp.st_mtime)	// Out of date
    rebuild(name);
}

// Call the compiler to get it built

void rebuild(char *fn)
{
char Output[STRING_LEN];
char *ptr;
FILE *file;
scScript scrpt;

// splice together the output filename

//strlwr(fn);
strcpy(Output,fn);
ptr=strstr(Output,".cpp");	// Kill the extension
*ptr=0;

strcat(Output,".vrm");		// Add VRM extension to copy of the string

remove(Output);                 // Remove any existing file before compiling
printf("\n");

// Compile the source code


scrpt=scCompile_File(fn);
if (scErrorNo)
    SCError(scErrorNo,"\nIn file %s:\n%s\n> %s\n",fn,scErrorMsg,scErrorLine);
if (!scrpt)
    SCError(5,"\nFile `%s' not found.\n",fn);


// Save the compiled code

file=fopen(Output,"wb");
if(!file)
    {
    perror("File IO error:");
    SCError(66,"\nFile `%s' could not be created\n",Output);
    }
fwrite(scrpt,scGet_Script_Size(scrpt),1,file);
fclose(file);
free(scrpt);
 

#ifdef USE_EXTERNAL_COMPILER

char splice[STRING_LEN];
char *ptr;

printf("\n");

// First, remove the VRM file.

//strlwr(fn);                     // Kill the extension
ptr=strstr(fn,".cpp");
*ptr=0;

strcpy(splice,fn);              // Add VRM extension to copy of the string
strcat(splice,".vrm");
remove(splice);

// Now, compile the source

sprintf(splice,"seerc %s.cpp -Oe -o %s.vrm",fn,fn);

if(system(splice))
    {
    printf("\nCompilation error in file %s.cpp!\n",fn);
    exit(1);
    }
#endif
}

#ifdef __BEOS__
long telldir(DIR *d)
{
return 0;
}

void seekdir(DIR *d, long loc)
{
return;
}

#endif
