/*
	IRE:		Core Data structures

	Version: 	2.0, 29/2/00

*/

#ifndef __IRCORE
#define __IRCORE

#define IKV "0.68"      // IRE Kernel Version

#define STORE 1024
#define W_SIZE 10

#define RANDOM_TILE 0xffff      // Appears as an eyesore

#define CHAR_U 0  // Directions for the sequence cache in CHlist
#define CHAR_D 1
#define CHAR_L 2
#define CHAR_R 3

//#define VSH 8
//#define VSW 8
#define MAPSIG 0xf15e0005

#define MAX_W 1024
#define MAX_H 1024
#define MAX_MEMBERS 128

#define MAP_POS(x,y) (ytab[y]+(x))

#include <allegro.h>


///////////////////////////////////////////////////////////////////////////
//  Script Processing Data Structures                                    //
///////////////////////////////////////////////////////////////////////////

// NPC Flags

typedef struct NPC_FLAGS
{
unsigned long 	 female	    :1;   // Male or Female
unsigned long 	 know_name  :1;   // Do I know you?
unsigned long 	 is_hero    :1;   // Is this the hero?
unsigned long 	 cant_eat   :1;   // Can it eat and drink? (Robots can't)
unsigned long 	 critical   :1;   // Has gone critical
unsigned long 	 no_shutdoor :1;  // Won't shut doors after themselves
unsigned long 	 symlink    :1;   // Owner is the NPC, not object (phones etc)
unsigned long 	 biological :1;   // Affected by poison etc
unsigned long 	 guard      :1;   // Policeman
unsigned long 	 spare      :22;
} NPC_FLAGS;

typedef struct NPC_RECORDING
{
char *page;
struct OBJECT *player;
struct NPC_RECORDING *next;
} NPC_RECORDING;

// Statistics

typedef struct STATS
{
int  hp;
int dex;
int  str;
int intel;
int  weight;
int quantity;
short prob; short tick;       // Probability, unused.  Animation timer.
NPC_FLAGS npcflags;
int damage;                   // Damage (caused by a weapon)
int  oldhp;                    // Previous hp, used internally
int pFlags;				// Personal Flags
struct OBJECT *owner;
int karma;                  // Have you been a good little boy?
int bulk;
NPC_RECORDING *npctalk;     // Used to record what you've said for continuity
NPC_RECORDING *lFlags;      // Local Flags

/*
 * READ AND OBEY
 *
 * If you add any more data entries, edit "has_changed()" in LOADSAVE.CPP
 * and add a check to the list if necessary, OR MAP DATA LOSS WILL OCCUR.
 *
 * (I.E.  The map editor won't often save the state of your new entry ;-)
 */

short vv[12]; // subtract 2 for every int you add to the structure
} STATS;


// Temporary data for a game in progress
// Add to the END of the list

typedef struct USEDATA
{
int user[20];
int poison;
int unconscious;
int potion[10]; // space for 10 potions
int dx,dy;      // Used by the pathfinder
} USEDATA;

// A single sprite

typedef struct S_POOL
{
char *name;
char *fname;
RLE_SPRITE *image;
int thumbcol;

short w,h;
unsigned short wxh;
short x,y;
} S_POOL;

// A single lightmap sprite

typedef struct L_POOL
{
char *fname;
BITMAP *image;
} L_POOL;

// A complete animation sequence

typedef struct SEQ_POOL
{
char *name;
S_POOL **seq;
int frames;
unsigned char flags;
int x,y;
unsigned char speed;
S_POOL *overlay;        // This is always drawn on top of the sprite
                        // but it has no physical existence
struct SEQ_POOL *jumpto;// Sequence to jump to on exit if necessary
char hook[32];          // VRM hook function to call if necessary
} SEQ_POOL;

// Virtual Runtime Module code

typedef struct VRM
	{
	char *name;
	char *fname;
	char *vrm;
	void *instance;
	int count;
	} VRM;

typedef struct PEM
	{
	char *name;
	char *file;
	char *code;
	int codelen;
	int size;
	char hidden;		// local function
	} PEM;

///////////////////////////////////////////////////////////////////////////
//  Graphics Kernel Data Structures                                      //
///////////////////////////////////////////////////////////////////////////

typedef struct S_FLAGS
{
unsigned long 	 on	:1;     // active?
unsigned long 	 willopen:1;    // Is is always impassible?
unsigned long 	 overlay:1;     // Is it an overlay?
unsigned long 	 solid:1;       // Is it solid to people?
unsigned long    fragile:1;     // Will it break if you drop it?
unsigned long    trigger:1;     // Is it a trigger?
unsigned long    invisible:1;   // Will it only appear in the editor?
unsigned long    party:1;       // Is it a Party Member? (affects solidity)
unsigned long    fixed:1;       // Is is nailed to the ground (can't move)
unsigned long    container:1;   // Is is a container?
unsigned long    translucent:1; // Is it translucent?
unsigned long    large :1;      // Object bigger than 1 tile
unsigned long    spikeproof:1;  // Isn't hurt by objects below
unsigned long    wield :1;      // Can it be wielded?
unsigned long    stepupdated :1;// Was just updated (used by stepping code)
unsigned long    rangeweapon:1; // If a weapon, is it close or ranged?
unsigned long    blocklight :1; // Does it blocklight?
unsigned long    tabletop   :1; // Can you drop things on it even if it's solid?
unsigned long    didinit    :1; // Have we done the INIT function?
unsigned long    didupdate  :1; // Have we moved just now (prevent runaway)?
unsigned long    person     :1; // Is a person
unsigned long    inpocket   :1; // Is it inside a container?
unsigned long    quantity   :1; // Can it be a pile of things?
unsigned long    watery     :1; // Can it go on water (or is it a water tile?)
unsigned long    shadow     :1; // Is it just a dark patch?
unsigned long 	 spare      :7;
} S_FLAGS;

typedef struct FUNCS            // This is called when you...
{
char use[32];           // USE it
int  ucache;
char *suse;             // This is a pointer to the string array, for PEscript

char talk[128];          // TALK to it
int  tcache;
char *stalk;

char kill[32];          // KILL/BREAK it
int  kcache;
char *skill;

char look[32];          // LOOK at it
int  lcache;
char *slook;

char stand[32];         // STAND on it
int  scache;
char *sstand;

char hurt[32];          // HURT/DAMAGE it
int  hcache;
char *shurt;

char init[32];          // Called when object first created
int  icache;
char *sinit;

// The ?cache number succeeding each entry is the index in the array of VRMs
// to the one that will be called, and is used for optimisation.

char contains[8][32];
int contents;
// Up to 8 objects can be put in the object when it's first created

char resurrect[32];          // Object it becomes when resurrected
char *sresurrect;
int dirty;                  // Changed from defaults?  (used in editor)
//short hackpadding;		// Best you don't know..
char wield[32];          // Called when object is wielded/removed
int  wcache;
char *swield;

} FUNCS;

// Tiles

typedef struct TILE
{
char *name;
S_FLAGS flags;
SEQ_POOL *form;     // Direct access to animation
char seqname[32];   // Name of animation (in editor)
int sptr,slen;      // animation info
int sdir,tick;      // animation info
int sdx,sdy;        // Scroll delta x,y
int sx,sy;          // scroll coordinates x,y
FUNCS *funcs;       // VRM Function table
char *desc;         // description
} TILE;

// AI objects

typedef struct SCHEDULE_065 // Backward compatability version
{
char ok,hour,minute;

char compat,abil,ity;	// These are for binary compatability with 0.65..

char vrm[32];
int call;
struct OBJECT *target;
void *compat4;			// ..they will disappear after 0.7
char targettype[32]; // Object type in case object is destroyed
} SCHEDULE_065;


typedef struct SCHEDULE
{
char ok,hour,minute;

char vrm[32];
int call;
struct OBJECT *target;
char targettype[32]; // Object type in case object is destroyed
} SCHEDULE;


typedef struct WIELD
{
struct OBJECT *head;
struct OBJECT *neck;
struct OBJECT *body;
struct OBJECT *legs;
struct OBJECT *feet;
struct OBJECT *arms;
struct OBJECT *l_hand;
struct OBJECT *r_hand;
struct OBJECT *l_finger;
struct OBJECT *r_finger;
struct OBJECT *spare1;
struct OBJECT *spare2;
struct OBJECT *spare3;
struct OBJECT *spare4;
} WIELD;

// Moving object entitiy

typedef struct OBJECT
{
char *name;
S_FLAGS flags;
int w,h,mw,mh;
int x,y,z;
char *personalname;
SCHEDULE *schedule;
void *UN_USED;
SEQ_POOL *form;
int sptr,slen,sdir; // animation info
STATS *maxstats;    // Statistics structure
STATS *stats;       // Current stats
FUNCS *funcs;       // VRM Function table
int curdir;
short dir[4];       // u,d,l,r
char *desc;         // Examine Description
char *shortdesc;    // Short Description
struct OBJECT *target; // Current target object (behaviour goal, if any)
int   tag;          // Tags are used for triggering things
int   activity;     // Current activity VRM
int   SPARE[8];     //
struct USEDATA *user;      // User data for a game in progress
unsigned char hblock[4]; // xywh
unsigned char vblock[4]; // xywh
unsigned char harea[4]; // xywh
unsigned char varea[4]; // xywh
unsigned int save_id;     // Assigned object number (during saving and loading)
unsigned int inside_id;   // What object is it inside? (during saving and loading)
int temp[8];     // Reserved space
int light;        // Light it casts
struct OBJECT *enemy;
struct OBJECT **wield;
struct OBJECT *pocket;
struct OBJECT *next;
} OBJECT;

#define BLK_X 0
#define BLK_Y 1
#define BLK_W 2
#define BLK_H 3

// World structure

typedef struct WORLD
{
int sig;					// Signature
int w,h;
unsigned short *physmap;	// Physical world
OBJECT *object;				// Linked list of objects
unsigned char *roof;		// Rooftop
OBJECT **objmap;			// Object Map
unsigned char *light;		// Static Lighting
unsigned char *lightst;
int *temp[29];
} WORLD;

typedef struct OBJLIST
{
OBJECT *ptr;
struct OBJLIST *next;
} OBJLIST;

typedef struct PEVM
	{
	OBJECT *reg[4];		// Local object registers
	char *name;			// Name of function (for debugging)
	char *file;			// Filename of function (for debugging)
	char *code;			// code and data
	int size;			// length of code and data
	int codelen;		// length of code and data
	char *ip;			// instruction pointer
	struct PEVM *next;	// for use as a linked-list (vmstack etc)
	} PEVM;

// Externals

extern OBJLIST *MasterList;
#ifdef __cplusplus
extern "C"
	{
#endif

extern int gamewinsize;
extern BITMAP *gamewin;
extern BITMAP *roofwin;
extern int VSW; // Tiles
extern int VSH;
extern int VSA;
extern int VSMIDX;
extern int VSMIDY;
extern int VSW32; // Pixels
extern int VSH32;
extern int VSA32;

	#ifdef __cplusplus
	}
	#endif

#endif

