/*
 *  - Large view of the world
 */


#include <stdio.h>
#include "igui.hpp"
#include "../ithelib.h"
#include "../console.h"
#include "../core.hpp"
#include "../gamedata.h"
#include "menusys.h"

// defines

#define MicroX 512
#define MicroY 288

// variables

static int pmx=0,pmy=0;
static int SafetyCatch=1;

extern int focus,BM_Id;
extern int L_bgtile;

extern long M_core;     // Total amount of core, from memory.cc
extern int s_proj;      // Flag, are sprites being displayed?
extern int l_proj;      // Flag, are layers being displayed?

// functions

extern int ReadMap(int x,int y);
extern void WriteMap(int x,int y, int tile);
extern void Toolbar();

static void OneToOne();
static void PanMap();
static void PanUp();
static void PanDown();
static void PanLeft();
static void PanRight();
static void SetL();
static void SetM();
static void Spray();
static void GoMap();
static void Nothing();

// GoFocal - The menu tab function.  This is called from the toolbar

void BM_GoFocal()
{
int temp;               // Used to get button handles

if(focus==4)            // Make sure we don't redraw if the user clicks again
	return;

focus=4;                // This is now the focus
Toolbar();              // Build up the menu system again from scratch
IG_SetFocus(BM_Id);     // And make the button stick inwards

// Add Panning buttons

IG_TextButton(516,104,__up,PanUp,NULL,NULL);         // Pan up button
IG_TextButton(500,128,__left,PanLeft,NULL,NULL);     // Pan left button
IG_TextButton(532,128,__right,PanRight,NULL,NULL);   // Pan right button
IG_TextButton(516,150,__down,PanDown,NULL,NULL);     // Pan down button

temp = IG_ToggleButton(500,256,"Can Edit ",Nothing,NULL,NULL,&SafetyCatch);
IG_SetInText(temp,"Read-Only");

// Set up Map windows

IG_BlackPanel(MicroX-2,MicroY-2,68,68);

if(curmap->w>=384 || curmap->h>=384)
    {
    IG_BlackPanel(62,62,386,386);
    PanMap();
    IG_AddKey(KEY_UP,PanUp);
    IG_AddKey(KEY_DOWN,PanDown);
    IG_AddKey(KEY_LEFT,PanLeft);
    IG_AddKey(KEY_RIGHT,PanRight);
    }
else
    OneToOne();

IG_Region(64,64,384,384,SetL,SetM,GoMap);

}

void OneToOne()
{
int temp,cx,cy,div,rx,ry;
IG_BlackPanel(62,62,386,386);

for(cy=0;cy<curmap->h;cy++)
    for(cx=0;cx<curmap->w;cx++)
        {
        temp=ReadMap(cx,cy);
        if(temp==RANDOM_TILE)
            putpixel(swapscreen,cx+64,cy+64,ITG_BLUE);
        else
            putpixel(swapscreen,cx+64,cy+64,TIlist[temp].form->seq[0]->thumbcol);
        }

div = curmap->w>>6; //div64
rx=0;
ry=0;
for(cy=0;cy<64;cy++)
	{
    for(cx=0;cx<64;cx++)
		{
		temp = ReadMap(rx,ry);
		if(temp==RANDOM_TILE)
			putpixel(swapscreen,MicroX+cx,MicroY+cy,ITG_BLUE);
		else
			putpixel(swapscreen,MicroX+cx,MicroY+cy,TIlist[temp].form->seq[0]->thumbcol);
		rx+=div;
		}
	rx=0;
	ry+=div;
	}
}

void PanMap()
{
int cx,cy,temp,cw,ch,div,rx,ry;
if(pmx+384 > curmap->w)
    pmx = (curmap->w-384);
if(pmy+384 > curmap->h)
    pmy = (curmap->h-384);
if(pmx < 0)
    pmx = 0;
if(pmy < 0)
    pmy = 0;

cw=curmap->w;
if(cw>384)
    cw=384;
ch=curmap->h;
if(ch>384)
    ch=384;

for(cy=0;cy<ch;cy++)
    for(cx=0;cx<cw;cx++)
        {
        temp=ReadMap(cx+pmx,cy+pmy);
        if(temp==RANDOM_TILE)
            putpixel(swapscreen,cx+64,cy+64,ITG_BLUE);
        else
            putpixel(swapscreen,cx+64,cy+64,TIlist[temp].form->seq[0]->thumbcol);
        }

div = curmap->w>>6; // div64
rx=0;
ry=0;
for(cy=0;cy<64;cy++)
	{
    for(cx=0;cx<64;cx++)
		{
		temp = ReadMap(rx,ry);
		if(temp==RANDOM_TILE)
			putpixel(swapscreen,MicroX+cx,MicroY+cy,ITG_BLUE);
		else
			putpixel(swapscreen,MicroX+cx,MicroY+cy,TIlist[temp].form->seq[0]->thumbcol);
		rx+=div;
		}
	rx=0;
	ry+=div;
	}
}

void PanUp()
{
pmy-=16;
PanMap();
}

void PanDown()
{
pmy+=16;
PanMap();
}

void PanLeft()
{
pmx-=16;
PanMap();
}

void PanRight()
{
pmx+=16;
PanMap();
}

void SetL()
{
int hx,hy,xx,yy,nodraw=0,col,div;
static int lastx=666,lasty=666;

if(SafetyCatch)
	return;

hx = (x-64)+pmx;
hy = (y-64)+pmy;

nodraw=0;
if(hx==lastx && hy==lasty)
	nodraw=1;
lastx=hx;lasty=hy;

if(L_bgtile==RANDOM_TILE)
	col=ITG_BLUE;
else
	col=TIlist[L_bgtile].form->seq[0]->thumbcol;

if(hx>(curmap->w-8))
	hx=(curmap->w-8);
if(hy>(curmap->h-8))
	hy=(curmap->h-8);

for(xx=0;xx<8;xx++)
	for(yy=0;yy<8;yy++)
		{
        WriteMap(hx+xx,hy+yy,L_bgtile);
		if(x+xx<448 && y+yy<448)
			putpixel(swapscreen,x+xx,y+yy,col);
		}
div=curmap->w>>6; // divide by 64
putpixel(swapscreen,MicroX+(hx/div),MicroY+(hy/div),col);
}

void SetM()
{
int hx,hy,xx,yy,nodraw=0,col,div;
static int lastx=666,lasty=666;

if(SafetyCatch)
	return;

hx = (x-64)+pmx;
hy = (y-64)+pmy;

nodraw=0;
if(hx==lastx && hy==lasty)
	nodraw=1;
lastx=hx;lasty=hy;

if(L_bgtile==RANDOM_TILE)
	col=ITG_BLUE;
else
	col=TIlist[L_bgtile].form->seq[0]->thumbcol;

if(hx>(curmap->w-32))
	hx=(curmap->w-32);
if(hy>(curmap->h-32))
	hy=(curmap->h-32);

for(xx=0;xx<32;xx++)
	for(yy=0;yy<32;yy++)
		{
        WriteMap(hx+xx,hy+yy,L_bgtile);
		if(x+xx<448 && y+yy<448)
			putpixel(swapscreen,x+xx,y+yy,col);
		}
div=curmap->w>>6; // divide by 64
putpixel(swapscreen,MicroX+(hx/div),MicroY+(hy/div),col);
}

void GoMap()
{
mapx = (x-64)+pmx;
mapy = (y-64)+pmy;
}

void Spray()
{
int hx,hy,xx,yy,nodraw=0,col,div;
static int lastx=666,lasty=666;

if(SafetyCatch)
	return;

hx = (x-64)+pmx;
hy = (y-64)+pmy;

nodraw=0;
if(hx==lastx && hy==lasty)
	nodraw=1;
lastx=hx;lasty=hy;

if(L_bgtile==RANDOM_TILE)
	col=ITG_BLUE;
else
	col=TIlist[L_bgtile].form->seq[0]->thumbcol;

if(hx>(curmap->w-8))
	hx=(curmap->w-8);
if(hy>(curmap->h-8))
	hy=(curmap->h-8);

for(xx=0;xx<8;xx++)
	for(yy=0;yy<8;yy++)
		{
		if(rand()%100>95)
			{
			WriteMap(hx+xx,hy+yy,L_bgtile);
			if(x+xx<448 && y+yy<448)
				putpixel(swapscreen,x+xx,y+yy,col);
			}
		}
//div=curmap->w>>6; // divide by 64
//putpixel(swapscreen,MicroX+(hx/div),MicroY+(hy/div),col);
}

void Nothing()
{
PanMap();
}
