//
//      IRE init systems
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// MR: added
#ifndef _WIN32
#include <unistd.h>
#endif

// defines

#define IRE_SYSTEM_MODULE
#define ALREADY_KNOW_ITG
#define UNDEFINED -1

// includes

#include "ithelib.h"
#include "media.h"
#include "console.h"
#include "init.h"
#include "loadfile.h"
#include "gamedata.h"
#include "script.hpp"
#include "oscli.h"
#include "seer/seer.h"
//#include "vrm_in.hpp"


// Global variables

extern char allegro_error[];

//SQ **SQlist;    long SQtot;     // Sequences in the SQlist, no. of sequences
//VRM *COlist;    long COtot;     // Array of VRM modules, no. of VRMs present
//CH *CHlist;     long CHtot;     // Array of characters, no. of characters
//SP *SPlist;     long SPtot;     // Array of sprites, number of sprites

VRM *COlist;                    // Virtual Runtime Module table
PEM *PElist;                    // PissEasy Script table
S_POOL *s_pool;                 // Pool of sprites loaded into memory
SEQ_POOL *seq_pool;             // Pool of animation sequences

typedef struct search_st
	{
	char *name;
	int num;
	} search_st;

static search_st *sprites;
static search_st *sequences;
static search_st *scripts;
static search_st *characters;

unsigned char digit[256][8];

unsigned short font_bground=0,font_fground=0x7fff;
char stemp[128];
char *rootname="main\0";

//extern int vscheduler,status,follower,loadproc,all_dead,playervrm,splashpe,trackstop,update_life,erase_object;
extern char mapname[];
//extern char big_font;

// Local variables

// functions

short FindProc(char *name);
extern void register_exporttable();
extern int sizeof_sprite(BITMAP *bmp);
extern BITMAP *load_lightmap(char *filename);
static int CMP_sort(void *a, void *b);
static int CMP_search(void *a, void *b);

// code

/*
 *      Init_Sprite() - Load in a single sprite, called by SCRIPT.CPP
 */

long Init_Sprite(int number)
{
char filename[1024];
BITMAP *img;

if(!loadfile(SPlist[number].fname,filename))
    ithe_panic("Could not load image: (file not found)",SPlist[number].fname);

//img = load_bitmap(filename,pal);
img = iload_bitmap(filename);
SPlist[number].image = get_rle_sprite(img);
SPlist[number].thumbcol=getpixel(img,0,0); // Get a thumbnail colour
destroy_bitmap(img);

//SPlist[number].image = load_bitmap(filename,pal);
if(!SPlist[number].image)
    ithe_panic("Could not load image: unsupported format?",filename);

/*
draw_sprite(swapscreen,SPlist[number].image,0,0);
Show();
WaitForAscii();
*/

// Now store the width and height of the sprite in the pool

SPlist[number].w=SPlist[number].image->w;
SPlist[number].h=SPlist[number].image->h;

// This is the width times the height, for fast access later


SPlist[number].wxh=SPlist[number].w*SPlist[number].h;

Plot(0);

//return sizeof_sprite(SPlist[number].image);
return 1;
}



/*
 *      Init_RoofTile() - Load in a single sprite, called by SCRIPT.CC
 */

long Init_RoofTile(int number)
{
char filename[1024];
BITMAP *img;

if(!loadfile(RTlist[number].fname,filename))
    ithe_panic("Could not load image: (file not found)",RTlist[number].fname);

//img = load_bitmap(filename,pal);
img = iload_bitmap(filename);

RTlist[number].image = get_rle_sprite(img);
destroy_bitmap(img);

if(!RTlist[number].image)
    ithe_panic("Could not load image: unsupported format?",filename);

// Now store the width and height of the sprite in the pool

RTlist[number].w=RTlist[number].image->w;
RTlist[number].h=RTlist[number].image->h;

// This is the width times the height, for fast access later

RTlist[number].wxh=RTlist[number].w*RTlist[number].h;

// Write a dot for this image.  This Plot was set up in script.cc

Plot(0);

//return sizeof_sprite(RTlist[number].image);
return 1;
}


/*
 *      Init_LightMap() - Load in a single sprite, called by SCRIPT.CC
 */

long Init_LightMap(int number)
{
char filename[1024];

if(!loadfile(LTlist[number].fname,filename))
    ithe_panic("Could not load lightmap image: (file not found)",LTlist[number].fname);

LTlist[number].image = load_lightmap(filename);

if(!LTlist[number].image)
    ithe_panic("Could not load image: unsupported format?",filename);

Plot(0);

return 1;
}


/*
 *      Init_Font - Load in the character set.
 */

void Init_Font()
{
ilog_quiet("Init_Font()\n");

/*
if(big_font)
    fp.open("font.big");            // Open the 'visually impaired' font
else
    fp.open("font.dat");            // Open the normal font
fp.read(digit,1,2048);           // Read it
fp.close();                      // Close it
*/
}


/*
 *    find_spr()  -  find the address of a sprite by its name
 */

S_POOL *find_spr(char *name)
{
int cx;
for(cx=0;cx<SPtot;cx++)
	if(!stricmp(SPlist[cx].name,name))
		return(&SPlist[cx]);
return NULL;
}


/*
 *    Init_VRM()  -  Used in the VRM engine to initialise a module
 */

void Init_VRM(int pos)
{
int len;
IFILE *fp;
char name[1024];

Plot(0);

// Open the VRM file, get the size, load it in as a huge string
// unless we're in the editor, in which case we just store the names

if(!in_editor)
	{
	if(!loadfile(COlist[pos].fname,name))
		ithe_panic("Could not load VRM file",COlist[pos].fname);
	fp=iopen(name);
	len=ifilelength(fp);
	COlist[pos].vrm = (char *)M_get(1,len+1);
	iread(COlist[pos].vrm,len,fp);
	iclose(fp);

	if(memcmp(COlist[pos].vrm,"SeeRVCPU",8))
		ithe_panic("This VRM was not compiled with SEER!",COlist[pos].fname);
// Do the read

	if(len != scGet_Script_Size(COlist[pos].vrm))
		ithe_panic(scErrorMsg,scErrorLine);
	}

if(!stricmp(COlist[pos].name,"all_dead"))
	all_dead = pos;

//if(!stricmp(COlist[pos].name,"follower"))
//	follower = pos;

if(!stricmp(COlist[pos].name,"loadproc"))
	loadproc = pos;

//if(!stricmp(COlist[pos].name,"player"))
//	playervrm = pos;

}

/*
 *    Submit a function, register it with the game if it is a System Function
 */

void Init_PE(int pos)
{
if(!stricmp(PElist[pos].name,"status"))
	status = pos;
if(!stricmp(PElist[pos].name,"erase"))
	erase_object = pos;
if(!stricmp(PElist[pos].name,"splash"))
	splashpe = pos;
if(!stricmp(PElist[pos].name,"scheduler"))
	vscheduler = pos;
if(!stricmp(PElist[pos].name,"update_life"))
	update_life = pos;
if(!stricmp(PElist[pos].name,"trackstop"))
	trackstop = pos;

}


/*
 *    Init_Funcs()  -  Assign each function in the characters to a VRM
 *						Calls InitFuncsFor(object)
 */

void Init_Funcs()
{
int ctr;

// Critical VRMs

if(all_dead == UNDEFINED)
	ithe_panic("Could not find system function 'all_dead' in the VRM list!",NULL);

//if(playervrm == UNDEFINED)
//	ithe_panic("Could not find system function 'player_action' in the VRM list!","Please make sure the player's behaviour VRM is called 'PLAYER'");

if(trackstop == UNDEFINED)
	ithe_panic("Could not find system function 'trackstop' in the VRM list!",NULL);

if(update_life == UNDEFINED)
	ithe_panic("Could not find system function 'update_life'!",NULL);

if(erase_object == UNDEFINED)
	ithe_panic("Could not find system function 'erase'!",NULL);

// Non-critical VRMs

if(vscheduler == UNDEFINED)
	ilog_quiet("Did not find system function 'scheduler', continuing anyway\n");
if(status == UNDEFINED)
	ilog_quiet("Did not find system function 'status', continuing anyway\n");
if(follower == UNDEFINED)
	ilog_quiet("Did not find system function 'follower', continuing anyway\n");
if(loadproc == UNDEFINED)
	ilog_quiet("Did not find system function 'loadproc', continuing anyway\n");
if(splashpe == UNDEFINED)
	ilog_quiet("Did not find system function 'splash', continuing anyway\n");

for(ctr=0;ctr<CHtot;ctr++)
	InitFuncsFor(&CHlist[ctr],0);
}

/*
 *    InitFuncsFor()  -  Assign each function in a single character to a VRM
 *						 Called by Init_Funcs and somewhere in loadsave.cpp
 */

void InitFuncsFor(OBJECT *o, int rebuild)
{
char *str;
int tmp,objtype;

if(rebuild)
	objtype=getnum4char(o->name);

    // Find the default USE function in the VRM list and put it in the Ucache
    str = o->funcs->use;
    if(str[0])
        {
        tmp = getnum4PE(str);
        if(tmp == UNDEFINED)
            {
            Bug("use: Did not find function '%s'.  Ignoring\n",str);
            o->funcs->ucache = -1;
			if(rebuild)
				{
				o->funcs->ucache = CHlist[objtype].funcs->ucache;
				strcpy(o->funcs->use,CHlist[objtype].funcs->use);
				}
            }
        else
            o->funcs->ucache = tmp;
        }
    else
		{
		o->funcs->ucache = -1;
		}

    // Find a STAND function in the VRM list and put it in the Scache
    // Switch off the trigger bit if none found

	str = o->funcs->stand;
	if(str[0])
		{
		tmp = getnum4PE(str);
		if(tmp == UNDEFINED)
			{
			Bug("stand: Did not find function '%s'.  Ignoring\n",str);
			o->funcs->scache = -1;
			if(rebuild)
				{
				o->funcs->scache = CHlist[objtype].funcs->scache;
				strcpy(o->funcs->stand,CHlist[objtype].funcs->stand);
				}
			else
				o->flags.trigger = 0;
			}
		else
			o->funcs->scache = tmp;
		}
	else
		o->flags.trigger = 0;

    // Find the default LOOK function in the VRM list and put it in the Lcache

    str = o->funcs->look;
    if(str[0])
        {
        tmp = getnum4PE(str);
        if(tmp==UNDEFINED)
            {
            Bug("look: Did not find function '%s'.  Ignoring\n",str);
            o->funcs->lcache = -1;
			if(rebuild)
				{
				o->funcs->lcache = CHlist[objtype].funcs->lcache;
				strcpy(o->funcs->look,CHlist[objtype].funcs->look);
				}
            }
        else
            o->funcs->lcache = tmp;
        }
    else
        o->funcs->lcache = -1;

    // Find the default KILLED function in the VRM list and put it in the Kcache
    // If none, the engine will just vanish the dead thing

    str = o->funcs->kill;
    if(str[0])
        {
        tmp = getnum4PE(str);
        if(tmp == UNDEFINED)
            {
            Bug("kill: Did not find function '%s'.  Ignoring\n",str);
            o->funcs->kcache = -1;
			if(rebuild)
				{
				o->funcs->kcache = CHlist[objtype].funcs->kcache;
				strcpy(o->funcs->kill,CHlist[objtype].funcs->kill);
				}
            }
        else
            o->funcs->kcache = tmp;
        }
    else
        o->funcs->kcache = -1;

    // Find the default HURT function in the VRM list and put it in the Hcache
    str = o->funcs->hurt;
    if(str[0])
        {
        tmp = getnum4PE(str);
        if(tmp == UNDEFINED)
            {
            Bug("hurt: Did not find function '%s'.  Ignoring\n",str);
            o->funcs->hcache = -1;
			if(rebuild)
				{
				o->funcs->hcache = CHlist[objtype].funcs->hcache;
				strcpy(o->funcs->hurt,CHlist[objtype].funcs->hurt);
				}
            }
        else
            o->funcs->hcache = tmp;
        }
    else
        o->funcs->hcache = -1;

    // Find the default WIELD function in the VRM list and put it in the Wcache
    str = o->funcs->wield;
    if(str[0])
        {
        tmp = getnum4PE(str);
        if(tmp == UNDEFINED)
            {
            Bug("wield: Did not find function '%s'.  Ignoring\n",str);
            o->funcs->wcache = -1;
			if(rebuild)
				{
				o->funcs->wcache = CHlist[objtype].funcs->wcache;
				strcpy(o->funcs->wield,CHlist[objtype].funcs->wield);
				}
            }
        else
            o->funcs->wcache = tmp;
        }
    else
        o->funcs->wcache = -1;

    // Find the default INIT function in the VRM list and put it in the Icache
    str = o->funcs->init;
    if(str[0])
        {
        tmp = getnum4PE(str);
        if(tmp == UNDEFINED)
            {
            Bug("init: Did not find function '%s'.  Ignoring\n",str);
            o->funcs->icache = -1;
			if(rebuild)
				{
				o->funcs->icache = CHlist[objtype].funcs->icache;
				strcpy(o->funcs->init,CHlist[objtype].funcs->init);
				}
            }
        else
            o->funcs->icache = tmp;
        }
    else
        o->funcs->icache = -1;

}


/*
 *		(re)calculate the size of a large object
 */

void Init_Areas(OBJECT *objsel)
{
// First the vertical

objsel->w = SQlist[objsel->dir[CHAR_D]].seq[0]->w;
objsel->h = SQlist[objsel->dir[CHAR_D]].seq[0]->h;

// Set up the size in tiles
objsel->mw = objsel->w>>5;
if(objsel->w & 0xf || !objsel->mw)
     objsel->mw++;

objsel->mh = objsel->h>>5;
if(objsel->h & 0xf || !objsel->mh)
     objsel->mh++;

if(objsel->varea[BLK_W]==0)
    {
    objsel->varea[BLK_W]=objsel->mw;
    objsel->varea[BLK_X]=0;
    }
if(objsel->varea[BLK_H]==0)
    {
    objsel->varea[BLK_H]=objsel->mh;
    objsel->varea[BLK_Y]=0;
    }
if(objsel->vblock[BLK_W]==0)
    {
    objsel->vblock[BLK_W]=objsel->mw;
    objsel->vblock[BLK_X]=0;
    }
if(objsel->vblock[BLK_H]==0)
    {
    objsel->vblock[BLK_H]=objsel->mh;
    objsel->vblock[BLK_Y]=0;
    }

// Then the horizontal

objsel->w = SQlist[objsel->dir[CHAR_R]].seq[0]->w;
objsel->h = SQlist[objsel->dir[CHAR_R]].seq[0]->h;

// Set up the size in tiles
objsel->mw = objsel->w>>5;
if(objsel->w & 0xf || !objsel->mw)
     objsel->mw++;

objsel->mh = objsel->h>>5;
if(objsel->h & 0xf || !objsel->mh)
     objsel->mh++;

if(objsel->harea[BLK_W]==0)
    {
    objsel->harea[BLK_W]=objsel->mw;
    objsel->harea[BLK_X]=0;
    }
if(objsel->harea[BLK_H]==0)
    {
    objsel->harea[BLK_H]=objsel->mh;
    objsel->harea[BLK_Y]=0;
    }
if(objsel->hblock[BLK_W]==0)
    {
    objsel->hblock[BLK_W]=objsel->mw;
    objsel->hblock[BLK_X]=0;
    }
if(objsel->hblock[BLK_H]==0)
    {
    objsel->hblock[BLK_H]=objsel->mh;
    objsel->hblock[BLK_Y]=0;
    }
}

/*
 *      Getnum4char - Find the index of the character in the CHlist array
 */

int getnum4char_slow(char *name)
{
int ctr;
if(!name)
    return -1;
for(ctr=0;ctr<CHtot;ctr++)
	if(!istricmp(name,CHlist[ctr].name))
		return ctr;
return -1;
}


/*
 *      Getnum4PE - Find the index of the PE script in the PElist array
 */

int getnum4PE_slow(char *name)
{
int ctr;

if(PEtot <1)
	ithe_panic("getnum4PE called before any scripts defined",NULL);

// Fast search first

for(ctr=0;ctr<PEtot;ctr++)
	if(!PElist[ctr].hidden)
		if(!istricmp(name,PElist[ctr].name))
			return ctr;

// Not found.  Is it a local function?

for(ctr=0;ctr<PEtot;ctr++)
	if(!istricmp(name,PElist[ctr].name))
		if(PElist[ctr].hidden)
			{
			Bug("PE script '%s' is local.. can't call it directly\n",name);
			return -1;
			}

// It simply doesn't exist..

Bug("PE script '%s' not found\n",name);
return -1;
}


/*
 *      Getnum4sprite - Find the index of the sequence in the SPlist array
 */

int getnum4sprite_slow(char *name)
{
int ctr;
for(ctr=0;ctr<SPtot;ctr++)
	if(!istricmp(name,SPlist[ctr].name))
		return ctr;
return -1;
}

/*
 *      Getnum4sequence - Find the index of the sequence in the SQlist array
 */

int getnum4sequence_slow(char *name)
{
int ctr;

for(ctr=0;ctr<SQtot;ctr++)
	{
	if(!istricmp(name,SQlist[ctr].name))
		return ctr;
	}

return -1;

}


static int CMP_sort(void *a, void *b)
{
return istricmp(((search_st *)a)->name,((search_st *)b)->name);
}

static int CMP_search(void *a, void *b)
{
return istricmp((char *)a,((search_st *)b)->name);
}

//
//	Build ordered lists for quick searching
//

void Init_Lookups()
{
int ctr;

if(SPtot < 1)
	ithe_panic("No sprites","Init_Lookups");
if(SQtot < 1)
	ithe_panic("No sequences","Init_Lookups");
if(CHtot < 1)
	ithe_panic("No characters","Init_Lookups");
if(PEtot < 1)
	ithe_panic("No scripts","Init_Lookups");

// Sprites

sprites = (search_st *)M_get(SPtot+1,sizeof(search_st));
for(ctr=0;ctr<SPtot;ctr++)
	{
	sprites[ctr].name=SPlist[ctr].name;
	sprites[ctr].num=ctr;
	}
qsort(sprites,SPtot,sizeof(search_st),CMP_sort);


// Sequences

sequences = (search_st *)M_get(SQtot+1,sizeof(search_st));
for(ctr=0;ctr<SQtot;ctr++)
	{
	sequences[ctr].name=SQlist[ctr].name;
	sequences[ctr].num=ctr;
	}
qsort(sequences,SQtot,sizeof(search_st),CMP_sort);

// Characters

characters = (search_st *)M_get(CHtot+1,sizeof(search_st));
for(ctr=0;ctr<CHtot;ctr++)
	{
	characters[ctr].name=CHlist[ctr].name;
	characters[ctr].num=ctr;
	}
qsort(characters,CHtot,sizeof(search_st),CMP_sort);

// Scripts

scripts = (search_st *)M_get(PEtot+1,sizeof(search_st));
for(ctr=0;ctr<PEtot;ctr++)
	{
	scripts[ctr].name=PElist[ctr].name;
	scripts[ctr].num=ctr;
	}
qsort(scripts,PEtot,sizeof(search_st),CMP_sort);

}

//
//	Search for a string
//

#define FIND(key,list,num) bsearch(key,list,num,sizeof(search_st),CMP_search)

/*
 *      Getnum4char - Find the index of the character in the CHlist array
 */

int getnum4char(char *name)
{
search_st *p;

p = FIND(name,characters,CHtot);
if(p)
	return p->num;
return -1;
}

/*
 *      Getnum4VRM - Find the index of the VRM in the COlist array
 */

int getnum4VRM(char *name)
{
int ctr;

for(ctr=0;ctr<COtot;ctr++)
	if(!istricmp(name,COlist[ctr].name))
		return ctr;
Bug("VRM '%s' not found\n",name);
return -1;
}

/*
 *      Getnum4PE - Find the index of the PE script in the PElist array
 */

int getnum4PE(char *name)
{
search_st *p;

p = FIND(name,scripts,PEtot);
if(p)
	{
	if(PElist[p->num].hidden)
		{
		Bug("PE script '%s' is local.. can't call it directly\n",name);
		return -1;
		}
	return p->num;
	}
return -1;
}


/*
 *      Getnum4sprite - Find the index of the sequence in the SPlist array
 */

int getnum4sprite(char *name)
{
search_st *p;

p = FIND(name,sprites,SPtot);
if(p)
	return p->num;
return -1;
}

/*
 *      Getnum4sequence - Find the index of the sequence in the SQlist array
 */

int getnum4sequence(char *name)
{
search_st *p;

p = FIND(name,sequences,SQtot);
if(p)
	return p->num;
return -1;
}

