/*
        IRE game engine

Copyright (c) 2001, IT-HE Software
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of IT-HE Software nor the names of its contributors may
be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.
IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

//#define SEER_CALLS	// How much is seer called?


#ifndef _WIN32
	#include <unistd.h>
	#ifdef __DJGPP__
		#include <conio.h>
	#endif
#else
	#include <conio.h>
#endif


#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>

#include <allegro.h>

#include "core.hpp"
#include "console.h"
#include "media.h"
#include "ithelib.h"
#include "oscli.h"      // Operating System Command Line Interpreter
#include "cookies.h"
#include "gamedata.h"
#include "loadfile.h"
#include "sound.h"
#include "init.h"
#include "linklist.hpp"


/*
 * Defines
 */

#define DATAFILES	// Use the DATAFILE subsystem

/*
 * Variables
 */

BITMAP *gamewin;        // Projection window
BITMAP *roofwin;        // Projection window
static BITMAP *s_soundbar,*s_slider1,*s_slider2;

/*
 * Functions
 */

static void credits();
void RFS_getescape();
static void Load_Soundbar();
static void read_config();
static void write_config();

extern void LoadResources(char *filename);
extern void Init_VRMs();
extern void InitVM();
extern void startgame();
extern void LoadMap();
extern "C" void scan_rar_file(char *filename);

#ifdef __DJGPP__
#include <crt0.h>
int _crt0_startup_flags =_CRT0_FLAG_FILL_SBRK_MEMORY | _CRT0_FLAG_FILL_DEADBEEF | _CRT0_FLAG_LOCK_MEMORY;
#endif

// Code

int main(int argc, char *argv[])
{
#ifdef FORTIFY
remove("fortify.out");
#endif

curmap = &worldcache[0]; // For multiple map support later

//worldcache[0]=

allegro_init(); // We need this ASAP

strcpy(resdir,"flat"); // Set to sensible default

M_init();         // Init memory systems

party = (OBJECT **)M_get(MAX_MEMBERS,sizeof(OBJECT *));
partyname = (char **)M_get(MAX_MEMBERS,sizeof(char *));

// Cache these linked lists to speed up loading levels
LLcache_register(&MasterList);
LLcache_register(&ActiveList);

strcpy(newbootlog,"bootlog.txt"); // This could be changed later in OSCLI
strcpy(oldbootlog,"bootlog.prv"); // If I remember to do so..

// First print it for the user

printf("IRE Kernel version: %s\n",IKV);
printf("Build date: %s at %s\n",__DATE__,__TIME__);
printf("IT-HE lib version: %s\n",ithe_get_version());
printf("\n");
#ifndef _WIN32
sleep(1);
#endif

// Now get the config settings (this may print stuff if RARfiles are added)

if(argc>=2)
	OSCLI(argc,argv);       // input from OSCLI

INI_file("game.ini");           // Read the ini file and act upon it

ilog_start(newbootlog,oldbootlog);

// Now write the stuff to the logfile

ilog_quiet("IRE Kernel version: %s\n",IKV);
ilog_quiet("Build date: %s at %s\n",__DATE__,__TIME__);
ilog_quiet("IT-HE lib version: %s\n",ithe_get_version());
ilog_quiet("\n");

// Now do the map data calculations
VSA=VSW*VSH;
VSMIDX=VSW/2;
//if(VSW&1) VSMIDX++; // Round up
VSMIDY=VSH/2;
//if(VSH&1) VSMIDY++; // Round up
VSW32=VSW*32;
VSH32=VSH*32;
VSA32=VSW32*VSH32;


#ifdef DATAFILES
datafile_warning=0;
checkdatafiles();
if(datafile_warning)
	{
    ilog_printf("Press a key to continue..\n");
#if defined(_WIN32) || defined(__DJGPP__)
	getch();
#else
	read(0,&datafile_warning,1); // blocking keypress (read byte from stdin)
#endif
    }
#endif

// Check the mapfile is there before we commit ourselves

if((unsigned char)mapname[0] == 0xff)
	ithe_safepanic("No map file was specified.","There should be a -mapfile parameter in one of the INI files.");

#ifdef DATAFILES
ilog_printf("Mount Datafiles:\n");
mountdatafiles();
#endif

init_media(1); // Enter graphics mode, for game

ilog_printf("Console started\n");
ilog_printf("\n");


if(!fileexists("resource.txt"))
    if(fileexists("main.txt"))
        ithe_panic("Resource file not found..","Please rename your main.txt file to resource.txt");
    else
        ithe_panic("Resource file not found..","File resource.txt is missing!");

ilog_quiet("VRMs\n");

Init_VRMs();                     // Pre-initialise VRMs

ilog_printf("Load Resources\n");
LoadResources("resource.txt");

Load_Soundbar();

ilog_printf("Load Sound and Music\n");
S_Load();

ilog_printf("VRM fixups\n");
Init_Funcs();                   // Assign the CHlist.func to appropriate VRM

ilog_printf("Starting PEVM microkernel\n");
InitVM();

ilog_printf("Init Random Number Generator\n");
srand(time(0));                 // Initialise rng

irecon_term();
ilog_quiet("Loading backing panel\n");

LoadBacking(swapscreen,"backings/panel.pcx");
Show(); // Update swapscreen
irecon_init(conx,cony,conwid,conlen);

ilog_printf("Load Map\n");
LoadMap();                       // Load in the game map

read_config();

ilog_printf("Start Game\n");
startgame();                     // Do things

write_config();

ilog_printf("Stop Game\n");

term_media();

#ifdef SEER_PROFILE
int ctr;
for(ctr=0;ctr<COtot;ctr++)
	ilog_quiet("%d   %s\n",COlist[ctr].count,COlist[ctr].name);
#endif

if(buggy)
    {
    printf("There were bugs in the VRM code.\n");
    printf("Check 'bootlog.txt' and search for 'BUG:'\n");
    }

return 0;			// Return a value to keep compiler happy
}
END_OF_MAIN();

/*
 *      SoundSettings - Set the sound and music levels
 */

void SoundSettings()
{
static int pos=0;
int key;
int startx,starty,sl1,sl2;

startx=32;
starty=64;

   do  {

       // Draw backing image

       draw_sprite(swapscreen,s_soundbar,startx,starty);
//		s_soundbar.block_put_sprite(startx,starty,swapscreen);

       // Draw sliders

       sl1 = startx + 32 +(mu_volume>>1);
       sl2 = startx + 32 +(sf_volume>>1);

       if(pos == 0)
              {
				draw_sprite(swapscreen,s_slider1,sl1,starty+16);
				draw_sprite(swapscreen,s_slider2,sl2,starty+40);
//              s_slider1.cel_put_sprite(sl1,starty+16,swapscreen);
//              s_slider2.cel_put_sprite(sl2,starty+40,swapscreen);
              }
       else
              {
				draw_sprite(swapscreen,s_slider2,sl1,starty+16);
				draw_sprite(swapscreen,s_slider1,sl2,starty+40);
//              s_slider2.cel_put_sprite(sl1,starty+16,swapscreen);
//              s_slider1.cel_put_sprite(sl2,starty+40,swapscreen);
              }

       rest(2);
//       itg_update(swapscreen);
		Show();
       key=GetKey();


       // Choose slider

       if(key == KEY_UP)
             pos = 0;
       if(key == KEY_DOWN)
             pos = 1;

       // Move appropriate slider

       if(key == KEY_LEFT)
               {
               if(pos == 0)
                      mu_volume-=2;
               else
                      sf_volume-=2;
               }

       if(key == KEY_RIGHT)
               {
               if(pos == 0)
                      mu_volume+=2;
               else
                      sf_volume+=2;
               }

       if(mu_volume<0)
            mu_volume=0;
       if(mu_volume>255)
            mu_volume=255;
       if(sf_volume<0)
            sf_volume=0;
       if(sf_volume>255)
            sf_volume=255;

		S_MusicVolume(mu_volume);
		S_SoundVolume(sf_volume);
		S_PollMusic();

       } while(key != KEY_ESC);
}


/*
 *      read_config - get previous volume levels from file
 */

void read_config()
{
char string[128];
FILE *config;

strcpy(string,"game.cfg");
config = fopen(string,"rb");
if(!config)
    return;

mu_volume = getw(config);
sf_volume = getw(config);

ilog_quiet("Set volume %d,%d\n",mu_volume,sf_volume);
S_MusicVolume(mu_volume);    // Update sound engine
S_SoundVolume(sf_volume);

fclose(config);
}

/*
 *      write_config - write current volume levels to file
 */

void write_config()
{
char string[128];
FILE *config;

strcpy(string,"game.cfg");
config = fopen(string,"wb");
if(!config)
    return;

putw(mu_volume,config);
putw(sf_volume,config);

fclose(config);
}


void RFS_getescape()
{
while((readkey() >> 8) != KEY_ESC);
}


/*
 *      LoadBacking - load a picture into a pre-existing bitmap
 */

void LoadBacking(BITMAP *img,char *fname)
{
char filename[1024];
BITMAP *x;

if(!loadfile(fname,filename))
    {
    Bug("Cannot open image file '%s'\n",fname);
    return;
    }

x = iload_bitmap(filename);
if(!x)
    {
    Bug("Cannot load image file '%s': unsupported format?\n",filename);
    return;
    }
    
draw_sprite(img,x,0,0);
destroy_bitmap(x);
}


void Load_Soundbar()
{
char filename[1024];

if(!loadfile("soundbar.pcx",filename))
	ithe_panic("Could not find soundbar image",NULL);
s_soundbar = iload_bitmap(filename);
if(!s_soundbar)
	ithe_panic("Could not load soundbar image",filename);

if(!loadfile("slider1.pcx",filename))
	ithe_panic("Could not find slider1 image",NULL);
s_slider1 = iload_bitmap(filename);
if(!s_slider1)
	ithe_panic("Could not load slider1 image",filename);

if(!loadfile("slider2.pcx",filename))
	ithe_panic("Could not find slider2 image",NULL);
s_slider2 = iload_bitmap(filename);
if(!s_slider2)
	ithe_panic("Could not load slider2 image",filename);
}
