//
//      IRE linked list management routines
//

#include <stdlib.h>
#ifndef _WIN32
#include <string.h>
#endif

#include "ithelib.h"
#include "core.hpp"
#include "gamedata.h"
#include "map.hpp"
#include "script.hpp"
#include "linklist.hpp"
#include "object.hpp"
#include "console.h"
#include "init.h"
#include "oscli.h"

// defines

#define ML_CACHE
//#define CACHE_DOUBLECHECK

//#define LOG_MLADD
//#define CHECK_MASTERLIST
//#define PROFILE_LLADD

// variables

static char stemp[128];       // Used for string operations
extern char in_editor;        // Which program we are in?
extern char AL_dirty;         // Has the ActiveList changed?
extern "C" char *lastvrmcall;	  // VRM object was detected in

// functions

OBJECT *OB_Alloc();           // Allocate an Object
void OB_Free(OBJECT *ptr);    // Free an Object

void OB_SetDir(OBJECT *objsel,int dir);     // set direction
void OB_SetSeq(OBJECT *objsel,char *name);  // change appearance
void OB_Init(OBJECT *objsel,char *name);  // (re)set an object

int LLcache_register(void *ptr);
void *LLcache_get(void *ptr);
void LLcache_update(void *ptr, void *newptr);
//int LLcache_invalidate(void *ptr, void *newptr);
int OB_ListCheck(OBJLIST *l,OBJLIST *o);
int LLcache_ALC();

int OB_Fore(OBJECT *objsel);   // Move object forwards in the list
int OB_Back(OBJECT *objsel);   // Move object back in the list

void FreePockets(OBJECT *obj);

extern OBJECT *GetRawSolidObject(int x,int y);
extern TILE *GetTile(int x,int y);
extern void NPC_ReadWipe(OBJECT *o);

extern char fullrestore;     // Be less thorough if we're only reloading
extern void Call_VRM(int i);
extern OBJECT *current_object;
extern OBJECT *victim;
OBJLIST *MasterList=NULL;

#define LLCACHE_MAX 16

static void *llcache[LLCACHE_MAX][2];

#define NPC_ReadWipe(x) ;

// code

/*
 *    OB_Check()  -  Verify the object pointer is valid
 */

int OB_Check(OBJECT *o)
{
OBJLIST *temp;
int ctr;

if(!o)
	return 0;	// NULL

for(temp=MasterList;temp;temp=temp->next)
	if(temp->ptr == o)
		return 1;

for(ctr=0;ctr<CHtot;ctr++)
	if(o == &CHlist[ctr])
		{
		Bug("%x refers to a Template\n",o);
		return 1;
		}

if(o == syspocket)	// This is independently created but still valid
	return 1;

Bug("FOREIGN OBJECT 0x%x DETECTED\n",o);
if(lastvrmcall)
	Bug("Was passed as a parameter from function %s\n",lastvrmcall);

ctr=0;
for(temp=ActiveList;temp;temp=temp->next)
	if(temp->ptr == o)
		ctr++;

if(ctr)
	Bug("PRESENT IN ACTIVELIST (%d times)\n",ctr);
else
	Bug("Not present in ActiveList\n");

if(probeInvalidObject)
	{
	Bug("Probing object:\n");
	Bug("o->name = %x:\n",o->name);
	Bug("o->name = %s:\n",o->name);
	Bug("o->personalname = %x:\n",o->personalname);
	Bug("o->personalname = %s:\n",o->personalname);
	}

return 0;
}


int OB_ListCheck(OBJLIST *l,OBJLIST *o)
{
OBJLIST *temp;
int ctr;

if(!o)
	return 0;	// NULL

for(temp=l;temp;temp=temp->next)
	if(temp == o)
		return 1;

ilog_printf("Il Problemo: dumping to log\n");

for(temp=l;temp;temp=temp->next)
	ilog_quiet("%x,%x\n",temp,temp->ptr);

return 0;
}


/*
 *    OB_Alloc()  -  Allocate a new Object in the linked list
 */

OBJECT *OB_Alloc()
{
OBJECT *temp;

temp=(struct OBJECT *)M_get(1,sizeof(OBJECT));
temp->next=NULL;

LL_Add(&curmap->object,temp);
ML_Add(&MasterList,temp);


temp->slen=0;
temp->sdir=0;
temp->sptr=0;
temp->tag=0;
temp->target = NULL;
temp->activity = -1;
temp->flags.on =1;

// Allocate sublayers

temp->personalname=(char *)M_get(32,sizeof(char));
temp->funcs = (FUNCS *)M_get(1,sizeof(FUNCS));
OB_Funcs(temp);

temp->stats = (STATS *)M_get(1,sizeof(STATS));
temp->stats->npctalk=NULL;
temp->maxstats = (STATS *)M_get(1,sizeof(STATS));
temp->wield = (OBJECT **)M_get(W_SIZE,sizeof(struct OBJECT *));
temp->schedule = (SCHEDULE *)M_get(24,sizeof(SCHEDULE));
temp->user = (USEDATA *)M_get(1,sizeof(USEDATA));

for(int ctr=0;ctr<24;ctr++)
	temp->schedule[ctr].hour=-1;

// Return the address of the new entry
return temp;
}

/*
 *    OB_Funcs - Set up the string pointers to point to the strings
 */

void OB_Funcs(OBJECT *temp)
{
temp->funcs->suse = &temp->funcs->use[0];
temp->funcs->stalk = &temp->funcs->talk[0];
temp->funcs->skill = &temp->funcs->kill[0];
temp->funcs->slook = &temp->funcs->look[0];
temp->funcs->sstand = &temp->funcs->stand[0];
temp->funcs->shurt = &temp->funcs->hurt[0];
temp->funcs->sstand = &temp->funcs->stand[0];
temp->funcs->sinit = &temp->funcs->init[0];
temp->funcs->sresurrect = &temp->funcs->resurrect[0];
temp->funcs->swield = &temp->funcs->wield[0];
}

/*
 *    OB_Free()  -  Destroy the specified entry in the list and relink it
 *                  May take a while
 */

void OB_Free(OBJECT *ptr)
{
// We cannot free the anchor (simplifies the logic)

if(ptr == curmap->object)
	ithe_panic("OB_Free: Attempted to free OB_anchor",NULL);

// We certainly cannot do this either

if(!ptr)
	ithe_panic("OB_Free: Attempted to free NULL pointer:",NULL);

// First remove the object
LL_Remove(&curmap->object,ptr);

// Then remove any objects inside it
FreePockets(ptr);

// Then Deallocate it
LL_Kill(ptr);
}


/*
 *    LL_Add - Generalised linked-list add
 */

void LL_Add(OBJECT **base,OBJECT *object)
{
OBJECT *temp,*end;
int ctr;

if(!base)
	ithe_panic("LL_Add: Attempted to add object to NULL",object->name);

if(!object)
	ithe_panic("LL_Add: Attempted to add NULL object to list",NULL);

// Terminate the object, since it will be the end of the list.
object->next = NULL;
temp =*base;

// No objects in the list?

if(!temp)
	{
	*base = object;
	return;
	}

// Only one?

if(!temp->next)
	{
	temp->next = object;
	return;
	}

// Handle for any other number of objects.


#ifdef PROFILE_LLADD
// Profiling
static int peak=0;
for(ctr=0;temp->next;temp=temp->next) ctr++;
if(++ctr>peak)
	{
	ilog_quiet("LL_Add: %d in list\n",ctr);
	peak=ctr;
	}
#else
// Release
for(;temp->next;temp=temp->next);
temp->next = object;
#endif
temp->next = object;
}


/*
 *    LL_Remove - Generalised linked-list free
 */

void LL_Remove(OBJECT **base,OBJECT *object)
{
OBJECT *temp;
char found;
int ctr;

if(!base)
    ithe_panic("LL_Remove: Attempted to remove object from NULL",object->name);

if(!object)
    ithe_panic("LL_Remove: Attempted to remove NULL object from list",NULL);

temp=*base;

// Look for it, to make sure it is there.

found=0;
for(;temp;temp=temp->next)
    if(temp == object)
        found = 1;

if(!found)
    ithe_panic("LL_Remove: Could not find object in the list",object->name);

temp=*base;

// Is it the first object?

if(temp == object)
    {
    *base = object->next;
    object->next = NULL;
    return;
    }

// Is it the second object?

if(temp->next == object)
    {
    temp->next = object->next;
    object->next = NULL;
    return;
    }

// Any other number

for(;temp->next;temp=temp->next)
    if(temp->next == object)
        {
        temp->next = object->next;
        object->next = NULL;
        return;
        }

ithe_panic("LL_Remove: Whoops!  You really messed that up Joe!",NULL);
}


/*
 *  LL_Kill - Release the memory held by an object and dependent structs.
 *            This is a low-level operation but has error checking.
 */

void LL_Kill(OBJECT *obj)
{
SCHEDULE *s;

if(!obj)
    ithe_panic("LL_Kill: Object is NULL",NULL);
if(obj->next)
    ithe_panic("LL_Kill: Object is part of a linked-list",obj->name);
if(obj->pocket)
    ithe_panic("LL_Kill: Object has items in pocket",obj->name);

NPC_ReadWipe(obj);      // Must be before STATS are removed

if(obj->stats)
	{
	M_free(obj->stats);
	obj->stats = NULL;
	}

if(obj->maxstats)
	{
	M_free(obj->maxstats);
	obj->maxstats = NULL;
	}

if(obj->funcs)
	{
	M_free(obj->funcs);
	obj->funcs = NULL;
	}

if(obj->wield)
	{
	M_free(obj->wield);
	obj->wield = NULL;
	}

if(obj->schedule)
	{
	M_free(obj->schedule);
	obj->schedule=NULL;
	}

if(obj->user)
	{
	M_free(obj->user);
	obj->user=NULL;
	}

// Check all dependencies

if(obj->activity != -1)
    ML_Del(&ActiveList,obj);

CheckDepend(obj);
ML_Del(&MasterList,obj);

if(current_object == obj)
    current_object = NULL;
if(victim== obj)
    victim = NULL;

#ifdef WITH_PREJUDICE
memset(obj,0xff,sizeof(OBJECT)); // F--- it for good
#endif
M_free(obj);
}

/*
 *      ML_Add - Add an object pointer to an object chain
 */

void ML_Add(OBJLIST **m,OBJECT *obj)
{
OBJLIST *t,*end,*newptr;
int ctr,found;

if(!obj)
    ithe_panic("ML_Add: Tried to add NULL to list\n",NULL);

#ifdef LOG_MLADD
if(m == &MasterList)
    ilog_quiet("Adding %x to Masterlist\n",obj);
if(m == &ActiveList)
    ilog_quiet("Adding %x to Activelist\n",obj);
#endif

// First object.

if(!*m)
    {
    *m = (OBJLIST *)M_get(1,sizeof(OBJLIST));
    (*m)->ptr = obj;
    (*m)->next = NULL;
#ifdef ML_CACHE
	LLcache_update(m,*m);
#endif
    AL_dirty = 1;
    return;
    }

// Don't add if it's already there

#ifdef CHECK_MASTERLIST
for(t = *m;t;t=t->next)
	if(t->ptr == obj)
		{
		Bug("Object %s already in list\n",obj->name);
		return;
		}
#endif

// More than one object in the list.. create it.

newptr = (OBJLIST *)M_get(1,sizeof(OBJLIST));
newptr->ptr = obj;
newptr->next = NULL;

// Check the cache.

#ifdef ML_CACHE
end=(OBJLIST *)LLcache_get(m);
if(end)
	{
#ifdef CACHE_DOUBLECHECK
	if(!OB_ListCheck(*m,end))
		{
			Bug("Cache Invalid (%x) %x:\n",end,end->ptr);
		if(m == &MasterList)
			Bug("Cache Invalid for MasterList\n");
		if(m == &ActiveList)
			Bug("Cache Invalid in ActiveList\n");
		LLcache_update(m,NULL);
		goto unlawful;
		}
#endif

	if(end->next)
		{
		Bug("Cache Expired\n");
		LLcache_update(m,NULL);
		}
	else
		{
		end->next = newptr;
		LLcache_update(m,newptr);
		AL_dirty = 1;
		return;
		}
	}
#endif

unlawful:

// Add the object to the list the slow way.

for(t = *m;t->next;t=t->next);
t->next = newptr;
#ifdef ML_CACHE
LLcache_update(m,newptr);
#endif
AL_dirty = 1;

return;
}


/*
 *    ML_Del - Free an object from an object chain
 */

void ML_Del(OBJLIST **m,OBJECT *object)
{
OBJLIST *temp,*t2;
char found;
int ctr;

if(!object)
    ithe_panic("ML_Del: Attempted to remove NULL object from list",NULL);

// If the list is empty, abort
if(!*m)
	return;

#ifdef LOG_MLADD
if(m == &MasterList)
    ilog_quiet("De-registering %x from Masterlist\n",object);
else
    ilog_quiet("De-registering %x from Activelist\n",object);
#endif

#ifdef ML_CACHE
LLcache_update(m,NULL);
#endif

temp=*m;

// Check
if(!temp)
	return;

// Is it the first object?

if(temp->ptr == object)
	{
	*m = temp->next;
	temp->next = NULL;
	M_free(temp);
	AL_dirty = 1;
#ifdef ML_CACHE
	LLcache_update(m,NULL);
#endif
	return;
	}

// Check
if(!temp->next)
	return;

// Is it the second object?

if(temp->next->ptr == object)
	{
	t2=temp->next;
	temp->next = temp->next->next;
	t2->next = NULL;
	M_free(t2);
	AL_dirty = 1;
#ifdef ML_CACHE
	LLcache_update(m,NULL);
#endif
	return;
	}

// Any other number

for(;temp->next;temp=temp->next)
	if(temp->next->ptr == object)
		{
		t2 = temp->next;
		temp->next = temp->next->next;
		t2->next = NULL;
		M_free(t2);
		AL_dirty = 1;
#ifdef ML_CACHE
		LLcache_update(m,NULL);
#endif
		return;
		}

return;
//ithe_panic("ML_Del: Oh no!","The chain has become corrupted");
}

/*
 *    ML_Query - Count instances of an object type by name
 */

int ML_Query(OBJLIST **m,char *name)
{
OBJLIST *temp,*t2;
int ctr,count;

if(!name)
	{
	Bug("ML_Query for NULL");
	return 0;
	}

temp=*m;

count=0;
for(;temp;temp=temp->next)
	if(temp->ptr)
		if(!stricmp(temp->ptr->name,name))
			count++;
return count;
}


/*
 *    ML_InList - Is an object in this object chain?
 */

int ML_InList(OBJLIST **m,OBJECT *o)
{
OBJLIST *temp,*t2;
int ctr,count;

if(!o)
	{
	Bug("ML_Check for NULL");
	return 0;
	}

for(temp=*m;temp;temp=temp->next)
	if(temp->ptr == o)
		return 1;
return 0;
}

void AL_Add(OBJLIST **a,OBJECT *o)
{
if(o->activity == -1)
	return;
//ilog_quiet("Go Active (%s)\n",o->name);
ML_Add(a,o);
AL_dirty = 1;
}




/* ========================== Non-core functionality ====================== */

/*
 *    OB_Init()  -  Initialize or re-initialise an Object.
 */

void OB_Init(OBJECT *objsel,char *name)
{
int ctr;
int seq;

ctr = getnum4char(name);                // Find the character's name
if(ctr == -1)
       ithe_panic("OB_Init: Could not find object in Section: CHARACTERS!",name);
objsel->name = CHlist[ctr].name;        // Update object's name from static

// Now the object will take on the properties of the new character

// Copy the direction cache
memcpy(objsel->dir,CHlist[ctr].dir,sizeof(CHlist[ctr].dir));

// Set up the stats and function systems
memcpy(objsel->stats,CHlist[ctr].stats,sizeof(STATS));
memcpy(objsel->maxstats,CHlist[ctr].maxstats,sizeof(STATS));
memcpy(objsel->funcs,CHlist[ctr].funcs,sizeof(FUNCS));
OB_Funcs(objsel);

// If critical maximum statistics are zero, improvise

if(objsel->maxstats->hp == 0)
    objsel->maxstats->hp = objsel->stats->hp;
if(objsel->maxstats->str == 0)
    objsel->maxstats->str = objsel->stats->str;
if(objsel->maxstats->dex == 0)
    objsel->maxstats->dex = objsel->stats->dex;
if(objsel->maxstats->intel == 0)
    objsel->maxstats->intel = objsel->stats->intel;

// Get the sequence index
objsel->curdir&=0xff;
seq = objsel->dir[objsel->curdir];

// Set up the animation stats
objsel->sdir = 1;
objsel->slen = SQlist[seq].frames;
objsel->sptr = 0;

// Set up the appearance
objsel->form = &SQlist[seq];
objsel->desc = CHlist[ctr].desc;                // Long text description
objsel->shortdesc = CHlist[ctr].shortdesc;      // Short text description

if(ML_InList(&ActiveList,objsel))
	ML_Del(&ActiveList,objsel);

// Set up the physics
if(!fullrestore)
    {
    objsel->flags = CHlist[ctr].flags;
    objsel->stats->npcflags = CHlist[ctr].stats->npcflags;
    objsel->light = CHlist[ctr].light;
	ActivityNum(objsel,CHlist[ctr].activity,NULL); // Set default action
    NPC_ReadWipe(objsel);
    }

//ML_Del(&ActiveList,objsel);     // Make sure its gone
//AL_Add(&ActiveList,objsel);     // Add it if necessary

objsel->flags.on =1;

    objsel->vblock[BLK_X]= CHlist[ctr].vblock[BLK_X];
    objsel->vblock[BLK_Y]= CHlist[ctr].vblock[BLK_Y];
    objsel->vblock[BLK_W]= CHlist[ctr].vblock[BLK_W];
    objsel->vblock[BLK_H]= CHlist[ctr].vblock[BLK_H];
    objsel->hblock[BLK_X]= CHlist[ctr].hblock[BLK_X];
    objsel->hblock[BLK_Y]= CHlist[ctr].hblock[BLK_Y];
    objsel->hblock[BLK_W]= CHlist[ctr].hblock[BLK_W];
    objsel->hblock[BLK_H]= CHlist[ctr].hblock[BLK_H];

    // Active-Area was not being dealt with correctly

    objsel->varea[BLK_X]= CHlist[ctr].varea[BLK_X];
    objsel->varea[BLK_Y]= CHlist[ctr].varea[BLK_Y];
    objsel->varea[BLK_W]= CHlist[ctr].varea[BLK_W];
    objsel->varea[BLK_H]= CHlist[ctr].varea[BLK_H];
    objsel->harea[BLK_X]= CHlist[ctr].harea[BLK_X];
    objsel->harea[BLK_Y]= CHlist[ctr].harea[BLK_Y];
    objsel->harea[BLK_W]= CHlist[ctr].harea[BLK_W];
    objsel->harea[BLK_H]= CHlist[ctr].harea[BLK_H];

// Set up the size

objsel->flags.large=0;  // Assume it's only 1 square

objsel->w    = objsel->form->seq[0]->w;
objsel->h    = objsel->form->seq[0]->h;

// Set up the size in tiles
objsel->mw = objsel->w>>5;
if(objsel->w & 0xf || !objsel->mw)
     objsel->mw++;

objsel->mh = objsel->h>>5;
if(objsel->h & 0xf || !objsel->mh)
     objsel->mh++;

if(objsel->mh>1 || objsel->mw>1)        // If it's bigger than 1 square..
     objsel->flags.large=1;

// If an Active Area has not been set aside, assume it covers the
// whole area

if(objsel->varea[BLK_W]==0)
    {
    objsel->varea[BLK_W]=objsel->mw;
    objsel->varea[BLK_X]=0;
    }
if(objsel->varea[BLK_H]==0)
    {
    objsel->varea[BLK_H]=objsel->mh;
    objsel->varea[BLK_Y]=0;
    }
if(objsel->harea[BLK_W]==0)
    {
    objsel->harea[BLK_W]=objsel->mw;
    objsel->harea[BLK_X]=0;
    }
if(objsel->harea[BLK_H]==0)
    {
    objsel->harea[BLK_H]=objsel->mh;
    objsel->harea[BLK_Y]=0;
    }

// If a special solid area has not been set aside, assume it covers the
// whole area

if(objsel->vblock[BLK_W]==0)
    {
    objsel->vblock[BLK_W]=objsel->mw;
    objsel->vblock[BLK_X]=0;
    }
if(objsel->vblock[BLK_H]==0)
    {
    objsel->vblock[BLK_H]=objsel->mh;
    objsel->vblock[BLK_Y]=0;
    }
if(objsel->hblock[BLK_W]==0)
    {
    objsel->hblock[BLK_W]=objsel->mw;
    objsel->hblock[BLK_X]=0;
    }
if(objsel->hblock[BLK_H]==0)
    {
    objsel->hblock[BLK_H]=objsel->mh;
    objsel->hblock[BLK_Y]=0;
    }
}

/*
 *    OB_SetDir()  -  Change the direction that a character is facing
 *                    objsel is the pointer in the list, dir the direction
 *                    it is facing.  See CHAR_L CHAR_R etc..
 */

void OB_SetDir(OBJECT *objsel,int dir)
{
int seq,cx,cy,w,h,x,y;
OBJECT *temp;
TILE *tile;
int olddir;

if(dir>3)
    {
    Bug("Attempt to set %s to face invalid direction %d!\n",objsel->name,dir);
    return;
    }

// Get the sequence index
seq = objsel->dir[dir];
olddir=objsel->curdir;  // In case we need to revert
objsel->curdir = dir;

#if 0
// If the object is sitting on a chair, turn the chair around

temp = GetRawObjectBase(objsel->x,objsel->y);
if(temp != objsel)
    for(;temp;temp=temp->next)
        if(temp->next == objsel)
            if(temp->flags.chair)
                OB_SetDir(temp,dir);
#endif

// Find out if we need to continue

if(objsel->form == &SQlist[seq])   // abort if we're already the candidate
   return;

if(!in_editor)                  // Not in the editor!
  if(objsel->flags.large)       // If it's a large object, find out if the
    {                           // new shape would actually fit
    // Find the polarity of the new object
    if(dir < CHAR_L)  //(U,D,L,R)
        {
        w = objsel->vblock[BLK_W];
        h = objsel->vblock[BLK_H];
        x = objsel->vblock[BLK_X];
        y = objsel->vblock[BLK_Y];
        }
    else
        {
        w = objsel->hblock[BLK_W];
        h = objsel->hblock[BLK_H];
        x = objsel->hblock[BLK_X];
        y = objsel->hblock[BLK_Y];
        }

    // Now make sure it doesn't hit something else
    // If it hits itself that's OK
    for(cx=0;cx<w;cx++)
        for(cy=0;cy<h;cy++)
            {
            temp = GetRawSolidObject(cx+x+objsel->x,cy+y+objsel->y);
            if(temp && temp!=objsel)
                {
                objsel->curdir = olddir; // Obstruction
                return;
                }
            tile=GetTile(cx+x+objsel->x,cy+y+objsel->y);
            if(tile->flags.solid)
                {
                objsel->curdir = olddir; // Obstruction
                return;
                }
            }
    }

// Set up new direction up

objsel->form = &SQlist[seq];
objsel->w    = objsel->form->seq[0]->w;
objsel->h    = objsel->form->seq[0]->h;

// Set up the animation stats
objsel->sdir = 1;
objsel->slen = SQlist[seq].frames;
if(objsel->sptr > objsel->slen)
    objsel->sptr = 0;

}

/*
 *    OB_SetSeq()-  Temporarily set the object's display frame to something
 *                  unusual.  The name is the SEQUENCE we select
 */

void OB_SetSeq(OBJECT *objsel,char *seqname)
{
int seq;

if(!objsel)
    {
    Bug("SetSequence passed NULL object\n");
    return;
    }

// Get the sequence index
seq = getnum4sequence(seqname);
if(seq == -1)
    {
    Bug("Could not find sequence %s in SECTION:SEQUENCES\n",seqname);
    return;
    }

objsel->form = &SQlist[seq];
objsel->w    = objsel->form->seq[0]->w;
objsel->h    = objsel->form->seq[0]->h;

// Set up the animation stats
objsel->sdir = 1;
objsel->slen = SQlist[seq].frames;
objsel->sptr = 0;


}

/*
 *    OB_Back()  -  Move an object backwards in the list
 */

int OB_Back(OBJECT *objsel)
{
OBJECT *temp;

if(!curmap->object->next)
	return 1;               // Not enough sprites to do this
if(!curmap->object->next->next)
	return 1;               // Not enough sprites to do this

if(!objsel)
	return 2;               // No sprite selected

if(objsel==curmap->object)
	return 3;               // Not permitted

// Look for a link to the selected object two objects away.

for(OBJECT *tt=curmap->object;tt->next->next;tt=tt->next)
	if(tt->next->next == objsel)
		{
		temp=objsel->next;              // Preserve original link
		objsel->next=tt->next;          // link into next part of list
		tt->next=objsel;                // link into the list itself
		objsel->next->next=temp;        // Restore old link
		return 0;                       // Exit
		}
return 0;
}

/*
 *    OB_Fore()  -  Move an object forward in the list
 */

int OB_Fore(OBJECT *objsel)
{
OBJECT *temp;

if(!curmap->object->next)
	return 1;               // Not enough sprites to do this
if(!curmap->object->next->next)
	return 1;               // Not enough sprites to do this

if(!objsel)
	return 2;               // No sprite selected

if(objsel==curmap->object)
	return 3;               // Not permitted

// Look for a link to the selected object two objects away.

for(OBJECT *tt=curmap->object;tt->next->next;tt=tt->next)
	if(tt->next == objsel)
		{
		temp=tt->next->next->next;      // Preserve original link
		tt->next=tt->next->next;        // Change first link
		tt->next->next=objsel;          // Change second link
		objsel->next=temp;              // Change third link
		return 0;
		}
return 0;
}


/*
 * OB_Previous - Pointer to previous object
 */

OBJECT *OB_Previous(OBJECT *a)
{
OBJECT *temp;
for(temp=curmap->object;temp->next;temp=temp->next)
    if(temp->next==a)
        return(temp);
ithe_panic("OB_Previous: Ah shit","You really messed that up Joe!");
return NULL;
}


void FreePockets(OBJECT *obj)
{
OBJECT *temp,*next;

// Sanity check
if(!obj)
    ithe_panic("FreePockets: NULL pointer encountered",NULL);

// Silently ignore object without pocket
if(!obj->pocket)
    return;

temp = obj->pocket;
while(temp)
    {
    next = temp->next;
    LL_Remove(&obj->pocket,temp);       // Unlink the object
    if(temp->pocket)                    // If it has objects inside
        FreePockets(temp);             // Recurse inside it
    LL_Kill(temp);                      // Free the memory
    temp = next;                        // Go to next object
    };
}

int LLcache_register(void *ptr)
{
int ctr;

for(ctr=0;ctr<LLCACHE_MAX;ctr++)
	if(llcache[ctr][0] == NULL)
		{
		llcache[ctr][0]=ptr;
		llcache[ctr][1]=NULL;
		return ctr;
		}
Bug("LLcache_register: Failed - Too many caches.\n");
return -1;
}

void *LLcache_get(void *ptr)
{
int ctr;

for(ctr=0;ctr<LLCACHE_MAX;ctr++)
	if(llcache[ctr][0] == ptr)
		return llcache[ctr][1];
return NULL;
}

void LLcache_update(void *ptr, void *newptr)
{
int ctr;

for(ctr=0;ctr<LLCACHE_MAX;ctr++)
	if(llcache[ctr][0] == ptr)
		{
//		ilog_printf("updating %d (%x)\n",ctr,newptr);
		llcache[ctr][1]=newptr;
		return;
		}
Bug("LLcache_upd: couldn't update %x : no such cache (%x)\n",newptr,ptr);
}

#if 0
int LLcache_invalidate(void *ptr, void *newptr)
{
int ctr,diag=0;

for(ctr=0;ctr<LLCACHE_MAX;ctr++)
	if(llcache[ctr][0] == ptr)
		{
		diag=1;
		if(llcache[ctr][1] == newptr)
			{
			ilog_printf("invalidating %d (%x)\n",ctr,newptr);
			llcache[ctr][1]=NULL;
			return 1;
			}
		}

Bug("LLcache_inv: couldn't invalidate %x\n",newptr);
if(!diag)
	Bug("Because I did not find the cache\n");
return 0;
}
#endif

int LLcache_ALC(char *fname, int line)
{
if(!OB_ListCheck(ActiveList,(OBJLIST *)llcache[1][1]))
	{
	ilog_printf("LL consistency check failed at: %s:%d\n",fname,line);
	return 0;
	}
return 1;
}
