//
//	Resource loaders for various data types
//

#include <stdio.h>
#include "ithelib.h"
#include "media.h"
#include <allegro.h>
#ifdef _WIN32 // What a crock
#include "jpeg/jpeglib.h"
#else
#include <jpeglib.h>
#endif
#include "core.hpp"

// Defines

#define STRUCT_CHECK // Strict checking for structure save/load


#ifdef STRUCT_CHECK
	#define S_START() char msg[128];int checksum=0;
	#define S_END(x) if(checksum != sizeof(x)) {sprintf(msg,"%s processed %d bytes instead of %d",#x,checksum,sizeof(x));ithe_panic("structure mismatch",msg);}
	#define S_GETWI(x) x=igetl_i(f);checksum+=4;
	#define S_GETWM(x) x=igetl_m(f);checksum+=4;
	#define S_GETSHI(x) x=igetsh_i(f);checksum+=2;
	#define S_GETSHM(x) x=igetsh_m(f);checksum+=2;
	#define S_GETB(x) x=igetb(f);checksum++;
	#define S_GETPTR(x) igetl_i(f);checksum+=4; // skip 4 bytes regardless
	#define S_READ(x,y) iread(x,y,f);checksum+=y;
	#define S_PUTWI(x) iputl_i(x,f);checksum+=4;
	#define S_PUTWM(x) iputl_m(x,f);checksum+=4;
	#define S_PUTSHI(x) iputsh_i(x,f);checksum+=2;
	#define S_PUTSHM(x) iputsh_m(x,f);checksum+=2;
	#define S_PUTB(x) iputb(x,f);checksum++;
	#define S_PUTPTR(x) iputl_i(NULL,f);checksum+=4; // 4 compatability bytes
	#define S_WRITE(x,y) iwrite(x,y,f);checksum+=y;
	#define S_HACK(x) checksum+=x; // Adjust for compatability
#else
	#define S_START()
	#define S_END(x)
	#define S_GETWI(x) x=igetl_i(f);
	#define S_GETWM(x) x=igetl_m(f);
	#define S_GETSHI(x) x=igetsh_i(f);
	#define S_GETSHM(x) x=igetsh_m(f);
	#define S_GETB(x) x=igetb(f);
	#define S_GETPTR(x) igetl_i(f); // skip 4 bytes regardless of true length
	#define S_READ(x,y) iread(x,y,f);
	#define S_PUTWI(x) iputl_i(x,f);
	#define S_PUTWM(x) iputl_m(x,f);
	#define S_PUTSHI(x) iputsh_i(x,f);
	#define S_PUTSHM(x) iputsh_m(x,f);
	#define S_PUTB(x) iputb(x,f);
	#define S_PUTPTR(x) iputl_i(NULL,f); // write 4 blank compatability bytes
	#define S_WRITE(x,y) iwrite(x,y,f);
	#define S_HACK(x)
#endif

// Variables

// Functions

static BITMAP *iload_cel(char *filename);
static BITMAP *iload_pcx(char *filename);
static BITMAP *iload_jpeg(char *filename);
static void UnPcx(char *in,char *out,int length);
static unsigned char *UnJPG(FILE *fp, int *w, int *h, int *bpp);
extern OBJECT *find_id(unsigned int id);

// Code

BITMAP *iload_bitmap(char *filename)
{
if(strstr(filename,".cel"))
	return iload_cel(filename);
if(strstr(filename,".pcx"))
	return iload_pcx(filename);
if(strstr(filename,".jpg"))
	return iload_jpeg(filename);
if(strstr(filename,".CEL"))
	return iload_cel(filename);
if(strstr(filename,".PCX"))
	return iload_pcx(filename);
if(strstr(filename,".JPG"))
	return iload_jpeg(filename);
return NULL;
}

/*
 *		load a CEL file using the IT-HE filesystem wrappers
 */

BITMAP *iload_cel(char *filename)
{
unsigned short w,h;                             // size of image
int len,ctr,x,y;
unsigned char *buf,*ptr;                        // conversion buffer
IFILE *fp;
BITMAP *img;
BITMAP *icb;
RGB p[256];

// Read the whole file

fp = iopen(filename);
// Good, we're still around

len=ifilelength(fp);

buf=(unsigned char *)M_get(1,len);              // Allocate room
iread(buf,len,fp);
iclose(fp);

// OK, let's go

if(buf[0] != 0x19 || buf[1] != 0x91)
    ithe_panic("CEL header invalid for image",filename);

w = buf[2]+(buf[3]<<8);
h = buf[4]+(buf[5]<<8);
len = w*h;

ptr = &buf[32];

// Build the conversion palette

for(ctr=0;ctr<256;ctr++)
    {
    p[ctr].r = *ptr++;
    p[ctr].g = *ptr++;
    p[ctr].b = *ptr++;
    }

/*
// Set palette entry 0 to current transparency colour for >8bpp
p[0].r = ire_transparent_r;
p[0].g = ire_transparent_g;
p[0].b = ire_transparent_b;
*/

select_palette(p); // Use this for bpp conversion later on

// Allocate the sprite we will put the .CEL into

img = create_bitmap(w,h);
if(!img)
    ithe_panic("Create bitmap failed for image",filename);
clear_to_color(img,ire_transparent);

// Allocate line conversion buffer

icb = create_bitmap_ex(8,w,2);
if(!icb)
    ithe_panic("Create icb failed for image",filename);

// Convert the sprite line by line

ptr = &buf[800];
for(y=0;y<h;y++)
	{
	// Copy raw data to line buffer
	memcpy(icb->line[0],ptr,w);
	// Translate to native BPP in destination image
	blit(icb,img,0,0,0,y,w,1);
	// Enforce any transparency, and increment ptr accordingly
	for(x=0;x<w;x++)
		if(!*ptr++)
			putpixel(img,x,y,ire_transparent);
	}

// Free the temporary buffers

destroy_bitmap(icb);
M_free(buf);
unselect_palette();

// Convert to main palette if 8bpp

if(ire_bpp == 8)
	{
	ptr=img->line[0];
	for(y=0;y<len;y++)
		{
        *ptr=makecol(p[*ptr].r,p[*ptr].g,p[*ptr].b);
		ptr++;
		}
	}

return img;
}


/*
 *		load a PCX file using the IT-HE filesystem wrappers
 */

BITMAP *iload_pcx(char *filename)
{
BITMAP *spr;					// Output
IFILE *fp;						// VFS filepointer
unsigned char *buf;				// Decoding buffer
unsigned char *raw;				// Raw buffer
long fl;						// file length
int w,h,len;					// Various data items
int ptr=0;						// For PCX-24 debanding
int ctr,ctr2,mode;
RGB pal[256];
unsigned int colour;

fp = iopen(filename);
// Good, we're still around

fl=ifilelength(fp)-128;			// PCX file is HEADER:DATA:PAL
raw = M_get(1,fl);
iseek(fp,8,SEEK_SET);			// Skip to the width

w = igetsh_i(fp)+1; // make 1-based from 0-based
h = igetsh_i(fp)+1;

iseek(fp,65,SEEK_SET);			// Skip to the type byte
mode = igetc(fp);

if(mode != 1 && mode != 3)
	{
    sprintf((char *)raw,"%d planes in file %s",mode,filename);
    ithe_panic("Funny number of bitplanes in PCX file:",raw);
	}

// Get the length of the raw bitmap

spr = create_bitmap(w,h);
len = w*h;

// Allocate decompression buffer
buf = M_get(1,len*mode);

switch(mode)
    {
    // 8 bits per pixel

    case 1:
    iseek(fp,-768L,SEEK_END);			// Seek to palette data (end-768)
	for(ctr=0;ctr<256;ctr++)
		{
		pal[ctr].r = igetb(fp);
		pal[ctr].g = igetb(fp);
		pal[ctr].b = igetb(fp);
		}
    iseek(fp,128,SEEK_SET);				// Seek back to start of RLE data
	iread(raw,fl,fp);
    UnPcx(raw,buf,len);					// Undo the RLE coding

	ptr=0;
	for(ctr=0;ctr<h;ctr++)
		{
		for(ctr2=0;ctr2<w;ctr2++)
			{
			colour = makecol(pal[buf[ptr]].r,pal[buf[ptr]].g,pal[buf[ptr]].b);
			if(buf[ptr])
				putpixel(spr,ctr2,ctr,colour);
			else
				putpixel(spr,ctr2,ctr,ire_transparent);
			ptr++;
			}
		}

    break;                                 // Finished

    // 24 bits per pixel, true colour

    case 3:
    iseek(fp,128,SEEK_SET);				// Seek back to start of RLE data
	iread(raw,fl,fp);
    UnPcx(raw,buf,len*3);				// Undo the RLE coding

    // Because PCX-24 is so strange, the RGB levels are banded.
    // First we have a complete mono image of the red component
    // Then a complete mono image of the Green component
    // Finally a complete mono image of the Blue component

	// The code below re-combines the three bands.
    // The result is fed into makecol, to get the natively-packed word
    // for the current video mode

	ptr=0;
	for(ctr=0;ctr<h;ctr++)
		{
		for(ctr2=0;ctr2<w;ctr2++)
			{
			colour = makecol(buf[ptr],buf[ptr+w],buf[ptr+(w*2)]);
			putpixel(spr,ctr2,ctr,colour);
			ptr++;
			}
		ptr+=(w*2);
		}

    break;

    default:
    sprintf((char *)buf,"%d planes in file %s",mode,filename);
    ithe_panic("Funny number of bitplanes in PCX file:",buf);
    }

M_free(buf);
M_free(raw);
iclose(fp);							// Close the file
              
return spr;
}



/*
 * UnPCX - Decode the RLE data into RAW bitmap info
 */

void UnPcx(char *in,char *out,int length)
{
unsigned char ob;						// Output Byte
int ctr,ctr2;

for(ctr=0;ctr<length;ctr++)		// Decode it all
	{
	ob=*in++;
	if(ob > 0xbf)						// If it's a run of colour
		{
		ctr2=(ob & 0x3f);				// Get run length
		ob=*in++;					// Get run colour
		memset(out,ob,ctr2);			// Do the run
		out+=ctr2;						// Increment pointer
		ctr+=ctr2-1;					// Adjust RLE pointer
		}
    else
		*out++=ob;						// It was a single colour
    }
return;                                 // Finished
}

/* load_wav: Taken from Allegro
 *  Reads a RIFF WAV format sample file, returning a SAMPLE structure, 
 *  or NULL on error.
 */
SAMPLE *iload_wav(char *filename)
{
   IFILE *f;
   char buffer[25];
   int i;
   int length, len;
   int freq = 22050;
   int bits = 8;
   int channels = 1;
   signed short s;
   SAMPLE *spl = NULL;

	*allegro_errno=0;

   f = iopen(filename);
   if (!f)
      return NULL;

   iread(buffer, 12, f);          /* check RIFF header */
   if (memcmp(buffer, "RIFF", 4) || memcmp(buffer+8, "WAVE", 4))
      goto getout;

   while (!ieof(f)) {
      if (iread(buffer, 4, f) != 4)
	 break;

      length = igetl_i(f);          /* read chunk length */

      if (memcmp(buffer, "fmt ", 4) == 0) {
	 i = igetsh_i(f);            /* should be 1 for PCM data */
	 length -= 2;
	 if (i != 1) 
	    goto getout;

	 channels = igetsh_i(f);     /* mono or stereo data */
	 length -= 2;
	 if ((channels != 1) && (channels != 2))
	    goto getout;

	 freq = igetl_i(f);         /* sample frequency */
	 length -= 4;

	 igetl_i(f);                /* skip six bytes */
	 igetsh_i(f);
	 length -= 6;

	 bits = igetsh_i(f);         /* 8 or 16 bit data? */
	 length -= 2;
	 if ((bits != 8) && (bits != 16))
	    goto getout;
      }
      else if (memcmp(buffer, "data", 4) == 0) {
	 len = length / channels;

	 if (bits == 16)
	    len /= 2;

	 spl = create_sample(bits, ((channels == 2) ? TRUE : FALSE), freq, len);

	 if (spl) { 
	    if (bits == 8) {
	       iread(spl->data, length, f);
	    }
	    else {
	       for (i=0; i<len*channels; i++) {
		  s = igetsh_i(f);
		  ((signed short *)spl->data)[i] = s^0x8000;
	       }
	    }

	    length = 0;

/*
	    if (*allegro_errno) {
	       destroy_sample(spl);
		ilog_quiet("\n>%d>%s\n",*allegro_errno,allegro_error);
	       spl = NULL;
	    }
*/
	 }
      }

      while (length > 0) {             /* skip the remainder of the chunk */
	 if (igetc(f) == EOF)
	    break;

	 length--;
      }
   }

   getout:
   iclose(f);
   return spl;
}

/*
 *  load_JPG - As display_JPG but for sprites
 *
 */

BITMAP *iload_jpeg(char *filename)
{
char msg[256];
IFILE *fp;                                      // VFS file pointer
int w,h,bpp,x,y,ptr;
BITMAP *out;
unsigned char *buf;

fp = iopen(filename);

buf = UnJPG(fp->fp, &w, &h, &bpp);
if(!buf)
	ithe_panic("Jpeg: Could not allocate memory to decompress",filename);

iclose(fp);

out=create_bitmap(w,h);
if(!out)
	ithe_panic("Jpeg: Out of memory loading .JPG file :",filename);

// It seems we have a result.  Convert to native colour format

ptr=0;

switch(bpp)
	{
	case 8: // Greyscale
	for(y=0;y<h;y++)
		for(x=0;x<w;x++)
			{
			if(buf[ptr])
				putpixel(out,x,y,makecol(buf[ptr],buf[ptr],buf[ptr]));
			else
				putpixel(out,x,y,ire_transparent);
			ptr++;
			}
	break;

    case 24:
	case 32:
	for(y=0;y<h;y++)
		for(x=0;x<w;x++)
			{
			putpixel(out,x,y,makecol(buf[ptr+0],buf[ptr+1],buf[ptr+2]));
			ptr+=3;
			}
//			putpixel(out,x,y,makecol(buf[ptr++],buf[ptr++],buf[ptr++]));
	break;

    default:
	sprintf(msg,"Jpeg: Unsupported bit depth %d in JPEG file",bpp);
    ithe_panic(msg,filename);
    break;
    }

M_free(buf);      // ok

return out;
}


/*
 *	Jpeg reader:
 */


// Error handler

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
char msgbuf[256];

  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  struct jpeg_error_mgr *myerr = cinfo->err;

  /* Display the message */
  (*cinfo->err->format_message)(cinfo,msgbuf);
  ithe_panic("Jpeg decompression error:",msgbuf);
}

//
// Here's the decompressor
//

unsigned char *UnJPG(FILE *fp, int *w, int *h, int *bpp)
{
struct jpeg_decompress_struct cinfo;
struct jpeg_error_mgr jerr;
JSAMPARRAY buffer;
int row_stride;
unsigned char *outbuf,*op;


cinfo.err = jpeg_std_error(&jerr);
jerr.error_exit = my_error_exit;

jpeg_create_decompress(&cinfo);

jpeg_stdio_src(&cinfo, fp);     // FILE *fp
jpeg_read_header(&cinfo, TRUE);

jpeg_calc_output_dimensions(&cinfo);

*w = cinfo.output_width;
*h = cinfo.output_height;

outbuf = (unsigned char *)M_get(*w**h,cinfo.output_components);
if(!outbuf)
      {
      ilog_quiet("alloc (%d*%d*%d) failed in load_jpeg\n",*w,*h,cinfo.output_components);
      return NULL;
      }
op=outbuf;

jpeg_start_decompress(&cinfo);

row_stride = cinfo.output_width * cinfo.output_components;
buffer = (*cinfo.mem->alloc_sarray) ((j_common_ptr) &cinfo, JPOOL_IMAGE,row_stride,1);


while (cinfo.output_scanline < cinfo.output_height)
      {
      jpeg_read_scanlines(&cinfo,buffer,1);
      memcpy(op,buffer[0],row_stride);
      op+=row_stride;
      }

*bpp = cinfo.output_components*8;       // 8 bit or 24 bit

jpeg_finish_decompress(&cinfo);
jpeg_destroy_decompress(&cinfo);

return outbuf;
}

/*
 *		load a CEL file using the IT-HE filesystem wrappers
 *		and without palette, so it can be used for lightmaps
 */

BITMAP *load_lightmap(char *filename)
{
unsigned short w,h;                             // size of image
int len,ctr;
unsigned char *buf,*iptr,*optr,l;				// conversion buffer
IFILE *fp;
BITMAP *img;

// Read the whole file

fp = iopen(filename);
// Good, we're still around

len=ifilelength(fp);
buf=(unsigned char *)M_get(1,len);              // Allocate room
iread(buf,len,fp);
iclose(fp);

// OK, let's go

if(buf[0] != 0x19 || buf[1] != 0x91)
    ithe_panic("CEL header invalid for image",filename);

w = buf[2]+(buf[3]<<8);
h = buf[4]+(buf[5]<<8);
len = w*h;

// Allocate the sprite we will put the .CEL into

select_palette(default_palette);

img = create_bitmap_ex(8,w,h);
if(!img)
    ithe_panic("Create bitmap failed for image",filename);
iptr=&buf[800];
optr=img->line[0];
for(ctr=0;ctr<len;ctr++)
	{
	l=*iptr++;
	if(!l)
		*optr++=1;
	else
		*optr++=l;
	}
M_free(buf);

unselect_palette();

return img;
}

/*
 *		Read and write an OBJECT structure
 */

void read_OBJECT(OBJECT *o,IFILE *f)
{
int flags;
S_START();

	S_GETPTR(o->name);
	S_GETWI(flags); // hack for the bitfield
	S_GETWI(o->w);
	S_GETWI(o->h);
	S_GETWI(o->mw);
	S_GETWI(o->mh);
	S_GETWI(o->x);
	S_GETWI(o->y);
	S_GETWI(o->z);
	S_GETPTR(o->personalname);
	S_GETPTR(o->schedule);
	S_GETPTR(o->UN_USED);
	S_GETWI(o->form);
	S_GETWI(o->sptr);
	S_GETWI(o->slen);
	S_GETWI(o->sdir);
	S_GETPTR(o->maxstats);
	S_GETPTR(o->stats);
	S_GETPTR(o->funcs);
	S_GETWI(o->curdir);
	S_GETSHI(o->dir[0]);
	S_GETSHI(o->dir[1]);
	S_GETSHI(o->dir[2]);
	S_GETSHI(o->dir[3]);
	S_GETPTR(o->desc);
	S_GETPTR(o->shortdesc);
	S_GETPTR(o->goals);
	S_GETWI(o->tag);
	S_GETPTR(o->goal);
	S_GETWI(o->SPARE[0]);
	S_GETWI(o->SPARE[1]);
	S_GETWI(o->SPARE[2]);
	S_GETWI(o->SPARE[3]);
	S_GETWI(o->SPARE[4]);
	S_GETWI(o->SPARE[5]);
	S_GETWI(o->SPARE[6]);
	S_GETWI(o->SPARE[7]);
	S_GETPTR(o->user);
	S_GETB(o->hblock[0]);
	S_GETB(o->hblock[1]);
	S_GETB(o->hblock[2]);
	S_GETB(o->hblock[3]);
	S_GETB(o->vblock[0]);
	S_GETB(o->vblock[1]);
	S_GETB(o->vblock[2]);
	S_GETB(o->vblock[3]);
	S_GETB(o->harea[0]);
	S_GETB(o->harea[1]);
	S_GETB(o->harea[2]);
	S_GETB(o->harea[3]);
	S_GETB(o->varea[0]);
	S_GETB(o->varea[1]);
	S_GETB(o->varea[2]);
	S_GETB(o->varea[3]);
	S_GETWI(o->save_id);
	S_GETWI(o->inside_id);
	S_GETWI(o->temp[0]);
	S_GETWI(o->temp[1]);
	S_GETWI(o->temp[2]);
	S_GETWI(o->temp[3]);
	S_GETWI(o->temp[4]);
	S_GETWI(o->temp[5]);
	S_GETWI(o->temp[6]);
	S_GETWI(o->temp[7]);
	S_GETWI(o->light);
	S_GETPTR(o->enemy);
	S_GETPTR(o->wield);
	S_GETPTR(o->pocket);
	S_GETPTR(o->next);

*(int *)&o->flags = flags; // nasty hack

S_END(OBJECT);
}

void write_OBJECT(OBJECT *o,IFILE *f)
{
unsigned int flags;
S_START();

flags = *(int *) &o->flags; // Coerce the flags into an integer

	S_PUTPTR(o->name);
	S_PUTWI(flags);
	S_PUTWI(o->w);
	S_PUTWI(o->h);
	S_PUTWI(o->mw);
	S_PUTWI(o->mh);
	S_PUTWI(o->x);
	S_PUTWI(o->y);
	S_PUTWI(o->z);
	S_PUTPTR(o->personalname);
	S_PUTPTR(o->schedule);
	S_PUTPTR(o->UN_USED);
	S_PUTWI(o->form);
	S_PUTWI(o->sptr);
	S_PUTWI(o->slen);
	S_PUTWI(o->sdir);
	S_PUTPTR(o->maxstats);
	S_PUTPTR(o->stats);
	S_PUTPTR(o->funcs);
	S_PUTWI(o->curdir);
	S_PUTSHI(o->dir[0]);
	S_PUTSHI(o->dir[1]);
	S_PUTSHI(o->dir[2]);
	S_PUTSHI(o->dir[3]);
	S_PUTPTR(o->desc);
	S_PUTPTR(o->shortdesc);
	S_PUTPTR(o->goals);
	S_PUTWI(o->tag);
	S_PUTPTR(o->goal);
	S_PUTWI(o->SPARE[0]);
	S_PUTWI(o->SPARE[1]);
	S_PUTWI(o->SPARE[2]);
	S_PUTWI(o->SPARE[3]);
	S_PUTWI(o->SPARE[4]);
	S_PUTWI(o->SPARE[5]);
	S_PUTWI(o->SPARE[6]);
	S_PUTWI(o->SPARE[7]);
	S_PUTPTR(o->user);
	S_PUTB(o->hblock[0]);
	S_PUTB(o->hblock[1]);
	S_PUTB(o->hblock[2]);
	S_PUTB(o->hblock[3]);
	S_PUTB(o->vblock[0]);
	S_PUTB(o->vblock[1]);
	S_PUTB(o->vblock[2]);
	S_PUTB(o->vblock[3]);
	S_PUTB(o->harea[0]);
	S_PUTB(o->harea[1]);
	S_PUTB(o->harea[2]);
	S_PUTB(o->harea[3]);
	S_PUTB(o->varea[0]);
	S_PUTB(o->varea[1]);
	S_PUTB(o->varea[2]);
	S_PUTB(o->varea[3]);
	S_PUTWI(o->save_id);
	S_PUTWI(o->inside_id);
	S_PUTWI(o->temp[0]);
	S_PUTWI(o->temp[1]);
	S_PUTWI(o->temp[2]);
	S_PUTWI(o->temp[3]);
	S_PUTWI(o->temp[4]);
	S_PUTWI(o->temp[5]);
	S_PUTWI(o->temp[6]);
	S_PUTWI(o->temp[7]);
	S_PUTWI(o->light);
	S_PUTPTR(o->enemy);
	S_PUTPTR(o->wield);
	S_PUTPTR(o->pocket);
	S_PUTPTR(o->next);

S_END(OBJECT);
}


/*
 *		Read and write a STATS structure
 */

void read_STATS(STATS *s,IFILE *f)
{
unsigned int flags,owner_id;
S_START();

	S_GETWI(s->hp);
	S_GETWI(s->dex);
	S_GETWI(s->str);
	S_GETWI(s->intel);
	S_GETWI(s->weight);
	S_GETWI(s->quantity);
	S_GETSHI(s->prob);
	S_GETSHI(s->tick);
	S_GETWI(flags);
	S_GETWI(s->damage);
	S_GETWI(s->oldhp);
	S_GETWI(s->pFlags);
	S_GETWI(owner_id);
	S_GETWI(s->karma);
	S_GETWI(s->bulk);
	S_GETPTR(s->npctalk);
	S_GETPTR(s->lFlags);
	S_READ(s->vv,(12*sizeof(short)));

if(owner_id != 0)
	{
	s->owner = find_id(owner_id);
	if(!s->owner)
		{
		ilog_quiet("Can't find owner_id %x\n",owner_id);
		}
	}

*(int *)&s->npcflags = flags; // nasty hack

S_END(STATS);
}

void write_STATS(STATS *s,IFILE *f)
{
unsigned int flags,owner_id;
S_START();

if(s->owner)
	owner_id = s->owner->save_id;
else
	owner_id = NULL;

flags = *(int *)&s->npcflags; // nasty hack

	S_PUTWI(s->hp);
	S_PUTWI(s->dex);
	S_PUTWI(s->str);
	S_PUTWI(s->intel);
	S_PUTWI(s->weight);
	S_PUTWI(s->quantity);
	S_PUTSHI(s->prob);
	S_PUTSHI(s->tick);
	S_PUTWI(flags);
	S_PUTWI(s->damage);
	S_PUTWI(s->oldhp);
	S_PUTWI(s->pFlags);
	S_PUTWI(owner_id);
	S_PUTWI(s->karma);
	S_PUTWI(s->bulk);
	S_PUTPTR(s->npctalk);
	S_PUTPTR(s->lFlags);
	S_WRITE(s->vv,(12*sizeof(short)));

S_END(STATS);
}

/*
 *		Read and write a FUNCS structure
 */

void read_OLDFUNCS(FUNCS *F,IFILE *f)
{
S_START();

	S_READ(F->use,32);
	S_GETWI(F->ucache);
	S_READ(F->talk,32);
	S_HACK(96); // Add 96 to 32 to make it 128
	S_GETWI(F->tcache);
	S_READ(F->kill,32);
	S_GETWI(F->kcache);
	S_READ(F->look,32);
	S_GETWI(F->lcache);
	S_READ(F->stand,32);
	S_GETWI(F->scache);
	S_READ(F->hurt,32);
	S_GETWI(F->hcache);
	S_READ(F->init,32);
	S_GETWI(F->icache);

	S_READ(F->contains[0],32);
	S_READ(F->contains[1],32);
	S_READ(F->contains[2],32);
	S_READ(F->contains[3],32);
	S_READ(F->contains[4],32);
	S_READ(F->contains[5],32);
	S_READ(F->contains[6],32);
	S_READ(F->contains[7],32);
	S_GETB(F->contents);
	S_READ(F->resurrect,32);
	S_GETB(F->dirty);
	S_READ(F->wield,32);
//	S_GETSHI(F->hackpadding);
	S_HACK(2);
	S_GETWI(F->wcache);

//S_END(FUNCS);  This is for compatability.. don't do any checking
}

void read_FUNCS(FUNCS *F,IFILE *f)
{
S_START();

	S_READ(F->use,32);
	S_GETWI(F->ucache);
	S_GETPTR(F->suse);
	S_READ(F->talk,128);
	S_GETWI(F->tcache);
	S_GETPTR(F->stalk);
	S_READ(F->kill,32);
	S_GETWI(F->kcache);
	S_GETPTR(F->skill);
	S_READ(F->look,32);
	S_GETWI(F->lcache);
	S_GETPTR(F->slook);
	S_READ(F->stand,32);
	S_GETWI(F->scache);
	S_GETPTR(F->sstand);
	S_READ(F->hurt,32);
	S_GETWI(F->hcache);
	S_GETPTR(F->shurt);
	S_READ(F->init,32);
	S_GETWI(F->icache);
	S_GETPTR(F->sinit);

	S_READ(F->contains[0],32);
	S_READ(F->contains[1],32);
	S_READ(F->contains[2],32);
	S_READ(F->contains[3],32);
	S_READ(F->contains[4],32);
	S_READ(F->contains[5],32);
	S_READ(F->contains[6],32);
	S_READ(F->contains[7],32);
//	S_GETB(F->contents);S_HACK(3);
	S_GETWI(F->contents);
	S_READ(F->resurrect,32);
	S_GETPTR(F->sresurrect);
//	S_GETB(F->dirty);S_HACK(3);
	S_GETWI(F->dirty);
	S_READ(F->wield,32);
	S_GETWI(F->wcache);
	S_GETPTR(F->swield);
//	S_GETSHI(F->hackpadding);

S_END(FUNCS);
}

void write_FUNCS(FUNCS *F,IFILE *f)
{
S_START();

	S_WRITE(F->use,32);
	S_PUTWI(F->ucache);
	S_PUTPTR(F->suse);
	S_WRITE(F->talk,128);
	S_PUTWI(F->tcache);
	S_PUTPTR(F->stalk);
	S_WRITE(F->kill,32);
	S_PUTWI(F->kcache);
	S_PUTPTR(F->skill);
	S_WRITE(F->look,32);
	S_PUTWI(F->lcache);
	S_PUTPTR(F->slook);
	S_WRITE(F->stand,32);
	S_PUTWI(F->scache);
	S_PUTPTR(F->sstand);
	S_WRITE(F->hurt,32);
	S_PUTWI(F->hcache);
	S_PUTPTR(F->shurt);
	S_WRITE(F->init,32);
	S_PUTWI(F->icache);
	S_PUTPTR(F->sinit);

	S_WRITE(F->contains[0],32);
	S_WRITE(F->contains[1],32);
	S_WRITE(F->contains[2],32);
	S_WRITE(F->contains[3],32);
	S_WRITE(F->contains[4],32);
	S_WRITE(F->contains[5],32);
	S_WRITE(F->contains[6],32);
	S_WRITE(F->contains[7],32);
	S_PUTWI(F->contents);

	S_WRITE(F->resurrect,32);
	S_PUTPTR(F->sresurrect);
	S_PUTWI(F->dirty);

	S_WRITE(F->wield,32);
	S_PUTWI(F->wcache);
	S_PUTPTR(F->swield);
//	S_PUTSHI(F->hackpadding);

S_END(FUNCS);
}

void read_WORLD(WORLD *w,IFILE *f)
{
S_START();

	S_GETWM(w->sig);
	S_GETWI(w->w);
	S_GETWI(w->h);
	S_GETPTR(w->physmap);
	S_GETPTR(w->object);
	S_GETPTR(w->roof);
	S_GETPTR(w->objmap);
	S_GETPTR(w->light);
	S_GETPTR(w->lightst);
	S_READ(w->temp,29*sizeof(int *));

S_END(WORLD);
}

void write_WORLD(WORLD *w,IFILE *f)
{
S_START();

	S_PUTWM(w->sig);
	S_PUTWI(w->w);
	S_PUTWI(w->h);
	S_PUTPTR(w->physmap);
	S_PUTPTR(w->object);
	S_PUTPTR(w->roof);
	S_PUTPTR(w->objmap);
	S_PUTPTR(w->light);
	S_PUTPTR(w->lightst);
	S_WRITE(w->temp,29*sizeof(int *));

S_END(WORLD);
}

