//
//      File loader
//

#include <allegro.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include "ithelib.h"
#include "loadfile.h"
#include "oscli.h"

// Defines

#define MAX_DATAFILES 32
//#define LOGSEEK // Debug the loadfile() function

static struct DFL_INFO
	{
	char *name;
	DATAFILE *data;
	int items;
	char **item;
	} DFL[MAX_DATAFILES];

// Variables

static int dflno=0;
char datafile_warning=0;

// Functions

extern int check_rar_file(char *filename);
extern void scan_rar_file(char *filename);

static void winslash(char *name);


// Code

char *loadfile(char *srcname, char *destname)
{
char buffer[1024];
int ctr;

#ifdef LOGSEEK
ilog_quiet("Given: '%s'\n",srcname);
#endif


if(!srcname)
	ithe_panic("loadfile: NULL source filename given",NULL);
if(!destname)
	ithe_panic("loadfile: NULL destination filename given",NULL);

// Try the basic name (in the running directory)

strcpy(destname,srcname);
strlwr(destname);
strslash(destname);

#ifdef LOGSEEK
ilog_quiet("check %s\n",destname);
#endif

if(iexist(destname))
	return destname;

// Try the full name (in the resource directory)

// First build the string without the resource directory, for later

strcpy(buffer,srcname);
strlwr(buffer);
strslash(buffer);

// Now add the resource directory and try that

strcpy(destname,resdir);
strcat(destname,"/");
strcat(destname,buffer);

#ifdef _WIN32
if(iexist(destname))
	return destname;
winslash(destname);	// Don't bother with it on other OSes
#endif

#ifdef LOGSEEK
ilog_quiet("check %s\n",destname);
#endif

if(iexist(destname))
	return destname;

#ifdef LOGSEEK
ilog_quiet("Not Found\n");
#endif
return NULL;
}

int fileexists(char *name)
{
char fn[1024];
if(!name)
	ithe_panic("fileexists: NULL filename given",NULL);
//if(strchr(name,'#'))
//	return(exists(name));
ilog_quiet("F-exists\n");
return((int)loadfile(name,fn));
}

int filelen(char *fname)
{
char fn[1024];

if(!fname)
	ithe_panic("filelen: NULL filename given",NULL);
if(strchr(fname,'#'))
	return file_size(fname);
if(loadfile(fname,fn))
	return file_size(fn);
else
	ithe_panic("filelen: file does not exist",fname);
}

void adddatafile(char *df)
{
if(dflno>=MAX_DATAFILES)
	{
	printf("\n\nToo many datafiles! (maximum %d)\n",MAX_DATAFILES);
	exit(1);
	}

DFL[dflno].name=M_get(1,strlen(df)+1);
strcpy(DFL[dflno++].name,df);
}

void checkdatafiles()
{
int ctr,code;
char temp[256];

//return;
ilog_printf("Checking datafiles:\n");

if(!dflno)
	{
	ilog_printf("  No datafiles, loading straight from disk\n");
	return;
	}

for(ctr=dflno-1;ctr>=0;ctr--)
	{
	code = check_rar_file(DFL[ctr].name);
	switch(code)
		{
		case 0:
	ilog_printf("  %s - Ok\n",DFL[ctr].name);
		break;

		case -1:
		ilog_printf("  %s - Warning: could not find datafile\n",DFL[ctr].name);
		datafile_warning = 1;
		break;

		case -2:
		ilog_printf("  %s - Warning: Not a proper RAR file\n",DFL[ctr].name);
		datafile_warning = 1;
		break;

		case -3:
		ilog_printf("  %s - Warning: RAR file must be uncompressed (use STORE method)\n",DFL[ctr].name);
		datafile_warning = 1;
		break;

		default:
		ilog_printf("  %s - Warning: Uck.. don't understand error code %d\n",DFL[ctr].name,code);
		datafile_warning = 1;
		break;
		}
	}
}

/*
 *		Convert slashes to windows format
 */

void winslash(char *name)
{
char *a;
do	{
	a=strchr(name,'/');
	if(a)
		*a='\\';
//      	*a='|';
	} while(a);
}

void mountdatafiles()
{
int ctr,ctr2,ctr3,ctr4;
char temp[256];

if(!dflno)
	return;

ilog_printf("Mounting datafiles:\n");

for(ctr=dflno-1;ctr>=0;ctr--)
	{
	ilog_printf("  %s\n",DFL[ctr].name);
	scan_rar_file(DFL[ctr].name);
	}
}

/*
 *      Convert a path from Absolute to Relative format
 */

void clean_path(char *fn)
{
char filename[1024];
char pathname[1024];

char *ptr;
int len;

strcpy(pathname,"/");
strcat(pathname,resdir);
strcat(pathname,"/");

strcpy(filename,"/");
strcat(filename,fn);

strslash(pathname);
strslash(filename);
strlwr(pathname);
strlwr(filename);

// OK, everything should be ready.

ptr = strstr(filename,pathname);
if(!ptr)
	return; // It's clean

// It's not clean.  Clean it.

ptr += strlen(pathname);
ptr--; // Remove initial '/'

len = (int)(ptr-filename);

strcpy(filename,&fn[len]);
strcpy(fn,filename);
}

/*
 *  .WAD manipulation functions.  Assumes a 40-byte signature in front of the
 *       data and works around this.  If you strip the first 40 bytes of the
 *       file away you have a true .WAD file that can load into DEU.
 */

void StartWad(IFILE *ofp)
{
iputl_i(0x44415750,ofp);// PWAD
iputl_i(0L,ofp);// Entries, patched later
iputl_i(0L,ofp);// Offset, ditto
}

/*
 *	Append the directory to the file to make a true .WAD
 */

void FinishWad(IFILE *ofp, WadEntry entries[])
{
unsigned int directory,num;
directory = itell(ofp);

// Write the entries

ilog_quiet("Start\n",num);
for(num=0;entries[num].name;num++)
	{
	iputl_i(entries[num].start-40,ofp); // Cancel out the signature
	iputl_i(entries[num+1].start-entries[num].start,ofp);
	iwrite(entries[num].name,8,ofp);
	}
ilog_quiet("Finished\n",num);
//num--;

// Back up the WAD and write the initial entries.  Since each IRE file has
// a 40-byte signature of it's own, we don't start from 4 like DEU does.

iseek(ofp,44,SEEK_SET);
iputl_i(num,ofp);
iputl_i(directory-40,ofp); // -40 to cancel the signature

// Finish

iclose(ofp);
}

int GetWadEntry(IFILE *ifp, char *name)
{
unsigned int sig,num,off,ctr;
char buf[]="nothing!";

iseek(ifp,40,SEEK_SET);
sig = igetl_i(ifp);
if(sig != 0x44415750)
	ithe_panic("Incorrect WAD substructure in file",ifp->truename);

num = igetl_i(ifp);
off = igetl_i(ifp);

iseek(ifp,off+40,SEEK_SET); // Cancel out the signature
for(ctr=0;ctr<num;ctr++)
	{
	off = igetl_i(ifp);
	igetl_i(ifp); // skip length
	iread(buf,8,ifp);
//	ilog_quiet("'%s'%s'\n",buf,name);
	if(!strcmp(buf,name))
		{
		iseek(ifp,off+40,SEEK_SET); // Cancel out the signature
		return 1;
		}
	}

return 0;
}
