//
//      loadsave.cc - Map IO for starting the game, saving and loading
//

//#define LOG_RW       // If you want to debug object load/save
#include <stdio.h>
#include <string.h>

#include "ithelib.h"
#include "media.h"
#include "console.h"
#include "core.hpp"
#include "gamedata.h"
#include "loadsave.hpp"
#include "oscli.h"
#include "cookies.h"
#include "object.hpp"
#include "nuspeech.hpp"
#include "loadfile.h"
#include "linklist.hpp"
#include "script.hpp"
#include "init.h"
#include "nuspeech.hpp"

// Defines

// Variables

extern "C" int map_W,map_H;	       // Default map size

static char stemp[128];
static int change_all=0;
static int temp_int,PStot;
static char sigtemp[40];       // Signature buffer
static IFILE *ofp;
static long objctr;
static int fastfind_id_num=0;
static OBJECT **fastfind_id_list=NULL;


// Functions

void load_map();
void load_z1(char *filename);
void save_z1(char *filename);
void wipe_z1();
void load_z2();

static void load_z1_old(char *filename);

static void SumContainers(OBJECT *cont,	long parent);
static void WriteContainers(OBJECT *cont);
static void WriteContainerSchedules(OBJECT *cont);
static void WriteContainerStats(OBJECT *cont);
static void WriteContainerFuncs(OBJECT *cont);
static void WriteContainerName(OBJECT *cont);

//static OBJECT *find_id(long id);
static void MoveToPocketEnd(OBJECT *object, OBJECT *container);
int SumObjects(OBJECT *cont, char *name, int total);
OBJECT *GetFirstObject(OBJECT *cont, char *name);
extern void Init_Areas(OBJECT *objsel);
static int has_changed(OBJECT *a);

static void fastfind_id_start();
static void fastfind_id_stop();
static OBJECT *fastfind_id(unsigned int id);

static int CMP_sort(const void *a, const void *b);
static int CMP_search(const void *a, const void *b);

extern void CallVMnum(int pos);
extern char *BestName(OBJECT *o);
extern "C" OBJECT *find_id(unsigned int id);

extern "C"
	{
	extern void read_OBJECT(OBJECT *o,IFILE *f);
	extern void write_OBJECT(OBJECT *o,IFILE *f);
	extern void read_STATS(STATS *s,IFILE *f);
	extern void write_STATS(STATS *s,IFILE *f);
	extern void read_FUNCS(FUNCS *F,IFILE *f);
	extern void read_OLDFUNCS(FUNCS *F,IFILE *f);
	extern void write_FUNCS(FUNCS *F,IFILE *f);
	extern void read_WORLD(WORLD *w,IFILE *f);
	extern void write_WORLD(WORLD *w,IFILE *f);
	}

// Code


/*
 *      LoadMap- Load the whole map from disk, for first-time startup
 */

void LoadMap()
{
OBJLIST	*ptr;
load_map();

strcpy(stemp,mapname);
strcat(stemp,".mz1");
load_z1(stemp);

load_z2();

load_z3();

ilog_printf("\n");
ilog_printf("Run Object Inits\n");

if(!in_editor)
	for(ptr=MasterList;ptr;ptr=ptr->next)
		if(!ptr->ptr->flags.didinit)
			{
			if(ptr->ptr->funcs->icache != -1)
				{
				//current_object = ptr->ptr;
				person = ptr->ptr;
				CallVMnum(ptr->ptr->funcs->icache);
				}
			ptr->ptr->flags.didinit=1;
			}
ilog_printf("\n");
ilog_printf("InitLand\n");
}

//
//      Functions for the tilemap IO
//

/*
 *      load_map - Load the tiles from disk for startup
 */

void load_map()
{
int yt,x,y,b,do_delete=0;
long mem;
char key;
char filename[1024];
char number[16];
IFILE *fp;

curmap->w = map_W;
curmap->h = map_H;

ilog_printf("Opening map:%s\n",mapname);
strcpy(stemp,mapname);
strcat(stemp,".map");

if(!loadfile(stemp,filename))
	{
	ilog_printf("Could not find file '%s'\n",stemp);
	ilog_printf("Hit Y to create a new file, or any other key to exit.\n");
        key = WaitForAscii();
	if(key!='y' && key!='Y')
		exit(1);

try_again:
	mem = map_W * map_H * (sizeof(short)+sizeof(char)+sizeof(char)+sizeof(char)+sizeof(OBJECT	*))/1024;
	ilog_printf("\n\n\n");
	ilog_printf("The new map will be %dx%d, and require %ldKB of memory.\n",map_W,map_H,mem);
	ilog_printf("Is this okay? (Y/N)\n");
	key = WaitForAscii();
	if(key!='y' && key!='Y')
		{
		ilog_printf("No.\n");
		ilog_printf("\n");
		do {
            number[0]=0;
			ilog_printf("Enter the new width of the map:\n");
			if(!irecon_getinput(number,15))
				exit(1);
			map_W = atoi(number);
            } while(map_W<1);

		do {
            number[0]=0;
			ilog_printf("Enter the new height of the map:\n");
			if(!irecon_getinput(number,15))
				exit(1);
			map_H = atoi(number);
            } while(map_H<1);

		curmap->w = map_W;
		curmap->h = map_H;
		goto try_again;
		}

	curmap->sig=MAPSIG;
	curmap->object =	(struct	OBJECT *)M_get(1,sizeof(OBJECT));

	curmap->object->next=NULL;
	curmap->physmap = (unsigned short *)M_get(curmap->w * curmap->h,sizeof(short));
	curmap->roof = (unsigned	char *)M_get(curmap->w * curmap->h,sizeof(char));
	curmap->objmap =	(OBJECT	**)M_get(curmap->w * curmap->h, sizeof(OBJECT *));
	curmap->light = (unsigned	char *)M_get(curmap->w * curmap->h,sizeof(char));
	curmap->lightst = (unsigned	char *)M_get(curmap->w * curmap->h,sizeof(char));
	}
else
	{
    fp = iopen(filename);
    if(!fp)
		ithe_panic("load_map: cannot open file",stemp);

	iread((unsigned char *)sigtemp,40,fp);
	if(strcmp(sigtemp,COOKIE_MAP))
		ithe_panic("load_map: file does not contain the right cookie",stemp);

//	iread((unsigned char *)curmap,sizeof(WORLD),fp);
//	temp_int=DecodeMapSig(curmap->sig);
//	if((unsigned)temp_int != MAPSIG)
	read_WORLD(curmap,fp);
	if(curmap->sig != MAPSIG)
		{
		sprintf(filename,"Current map version is 0x%x, but this is 0x%x\n",MAPSIG,curmap->sig);
		ithe_panic(filename,stemp);
//		ilog_printf("Current map version is 0x%x, but this is 0x%x\n",MAPSIG,curmap->sig);
		}
	curmap->physmap = (unsigned short *)M_get(curmap->w * curmap->h,sizeof(short));
	iread((unsigned char *)curmap->physmap,curmap->w*curmap->h*sizeof(short),fp);
	iclose(fp);

	curmap->object =	(struct	OBJECT *)M_get(1,sizeof(OBJECT));
	curmap->object->name="Linkedlist entrypoint object";
	curmap->object->next=NULL;

	// Set up matrix of linked lists

	curmap->objmap =	(OBJECT	**)M_get(curmap->w * curmap->h, sizeof(OBJECT *));

	// Build quick-access table for the matrix

	ilog_quiet("Build qat using width %d\n",curmap->w);

	for(yt=0;yt<curmap->h;yt++)
	    ytab[yt]=curmap->w *	yt;

	curmap->roof = (unsigned	char *)M_get(curmap->w * curmap->h,sizeof(char));
	curmap->light= (unsigned	char *)M_get(curmap->w * curmap->h,sizeof(char));
	curmap->lightst= (unsigned	char *)M_get(curmap->w * curmap->h,sizeof(char));
	}

ilog_quiet("Check map integrity\n");

// Check map state and correct if necessary

for(y=0;y<curmap->h;y++)
	for(x=0;x<curmap->w;x++)
		{
		b=curmap->physmap[MAP_POS(x,y)];
		if(b>=TItot && b != 0xffff)	// 0xffff is random and allowed
			{
			if(in_editor)
				{
				if(!do_delete)
				    {
				    ilog_printf("Invalid tiles detected in the map.\n");
				    ilog_printf("Do you want to:\n");
				    ilog_printf("1. Set all unknown tiles to 0\n");
				    ilog_printf("2. Quit so you can add the tile to resource.txt\n");
				    do
						{
						key=WaitForAscii();
						if(key == '1')
						    do_delete=1;
						if(key == '2')
						    exit(1);
						} while(key != '1');
				    }
				}
			else
				{
				curmap->physmap[MAP_POS(x,y)] = 0xffff; // surprise!!
				do_delete = 2;
				}
			if(do_delete == 1)
				curmap->physmap[MAP_POS(x,y)]=0;
			}
		}
if(do_delete ==	2)
    Bug("Invalid map tiles detected\n");
}

/*
 *      save_map - Save the tiles to disk
 */

void save_map()
{
ofp=NULL;

strcpy(stemp,mapname);
strcat(stemp,".map");

ofp = iopen_write(stemp);

// Since error checking will have been done previously it is safe to explode
// if something has gone wrong.

iwrite((unsigned char *)COOKIE_MAP,40,ofp);
curmap->sig=MAPSIG; // Just to make sure
write_WORLD(curmap,ofp);
iwrite((unsigned char *)curmap->physmap,curmap->w*curmap->h*sizeof(short),ofp);
iclose(ofp);
}

//
//      Z1 functions for the object layer IO
//


/*
 *      load_z1 - Load the object layer, or a savegame
 */

void load_z1(char *fname)
{
void *Ltemp[12];
IFILE *fp;
OBJECT *temp,*parent,*next,*temp2;
STATS buf;
char dodelete=0,bad_object=0,is_open=0;
int len,ctr,ctr2,unknown=0,qty;
int form,sptr,slen,sdir,newpos;
int key;
char filename[1024];
char objname[64];
unsigned int id;
SCHEDULE_065 compat065; // Loader for 0.65 schedule data (remove after 0.7)

if(!loadfile(fname,filename))
	{
	ilog_printf("Could not find file '%s'\n",fname);
	ilog_printf("Hit Y to create a new file, or any other key to exit.\n");
	key=WaitForAscii();
	if(key!='y' && key!='Y')
		exit(1);
	return;
	}
else
	{
	fp=iopen(filename);
	is_open=1;
	iread((unsigned char *)sigtemp,40,fp);
	// If it's not a current map, check for a previous revision
	if(strcmp(sigtemp,COOKIE_MZ1))
		{
		if(!strcmp(sigtemp,COOKIE_MZ1_OLD)) // Too old, don't want it
			ithe_panic("Maps before 0.65 no longer supported","Load into the 0.65 editor to convert");
		// Bad.  Crash out.
		ilog_printf("File '%s' is not a z1 map.. it does not contain the right cookie\n",stemp);
		exit(1);
		}

	if(!GetWadEntry(fp,"OBJECTS "))
		ithe_panic("'OBJECTS ' entry missing from objects layer.. cannot continue.",filename);

	PStot = igetl_i(fp);
	if(PStot)	// There are some items in the file
		{
		temp_int = igetl_i(fp);
		if(temp_int != sizeof(OBJECT))
			{
			ilog_printf("File '%s' has incorrect data structure.\n",stemp);
			ilog_printf("Current sizeof is %d, but this is %d\n",sizeof(OBJECT),temp_int);
			exit(1);
			}

ilog_printf("    Creating\n");
strcpy(stemp,"");

		for(ctr=0;ctr<PStot;ctr++)
			{
			temp = OB_Alloc();
//#if 1

			// Preserve the alloc'd values

			Ltemp[0] = temp->funcs;
			Ltemp[1] = temp->maxstats;
			Ltemp[2] = temp->stats;
			Ltemp[3] = temp->wield;
			Ltemp[4] = temp->enemy;
			Ltemp[5] = temp->schedule;
//			Ltemp[6] = temp->goals;
			Ltemp[7] = temp->personalname;
			Ltemp[8] = temp->user;
//			Ltemp[9] = temp->goal;

			// the alloc'd values get trashed here

			read_OBJECT(temp,fp);
			iread((unsigned char *)stemp,32,fp);		// Get name
			stemp[32]=0;					// Make sure

			// restore preserved values

			temp->funcs	= (FUNCS *)Ltemp[0];
			temp->maxstats = (STATS *)Ltemp[1];
			temp->stats	= (STATS *)Ltemp[2];
			temp->wield	= (OBJECT **)Ltemp[3];
			temp->enemy	= (OBJECT *)Ltemp[4];
			temp->schedule = (SCHEDULE *)Ltemp[5];
//			temp->goals = (GOAL *)Ltemp[6];
			temp->personalname=	(char *)Ltemp[7];
			temp->user=	(USEDATA *)Ltemp[8];
//			temp->goal = (GOAL *)Ltemp[9];
			temp->stats->npctalk = NULL;
			temp->stats->lFlags = NULL;

// If we're not reloading, eradicate behaviour

			if(!fullrestore)
				{
				if(ML_InList(&ActiveList,temp))
					ML_Del(&ActiveList,temp);
	//            temp->behave = -1;
				}

			// Look up the name from the script

			bad_object=0;
			temp_int = getnum4char(stemp);  // Find character
			if(temp_int == -1)
				{
				bad_object=1;
				if(!in_editor)
					ithe_panic("Load_Z1: Could not find this name in 'section: character'!",stemp);
				if(change_all)
					temp_int=0;
				else
					{
					dodelete=0;
					irecon_printf("Could not find definition of object '%s' in the scriptfile.\n",stemp);
					irecon_printf("Do you want to..\n");
					irecon_printf("1.  Change this object to '%s'\n",CHlist[0].name);
					irecon_printf("2.  Change ALL unknown objects to '%s'\n",CHlist[0].name);
					irecon_printf("3.  Delete this object\n");
					irecon_printf("4.  Delete ALL unknown objects\n");
					irecon_printf("5.  Enter a name to change this object to\n");
					irecon_printf("6.  Quit (so you can add the object to the scriptfile)\n");
					do
						{
						key=WaitForAscii();
						if(key == '1')
							temp_int=0;
						if(key == '2')
							{
							temp_int=0;
							change_all=1;
							}
						if(key == '3')
							dodelete=1;
						if(key == '4')
							{
							dodelete=1;
							change_all=1;
							}
						if(key == '5')
							{
							do  {
								irecon_printf("Enter the object name:\n");
								if(!GetStringInput(objname,64))
									exit(1);
//								scanf("%s",stemp);
								temp_int = getnum4char(objname);  // Find character
								if(temp_int == -1)
									irecon_printf("Not found in main.txt\n");
								} while(temp_int == -1);
							}
						if(key == '6')
							exit(1);
						} while(key <'1' || key>'6');
					irecon_printf("%c\n",key);
					}
				}

        // Now set up this name using the index in the NPC dbase

			if(dodelete && bad_object)
				temp_int = 0;
			temp->name = CHlist[temp_int].name;	// Get name in script
			if(!fullrestore)
				temp->flags= CHlist[temp_int].flags;// Copy flags
			temp->desc=	CHlist[temp_int].desc;// Copy descriptions
			temp->shortdesc= CHlist[temp_int].shortdesc;

			temp->next=NULL;
			temp->pocket=NULL;

#ifdef LOG_RW
	ilog_quiet("load: %s id = %d  iid = %d map %d,%d\n",temp->name,temp->save_id,temp->inside_id,temp->x,temp->y);
#endif

			form = (int)temp->form;
			sptr = temp->sptr;
			slen = temp->slen;
			sdir = temp->sdir;

			OB_Init(temp,temp->name);		// reinitialise object

			temp->stats->oldhp=temp->stats->hp;
			// Inherit any new flags in the editor
			if(!fullrestore)
				temp->stats->npcflags= CHlist[temp_int].stats->npcflags;

			if(fullrestore)
				{
	//                        boot2("***:%s\n",SQlist[form].name);
				OB_SetSeq(temp,SQlist[form].name);
				temp->sptr = sptr;
				temp->slen = slen;
				temp->sdir = sdir;
				}

// Store whether the object is fixed or not.
// Seekstate is a transient flag only used when non-fixed so store it there
//                    temp->flags.seekstate = CHlist[temp_int].flags.fixed;
//                    temp->flags.seekstate=temp->flags.fixed;
//                    if(temp->behave != -1)
//                        AL_Add(&ActiveList,temp); // Make it active
			if(!temp->inside_id)	      // Not inside another
				{
				MoveToMap(temp->x,temp->y,temp);
				Init_Areas(temp);
				}
			if(dodelete && bad_object)
				DestroyObject(temp);
//#endif
			}
		}
	}

// Now sort out the nested objects business.
// Everything still left in the objlist needs to be inside something else
// This has to be done using a WHILE loop instead of a FOR loop, because
// temp->next gets changed!

//exit(1);

fastfind_id_start();

ilog_printf("    DeNesting\n");


temp = curmap->object->next;
while(temp)
	{
	next = temp->next;
	parent = fastfind_id(temp->inside_id);
	if(parent)
		{
#ifdef LOG_RW
	ilog_quiet("Object ID %d (%s) is inside %d\n", temp->save_id,temp->name,temp->inside_id);
#endif
//        temp->flags.seekstate=temp->flags.fixed; // store it here for now
//        temp->flags.fixed=1;            // Send it to sleep until it leaves
		MoveToPocketEnd(temp,parent);
		}
	else
		{
		ilog_quiet("Oh shit!  %s is orphaned.  ID = %d Parent = %d\n",temp->name,temp->save_id,temp->inside_id);
		Bug("%s [%d] has an identity crisis\n",temp->name,temp->save_id);
		DestroyObject(temp);
		}
	temp = next;
	};

// Abort if failed to open the map (and just used defaults)

if(!is_open)
	return;

// Auxiliary block 2: statistics

if(!GetWadEntry(fp,"OBJSTATS"))
	Bug("Warning.. no statistics entry in level data\n");
else
	{
	ilog_printf("    Loading Statistics\n");
	len = igetl_i(fp);//        len = sizeof(STATS);
	for(ctr=0;ctr<PStot;ctr++)
		{
		id = igetl_i(fp);
		if(id == 0xffffffff)
			temp = NULL;
		else
			temp = fastfind_id(id);
		if(!temp)
			{
			// If the object's stats are same as the default id=0xffffffff
			// and no data follows.

			// If it's not deliberately blank, skip supposed length
			if(id != 0xffffffff)
				{
				Bug("Bugger #2\n");
				iseek(fp,len,SEEK_CUR);
				}
			}
		else
			{
			newpos=itell(fp)+len;
			memcpy(&buf,temp->stats,sizeof(STATS));
			read_STATS(temp->stats,fp);
			iseek(fp,newpos,SEEK_SET); // for compatability

			if(reset_stats)
				{
				qty = temp->stats->quantity;
				temp2 = temp->stats->owner;
				memcpy(temp->stats,&buf,len);
				temp->stats->owner = temp2;
				temp->stats->quantity = qty;
				}
			temp->stats->oldhp=temp->stats->hp; // JM: Do this always
			// Inherit any new flags in the editor
			if(!fullrestore)
				temp->stats->npcflags= CHlist[temp_int].stats->npcflags;
			temp->stats->npctalk=NULL;
			temp->stats->lFlags=NULL;
			}
		}
	}

// Auxiliary block 3: individual names

if(!GetWadEntry(fp,"P_NAMES "))
	Bug("Warning.. no personal names entry in level data\n");
else
	{
	ilog_printf("    Loading Personal Names\n");
	for(ctr=0;ctr<PStot;ctr++)
		{
		id = igetl_i(fp);
		temp = fastfind_id(id);
		if(!temp)
			{
			Bug("Bugger #3\n");
			iseek(fp,32,SEEK_CUR);
			}
		else
			iread((unsigned char *)temp->personalname,32,fp);
		}
	}

// Auxiliary block 4: schedules

if(!GetWadEntry(fp,"SCHEDULE"))
	Bug("Warning.. no schedules entry in level data\n");
else
	{
	ilog_printf("    Loading Schedules\n");
	len=igetl_i(fp);

	switch(len)
		{
		case sizeof(SCHEDULE):
		ilog_quiet("Current version (%s)\n",IKV);
		break;

		case sizeof(SCHEDULE_065):
		ilog_quiet("Old version (0.65)\n");
		break;

		default:
		ilog_quiet("Unknown version: risking it\n");
		break;
		};

	for(ctr=0;ctr<PStot;ctr++)
		{
		id = igetl_i(fp);
		if(id == 0xffffffff)
			temp=NULL;
		else
			temp=fastfind_id(id);

		if(!temp)
			{
			if(id != 0xffffffff)
				{
				Bug("Bugger #4\n");
				iseek(fp,len*24,SEEK_CUR);
				}
			}
		else
			{
			if(!temp->schedule)
				temp->schedule = (SCHEDULE *)M_get(24,sizeof(SCHEDULE));
			for(ctr2=0;ctr2<24;ctr2++)
				{
				temp->schedule[ctr2].hour=-1; // Mark unused by default
				// Backward compatability.. remove after 0.65
				if (len == sizeof(SCHEDULE_065))
					{
					iread((unsigned char *)&compat065,len,fp);
					if(compat065.hour != -1)
						{
						temp->schedule[ctr2].hour = compat065.hour;
						temp->schedule[ctr2].minute= compat065.minute;
						temp->schedule[ctr2].ok= compat065.ok;
						temp->schedule[ctr2].target = (OBJECT *)find_id((int)compat065.target);
						temp->schedule[ctr2].call = getnum4PE(compat065.vrm);
						if(temp->schedule[ctr2].call == -1)
							{
							Bug("Don't know '%s'\n",compat065.vrm);
							}
						strcpy(temp->schedule[ctr2].targettype,compat065.targettype);
						strcpy(temp->schedule[ctr2].vrm,compat065.vrm);
//						ilog_quiet("%d:%d %s\n",temp->schedule[ctr2].hour,temp->schedule[ctr2].minute,temp->schedule[ctr2].target?temp->schedule[ctr2].target->name:"<nothing>");
//						ilog_quiet("[%d] %s (%s)\n",temp->schedule[ctr2].ok,temp->schedule[ctr2].vrm,temp->schedule[ctr2].targettype);
						}
					}
				else		// 0.7 Schedule reader
					{
					iread((unsigned char *)&temp->schedule[ctr2],len,fp);
					if(temp->schedule[ctr2].hour != -1)
						{
						temp->schedule[ctr2].target = (OBJECT *)find_id((int)temp->schedule[ctr2].target);
						temp->schedule[ctr2].call = getnum4PE(temp->schedule[ctr2].vrm);
						if(temp->schedule[ctr2].call == -1)
							{
							Bug("Don't know '%s'\n",temp->schedule[ctr2].vrm);
							}
						}
					}
				}
			}
		}
	}

// Auxiliary block 5: functions

if(!GetWadEntry(fp,"OBJFUNCS"))
	Bug("Warning.. no functions entry in level data\n");
else
	{
	ilog_printf("    Loading Custom Functions\n");
	len = igetl_i(fp);
	if(len<sizeof(FUNCS))
		ilog_quiet("    Loading old-style functions\n");
	ctr2 = 1;
	for(;ctr2;)
		{
		id = igetl_i(fp);
		if((unsigned)id == 0xffffffff)
			ctr2=0;
		else
			{
			temp = fastfind_id(id);
			if(!temp || reset_funcs)
				{
				if(!reset_funcs)
					Bug("Bugger #5\n");
				iseek(fp,len,SEEK_CUR);
				}
			else
				{
				newpos=itell(fp)+len;
				if(len<sizeof(FUNCS))
					read_OLDFUNCS(temp->funcs,fp);
				else
					read_FUNCS(temp->funcs,fp);
				iseek(fp,newpos,SEEK_SET); // For compatability
				InitFuncsFor(temp,1);       // Regenerate cached entries
				InitFuncsFor(temp,0);       // Regenerate cached entries
				}
			}
		}
	}

fastfind_id_stop();

iclose(fp);

// If we're in the game, deck the empty schedule pointers so we can quickly
// see if the object has any events or not

OBJLIST	*ptr;
char ok=0;

if(!in_editor)
	for(ptr=MasterList;ptr;ptr=ptr->next)
		{
		if(ptr->ptr->schedule)
			{
			ok=0;
			for(ctr=0;ctr<24;ctr++)
				if(ptr->ptr->schedule[ctr].hour != -1)
					ok=1;
			if(!ok)
				{
				M_free(ptr->ptr->schedule);
				ptr->ptr->schedule = NULL;
				}
			}
		}
}

/*
 *      save_z1 - Write the object layer, or a savegame
 */

void save_z1(char *filename)
{
OBJECT *temp,*owner,*old;
int vx,vy,ctr,ok,objtot;
SEQ_POOL *form;
WadEntry WadEntries[]=	{
					{0,"OBJECTS "},
					{0,"OBJSTATS"},
					{0,"P_NAMES "},
					{0,"SCHEDULE"},
					{0,"OBJFUNCS"},
					{0,NULL},
				};
int wEntry=0;

// Open the file.
ofp=iopen_write(filename);
// Write the main header
iwrite((unsigned char *)COOKIE_MZ1,40,ofp);	    // Cookie
// Write the WAD header
StartWad(ofp);

// First, set each object so it has an ID of its own and knows which object
// it is inside (if any)

objctr=1;
for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		temp = GetRawObjectBase(vx,vy);	// RAWobjectbase ignores large objs
		for(;temp;temp=temp->next)
			{
			temp->save_id=objctr;
			temp->inside_id=0;
			if(temp->pocket)
				SumContainers(temp->pocket,objctr);
			objctr++;
			}
		}
objtot=objctr-1;

// OBJECTS
WadEntries[wEntry++].start=itell(ofp); // Fix up the start position

iputl_i(objtot,ofp);		    // Number of objects
iputl_i(sizeof(OBJECT),ofp);	    // Size of data structure

for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		temp = GetRawObjectBase(vx,vy);
		for(;temp;temp=temp->next)
			{
			// Substitute id number for FORM pointer
			form = temp->form;
			temp->form = (SEQ_POOL *)getnum4sequence(temp->form->name);
			write_OBJECT(temp,ofp);
			iwrite((unsigned char *)temp->name,32,ofp);

			// then switch it back again
			temp->form = form;

#ifdef LOG_RW
	    ilog_quiet("Object ID %d (%s) is outside at %d,%d\n", temp->save_id,temp->name,temp->x,temp->y);
#endif
			if(temp->pocket)
				WriteContainers(temp->pocket);
			}
		}

// OBJSTATS
WadEntries[wEntry++].start=itell(ofp); // Fix up the start position

iputl_i(sizeof(STATS),ofp);
for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		temp = GetRawObjectBase(vx,vy);
		for(;temp;temp=temp->next)
		if(has_changed(temp))
			{
			// Put object's ID number
			iputl_i(temp->save_id,ofp);
			write_STATS(temp->stats,ofp);
			if(temp->pocket)
				WriteContainerStats(temp->pocket);
			}
		else // Is the same, just ignore it
			{
			iputl_i(0xffffffff,ofp);
			if(temp->pocket)
				WriteContainerStats(temp->pocket);
			}
		}

// P_NAMES
WadEntries[wEntry++].start=itell(ofp); // Fix up the start position

for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		temp = GetRawObjectBase(vx,vy);
		for(;temp;temp=temp->next)
			{
			iputl_i(temp->save_id,ofp);
			iwrite((unsigned char *)temp->personalname,32,ofp);
			if(temp->pocket)
				WriteContainerName(temp->pocket);
			}
		}

// SCHEDULE
WadEntries[wEntry++].start=itell(ofp); // Fix up the start position

iputl_i(sizeof(SCHEDULE),ofp);

for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		temp = GetRawObjectBase(vx,vy);
		for(;temp;temp=temp->next)
			{
			ok=0;
			if(temp->schedule)
				for(ctr=0;ctr<24;ctr++)
				if(temp->schedule[ctr].hour != -1)
					ok=1;
			if(!ok)
				iputl_i(0xffffffff,ofp);
			else
				{
				ilog_quiet("Am Saving\n");
				iputl_i(temp->save_id,ofp);
				for(ctr=0;ctr<24;ctr++)
					if(temp->schedule[ctr].hour == -1)
						iwrite((unsigned char *)&temp->schedule[ctr],sizeof(SCHEDULE),ofp);
					else
						{
						// Convert the pointer to an ID number for the saving
						old = temp->schedule[ctr].target;
						if(old)
							temp->schedule[ctr].target=(OBJECT *)temp->schedule[ctr].target->save_id;
						// Save it
						iwrite((unsigned char *)&temp->schedule[ctr],sizeof(SCHEDULE),ofp);
						// Restore the original
						temp->schedule[ctr].target=old;
						}
				}
			if(temp->pocket)
				WriteContainerSchedules(temp->pocket);
			}
		}

// OBJFUNCS
WadEntries[wEntry++].start=itell(ofp); // Fix up the start position

iputl_i(sizeof(FUNCS),ofp);

for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		temp = GetRawObjectBase(vx,vy);
		for(;temp;temp=temp->next)
			{
			if(temp->funcs->dirty)
				{
				iputl_i(temp->save_id,ofp);
				write_FUNCS(temp->funcs,ofp);
//				iwrite((unsigned char *)temp->funcs,sizeof(FUNCS),ofp);
				}
		if(temp->pocket)
			WriteContainerFuncs(temp->pocket);
		}
	}
iputl_i(0xffffffff,ofp); // sentinel value for Custom Functions (blk5)

// Finish off
WadEntries[wEntry++].start=itell(ofp); // To calculate length of last entry

FinishWad(ofp,WadEntries);

// The file is now closed.

}

/*
 *      wipe_z1 - deallocate the z1 layer entirely
 *                called prior to load_z1 in game, not in editor
 */

void wipe_z1()
{
OBJECT *temp,*anchor;
OBJLIST *o;
int vx,vy;

ilog_quiet("wiping object matrix\n");

for(vy=0;vy<curmap->h;vy++)
	for(vx=0;vx<curmap->w;vx++)
		{
		anchor = GetRawObjectBase(vx,vy); // RAWobjectbase ignores large objs
		if(anchor)
			for(temp = anchor;temp;temp=temp->next)	temp->flags.on=0;
		}

ilog_quiet("Emptying Syspocket\n\n");
FreePockets(curmap->object);

pending_delete=1;	// Force garbage collection
DeletePending();
player=NULL;

ilog_quiet("\nwiping party members\n\n");

for(vx=0;vx<MAX_MEMBERS;vx++)
	party[vx]=NULL;

}

//
//      Z2 rooftops layer map IO
//

/*
 *      load_z2 - Load the rooftops from disk
 */

void load_z2()
{
IFILE *fp;
char filename[1024];
int key;

ilog_printf("  Rooftops\n",mapname);

//      Objects layer 2 - roof

strcpy(stemp,mapname);
strcat(stemp,".mz2");

if(!loadfile(stemp,filename))
	{
	ilog_printf("Could not find file '%s'\n",stemp);
	irecon_printf("Hit Y to create a new file, or any other key to exit.");
	key=WaitForAscii();
	if(key!='Y'&&key!='y')
		exit(1);
	}
else
	{
	fp=iopen(filename);
	iread((unsigned char *)sigtemp,40,fp);
	if(strcmp(sigtemp,COOKIE_MZ2))
		{
		ilog_printf("File '%s' is not a z2 map.. it does not contain the right cookie\n",stemp);
		exit(1);
		}

	iread((unsigned char *)curmap->roof,curmap->w*curmap->h*sizeof(unsigned char),fp);
	iclose(fp);
	}
}

/*
 *      save_z2 - Save the rooftops to disk
 */

void save_z2()
{
//      Objects layer 2 - statlc lights

strcpy(stemp,mapname);
strcat(stemp,".mz2");

ofp = iopen_write(stemp);

if(!ofp)
	{
	if(!in_editor)
		{
		ilog_printf("Oh no!\n");
		ilog_printf("Could not create file '%s'\n",stemp);
		}
	return;
	}

iwrite((unsigned char *)COOKIE_MZ2,40,ofp);
iwrite((unsigned char *)curmap->roof,curmap->w*curmap->h,ofp);
iclose(ofp);
return;
}

//
//      Z3 static light layer map IO
//

/*
 *      load_z3 - Load the static lighting from disk
 */

void load_z3()
{
IFILE *fp;
char filename[1024];
int key;

//      Objects layer 3 - static lights

ilog_printf("  Lights\n",mapname);

strcpy(stemp,mapname);
strcat(stemp,".mz3");

if(!loadfile(stemp,filename))
	{
	ilog_printf("Could not find file '%s'\n",stemp);
	irecon_printf("Hit Y to create a new file, or any other key to exit.");
	key=WaitForAscii();
	if(key!='Y'&&key!='y')
		exit(1);
	}
else
	{
	fp=iopen(filename);
	iread((unsigned char *)sigtemp,40,fp);
	if(strcmp(sigtemp,COOKIE_MZ3))
		{
		ilog_printf("File '%s' is not a z2 map.. it does not contain the right cookie\n",stemp);
		exit(1);
		}

	iread((unsigned char *)curmap->light,curmap->w*curmap->h*sizeof(unsigned char),fp);
	iclose(fp);
	}
ilog_printf("  Done\n",mapname);
}

/*
 *      save_z3 - Save the static lighting to disk
 */

void save_z3()
{
//      Objects layer 3 - static lights

strcpy(stemp,mapname);
strcat(stemp,".mz3");

ofp = iopen_write(stemp);

if(!ofp)
	{
	if(!in_editor)
		{
		ilog_printf("Oh no!\n");
		ilog_printf("Could not create file '%s'\n",stemp);
		}
	return;
	}

iwrite((unsigned char *)COOKIE_MZ3,40,ofp);
iwrite((unsigned char *)curmap->light,curmap->w*curmap->h,ofp);
iclose(ofp);
return;
}

// Miscellaneous State

/*
 *      load_ms - Read miscellaneous savegame data
 */

void load_ms(char *filename)
{
FILE *ifp;
char buf[256];
int ctr,num,ctr2,ctr3;
char priority;
OBJECT *o,*o2;
unsigned int t;

// Open the file.

ifp=fopen(filename,"rb");
if(!ifp)
    return;

//      Write the header

fread(buf,1,40,ifp);			 // Skip title
fread(buf,1,40,ifp);			 // Cookie
if(strcmp(COOKIE_MS,buf))
    {
    fclose(ifp);
    return;
    }

//player = find_id(getl(ifp));			// Find the player
t = getw(ifp);
player=find_id(t);
ilog_quiet("player id = %d\n",t);
if(!player)
    ithe_panic("load_ms: Player AWOL in savegame",NULL);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
    {
    t =	getw(ifp);
    if(t)
	{
	party[ctr] = find_id(t);	      // ID numbers change to objptrs
	party[ctr]->flags.party	= 1;
	}
    else
	party[ctr] = NULL;
    }

if(getw(ifp) !=	STORE)		// Oh no!
    {
    fclose(ifp);
    return;
    }
fread(storage,1,STORE,ifp);

// Load user flags

Wipe_tFlags();
t=getw(ifp);
for(ctr=0;ctr<t;ctr++)
    {
    fread(buf,1,getc(ifp),ifp);
    Set_tFlag(buf,1);
    }

// Load time/date

game_minute = getw(ifp);
game_hour   = getw(ifp);
game_day    = getw(ifp);
game_month  = getw(ifp);
game_year   = getw(ifp);
days_passed = getw(ifp);
day_of_week = getw(ifp);

// Load goal data

num = getw(ifp);
for(ctr=0;ctr<num;ctr++)
	{
	t = getw(ifp);
	o = find_id(t);
	if(!o)
		ithe_panic("Oh well whatever","Never mind");
	fread(buf,1,32,ifp);
	t = getw(ifp);
	ActivityName(o,buf,find_id(t));
	}

// Load usedata

num = getw(ifp);
ctr2 = getw(ifp);
for(ctr=0;ctr<num;ctr++)
    {
    t =	getw(ifp);
    o =	find_id(t);
    if(!o)
	  ithe_panic("Load_miscellaneous_state (usedata): Can't find object","Savegame corrupted");
    fread(o->user,1,ctr2,ifp);
    }

// load record of what we've said to NPCs

num = getw(ifp);
ilog_quiet("Reading %d npctalks\n",num);
for(ctr=0;ctr<num;ctr++)
	{
ilog_quiet("%d\n",ctr);
	t = getw(ifp);
ilog_quiet("id = %d\n",ctr);
	o = find_id(t);
	if(!o)
		ithe_panic("Load_miscellaneous_state (NPCtalk): Can't find object","Savegame corrupted");
ilog_quiet("o = %s\n",o->name);
	t = getw(ifp); // number of entries for this NPC
ilog_quiet("entries for NPC = %d\n",t);
	for(ctr2=0;ctr2<t;ctr2++)
		{
		ctr3=getw(ifp);
		fread(buf,1,(unsigned char)getc(ifp),ifp);
ilog_quiet("Load_NPCtalk: %d\n",ctr3);
		o2 = find_id(ctr3);
		if(!o2)
			ithe_panic("Load_miscellaneous_state (NPCtalk): Can't find 'player'","Savegame corrupted");
		NPC_BeenRead(o,buf,o2);
		}
	}

// load record of local NPC flags

num = getw(ifp);
for(ctr=0;ctr<num;ctr++)
	{
	t = getw(ifp);
	o = find_id(t);
	if(!o)
		ithe_panic("Load_miscellaneous_state (lFlags): Can't find object","Savegame corrupted");
	t = getw(ifp); // number of entries for this NPC
	for(ctr2=0;ctr2<t;ctr2++)
		{
		ctr3=getw(ifp);
		fread(buf,1,(unsigned char)getc(ifp),ifp);
		o2 = find_id(ctr3);
		if(!o2)
			ithe_panic("Load_miscellaneous_state (lFlags): Can't find 'player'","Savegame corrupted");
		NPC_set_lFlag(o,buf,o2,1);
		}
	}

// Load wielded objects

num = getw(ifp);
ctr2 = getw(ifp);
for(ctr=0;ctr<num;ctr++)
    {
    t = getw(ifp);
    o =	find_id(t);
    if(!o)
        ithe_panic("Load_miscellaneous_state (wieldtab): Can't find object","Savegame corrupted");
    for(ctr3=0;ctr3<ctr2;ctr3++)
        {
        o->wield[ctr3]=NULL;    // Default to blank
        t = getw(ifp);
        if(t != 0xffffffff)
            {
            o2 = find_id(t);
            if(!o2)
	        ithe_panic("Load_miscellaneous_state (wieldtab2): Can't find object","Savegame corrupted");
            o->wield[ctr3] = o2;
            }
        }
    }

fclose(ifp);
}

/*
 *      save_ms - Write miscellaneous savegame data
 */

void save_ms(char *filename, char *title)
{
long i;
int ctr,ctr2,ctr3,ok;
OBJLIST	*o;
USEDATA	empty;
NPC_RECORDING *n;
FILE *ofp;

ofp=fopen(filename,"wb");
if(!ofp)
    {
    if(!in_editor)
	{
	ilog_printf("Oh no!\n");
	ilog_printf("Could not create file '%s'\n",filename);
	}
    return;
    }

//      Write the header

fwrite(title,1,40,ofp);			        // Description
fwrite(COOKIE_MS,1,40,ofp);			// Cookie

putw(player->save_id,ofp);			// The Player
ilog_quiet("player id = %d\n",player->save_id);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
    {
    i=0;			// Get the member's ID without blowing up
    if(party[ctr])		// if the member does not exist.
	i=party[ctr]->save_id;
    putw(i,ofp);		// Write the ID number, or 0 for N/A
    }

putw(STORE,ofp);
fwrite(storage,1,STORE,ofp);

// Flags: find all flags which are TRUE

i=0;
for(ctr=0;ctr<MAXFLAGS;ctr++)
    if(tFlag[ctr])
	{
	if(!tFlagIdentifier[ctr])
	    {
	    Bug("Flag is TRUE but has no identifier!\n");
	    break;
	    }
	i++;
	}
putw(i,ofp);

// Now write the names to disk

for(ctr=0;ctr<MAXFLAGS;ctr++)
    if(tFlag[ctr])
	{
	i = strlen(tFlagIdentifier[ctr]);
	putc((char)i,ofp);
	fwrite(tFlagIdentifier[ctr],1,i,ofp);
	}

// Save the time and date

putw(game_minute,ofp);
putw(game_hour,ofp);
putw(game_day,ofp);
putw(game_month,ofp);
putw(game_year,ofp);
putw(days_passed,ofp);
putw(day_of_week,ofp);

// Save the goal data

i=0;
for(o=ActiveList;o;o=o->next)
   if(o->ptr->activity)
   if(o->ptr->save_id > 0)
       i++;

putw(i,ofp);
for(o=ActiveList;o;o=o->next)
	if(o->ptr->activity != -1)
		if(o->ptr->save_id > 0)
			{
			putw(o->ptr->save_id,ofp);
			fwrite(PElist[o->ptr->activity].name,1,32,ofp);
			if(!o->ptr->target)
				putw(0xffffffff,ofp);
			else
				putw(o->ptr->target->save_id,ofp);
			}

// Save usedata

memset(&empty,0,sizeof(USEDATA));

// count blank ones

i=0;
for(o=MasterList;o;o=o->next)
   if(o->ptr->save_id > 0)
       if(memcmp(o->ptr->user,&empty,sizeof(USEDATA)))
           i++;

putw(i,ofp);
putw(sizeof(USEDATA),ofp);

for(o=MasterList;o;o=o->next)
   if(o->ptr->save_id > 0)
       if(memcmp(o->ptr->user,&empty,sizeof(USEDATA)))
           {
           putw(o->ptr->save_id,ofp);
           fwrite(o->ptr->user,1,sizeof(USEDATA),ofp);
           }

// Count number of NPC's we've spoken to.  This can become fragmented with
// blanks so the detection routine is a little strange

i=0;
for(o=MasterList;o;o=o->next)
	if(o->ptr->stats->npctalk)
		{
		ok=0; // Did we speak with words, or silence?
		for(n=o->ptr->stats->npctalk;n;n=n->next)
			if(n->player != NULL && n->page[0] != '\0')
				ok=1;
		if(ok)
			i++;
		}

putw(i,ofp);    // Store the count
ilog_quiet("Estimate %d npctalks\n",i);

// Write all the pages we've seen for each NPC

for(o=MasterList;o;o=o->next)
	if(o->ptr->stats->npctalk)
		{
		i=0;
		for(n=o->ptr->stats->npctalk;n;n=n->next)
			if(n->player != NULL && n->page[0] != '\0')
				i++;
		ilog_quiet("writing %d entries for NPC\n",i);
		if(i>0)
			{
			putw(o->ptr->save_id,ofp);
			putw(i,ofp);
			for(n=o->ptr->stats->npctalk;n;n=n->next)
				if(n->player != NULL && n->page[0] != '\0')
					{
//		ilog_quiet("Save_NPCtalk: player type %s\n",n->player->name);
///		ilog_quiet("Save_NPCtalk: player name %s\n",n->player->personalname);
		ilog_quiet("Save_NPCtalk: %d\n",n->player->save_id);
					putw(n->player->save_id,ofp);
					i=strlen(n->page)+1;
					putc((unsigned char)i,ofp);
					fwrite(n->page,1,i,ofp);
					}
			}
		}

// Count number of NPC's with lFlags
// Again this may be fragmented.

i=0;
for(o=MasterList;o;o=o->next)
	if(o->ptr->stats->lFlags)
		{
		ok=0; // Did we speak with words, or silence?
		for(n=o->ptr->stats->lFlags;n;n=n->next)
			if(n->player != NULL && n->page[0] != '\0')
				ok=1;
		if(ok)
			i++;
		}

putw(i,ofp);    // Store the count

// Write all the lFlags for each NPC

for(o=MasterList;o;o=o->next)
	if(o->ptr->stats->lFlags)
		{
		i=0;
		for(n=o->ptr->stats->lFlags;n;n=n->next)
			if(n->player != NULL && n->page[0] != '\0')
				i++;
		if(i>0)
			{
			putw(o->ptr->save_id,ofp);
			putw(i,ofp);
			for(n=o->ptr->stats->lFlags;n;n=n->next)
				if(n->player != NULL && n->page[0] != '\0')
					{
					putw(n->player->save_id,ofp);
					i=strlen(n->page)+1;
					putc((unsigned char)i,ofp);
					fwrite(n->page,1,i,ofp);
					}
			}
		}

// Count number of characters wielding objects

i=0;
for(o=MasterList;o;o=o->next)
	{
	ctr2=0;
	for(ctr=0;ctr<9;ctr++)
		if(o->ptr->wield[ctr])
			ctr2=1;
	if(ctr2)
		i++;
	}

putw(i,ofp);    // Store the count
putw(9,ofp);    // 9 objects in the wieldtab in this version, maybe more later

// Store the wielded objects

for(o=MasterList;o;o=o->next)
	{
	ctr2=0;
	for(ctr=0;ctr<9;ctr++)
		if(o->ptr->wield[ctr])
			ctr2=1;
	if(ctr2)
		{
		putw(o->ptr->save_id,ofp);
		for(ctr3=0;ctr3<9;ctr3++)
			if(o->ptr->wield[ctr3])
				putw(o->ptr->wield[ctr3]->save_id,ofp);  // Store ID number
			else
				putw(0xffffffff,ofp);    // No object here
		}
	}

fclose(ofp);

}

/*
 *      query_ms - Query miscellaneous savegame data, get title
 */

char *query_ms(char *filename)
{
char *title;
char cookie[40];
FILE *ifp;

title=strLocalBuf(); // Allocate a 'local' buffer

ifp=fopen(filename,"rb");
if(!ifp)
    return NULL;

//      Write the header

fread(title,1,40,ifp);			 // Skip title
fread(cookie,1,40,ifp);			 // Cookie
fclose(ifp);

if(strcmp(COOKIE_MS,cookie))
    return NULL;

return title;
}


//======================================================================
//===========  Auxiliary functions for map IO systems ==================
//======================================================================

/*
 *    MoveToPocketEnd()  -  Move an object into the end of someone's pocket
 *                          Only for objects in the main objlist
 */

void MoveToPocketEnd(OBJECT *object, OBJECT *container)
{
OBJECT *temp;

// Check for silliness.  This error has never happened to me, but if it
// does I thought that you should know.

if(!object || !container)
    {
    Bug("MoveToPocketEnd(%x,%x) attempted\n",object,container);
    return;
    }

// Look for the object in the list

for(temp=curmap->object;temp->next;temp=temp->next)
	if(temp->next==object)
		{
		LL_Remove(&curmap->object->next,object);
		LL_Add(&container->pocket,object);	// Add to end
		object->flags.inpocket=1;
		object->flags.fixed=1;
		return;
		}
ithe_panic("MoveToPocketEnd: This object was not in the list:",object->name);
}


/*
 *      SumContainers - Set up objects in pockets so they point to their
 *                      parent.  Recursive, called by save_z1
 */

void SumContainers(OBJECT *cont, long parent)
{
for(OBJECT *temp = cont;temp;temp=temp->next)
    {
    temp->save_id = ++objctr;
    temp->inside_id = parent;
    if(temp->pocket)
	SumContainers(temp->pocket,temp->save_id);
    }
return;
}


/*
 *      WriteContainers - Write contents containers to disk.
 *                        Recursive, called by save_z1
 */

void WriteContainers(OBJECT *cont)
{
SEQ_POOL *form;
OBJECT *temp;

for(temp = cont;temp;temp=temp->next)
	{
#ifdef LOG_RW
    ilog_quiet("Object ID %d (%s) is inside %d\n", temp->save_id,temp->name,temp->inside_id);
#endif
	// Substitute id number for FORM pointer
	form = temp->form;
	temp->form = (SEQ_POOL *)getnum4sequence(temp->form->name);

//    fwrite(temp,1,sizeof(OBJECT),ofp);
	write_OBJECT(temp,ofp);
	iwrite((unsigned char *)temp->name,32,ofp);

	// then switch it back again
	temp->form = form;
	if(temp->pocket)
		WriteContainers(temp->pocket);
	}

return;
}

/*
 *      WriteContainerSchedules - Write Schedules of container contents
 */

void WriteContainerSchedules(OBJECT *cont)
{
OBJECT *old,*temp;
char ok;
int ctr;

for(temp = cont;temp;temp=temp->next)
	{
	ok=0;
	if(temp->schedule)
		for(ctr=0;ctr<24;ctr++)
			if(temp->schedule[ctr].hour	!= -1)
				ok=1;

	if(!ok)
		iputl_i(0xffffffff,ofp);
	else
		{
		iputl_i(temp->save_id,ofp);
		for(ctr=0;ctr<24;ctr++)
			{
			// Convert the pointer to an ID number for the saving
			old	= temp->schedule[ctr].target;
			if(old)
				temp->schedule[ctr].target=(OBJECT	*)temp->schedule[ctr].target->save_id;
			// Save it
			iwrite((unsigned char *)&temp->schedule[ctr],sizeof(SCHEDULE),ofp);
			// Restore the original
			temp->schedule[ctr].target=old;
			}
		}
	if(temp->pocket)
		WriteContainerSchedules(temp->pocket);
	}

return;
}

/*
 *      WriteContainerStats - Write Stats of container contents
 */

void WriteContainerStats(OBJECT	*cont)
{
OBJECT *owner,*temp;

for(temp = cont;temp;temp=temp->next)
	if(has_changed(temp))
		{
		iputl_i(temp->save_id,ofp);
		write_STATS(temp->stats,ofp);
		if(temp->pocket)
			WriteContainerStats(temp->pocket);
		}
	else
		{
		iputl_i(0xffffffff,ofp);
		if(temp->pocket)
			WriteContainerStats(temp->pocket);
		}

return;
}

/*
 *      WriteContainerFuncs - Write custom functions of container contents
 */

void WriteContainerFuncs(OBJECT	*cont)
{
OBJECT *temp;
for(temp = cont;temp;temp=temp->next)
	{
	if(temp->funcs->dirty)
		{
		iputl_i(temp->save_id,ofp);
		write_FUNCS(temp->funcs,ofp);
		}
	if(temp->pocket)
		WriteContainerFuncs(temp->pocket);
	}

return;
}


/*
 *      WriteContainerNames - Write personalname of container contents
 */

void WriteContainerName(OBJECT *cont)
{
OBJECT *temp;

for(temp = cont;temp;temp=temp->next)
	{
	iputl_i(temp->save_id,ofp);
	iwrite((unsigned char *)temp->personalname,32,ofp);
	if(temp->pocket)
		WriteContainerName(temp->pocket);
	}

return;
}


/*
 *      Find out whether an object's STATS have changed from the defaults
 */

int has_changed(OBJECT *a)
{
int pos;
STATS *b,*c;

#define	CHECK(x) if(b->x != c->x) return 1;

if(!a)
    {
    Bug("has_changed was passed NULL\n");
    return 0;
    }

pos = getnum4char(a->name);
if(pos == -1)
    {
    Bug("Entity %s does not officially exist\n",a->name);
    return 0;
    }

b = a->stats;
c = CHlist[pos].stats;

// It will return if any of these checks discovers a Change

CHECK(hp);
CHECK(dex);
CHECK(str);
CHECK(intel);
CHECK(weight);
CHECK(damage);
CHECK(quantity);
CHECK(karma);
CHECK(bulk);

// Check Flags (using a combination of cunning and extreme violence)
if(*(unsigned int*)&b->npcflags != *(unsigned int *)&c->npcflags)
    return 1;

// It has an owner: this is a Change
if(b->owner != NULL)
    return 1;

// We do not need to check the NPC_RECORDs for changes, they are savegame only
// and will be dealt with separately.

return 0;
}


// Fast Find object ID

static void fastfind_id_start()
{
OBJLIST *t;
int ctr;

if(fastfind_id_list)
	return;

ctr=0;
for(t=MasterList;t;t=t->next)
	if(t->ptr)
		ctr++;

fastfind_id_num = ctr;
fastfind_id_list = (OBJECT **)M_get(ctr,sizeof(OBJECT *));

ctr=0;
for(t=MasterList;t;t=t->next)
	if(t->ptr)
		fastfind_id_list[ctr++]=t->ptr;

qsort(fastfind_id_list,fastfind_id_num,sizeof(OBJECT *),CMP_sort);
}


static void fastfind_id_stop()
{
if(!fastfind_id_list)
	return;
M_free(fastfind_id_list);
fastfind_id_list = NULL;
}


static OBJECT *fastfind_id(unsigned int id)
{
OBJECT **xx;

if(!fastfind_id_list)
	ithe_panic("FastFind_ID called before FastFind_ID_start",NULL);

xx=(OBJECT **)bsearch((const void *)id,fastfind_id_list,fastfind_id_num,sizeof(OBJECT *),CMP_search);
if(xx)
	return *xx;
return NULL;
}

static int CMP_sort(const void *a, const void *b)
{
return (*(OBJECT **)a)->save_id - (*(OBJECT **)b)->save_id;
}

static int CMP_search(const void *a, const void *b)
{
return (int)a - (*(OBJECT **)b)->save_id;
}

