/*
 *      IRE game runner main module
 */

// Debugging switches

//#define DEBUG_SCHEDULE
//#define DEBUG_LARGE_OBJECTS
//#define DEBUG_SOLID_OBJECTS
//#define DEBUG_RESUMEACTIVITY
//#define CHECK_ACTIVELIST
//#define CHECK_MASTERLIST

#include <stdlib.h>
#include <sys/types.h>

#if !defined(__BEOS__) && !defined(_WIN32)
#include <sys/mman.h>
#endif
#include <string.h>
#include <time.h>

#ifdef _WIN32
#include <process.h>
#endif

#include "ithelib.h"
#include "media.h"
#include "core.hpp"
#include "gamedata.h"
#include "script.hpp"
#include "init.h"
#include "oscli.h"
#include "console.h"
#include "linklist.hpp"
#include "project.hpp"
#include "bitmap.h"
#include "object.hpp"
//#include "keys.hpp"
#include "loadsave.hpp"
#include "sound.h"
#include "nuspeech.hpp"
#include "map.hpp"
#include "pe/pe_api.hpp"
#include "darken.h"

// defines

#define VIEWX 16
#define VIEWY 16
#define CURSOR_CHAR "_"

// variables


extern char *darkmap,*darkrect;
extern char AL_dirty;
extern int destroyhp;
int do_lightning=0,waiting=0;
char user_input[128];


unsigned long fpsmetric=0;
time_t startmetric, stopmetric;
double timemetric;


char *savegame="savemisc.000";

// functions

void DrawMap(int x,int y);              // Draw the map and all the things
extern void p_light(char *screen);
extern char *BestName(OBJECT *o);

extern void Call_VRM(int v);
extern "C" int DebugVM;
int getYN(char *q);
int get_key();
int get_key_debounced();
int get_num(int no);

OBJECT *GetSolidObject(int x,int y);
OBJECT *GetTopObject(int x,int y);
void GetNearDirection(char *message);
void CheckTime();
void ResumeSchedule(OBJECT *o);

//static void SetDarkness(int r);

//extern void Project_Darkness(int x, int y);                // Render darkness
extern void gen_largemap();
extern void SoundSettings();
extern void StoreObjects();
void Restart();
extern int choose_leader(int x);
extern "C" void (*lightning)(BITMAP *img);

// code

/*
 *      IRE runner
 */

void startgame()
{
char running;
OBJECT *current;
OBJECT *temp;
TILE *chktile;
//OBJECT **ptr;
OBJLIST *active,*next;
char non_vehicle_in_party = 0;
int vx,vy,ctr;
char msg[80];

game_hour = ig_hour;
game_minute = ig_min;
game_day = ig_day;
game_month = ig_month;
game_year = ig_year;

ilog_quiet("Allocating party name registrar\n");

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	partyname[ctr]=(char *)M_get(1,128); // Should be plenty

ilog_quiet("Allocating storage dump\n");

storage = (int *)M_get(STORE,sizeof(int));

syspocket = curmap->object;
curmap->object->flags.on=1;

running = 1;
player = NULL;
memset(party,0,sizeof(OBJECT *) * MAX_MEMBERS);

ilog_quiet("Set up conversation system\n");
NPC_Init();

ilog_quiet("Init projector\n");
Init_Projector(VSW32,VSH32);

ilog_quiet("Seeking player\n");

for(int matx=0;matx<curmap->w;matx++)
	for(int maty=0;maty<curmap->h;maty++)
		for(temp=curmap->objmap[ytab[maty]+matx];temp;temp=temp->next)
			if(!stricmp(temp->name,"player"))
				player = temp;

if(!player)
	ithe_panic("Could not find any players at all!",NULL);

party[0] = player;      // Init party
strcpy(partyname[0],BestName(player));
player->flags.party=1;  // Make sure

//ilog_quiet("%d bytes free as we start the game:\n",coreleft());

ilog_quiet("Player =%x\n",player);
ilog_quiet("Player action =%d\n",player->activity);
if(player->activity > 0)
	ilog_quiet("Player action: %s\n",PElist[player->activity].name);
else
	ilog_quiet("No player action\n");

ilog_quiet("Calling PE: mainproc\n");

CallVM("mainproc");

S_MusicVolume(mu_volume);

ilog_quiet("Init main game engine\n");

ilog_printf("Starting main game engine\n");

CentreMap(player);      // Centre on the player

if(IsTileWater(0,0))
	Bug("Tile at 0,0 is water.. may cause problems\n");

fpsmetric=0;
time(&startmetric);

// The main loop begins here

do {
	if(show_vrm_calls)
		{
		ilog_quiet("\n");
		ilog_quiet("Start of turn\n");
		ilog_quiet("\n");
		}

	CallVMnum(vscheduler);     // Call initial function
	CallVMnum(status);
	gen_largemap();
	DrawMap(mapx,mapy);
	irecon_update();

	irekey = 0;
	irekey = get_key();

	if(irekey == (KEY_F3 | 0x80))
		{
		irecon_cls();
		irecon_printf("Create Object.. Enter name:\n");
		strcpy(user_input,"");
		if(irecon_getinput(user_input,32))
			if(getnum4char_slow(user_input)>0)
				{
				temp = OB_Alloc();
				temp->curdir = temp->dir[0];
				OB_Init(temp,user_input);
				OB_SetDir(temp,1);
				MoveToMap(player->x,player->y,temp);
				CreateContents(temp);
				}
			else
				{
				irecon_printf("unknown object\n");
				}
		}

	if(irekey == (KEY_F2 | 0x80))
		{
		DebugVM=!DebugVM;
		irecon_printf("VM Debugging now %s\n",DebugVM?"ON":"OFF");
		}


	if(waiting)
		{
		DrawMap(mapx,mapy);
		Show();
		S_PollMusic();
		waiting--;
		}

// Move all things

	if(show_vrm_calls)
		ilog_quiet("Moving Things..\n");

//	current=current_object;

// Move the player first

	current_object = NULL;//player;   // Set up parameters for the VRM
	person = player;           // Who am I?
	CallVMnum(update_life);
	if(player->activity != -1)
		CallVMnum(player->activity);
	else
		Bug("Player doesn't know what to do\n");

// Reset update flags
	for(active=ActiveList;active;active=active->next)
		{
		active->ptr->flags.didupdate=0;
#ifdef CHECK_ACTIVELIST
		if(!OB_Check(active->ptr))
			{
			Bug("Uck: object->name = %x\n",active->ptr->name);
			Bug("Uck: object->name = %s\n",active->ptr->name);
			Bug("Uck: object->personalname = %x\n",active->ptr->personalname);
			Bug("Uck: object->personalname = %s\n",active->ptr->personalname);
			ithe_panic("Encountered reality dysfunction",NULL);
			}
		if(active->ptr->activity == -1)
			{
			Bug("ActiveList: %s is not active\n",active->ptr->name);
			ML_Del(&ActiveList,active->ptr);
			active=ActiveList;
			}
#endif
		}

#ifdef CHECK_MASTERLIST
	for(active=MasterList;active;active=active->next)
		{
		if(!active->ptr)
			{
			Bug("MasterList: NULL detected\n");
			ML_Del(&MasterList,active->ptr);
			active=MasterList;
			}
		// If this happens we are badly ____ed (or SEER is broken again)
		if(active->ptr == (OBJECT *)0xdeadbeef)
			{
			Bug("MasterList: DEAD BEEF detected\n");
			ML_Del(&MasterList,active->ptr);
			active=MasterList;
			}
		if(active->ptr == (OBJECT *)0xbeefdead)
			{
			Bug("MasterList: BEEF DEAD detected\n");
			ML_Del(&MasterList,active->ptr);
			active=MasterList;
			}
		}
#endif

	player->flags.didupdate=1; // Player can't move again

// Now move all other objects
	active=ActiveList;
	if(active)
		do {
			next = active->next;
			if(active->ptr->flags.on)
				if(!active->ptr->flags.didupdate) // Don't do this one again if we need
					{                               // to restart the update
#ifdef CHECK_ACTIVELIST
	#ifndef __BEOS__ // BeOS can give addresses < 1000h and they are valid
					if((unsigned int)active->ptr->stats < 0x1000)
						{
						Bug("CORRUPTION of object [%x]\n",active->ptr);
						break;
						}
	#endif
#endif
					active->ptr->stats->oldhp = active->ptr->stats->hp;
					AL_dirty=0;                     // Mark list as clean
					current_object = NULL;//active->ptr;   // Set up parameters for the VRM
					person = active->ptr;           // Who am I?
					active->ptr->flags.didupdate=1; // Mark object as moved
					if(active->ptr->stats->npcflags.biological)
						CallVMnum(update_life);		// Only for carbon-based life
					CallVMnum(active->ptr->activity);
					if(AL_dirty)                    // If list is dirty we need to
						next=ActiveList;            // start over.
					}
			active = next;
			} while(active);

	Fortify_CheckAllMemory(); // Debugging tool.  No effect unless compiled in

	game_minute++;

	CheckTime();

	for(active=MasterList;active;active=active->next)
		if(active->ptr->stats->hp>0)
			if(active->ptr->schedule)
				if(active->ptr->flags.on && !active->ptr->flags.party)
					for(vx=0;vx<24;vx++)
						{
						if(active->ptr->schedule[vx].hour == game_hour)
							{
							if(active->ptr->schedule[vx].minute == game_minute)
								{
#ifdef DEBUG_SCHEDULE
irecon_printf("%s does %s to %s\n",active->ptr->name,active->ptr->schedule[vx].vrm,active->ptr->schedule[vx].target?active->ptr->schedule[vx].target->name:"Nothing");
#endif
								lastvrmcall="Activity Engine";
								ActivityNum(active->ptr,active->ptr->schedule[vx].call,active->ptr->schedule[vx].target);
								lastvrmcall="N/A";
								}
							}
						}
// Determine roof state

	show_roof = 1;       // Assume we don't need to (0 = show, 1 = hide)

	if(curmap->roof[(player->y*curmap->w)+player->x])
		if(RTlist[curmap->roof[(player->y*curmap->w)+player->x]].name)
			show_roof=0;

//
// Kill dead things
//

	if(show_vrm_calls)
		ilog_quiet("Admin death and destruction..\n");

	if(player->stats->hp < 0)
		CallVMnum(status);			// show the player their dead HP

	// Make sure we have at least one intelligent being in the party otherwise
	// there might be just the car or boat they were in when they died

		non_vehicle_in_party = 0;    // Assume worst-case scenario

		for(ctr=0;ctr<MAX_MEMBERS-1;ctr++)
			if(party[ctr])
				if(party[ctr]->stats->hp > 0)            // Still living
					if(party[ctr]->stats->intel > 0)     // Not a vehicle
						non_vehicle_in_party=1;

	// Any party members who are found to be dead get their membership revoked

	for(ctr=0;ctr<MAX_MEMBERS-1;ctr++)
		if(party[ctr])
			if(party[ctr]->stats->hp <1)
				{
				irecon_printf("%s is dead!\n",partyname[ctr]);
				if(player == party[ctr])         // Sync map before player
					CentreMap(player);        // gets erased
				SubFromParty(party[ctr]);
				}

	// Is there no-one at the helm of our vehicle?

	if(non_vehicle_in_party == 0)
		if(player)
			if(player->stats->intel < 1)
				{
				CentreMap(player);
				SubFromParty(player);
				player=NULL;
				}

// Call the game_over code.  This may reinit the world, or just resurrect
// the player elsewhere.

	if(!player)
		{
		DrawMap(mapx,mapy);
		if(show_vrm_calls)
			ilog_quiet("all fall down\n");
		Call_VRM(all_dead);
		}

	DeletePending();        // Destroy And Be Free!  Destroy And Be Free!

// Check internal system keys

	if(irekey == KEY_F4)       // F4 - sound volume
		SoundSettings();

	if(irekey == KEY_F2)
		if(getYN("Save game"))
			{
			savegame = "savemap1.000";
			save_z1(savegame);
			savegame = "savemisc.000";
			save_ms(savegame,"Default title");
			}

	if(irekey == KEY_F3)
		{
		savegame="savemisc.000";
		if(query_ms(savegame))
			{
			strcpy(msg,"Restore your game called \"");
			strcat(msg,query_ms("savemisc.000"));
			strcat(msg,"\"");
			if(getYN(msg))
				{
				irecon_printf("Please wait..\n");
				irecon_update();
				Show();
				ilog_quiet("\nRELOAD\n\n");
//                ilog_quiet("before memory %d\n",coreleft());
				wipe_z1();
				fullrestore = 1;
				savegame = "savemap1.000";
				load_z1(savegame);
				savegame = "savemisc.000";
				load_ms(savegame);
				fullrestore = 0;
				CallVM("mainproc");
				Call_VRM(loadproc);
//                ilog_quiet("after memory %d\n",coreleft());
				}
			}
		else
			irecon_printf("Could not find savegame '%s'\n",savegame);
		}

	if(irekey == KEY_ESC)
		{
		if(getYN("Quit game"))
			running = 0;
		}

/*
	if(irekey == KEY_V)
		{
		ilog_quiet("Masterlist: \n");
		for(active=MasterList;active;active=active->next)
			ilog_quiet("0x%x (%s)\n",active->ptr,active->ptr->name);
		}
*/
	if(irekey == KEY_W)
		{
		irecon_printf("Wait how many minutes?\n");
		irecon_update();
		Show();
		get_num(60);
		waiting = atoi(user_input);
		}

	if(irekey == KEY_F10)
		{
		irecon_printf("Taking screenshot..\n");
		irecon_update();
		Show();
		BMPshot();
		}
	CentreMap(player);
	} while(running);

time(&stopmetric);

irecon_term();

ilog_quiet("Game is quit by user\n");

timemetric = difftime(stopmetric, startmetric);
ilog_quiet("Measured %f fps\n",(double)fpsmetric/timemetric);

}



/* ============================== Map Manipulation ======================== */

/*
 *      DrawMap - Display the map on screen
 */

void DrawMap(int x,int y)
{
if(x+VSW > curmap->w)
	x=curmap->w-VSW;
if(y+VSH > curmap->h)
	y=curmap->h-VSH;

Project_Map(x,y);
Project_Sprites(x,y);
if(show_roof)
	Project_Roof(x,y);
// Do the shading
if(use_light)
	render_lighting(x,y);
// Block invisible tiles
if(vis_blanking)
	Hide_Unseen();
// Combine the roof with the main window
if(show_roof)
	darken_blit((unsigned char *)gamewin->line[0],(unsigned char *)roofwin->line[0],gamewinsize);
// Add a lightning effect if we need it
if(do_lightning)
	{
	if(use_light)
		lightning(gamewin);
	do_lightning--;
	}

// visual debugger for Large Objects.   Not for production, as you will
// see very quickly if you enable it.

#ifdef DEBUG_LARGE_OBJECTS
for(int vy=0;vy<VSW;vy++)
    for(int vx=0;vx<VSH;vx++)
        if(largemap[vx+VIEWDIST][vy+VIEWDIST])
            rectfill(gamewin,(vx<<5),(vy<<5),(vx<<5)+32,(vy<<5)+32,makecol(255,255,255));
#endif
#ifdef DEBUG_SOLID_OBJECTS
for(int vy=0;vy<VSW;vy++)
    for(int vx=0;vx<VSH;vx++)
        if(solidmap[vx+VIEWDIST][vy+VIEWDIST])
            rectfill(gamewin,(vx<<5),(vy<<5),(vx<<5)+32,(vy<<5)+32,makecol(255,255,255));
#endif

// Finally, draw the thing, possibly darker than normal if we're fading
if(Fader)
	draw_lit_sprite(swapscreen,gamewin,VIEWX,VIEWY,Fader);
else
	draw_sprite(swapscreen,gamewin,VIEWX,VIEWY);

fpsmetric++;
}


/*
 *      getYN - Ask a question, return TRUE if the answer is Y
 */

int getYN(char *q)
{
irecon_printf("%s- Are you sure?  ",q);
irecon_update();
Show();
if(get_key() != KEY_Y)
    {
    irecon_printf("No.\n");
    irecon_update();
    Show();
    return 0;
    }
else
    {
    irecon_printf("Yes.\n");
    irecon_update();
    Show();
    return 1;
    }
}

/*
 *      get_key - Get a key from the user, animate things in the background
 */

int get_key()
{
int input=0;

if(waiting)
	return 0;

do  {
	DrawMap(mapx,mapy);
	Show();
	S_PollMusic();
	input = GetKey();
	} while(!input);
return input;
}

/*
 *      get_key_debounced - Get a key, wait for it to be released
 */

int get_key_debounced()
{
int input=0;

//FlushKeys();
input = get_key();
FlushKeys();

return input;
}

/*
 *      get_num - User types in a number, animate things in the background
 */

int get_num(int no)
{
int input=0;
int bufpos=0;
char buf[128];

itoa(no,buf,10);
bufpos=strlen(buf);
buf[bufpos]=0;
do  {
	DrawMap(mapx,mapy);
	irecon_clearline();
	if(buf[0] != 0)
		irecon_print(buf);         // Printing an empty line makes a newline
	irecon_print(CURSOR_CHAR);
	irecon_update();
	Show();
	S_PollMusic();
	if(keypressed())
		{
		input = readkey();
		if((input>>8) == KEY_F10)
			BMPshot();
		if((input&0xff) >= '0' && (input&0xff) <= '9')
			{
			if(bufpos<conwid)
				{
				buf[bufpos++]=input&0xff;
				buf[bufpos]=0;
				}
			}
		if((input>>8) == KEY_DEL || (input>>8) == KEY_BACKSPACE)
			if(bufpos>0)
				{
				bufpos--;
				buf[bufpos]=0;
				}
		}
	else
		input=0;
	} while((input>>8) != KEY_ESC && (input>>8) != KEY_ENTER);
	irecon_clearline();
irecon_print(buf);
irecon_print("\n");
if((input>>8) == KEY_ESC)
	return 0;
strcpy(user_input,buf);
return 1;
}

/*
 *      RedrawMap - Redraw the map and update
 */

void RedrawMap()
{
CentreMap(player);
DrawMap(mapx,mapy);
Show();
if(GetKey()==KEY_F10)
	BMPshot();
}


/*
 *      Restart - Restart the game
 */

void Restart()
{
char test[32];
OBJECT *temp;
int ctr;

game_hour = ig_hour;
game_minute = ig_min;
game_day = ig_day;
game_month = ig_month;
game_year = ig_year;

Wipe_tFlags();

wipe_z1();

fullrestore = 0;
strcpy(test,mapname);
strcat(test,".mz1");
load_z1(test);
fullrestore = 0;

// Find player

player = NULL;
for(int matx=0;matx<curmap->w;matx++)
    for(int maty=0;maty<curmap->h;maty++)
        for(temp=curmap->objmap[ytab[maty]+matx];temp;temp=temp->next)
            if(!stricmp(temp->name,"player"))
                player = temp;
if(!player)
    ithe_panic("Cannot find the player anymore",NULL);

ilog_quiet("player = %x\n",player);
ilog_quiet("player->stats->hp = %d\n",player->stats->hp);

CallVM("mainproc");
Call_VRM(loadproc);

for(ctr=0;ctr<MAX_MEMBERS-1;ctr++)
    {
    party[ctr]=0;
    partyname[ctr][0]='\0';
    }
party[0]=player;
strcpy(partyname[0],BestName(player));
player->flags.party=1;
}

void CheckTime()
{
if(game_minute >= min_hour)
    {
    game_hour += game_minute/min_hour;
    game_minute %= min_hour;
    }

if(game_hour >= hour_day)
    {
    game_day += game_hour/hour_day;
    game_hour %= hour_day;
    }

if(game_day >= day_month)
    {
    game_month += game_day/day_month;
    game_day %= day_month;
    }

if(game_month >= month_year)
    {
    game_year += game_month/month_year;
    game_month %= month_year;
    }

days_passed = ((((game_year - ig_year)*month_year)+game_month)*day_month)+game_day;
day_of_week = days_passed % day_week;

}

void ResyncEverything()
{
OBJLIST *t;
for(t=MasterList;t;t=t->next)
    ResumeSchedule(t->ptr);
}


/*
 *      Make an NPC do whatever they were doing before they were distracted
 */

void ResumeSchedule(OBJECT *o)
{
int i,h,ctr;

#ifdef DEBUG_RESUMEACTIVITY
#define DEBUG_RESACT_PRINT(x,y) ilog_quiet(x,y);
#else
#define DEBUG_RESACT_PRINT(x,y) ;
#endif


// Make sure it's valid

if(!o)
	{
	DEBUG_RESACT_PRINT("ResumeSchedule was passed %x\n",NULL);
	return;
	}

if(!o->flags.on)
	{
	DEBUG_RESACT_PRINT("%s is switched off\n",BestName(o));
	return;
	}

if(o->flags.party)
	{
	DEBUG_RESACT_PRINT("%s is of the party\n",BestName(o));
	return;
	}

if(o->stats->hp<=0)
	{
	DEBUG_RESACT_PRINT("%s is dead\n",BestName(o));
	return;
	}

if(!o->schedule)
	{
	DEBUG_RESACT_PRINT("%s has no schedule\n",BestName(o));
	return;
	}

DEBUG_RESACT_PRINT("Game Hour is %d\n",game_hour);

i = -1;
h = -1;
for(ctr=0;ctr<24;ctr++)
	{
	DEBUG_RESACT_PRINT("Event is %d\n",o->schedule[ctr].hour);
	if(o->schedule[ctr].hour < game_hour)
		if(o->schedule[ctr].hour > h)
			{
			i = ctr;
			h = o->schedule[ctr].hour;
			}
	if(o->schedule[ctr].hour == game_hour && o->schedule[ctr].minute < game_minute)
			{
			i = ctr;
			h = o->schedule[ctr].hour;
			}
	}

if(i>=0)
	{
	if(o->schedule[i].call == o->activity && o->schedule[i].target == o->target)
		{
		ilog_quiet("%s trying to resume current task\n",o->name);
		return;
		}
	ActivityNum(o,o->schedule[i].call,o->schedule[i].target);
	#ifdef DEBUG_RESUMEACTIVITY
	ilog_quiet("%s resumes doing %s to %s\n",o->name,o->schedule[i].vrm,o->schedule[i].target?o->schedule[i].target->name:"Nothing");
	#endif
	}
else
	DEBUG_RESACT_PRINT("%s couldn't decide what to do\n",BestName(o));
}




