/*
 *      Object manipulation routines, world management etc
 */

#include <string.h>

#include "ithelib.h"
#include "core.hpp"
#include "linklist.hpp"
#include "console.h"
#include "oscli.h"
#include "script.hpp"
#include "pe/pe_api.hpp"
#include "object.hpp"
#include "bitmap.h"
#include "init.h"
#include "gamedata.h"
#include "map.hpp"

// defines

#define TOO_MUCH 1000

#define CHECK_FOR_MOVE_UP_ARSE(X,Y); if(X==Y)\
    {\
    Bug("Object '%s' tried to vanish up it's own arse\n",Y->name);\
    return;\
    }

#define CHECK_FOR_PLAYER_GETTING_LOST(X); if(X == player)\
    {\
    return;\
    }

// variables

static char str[90];
char pending_delete=0;
char AL_dirty=0;
static int object_search_rec=0;

extern char in_editor;
extern "C" char *lastvrmcall;

// functions

extern void CallVM(char *function);
extern void Init_Areas(OBJECT *objsel);
extern void wipe_refs(void *o);
extern void ResumeSchedule(OBJECT *o);

int AddToParty(OBJECT *new_member);
TILE *GetTile(int x,int y);
int isSolid(int x,int y);
int ChooseLeader(int x);
void SubFromParty(OBJECT *member);
void gen_largemap();

void AddQuantity(OBJECT *container,char *type,int request);
int TakeQuantity(OBJECT *container,char *type,int request);
int MoveQuantity(OBJECT *src,OBJECT *dest,char *objectname,int qty);


/* Linked-list world functions */

void MoveToPocket(OBJECT *object, OBJECT *container);
void TransferToPocket(OBJECT *object, OBJECT *container);
int MoveFromPocket(OBJECT *object, OBJECT *container,int x,int y);

/* Matrix world functions */

void MoveToMap(int x, int y, OBJECT *object);
void TakeObject(int x,int y, OBJECT *container);
int MoveObject(OBJECT *object,int x,int y);
void TransferObject(OBJECT *object,int x,int y);
OBJECT *GetObject(int x,int y);
OBJECT *GetObjectBase(int x,int y);
OBJECT *GetRawObjectBase(int x,int y);
OBJECT *GetSolidObject(int x,int y);
void ForceDropObject(int x,int y, OBJECT *object);
void FindEmptySquare(int *x,int *y);
OBJECT *LookInAllPocketsForTag(int tag);
OBJECT *LookInAllPocketsFor(OBJECT *obj);
int InPocket(OBJECT *obj);
void CreateContents(OBJECT *obj);
void MoveToTop(OBJECT *object);
void UpdatePocketCoords(OBJECT *o, int x, int y);

static OBJECT *SearchContainerTag(OBJECT *cont, int tag);
static OBJECT *SearchContainer(OBJECT *cont, OBJECT *obj);
static int WeighObjectR(OBJECT *list);
static int GetBulkR(OBJECT *list);
OBJECT *GetFirstObject(OBJECT *cont, char *name);
int SumObjects(OBJECT *cont, char *name, int total);

void CheckRefs(void **p1, void *ptr1);

extern char *BestName(OBJECT *o);

// code


/*
 *    MoveToPocket()  -  Move an object into someone's pocket
 */

void MoveToPocket(OBJECT *object, OBJECT *container)
{
// Check for silliness.  This error has never happened to me, but if it
// does I thought that you should know.

if(!object || !container)
    {
    Bug("MoveToPocket(%x,%x) attempted\n",object,container);
    return;
    }

/*
if(!object->flags.on)      // It's going to be deleted.  Skip it.
    return;*/
if(!container->flags.on)   // NOOO!
    return;


CHECK_FOR_MOVE_UP_ARSE(container,object);
CHECK_FOR_PLAYER_GETTING_LOST(object);

TransferToPocket(object,container);
}

/*
 *    TransferToPocket()  -  Force an object into someone's pocket
 */

void TransferToPocket(OBJECT *object, OBJECT *container)
{
OBJECT *temp;
int ox,oy;

// Look for the object in nowhere land (the system pockets)

for(temp=curmap->object;temp->next;temp=temp->next)
        {
	if(temp->next==object)
		{
                // Good we found it.  We are looking at the entry before it.
                object->flags.inpocket=1;          // It has left our universe
                temp->next = object->next;         // object no longer in list
                object->next = container->pocket;  // Get of pocket->next
                container->pocket = object;        // Object now in pocket
//                ML_Del(&ActiveList,object);        // Idle while in pocket
		return;                            // Done
		}
        }

// Is it lying on the ground?

if(!InPocket(object))
    {
    ox = object->x;
    oy = object->y;
    TransferObject(object,0,0);
    temp = LookInAllPocketsFor(container);
    if(temp == object)
        {
        TransferObject(object,ox,oy);
        Bug("EMERGENCY: Object '%s' vanished up it's own arse\n",object->name);
        return;
        }
    TakeObject(0,0,container);
    object->x = ox;
    object->y = oy;
    return;
    }

// It must be in someone's pocket then

temp = LookInAllPocketsFor(object);
if(temp)
    {
    LL_Remove(&temp->pocket,object);
    object->next = container->pocket;  // Get of pocket->next
    container->pocket = object;        // Object now in pocket
    object->flags.inpocket=1;
    return;
    }

ithe_panic("move_to_pocket: This object could not be found:",object->name);
}


/*
 *    MoveFromPocket()  -  Move an object out of someone's pocket
 */

int MoveFromPocket(OBJECT *object, OBJECT *container, int x,int y)
{
OBJECT *temp;
int ok,wok=0;

// Check for silliness

if(!object || !container)
    return 0;
if(!container->pocket)
    return 0;

//if(!object->flags.on)      // It's going to be deleted.  Skip it.
//    return 0;
if(!container->flags.on)   // NOOO!
    return 0;

// If there is something solid there, abort, unless it's a container

if(isSolid(x,y))
    {
    temp=GetSolidObject(x,y);   // Find the obstruction
    if(!temp)                   // None? Must be a tile then
        {
//        C_printf("[!]\n");
        return 1;
        }
/*        {
        wok=0;
        tile = GetTile(x,y);
        C_printf("tile flags: %x\n",*(int*)(&tile->flags));

        // If the tile is water, then some objects can be dropped in it
        if(tile->flags.watery && !object->flags.person)
            {
            object->flags.shadow=1;
            wok=1;
            }
        if(tile->flags.watery && object->flags.watery)
            wok=1;
        if(!wok)
            return 0;
        }
*/
    // If the Water says it's ok, then do it, else perform more checks

    if(!wok)
        {
        ok=1;
        if(!temp->flags.tabletop)   // You can only drop on tabletops
            ok=0;                   // Disable
        if(temp->flags.container)   // Unless there is a bag there
            ok=1;                   // Re-enable
        if(!ok)                     // Quit if disabled
            return 0;
        }
    }

// Go through to be sure it is there.

for(temp = container->pocket;temp;temp=temp->next)
    if(temp == object)
        {
        LL_Remove(&container->pocket,object);
        DropObject(x,y,object);                     // Drop into the map
        return 1;
        }

return 0;
}

/*
 *    ForceFromPocket()  -  MoveFromPocket without checking for obstructions
 */

int ForceFromPocket(OBJECT *object, OBJECT *container, int x,int y)
{
OBJECT *temp;

// These checks are to prevent a crash

if(!object || !container)
    return 0;
if(!container->pocket)
    return 0;

//if(!object->flags.on)      // It's going to be deleted.  Skip it.
//    return 0;
if(!container->flags.on)   // NOOO!
    return 0;

for(temp = container->pocket;temp;temp=temp->next)
    if(temp == object)
        {
        LL_Remove(&container->pocket,object);
        ForceDropObject(x,y,object);                     // Drop into the map
        return 1;
        }

return 0;
}

/*
 *    MoveToMap(int x,int y,OBJECT *object)  -  Move object to the world
 */

void MoveToMap(int x, int y, OBJECT *object)
{
OBJECT *temp;

// Check for silliness.

if(!object)
    return;

LL_Remove(&curmap->object,object);
ForceDropObject(x,y,object);
}

/*
 *    DropObject(int x,int y,OBJECT *object)  -  Drop object at x,y
 */

int DropObject(int x,int y, OBJECT *object)
{
OBJECT *temp;
OBJECT *anchor;
int sx,sy,id;

object->next=NULL; // It will be put onto the end of the list

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
   {
   // Terminate melodramatically
//   sprintf(str,"DropObject(%d,%d,%s) Attempted to put outside map boundary",x,y,object->name);
//   panic(cfa,str,"Ai! ai!  A Balrog!  A Balrog is come!  'Tis Durin's Bane!");
   return 0;
   }

// Find the list we want to modify

anchor=GetObjectBase(x,y);
//anchor=curmap.objmap[ytab[y]+x];

// If there's a bag, drop it inside
temp=anchor;
if(temp)
    for(;temp;temp=temp->next)
        if(temp->flags.container)
            {
            FindEmptySquare(&sx,&sy);           // Find scratch space
            curmap->objmap[MAP_POS(sx,sy)]=object;  // Dump the object there
            TakeObject(sx,sy,temp);             // Shift it into the bag
            return 1;
            }

// Modify the object's X,Y coords to sync with the matrix

object->x=x;
object->y=y;

id = getnum4char(object->name);
if(id != -1)
    object->flags.fixed=CHlist[id].flags.fixed;
object->flags.inpocket=0;
LL_Add(&curmap->objmap[MAP_POS(x,y)],object);
gen_largemap();
return 1;
}


/*
 *    ForceDropObject(int x,int y,OBJECT *object)  -  Drop object at x,y
 */

void ForceDropObject(int x,int y, OBJECT *object)
{
int id;
if(!object)
    return;

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
   {
   // Terminate melodramatically
//   sprintf(str,"ForceDropObject(%d,%d,%s) Attempted to put outside map boundary",x,y,object->name);
//   panic(cfa,str,"Ai! ai!  A Balrog!  A Balrog is come!  'Tis Durin's Bane!");
   return;
   }

// Modify the object's X,Y coords to sync with the matrix

object->x=x;
object->y=y;

id = getnum4char(object->name);
if(id != -1)
    object->flags.fixed=CHlist[id].flags.fixed;
object->flags.inpocket=0;
LL_Add(&curmap->objmap[MAP_POS(x,y)],object);
gen_largemap();
}


/*
 *    TakeObject(int x,int y,OBJECT *container)  -  take object from x,y into
 */

void TakeObject(int x,int y, OBJECT *container)
{
OBJECT *temp;

if(!container)
    return;

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
   {
   // Terminate melodramatically
   sprintf(str,"TakeObject(%d,%d) Attempted to get outside map boundary",x,y);
   ithe_panic(str,"Ai! ai!  A Balrog!  A Balrog is come!  'Tis Durin's Bane!");
   }

// Find the list we want to modify

temp=GetObject(x,y);
if(!temp)
    return;

//CHECK_FOR_MOVE_UP_ARSE(container,temp);
//CHECK_FOR_PLAYER_GETTING_LOST(temp);

LL_Remove(&curmap->objmap[MAP_POS(x,y)],temp);

temp->next=container->pocket;
container->pocket=temp;

temp->flags.fixed=1;            // Object sleeps while in pocket
temp->flags.inpocket=1;

gen_largemap();
}


/*
 *    MoveObject(OBJECT *object, int x,int y)  -  move object to x,y
 *    Performs error checking and some game functions, calls TransferObject
 */

int MoveObject(OBJECT *object,int x,int y)
{
OBJECT *temp;
TILE *tile;
int ox,oy,sx,sy,blockx,blocky,blockw,blockh;
char allow=1;

// Prevent the impossible

if(!object)
	return 0;

if(object->x == x)
	if(object->y == y)
		return 1;     // Oh yes, we've done it, honest ;-)

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
	return 0;

if(object->flags.fixed) // Can't move
	return 0;

// If it's a large object, check all of it, but allow if it's only
// hitting itself.

  if(object->flags.large)
    {
    if(object->curdir > CHAR_D) //(U,D,L,R)
        {
        blockx=object->hblock[BLK_X];
        blocky=object->hblock[BLK_Y];
        blockw=object->hblock[BLK_W];
        blockh=object->hblock[BLK_H];
        }
    else
        {
        blockx=object->vblock[BLK_X];
        blocky=object->vblock[BLK_Y];
        blockw=object->vblock[BLK_W];
        blockh=object->vblock[BLK_H];
        }
    for(sy=0;sy<blockh;sy++)
       for(sx=0;sx<blockw;sx++)
           {
           if(isSolid(sx+x+blockx,sy+y+blocky))
               if(GetSolidObject(sx+x+blockx,sy+y+blocky) != object)
                   return 0;
           tile=GetTile(sx+x+blockx,sy+y+blocky);
           if(tile->flags.solid)
              return 0;
           }
    }

#if 0
if(object->flags.large)
    {
    for(sy=0;sy<object->mh;sy++)
       for(sx=0;sx<object->mw;sx++)
           if(isSolid(sx+x,sy+y))
               if(GetSolidObject(sx+x,sy+y) != object)
                   return 0;
    }
#endif

// If there is something in the way, abort.
// If it turns out to be two party members, swap them over

if(isSolid(x,y))
    {
    tile=GetTile(x,y);
    temp=GetSolidObject(x,y);   // Find the obstruction
    allow=0;

    if(temp == object)                  // The object has hit itself..
        {
        TransferObject(object,x,y);     // Allow the move anyway
        return 1;
        }

      // It is legal for unliving objects and boats to enter water

      if(tile->flags.watery && !temp)
          if(!object->flags.person)       // Live people won't swim
              allow=1;

    if(temp)                    // Before we touch the flags, sanity check
       {
//       C_printf("Hit %s\n",temp->name);

       // If there is a boat over water, allow anything to rest on it
       // Also drives the shore and bridges
       if(tile->flags.watery && temp->flags.watery)
           allow=1;

       // If the obstruction is a party member, swap them with the player
       if(temp->flags.party)
       if(object==player)
          {
          // Store the player's old coordinates for the follower
          ox=player->x;
          oy=player->y;
          // Find a square we can move the player temporarily
          FindEmptySquare(&sx,&sy);
          // Got one.   Swap the two party members
          TransferObject(player,sx,sy);
          TransferObject(temp,ox,oy);
          TransferObject(player,x,y);
          // The two party members are reversed now
          return 1;
          }
      if(temp->flags.tabletop)            // Pushing something into a table
          if(!object->flags.person)       // Other people can't either
              if(object != player)        // Just in case player is boat/car
                  allow=1;                // and not a 'person'

        if(temp->flags.container && !object->flags.spikeproof)
          {
          MoveToPocket(object,temp);
          return 1;
          }
       }
    // Blocked
    }

// If we're about to move into a bag (the player must not)

if(!object->flags.person && !object->flags.spikeproof)
	{
	temp=GetObject(x,y);
	if(temp)
		if(temp->flags.container)
			{
			MoveToPocket(object,temp);
			return 1;
			}
	}

if(allow)
	TransferObject(object,x,y);
return allow;
}


/*
 *    TransferObject(OBJECT *object, int x,int y)  -  move object to x,y
 *                                                    without solidity checks
 */

void TransferObject(OBJECT *object,int x,int y)
{
TILE *srctile,*desttile;
OBJECT *temp,*dest,*src;
char destwater,srcwater;
int id;

if(!object)
	return;

// Check to prevent object disappearing up it's own arse

if(object->x == x)
	if(object->y == y)
		return;

// silently ignore object request to leave the map

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
	return;

if(InPocket(object))
	{
	if(!ForceFromPocket(object,syspocket,x,y))
		{
		temp = LookInAllPocketsFor(object);
		if(!temp)    // Oh no it bloody isn't
			{
			sprintf(str,"move_object(%s,%d,%d) Object moved without using move_object",object->name,object->x,object->y);
			ithe_panic(str,"Fly!  This is a foe beyond any of you.  I must hold the narrow way.  Fly!");
			return;
			}
		ForceFromPocket(object,temp,0,0);
		}
	}

// Get tile information before object moves
srctile = GetTile(object->x,object->y);
desttile = GetTile(x,y);
src = GetObjectBase(object->x,object->y);
dest = GetObjectBase(x,y);

srcwater=0; destwater=0;

// Remove the object from the appropriate linked list in curmap.
LL_Remove(&curmap->objmap[MAP_POS(object->x,object->y)],object);

// Keep the object's internal data consistent
object->x=x;       // Very important!
object->y=y;       // Very important!

UpdatePocketCoords(object,x,y); // Update coordinates of all objects inside

// Update the animation frame, if it's stepped
object->flags.stepupdated=1;

// Add to the target list
LL_Add(&curmap->objmap[MAP_POS(x,y)],object);

// Rebuild Large Objects table
gen_largemap();

// If the last tile was water, then say so.
if(srctile->flags.watery)
	srcwater=1;
 // If the object was resting on a boat, then say it wasn't
for(;src;src=src->next)
	if(src->flags.watery)
		if(src != object)       // Don't count the boat we're moving
			srcwater=0;

// If the new tile is water, then say so.
if(desttile->flags.watery)
	destwater=1;
 // If the object is going to rest on a boat, then say it isn't
for(;dest;dest=dest->next)
	if(dest->flags.watery)
		destwater=0;

// If the object was in water, take away the shadow effect where necessary
if(srcwater && !destwater)
	if(object->flags.shadow)
		{
		id = getnum4char(object->name);
		if(id == -1)
			object->flags.shadow = 0;   // educated guess
		else
			object->flags.shadow = CHlist[id].flags.shadow;
		}

// If the object is actually going into the water, do things

if(!srcwater && destwater)
	if(!object->flags.spikeproof)
		{
		if(!object->flags.watery)
			object->flags.shadow=1; // If it can't float, it sinks

		temp = current_object;
		current_object = object;
		CallVMnum(splashpe);
		current_object = temp;
		}

return;
}


/*
 *    GetObjectBase(int x,int y)  -  return pointer to list at x,y
 */

OBJECT *GetObjectBase(int x,int y)
{
OBJECT *anchor;
char ok;

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
   {
   return NULL;
//   sprintf(str,"GetObjectBase(%d,%d,%s) Attempted outside map boundary",x,y);
//   panic(cfa,str,"Ai! ai!  A Balrog!  A Balrog is come!  'Tis Durin's Bane!");
   }

// Find the list

anchor=curmap->objmap[ytab[y]+x];

if(!anchor)
  if(x>=mapx && x<(mapx+VSW))
    if(y>=mapy && y<(mapy+VSH))
         anchor=largemap[VIEWDIST+x-mapx][VIEWDIST+y-mapy];

// Validate the chosen object (make sure it's not pending-delete)

do	{
	ok=1;
	if(!anchor)			// No object, return failure
		return NULL;

	if(!anchor->flags.on)	// Object to be deleted, choose another
		{
		ok=0;
		anchor=anchor->next;
		}
	} while(!ok);

/*
if(anchor)
  if(!anchor->flags.on)
    anchor=NULL;
*/

return anchor;
}


/*
 *    GetRawObjectBase(int x,int y)  -  return pointer to list at x,y
 *                                      Excludes large objects.
 */

OBJECT *GetRawObjectBase(int x,int y)
{
OBJECT *anchor;

// Find the list we want to modify

if(x<0 || x>=curmap->w || y<0 || y>=curmap->h)
   {
   return NULL;
//   sprintf(str,"GetObjectBase(%d,%d,%s) Attempted outside map boundary",x,y);
//   panic(cfa,str,"Ai! ai!  A Balrog!  A Balrog is come!  'Tis Durin's Bane!");
   }

anchor=curmap->objmap[ytab[y]+x];
return anchor;
}


/*
 *    GetObject(int x,int y)  -  return pointer to object at x,y
 */

OBJECT *GetObject(int x,int y)
{
OBJECT *temp,*prev;
OBJECT *anchor_and_large;
OBJECT *anchor_small;

// Get two separate opinions of what is in that square.
// Find out where they differ and thereby make plain the truth.
// A little bit like the way RGB is coded into just two signals in a PAL tv

anchor_and_large=GetObjectBase(x,y);
anchor_small=GetRawObjectBase(x,y);

// If this square's list is empty, say so

if(!anchor_and_large)              // No large or small objects found
    return anchor_and_large;       // Might already be in EAX

if(!anchor_small)                  // No small objects here
    return anchor_and_large;       // Return any large object

// Find the last object
prev = NULL;
for(temp=anchor_and_large;temp->next;temp=temp->next)
    {
    if(temp->flags.on)                          // Store last real object
        prev=temp;
    };    // Just seek

// If the candidate is not real regress

if(!temp->flags.on)
    temp = prev;

// If we're in the game and the candidate is invisible, regress

/*
if(!in_editor)
    if(temp->flags.invisible)
        temp = prev;
*/
return temp;
}


/*
 *    GameGetObject(int x,int y)  -  return pointer to object at x,y
 *                                   if it is visible
 */

OBJECT *GameGetObject(int x,int y)
{
OBJECT *temp,*prev;
OBJECT *anchor_and_large;
OBJECT *anchor_small;

// Get two separate opinions of what is in that square.
// Find out where they differ and thereby make plain the truth.
// A little bit like the way RGB is coded into just two signals in a PAL tv

anchor_and_large=GetObjectBase(x,y);
anchor_small=GetRawObjectBase(x,y);

// If this square's list is empty, say so

if(!anchor_and_large)              // No large or small objects found
    return anchor_and_large;       // Might already be in EAX

if(!anchor_small)                  // No small objects here
    return anchor_and_large;       // Return any large object

CHECK_OBJECT(anchor_and_large);
CHECK_OBJECT(anchor_small);

// Find the last object
prev = NULL;
for(temp=anchor_and_large;temp->next;temp=temp->next)
    {
    if(temp->flags.on && temp->flags.invisible == 0)
        prev=temp;                         // Store last real object
    };    // Just seek

// If the candidate is not real, regress

CHECK_OBJECT(temp);

if(!temp->flags.on)
    temp = prev;

// If candidate is invisible, regress

if(temp->flags.invisible)
    temp = prev;

CHECK_OBJECT(temp);
return temp;
}


/*
 *      GetSolidObject - return a pointer to the last solid object
 *                       Based on GameGetObject
 */

OBJECT *GetSolidObject(int x,int y)
{
OBJECT *temp,*prev;
OBJECT *anchor_and_large;
OBJECT *anchor_small;

// Get two separate opinions of what is in that square.
// Find out where they differ and thereby make plain the truth.
// This is tricker than GameGetObject.. we have to find the solid ones

temp=GetObjectBase(x,y);
for(anchor_and_large=NULL;temp;temp=temp->next)
	if(temp->flags.solid)
		{
		anchor_and_large=temp;
		break;
		}
anchor_small=GetRawSolidObject(x,y);

// If this square's list is empty, say so

if(!anchor_and_large)              // No large or small objects found
	return anchor_and_large;       // Might already be in EAX

if(!anchor_small)                  // No small objects here
	return anchor_and_large;       // Return any large object

CHECK_OBJECT(anchor_and_large);
CHECK_OBJECT(anchor_small);

// Find the last object
prev = NULL;
for(temp=anchor_and_large;temp->next;temp=temp->next)
	{
	if(temp->flags.on && temp->flags.invisible == 0 && temp->flags.solid)
		prev=temp;                         // Store last real object
	};    // Just seek

// If the candidate is not real, regress

CHECK_OBJECT(temp);

if(!temp->flags.on)
	temp = prev;

// If candidate is invisible, regress

if(temp->flags.invisible)
	temp = prev;

// If candidate isn't solid, regress

if(!temp->flags.solid)
	temp = prev;

CHECK_OBJECT(temp);

return temp;
}

/*
 *      GetRawSolidObject - return a pointer to the first solid object
 */

OBJECT *GetRawSolidObject(int x,int y)
{
OBJECT *temp;

temp=GetRawObjectBase(x,y);

if(!temp)
    return temp;

for(;temp;temp=temp->next)
    if(temp->flags.solid)
      if(temp->flags.on)
        if(!temp->flags.invisible)
          return temp;
return NULL;
}


/*
 *      WeighObject - calculate how much weight is in the list
 *                    calls WeighObjectR
 */

int WeighObject(OBJECT *obj)
{
int w=0;
if(!obj)
     return 0;
if(!obj->flags.on)
    return 0;

w = obj->stats->weight;
if(obj->pocket)
	{
	object_search_rec=0;
	return w+WeighObjectR(obj->pocket);
	}

return w;
}

/*
 *      WeighObjectR - recursive engine for WeightObject
 */

int WeighObjectR(OBJECT *list)
{
OBJECT *temp;
int w=0;
static int complain=0;

// Defence against objects inside their asses

if(object_search_rec>TOO_MUCH)
	{
	if(!complain)
		{
		Bug("OH NO!  Too much recursion in WeighObject!\n");
		complain=1;
		}
	return 0;
	}
object_search_rec++;
complain=0;

// Okay, do the thing

for(temp = list;temp;temp=temp->next)
    if(temp->flags.on)
        {
        w+=temp->stats->weight;
        if(temp->pocket)
            w+=WeighObjectR(temp->pocket);
        }
return w;
}

/*
 *      GetBulk - calculate how much bulk is in the list
 *                calls GetBulkR
 */

int GetBulk(OBJECT *obj)
{
int w=0;
if(!obj)
     return 0;
if(!obj->flags.on)
    return 0;

w = obj->stats->bulk;
if(obj->pocket)
    return w+GetBulkR(obj->pocket);

return w;
}

/*
 *      GetBulkR - recursive engine for GetBulk
 */

int GetBulkR(OBJECT *list)
{
OBJECT *temp;
int w=0;

for(temp = list;temp;temp=temp->next)
    if(temp->flags.on)
        {
        w+=temp->stats->bulk;
        if(temp->pocket)
            w+=GetBulkR(temp->pocket);
        }
return w;
}


/*
 *      AddToParty - append a new member to the party array
 */

int AddToParty(OBJECT *new_member)
{
int ctr;

//ilog_quiet("PARTY: new member: %s\n",new_member->name);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	if(!party[ctr])
		{
		party[ctr] = new_member;
		strcpy(partyname[ctr],BestName(new_member));
ilog_quiet("addtoparty (%d)\n",follower);
		ActivityNum(new_member,follower,NULL);
		new_member->flags.party = 1;
		return ctr;
		}
return -1;
}


/*
 *      SubFromParty - remove a party member
 */

void SubFromParty(OBJECT *member)
{
int ctr,ctr2,pos;

if(!member)
    return;

//ilog_quiet("PARTY: sub member: %s\n",member->name);

pos=-1;
for(ctr=0;ctr<MAX_MEMBERS;ctr++)
    if(party[ctr] == member)
        pos=ctr;

if(pos == -1)
    return;

if(member->stats->hp>0)
	{
	member->flags.party = 0;
	ResumeSchedule(member);
    }
else
	Inactive(member);  // Prevent the dead guy from following you about

// If the leader is leaving, call an election

if(player == member)
     {
     player = NULL;
     for(ctr2=0;ctr2<MAX_MEMBERS-1;ctr2++)
         if(party[ctr2])
             if(party[ctr2]->stats->hp>0)
                 ChooseLeader(ctr2);
     }

party[pos]=NULL;
partyname[pos][0]='\0';

   // sort the party to remove any holes

for(ctr=0;ctr<MAX_MEMBERS-1;ctr++)
    if(!party[ctr])
        for(ctr2=ctr;ctr2<MAX_MEMBERS-1;ctr2++)
            if(party[ctr2])
                {
                party[ctr] = party[ctr2];
                party[ctr2] = NULL;
                }

}


/*
 *      ChooseLeader- Set the party leader
 */

int ChooseLeader(int x)
{
if(!party[x])
    {
    Bug("Election: party member %d not present\n",x);
    return 0;
    }

//ilog_quiet("PARTY: ChooseLeader: %s\n",party[x]->name);

for(int ctr=0;ctr<MAX_MEMBERS;ctr++)
    if(party[ctr])
        if(party[ctr]->stats->hp > 0)
            {
            ML_Del(&ActiveList,party[ctr]);
//            party[ctr]->behave = follower;
//            AL_Add(&ActiveList,party[ctr]);
            }
player = party[x];
ActivityName(player,"player_action",NULL);

return 1;
}


/*
 *      FindEmptySquare- Find some scratch space for certain map operations
 */

void FindEmptySquare(int *x,int *y)
{
int sx,sy;

for(sy=0;sy<curmap->h;sy++)
    for(sx=0;sx<curmap->w;sx++)
        if(!curmap->objmap[ytab[sy]+sx])
            {
            *x=sx;
            *y=sy;
            return;
            }
ithe_panic("FindEmptySquare failed","Map probably corrupted");
}


/*
 *      gen_largemap- Build the table of objects larger than 1 square
 */

void gen_largemap()
{
int x,y,xctr,yctr;
int sx,sy,ex,ey,bx,by,xo,yo;
int blockx,blocky,blockw,blockh;

OBJECT *temp;

sx=mapx-VIEWDIST;
sy=mapy-VIEWDIST; // was -4
ex=mapx+VSW+4;
ey=mapy+VSH+4;
xo=0;
yo=0;
if(sx<0)
	{
	xo=0-sx;
	sx=0;
	}
if(sy<0)
	{
	yo=0-sy;
	sy=0;
	}
if(ex>=curmap->w)
	ex=curmap->w-(VSW+4);
if(ey>=curmap->h)
	ey=curmap->h-(VSH+4);

// Clear the grids
memset(&largemap[0][0],0,(sizeof(OBJECT *)*(LMSIZE*LMSIZE)));
memset(&solidmap[0][0],0,(sizeof(OBJECT *)*(LMSIZE*LMSIZE)));

by=yo;
bx=xo;
for(y=sy;y<ey;y++)
    {
    for(x=sx;x<ex;x++)
        {
        temp = curmap->objmap[ytab[y]+x];
        for(;temp;temp=temp->next)
          if(temp->flags.on)
            {
            if(temp->flags.large)       // Found a large object?
                {
                if(temp->flags.solid)   // Is it solid? If so, find out where
                    {
                    if(temp->curdir > CHAR_D) //(U,D,L,R)
                        {
                        blockx=temp->hblock[BLK_X];
                        blocky=temp->hblock[BLK_Y];
                        blockw=temp->hblock[BLK_W];
                        blockh=temp->hblock[BLK_H];
                        }
                    else
                        {
                        blockx=temp->vblock[BLK_X];
                        blocky=temp->vblock[BLK_Y];
                        blockw=temp->vblock[BLK_W];
                        blockh=temp->vblock[BLK_H];
                        }

                    // Reset border size
                    temp->w = blockw<<5;
                    temp->h = blockh<<5;

                    // Colour in the squares where the solid parts are

                    for(xctr=0;xctr<blockw;xctr++)
                        for(yctr=0;yctr<blockh;yctr++)
                            solidmap[bx+xctr+blockx][by+yctr+blocky]=temp;
                    }


                // Fill in the active area of the object

                if(temp->curdir > CHAR_D) //(U,D,L,R)
                    {
                    blockx=temp->harea[BLK_X];
                    blocky=temp->harea[BLK_Y];
                    blockw=temp->harea[BLK_W];
                    blockh=temp->harea[BLK_H];
                    }
                else
                    {
                    blockx=temp->varea[BLK_X];
                    blocky=temp->varea[BLK_Y];
                    blockw=temp->varea[BLK_W];
                    blockh=temp->varea[BLK_H];
                    }

                for(xctr=0;xctr<blockw;xctr++)            // For the width,
                    for(yctr=0;yctr<blockh;yctr++)        // For the height,
                        largemap[bx+blockx+xctr][by+blocky+yctr]=temp;
                largemap[bx+blockx][by+blocky]=NULL; // Delete core section
                }
            else
                if(temp->flags.solid)           // If it's small and solid
                        solidmap[bx][by]=temp;  // Fill in the single square
            }
        bx++;
        }
    bx=xo;
    by++;
    }

}

/*
 *      DestroyObject - Perform the complex task of destroying an arbitrary
 *                      object without harming the fabric of the universe
 */

void DestroyObject(OBJECT *obj)
{
OBJECT *temp,*pocket;
char str[256];
int ctr;

if(!obj)                // No!
        {
        ithe_panic("DestroyObject: Attempted to destroy NULL",NULL);
        return;
        }

// First see if it is where it seems to be

if(!InPocket(obj))
    {
    LL_Remove(&curmap->objmap[MAP_POS(obj->x,obj->y)],obj);
    FreePockets(obj);
    LL_Kill(obj);
    return;
    }

// Next look in the pockets

pocket = LookInAllPocketsFor(obj);
if(pocket)
    {
    // Delete it from the wielding list if it's there
    if(player)
    for(ctr=0;ctr<W_SIZE;ctr++)
        if(player->wield[ctr] == obj)
            player->wield[ctr] = NULL;

    // Remove it from the system
    LL_Remove(&pocket->pocket,obj);
    FreePockets(obj);
    LL_Kill(obj);
    return;
    }

// In case of non-fatal bug, look for it in the linked list buffer

temp = curmap->object;
for(;temp;temp=temp->next)
    if(temp == obj)
        {
        Bug("Removing loose object %s\n",obj->name);
        LL_Remove(&curmap->object,obj);
        FreePockets(obj);
        LL_Kill(obj);
        return;
        }

// Oh dear

sprintf(str,"obj %s at %d,%d (id %d iid %d)",obj->name,obj->x,obj->y,obj->save_id,obj->inside_id);
ithe_panic("DestroyObject: Object has no physical existence",str);
}

/*
 *      CreateContents - Create the default contents of an item
 */

void CreateContents(OBJECT *objsel)
{
int num;
OBJECT *temp;
for(int ictr=0;ictr<objsel->funcs->contents;ictr++)
    if(objsel->funcs->contains[ictr])
        {
        temp = OB_Alloc();
        temp->curdir = temp->dir[0];
        OB_Init(temp,objsel->funcs->contains[ictr]);
        OB_SetDir(temp,1);
        MoveToPocket(temp,objsel);
        }

// Call initial script if in the game

if(!in_editor)
	if(!objsel->flags.didinit)
		{
		if(objsel->funcs->icache != -1)
			CallVMnum(objsel->funcs->icache);
		objsel->flags.didinit=1;
		}

}

/*
 *      MoveToTop - Move specified object to the top of the pile
 */

void MoveToTop(OBJECT *object)
{
int ox,oy;
// Move object to the top of the heap by moving it to 0,0 and back

if(InPocket(object))   // See if on ground or in pocket
    return;            // Either in a pocket or world not consistent

ox = object->x;
oy = object->y;
TransferObject(object,0,0);
TransferObject(object,ox,oy);
}

/*
 *      MoveToFloor - Move specified object to above the last FIXED object
 *                    CAUTION: does low-level linked-list manipulation
 */

void MoveToFloor(OBJECT *object)
{
OBJECT *temp,*after;

// Move object to the top of the heap by moving it to 0,0 and back

if(InPocket(object))   // See if on ground or in pocket
    return;            // Either in a pocket or world not consistent

// Disassociate from the map
LL_Remove(&curmap->objmap[MAP_POS(object->x,object->y)],object);

after=NULL;
temp = GetRawObjectBase(object->x,object->y);
for(;temp;temp=temp->next)
   if(temp->flags.fixed)
        after=temp;

if(!after)      // If there are no fixed objects, stick it on the front
    {
    object->next = GetRawObjectBase(object->x,object->y);
    curmap->objmap[MAP_POS(object->x,object->y)] = object;
    return;
    }

// Otherwise stick it after the object we've chosen

object->next = after->next;
after->next = object;
}

/*
 *      TakeQuantity - Subtract an amount from a container with several
 *                     of these objects in, e.g. player pays 100 gold coins,
 *                     the coins may be scattered througout his pockets.
 */

int TakeQuantity(OBJECT *container,char *type,int request)
{
int total,found;
OBJECT *temp;
int o;

// First, see if the object can support quantity values.
o = getnum4char(type);
if(o == -1)
    return 0;

o = CHlist[o].flags.quantity;   // o is the boolean result

total = SumObjects(container->pocket, type, 0);

if(!request)
    request = 1;

if(total<request)
    return 0;

found=0;
do {
   temp = GetFirstObject(container->pocket,type);
   if(!temp)
       {
       Bug("take_quantity: financial irregularities detected in object %s\n",container->name);
       return 0;
       }

   // If the object doesn't support Quantities, we change it so it does
   // and set the quantity of the object to 1 temporarily, since it will
   // be destroyed later, anyway

   if(!o)
       temp->stats->quantity = 1;

   // If less than requested sum, take it all and destroy it.

   if(temp->stats->quantity <= request)
       {
       found += temp->stats->quantity;
       temp->flags.on = 0;     // Mark pending delete
       pending_delete = 1;     // Inform garbage colletion
       }
   else
       {
       temp->stats->quantity -=request; // Otherwise just take what we need
       found=request;
       }
   } while(found<request);

return 1;
}

/*
 *      AddQuantity - Add a number of objects to the container.
 *                    E.g. pay the player 100 gold coins, and group them
 *                    together by adding to the first pile we come across.
 *                    If we try to add 0, create a new object regardless
 */

void AddQuantity(OBJECT *container,char *type,int request)
{
OBJECT *temp;
int o;

// First, see if the object can support quantity values.
o = getnum4char(type);
if(o == -1)
    {
    Bug("Cannot create unknown object %s\n",type);
    return;
    }
o = CHlist[o].flags.quantity;   // o is the boolean result

temp = GetFirstObject(container->pocket,type);
if(request == 0)
    temp = NULL;

// If the object exists


if(temp)
  if(o)
     {
     temp->stats->quantity+=request;
     return;
     }

// If it supports multiply quantites, just create one and set the quantity

if(o)
    {
    temp = OB_Alloc();
    OB_Init(temp,type);
    temp->stats->quantity=request;
    MoveToPocket(temp,container);
    return;
    }

// If not, make several

for(o=0;o<request;o++)
    {
    temp = OB_Alloc();
    OB_Init(temp,type);
    CreateContents(temp);
    MoveToPocket(temp,container);
    }

return;
}

/*
 *      MoveQuantity - Transfer an amount between containers with several
 *                     of these objects in, e.g. player pays 100 gold coins,
 *                     the coins may be scattered througout his pockets.
 */

int MoveQuantity(OBJECT *src,OBJECT *dest,char *type,int request)
{
int total,found;
OBJECT *temp;
int o,ctr;

// First, see if the object can support quantity values.
o = getnum4char(type);
if(o == -1)
    return 0;

o = CHlist[o].flags.quantity;   // o is the boolean result

total = SumObjects(src->pocket, type, 0);

if(!request)
    request = 1;

if(total<request)
    return 0;

// If the objects cannot be piles, trasfer them one at a time

if(!o)
    {
    for(ctr=0;ctr<request;ctr++)
        {
        temp = GetFirstObject(src->pocket,type);
        ForceFromPocket(temp,src,0,0);
        TransferToPocket(temp,dest);
        }
    return 1;
    }

// If they can, delete the specified amount in the src and create in the dest

found=0;
do {
   temp = GetFirstObject(src->pocket,type);
   if(!temp)
       {
       Bug("take_quantity: financial irregularities detected in object %s\n",src->name);
       return 0;
       }

   // If less than requested sum, take it all and destroy it.

   if(temp->stats->quantity <= request)
       {
       found += temp->stats->quantity;
       temp->flags.on = 0;     // Mark pending delete
       pending_delete = 1;     // Inform garbage colletion
       AddQuantity(dest,type,request);
       }
   else
       {
       temp->stats->quantity -=request; // Otherwise just take what we need
       found=request;
       AddQuantity(dest,type,request);
       }
   } while(found<request);

return 1;
}

/*
 * Set up the Object's current activity by number
 */

void ActivityNum(OBJECT *o,int activity, OBJECT *target)
{
if(activity < 0)
	return;

//ilog_quiet("Assigning %s (%d) to %s\n",PElist[activity].name,activity,o->name);

if(!o)
	{
	ilog_quiet("Tried to assign activity %d to NULL\n",activity);
	ithe_panic("Tried to assign activity to NULL",NULL);
	}
o->activity = activity;
o->target = target;

if(!ML_InList(&ActiveList,o))
	AL_Add(&ActiveList,o);
}

/*
 * Set up the Object's current activity by name
 */

void ActivityName(OBJECT *o,char *activity, OBJECT *target)
{
if(!o)
	{
	Bug("Tried to assign activity %s to NULL\n",activity);
	return;
	}

o->activity = getnum4PE(activity);
if(o->activity <0)
	{
	Bug("Couldn't find activity script called '%s' for %s\n",activity,BestName(o));
	Inactive(o); // Can't do it.  Stop the object
	return;
	}
o->target = target;

if(!ML_InList(&ActiveList,o))
	AL_Add(&ActiveList,o);

//ML_Del(&ActiveList,o); // just to be sure
//AL_Add(&ActiveList,o);
}

/*
 * Wipe the Object's current activity
 */

void Inactive(OBJECT *o)
{
if(!o)
	ithe_panic("Tried to delete activity from NULL",NULL);

//ilog_quiet("InActivate %s\n",o->name);

o->activity = -1;
o->target = NULL;

ML_Del(&ActiveList,o);
}


/*
 *      LookInAllPockets - Return the container which holds the object.
 *                         Calls SearchContainer.
 */

OBJECT *LookInAllPocketsFor(OBJECT *obj)
{
int vx,vy;
OBJECT *temp,*search;
for(vy=0;vy<curmap->h;vy++)
    for(vx=0;vx<curmap->w;vx++)
        for(temp = GetRawObjectBase(vx,vy);temp;temp=temp->next)
            if(temp->pocket)
                {
                search = SearchContainer(temp,obj);
                if(search)
                    return search;
                }
return NULL;
}


/*
 *      SearchContainer - Return the pocket which contains the object.
 *                        Recursive, called by LookInAllPockets
 */

OBJECT *SearchContainer(OBJECT *cont, OBJECT *obj)
{
OBJECT *search;

for(OBJECT *temp = cont->pocket;temp;temp=temp->next)
    {
    if(temp == obj)
        return cont;
    if(temp->pocket)
        {
        search = SearchContainer(temp,obj);
        if(search)
            return search;
        }
    }
return NULL;
}

/*
 *      InPocket - Is the object in somebody's pocket?
 */

int InPocket(OBJECT *obj)
{
OBJECT *temp;

if(!obj)
    ithe_panic("InPocket: Given a NULL pointer",NULL);

if(obj->flags.inpocket)
    return 1;

temp = GetRawObjectBase(obj->x,obj->y);
for(;temp;temp=temp->next)
    if(temp == obj)
        return 0;

// It isn't where it thinks it is, so we assume (hope) it's in a pocket

return 1;
}

/*
 *      LookInAllPocketsForTag - Return the object which has that tag number.
 *                               Calls SearchContainerTag.
 */

OBJECT *LookInAllPocketsForTag(int tag)
{
int vx,vy;
OBJECT *temp,*search;
for(vy=0;vy<curmap->h;vy++)
    for(vx=0;vx<curmap->w;vx++)
        for(temp = GetRawObjectBase(vx,vy);temp;temp=temp->next)
            {
            if(temp->tag == tag)
                return temp;
            if(temp->pocket)
                {
                search = SearchContainerTag(temp,tag);
                if(search)
                    return search;
                }
            }
return NULL;
}


/*
 *      SearchContainerTag - Return the object with this tag.
 *                           Recursive, called by LookInAllPockets
 */

OBJECT *SearchContainerTag(OBJECT *cont, int tag)
{
OBJECT *search;

for(OBJECT *temp = cont->pocket;temp;temp=temp->next)
    {
    if(temp->tag == tag)
        return temp;
    if(temp->pocket)
        {
        search = SearchContainerTag(temp,tag);
        if(search)
            return search;
        }
    }
return NULL;
}

/*
 *     SumObjects - Return the total of a multiple object, e.g. money.
 *                  Recursive.
 */

int SumObjects(OBJECT *cont, char *name, int total)
{
for(OBJECT *temp = cont;temp;temp=temp->next)
    {
    if(temp->flags.on)
      if(!stricmp(temp->name,name))
        {
        if(temp->flags.quantity)
            total+=temp->stats->quantity;
        else
            total++;
        }
    if(temp->pocket)
        total+=SumObjects(temp->pocket,name,total);
    }
return total;
}

/*
 *     GetFirstObject - Find the first instance of a specified object.
 *                      Recursive.
 */

OBJECT *GetFirstObject(OBJECT *cont, char *name)
{
OBJECT *search;

for(OBJECT *temp = cont;temp;temp=temp->next)
    {
    if(temp->flags.on)
      if(!stricmp(temp->name,name))
        return temp;
    if(temp->pocket)
        {
        search=GetFirstObject(temp->pocket,name);
        if(search)
            return search;
        }
    }
return NULL;
}

/*
 *      CheckDepend - Remove all references to a pointer
 */

void CheckDepend(OBJECT *ptr)
{
int ctr,ctr2;
OBJLIST *t;
NPC_RECORDING *n;

// All the objects in the entire universe

t = MasterList;
if(!t)
    return;

for(t=MasterList;t;t=t->next)
	if(!t->ptr)
		{
		Bug("NULL in masterlist at %x\n",t);
		ML_Del(&MasterList,NULL);
		t=MasterList;
		}
	else
		{
		// If the object's enemy is the One That Must Be Cleansed, remove it

		if(t->ptr->enemy == ptr)
			{
//			ilog_printf("%s enemy %s deleted\n",t->ptr->name,ptr->name);
			t->ptr->enemy=NULL;
			}

  // If the object is owned by the One That Must Be Cleansed, it goes public

		if(t->ptr->stats)
			{
			if(t->ptr->stats->owner == ptr)
				{
//				ilog_printf("%s owner %s deleted\n",t->ptr->name,ptr->name);
				t->ptr->stats->owner=NULL;
				}
			}

  // If the object's activity involves the One That Must Be Cleansed, delete it

		if(t->ptr->target)
			{
			t->ptr->target = NULL;
//			ilog_printf("%s target %s deleted\n",t->ptr->name,ptr->name);
			}

  // If the object's schedule involves the One That Must Be Cleansed, kill entry

		if(t->ptr->schedule)
			for(ctr=0;ctr<24;ctr++)
				if(t->ptr->schedule[ctr].hour != -1)
					if(t->ptr->schedule[ctr].target == ptr)
						t->ptr->schedule[ctr].hour = -1;

 // If the object's conversations involve the One That Must Be Cleansed...
		if(t->ptr->stats)
			{
			if(t->ptr->stats->npctalk)
				for(n=t->ptr->stats->npctalk;n;n=n->next)
					if(n->player == ptr)
						{
						n->player=NULL;
						n->page[0]='\0';
						}
			if(t->ptr->stats->lFlags)
				for(n=t->ptr->stats->lFlags;n;n=n->next)
					if(n->player == ptr)
						{
						n->player=NULL;
						n->page[0]='\0';
						}
			}
  }


// All globally-defined objects

if(!in_editor)
	wipe_refs(ptr);

// If a party member was boiled to vapour kick 'em out of the party

if(!in_editor)
for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	if(party[ctr])
		if(!party[ctr]->flags.on)
			party[ctr] = NULL;

return;
}

void CheckRefs(void **p1, void *ptr1)
{
int ctr;
OBJECT ***p,*ptr;
NPC_RECORDING *n;

p=(OBJECT ***)p1;
ptr=(OBJECT *)ptr1;

if(**p == ptr)
	**p = NULL;
if(**p)
	{
	if((**p)->enemy == ptr)
		(**p)->enemy=NULL;
	if((**p)->stats)
		if((**p)->stats->owner == ptr)
			(**p)->stats->owner=NULL;
	if((**p)->target)
		(**p)->target=NULL;
	if((**p)->schedule)
		for(ctr=0;ctr<24;ctr++)
			if((**p)->schedule[ctr].hour != -1)
				if((**p)->schedule[ctr].target == ptr)
					(**p)->schedule[ctr].hour = -1;

 // If the object's conversations involve the One That Must Be Cleansed...
	if((**p)->stats)
		{
		if((**p)->stats->npctalk)
			for(n=(**p)->stats->npctalk;n;n=n->next)
				if(n->player == ptr)
					{
					n->player=NULL;
					n->page[0]='\0';
					}
		if((**p)->stats->lFlags)
			for(n=(**p)->stats->lFlags;n;n=n->next)
				if(n->player == ptr)
					{
					n->player=NULL;
					n->page[0]='\0';
					}
		}
	}
}

/*
 *      DeletePending() - Remove all Pending-Delete objects (Garbage collect)
 */

void DeletePending()
{
OBJLIST *t;
if(!pending_delete)
    return;

t = MasterList;
if(!t)
    return;

for(t=MasterList;t;t=t->next)
  {
  if(!t->ptr->flags.on)
      {
      DestroyObject(t->ptr);  // Free the object
      t=MasterList;           // Restart the check
      }
  }
pending_delete=0;             // None
}

/*
 *      IsOnscreen() - is an NPC on screen?
 */

int IsOnscreen(char *pname)
{
OBJECT *temp;

for(int y=0;y<VSH;y++)
    for(int x=0;x<VSW;x++)
        for(temp = GetObjectBase(x+mapx,y+mapy);temp;temp=temp->next)
            if(temp->personalname)
                if(!stricmp(temp->personalname,pname))
                    return 1;
return 0;
}

/*
 *  Update the coordinates of all objects in this one's pocket
 */

void UpdatePocketCoords(OBJECT *o, int x, int y)
{
OBJECT *temp;
for(temp=o->pocket;temp;temp=temp->next)
    {
    temp->x=x;
    temp->y=y;
    if(temp->pocket)
        UpdatePocketCoords(temp,x,y);
    }
}


char *BestName(OBJECT *o)
{
// If it is Not, say so

if(!o)
      return "Nothing";

// If it has a personal name, and you know it, use it

if(o->personalname)
    if(o->personalname[0] != '\0')
        if(o->stats->npcflags.know_name)
            if(stricmp(o->personalname,"-"))
                return o->personalname;

/*
// If it is you, say as much

if(o == player)
     return "yourself";
*/

// Oh well.  Use the short description

return o->shortdesc;
}


/*
 *      Spillcontents - If a container is destroyed it bursts
 */

void spillcontents(OBJECT *t,int x,int y)
{
OBJECT *temp;
if(!t)
    return;

for(int ctr=0;ctr<9;ctr++)
    if(t->wield[ctr])
        t->wield[ctr]->flags.invisible=0;

while(t->pocket)
    {
    temp=t->pocket;
    t->pocket=t->pocket->next;
    ForceDropObject(x,y,temp);
    };
}


/*
 *      check_hurt - Poll the things to see if they've been harmed or killed
 */

void CheckHurt(OBJECT *start)
{
OBJECT *temp,*oldcurrent,*next=NULL;

// Check for things that are hurt

if(!start)
	{
	ilog_quiet("CheckHurt NULL\n");
    return;
	}

oldcurrent=current_object;

for(temp=start;temp;temp=next)
    {
    next=temp->next;
    if(temp->flags.on)
        {
        if(temp->stats->hp < temp->stats->oldhp)
            {
            if(temp->stats->hp>0)
                if(temp->funcs->hcache != -1)
                    {
                    current_object = temp;
                    CallVMnum(temp->funcs->hcache);
                    current_object = oldcurrent;
                    }
            temp->stats->oldhp=temp->stats->hp;
            }

// "They're all dead.  They just don't know it yet." - Eric Draven, The Crow

        if(temp->stats->hp < 1)
            {
            if(temp->funcs->kcache != -1)      // If it can be killed..
                {
                current_object = temp;
                CallVMnum(temp->funcs->kcache); // ..then kill it.
                if(current_object)
                    temp->stats->oldhp=temp->stats->hp;
                current_object = oldcurrent;
                }
            else                                  // If it cannot be killed
            if(temp->stats->hp < destroyhp)       // and it is very dead..
                {                                 // ..it shall be Made Not
                current_object = temp;
				ilog_quiet("DoErase: %x\n",current_object);
                CallVMnum(erase_object);       // wipe it out
                if(oldcurrent == temp)
                    oldcurrent = NULL;
                if(victim == temp)
                    victim = NULL;
                current_object = oldcurrent;
                }
            }
        }
    }
}

/*
 *      find_id - Look for an ID number on the map and return the objptr
 */

extern "C" OBJECT *find_id(unsigned int id)
{
OBJLIST *t;
for(t=MasterList;t;t=t->next)
	if(t->ptr)
		if(t->ptr->save_id == id)
			return t->ptr;

return NULL;
}


