
/*
 *    Operating System Command Line Interpreter
 */

#include <stdio.h>
#include <stdlib.h>

// MR: added
#ifndef _WIN32
#include <strings.h>
#endif

#include <string.h>
#include "media.h"
#include "oscli.h"
#include "ithelib.h"
#include "loadfile.h"
#include "cookies.h"
#include "core.hpp"

// Defines

#define find(xxx) OSCLI_find(ArgC,ArgV,xxx)
#define found(xxx) OSCLI_found(ArgC,ArgV,xxx)

// Variables:  These are all controlled by commandline switches

// Mapfile, filename of the current map, begins with 0xff if not defined
char mapname[]=COOKIE_MapName; // belated anti-Vietnam message
char newbootlog[]=COOKIE_NewBootLog;
char oldbootlog[]=COOKIE_OldBootLog;
char resdir[]=COOKIE_ResDir; // Resource Directory
char loadingpic[]=COOKIE_LoadingPic;
int  logx=0,logy=0,loglen=60;  // Graphical boot log, X,Y and length in lines
int  conx=0,cony=0,conlen=60,conwid=80;

char probeInvalidObject=0;
char gamespec =0;
char VideoMode=-1;
char debug_nosound=0;
char no_panic=0;
char ire_notimer=0;
char show_vrm_calls=0;
char graflog=-1;        // Graphical bootlog, true of false
char quickstart=0;      // Prevent logging to disk
char mus_chan=16;       // Music channels
char sfx_chan=4;        // Sound channels
char use_light=1;       // Use lighting by default
char vis_blanking=1;    // Use visibility blanking by default
char skip_unk=0;        // Don't skip unknown keywords in the scriptfile
int map_W=512,map_H=512;        // Map size
int driftlevel=20;      // Frequency change in Hz times -10 to +10
char reset_stats=0;
char reset_funcs=0;
char enable_censorware=0;
int destroyhp=-100;     // When this number is reached, object is obliterated
char show_imap=0;       // Show a map of the tracking engine
int imapx=384,imapy=240;
int ig_min,ig_hour,ig_day,ig_month,ig_year,ig_dow;
int min_hour=60,hour_day=24,day_week=7,day_month=30,month_year=12;
char DebugVM=0;
char in_editor; // Which program we are in (some modules are shared)

// Limits for the script editor

extern long spr_alloc;
extern long seq_alloc;
extern long vrm_alloc;
extern long chr_alloc;
extern long til_alloc;
extern long wav_alloc;
extern long mus_alloc;
extern long pef_alloc;

// Functions

static short OSCLI_find(short ArgC,char **ArgV,char *param);
static char OSCLI_found(short ArgC,char **ArgV,char *param);
void OSCLI(int ArgC,char **ArgV);

// Code

// OSCLI, process a list of switches

void OSCLI(int ArgC,char **ArgV)
{
int cpt,len;

if(found("?") || found("h") || found("help"))
	{
	printf("\n");
	puts("-list [listfile]    Execute a list of commands (like @ in doom)");
	puts("-game [gamefile]    (same as -list)");
	puts("-file [RARfile(s)]  Use an external .RAR file");
	puts("-mapfile [mapfile]  Load a different map (No file extension!)");
	puts("-nopanic            no fancy error messages");
	puts("-debugvrm           List all vrm calls in BOOTLOG.TXT");
	puts("-debugVM            Allow debugging keys in the PE Virtual Machine");
	puts("-showmodes          List all available video modes");
	puts("-videomode          Set number of bpp, or 0 to ask user (recommended)");
	puts("-nolight            Don't do any lighting (for slow systems)");
	puts("-skipunknowns       Ignore unknown keywords in the script, don't stop");
	puts("-show_imap          Display a map of the pathfinder engine");
if(in_editor)
		{
		puts("-reset_stats        Reset all objects in the editor to their default statistics");
		puts("-reset_funcs        Reset all objects in the editor to their default functions");
		}
	puts("-L                  Display license");
	puts("-?                  This screen");
	puts("\nTo cause a stack trace on a Panic, set SUPERPANIC=1 in your environment space");
	puts("\nTo send ilog_quiet to the printer set LPTDEBUG=1 in your environment space");
	exit(1);
	}

if(found("L"))
	{
	puts("");
	puts("Copyright (c) 2001, IT-HE Software");
	puts("");
	puts("Redistribution and use in source and binary forms, with or without");
	puts("modification, are permitted provided that the following conditions are met:");
	puts("");
	puts("Redistributions of source code must retain the above copyright notice, this");
	puts("list of conditions and the following disclaimer.");
	puts("");
	puts("Redistributions in binary form must reproduce the above copyright notice,");
	puts("this list of conditions and the following disclaimer in the documentation");
	puts("and/or other materials provided with the distribution.");
	puts("");
	puts("Neither the name of IT-HE Software nor the names of its contributors may");
	puts("be used to endorse or promote products derived from this software without");
	puts("specific prior written permission.");

	puts("THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''");
	puts("AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE");
	puts("IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE");
	puts("ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR");
	puts("ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL");
	puts("DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR");
	puts("SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER");
	puts("CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT");
	puts("LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY");
	puts("OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH");
	puts("DAMAGE.");
	exit(0);
	}

cpt=find("mapfile");
if(cpt>=0&&(unsigned char)mapname[0]==0xff)
if(ArgV[cpt+1])
		{
		strcpy(mapname,ArgV[cpt+1]);
		len = strlen(mapname);
		for(cpt=0;cpt<len;cpt++)
			if(mapname[cpt]<32)
				mapname[cpt]=0;
		}

cpt=find("edit");
if(cpt>=0)
if(ArgV[cpt+1])
    strcpy(mapname,ArgV[cpt+1]);

if(found("nopanic"))
    no_panic = 1;

if(found("skipunknowns"))
    skip_unk= 1;

if(found("debugvrm") || found("debugvrms"))
    show_vrm_calls= 1;

if(found("debugvm"))
    DebugVM= 1;

if(found("nolight"))
    use_light = 0;

if(found("noblanking"))
    vis_blanking = 0;

if(found("nosound"))
    debug_nosound = 1;

if(found("reset_stats"))
    reset_stats = 1;
if(found("reset_funcs"))
    reset_funcs = 1;

if(found("censorware"))
	enable_censorware = 1;

if(!gamespec)
    {
    cpt=find("game");
    if(cpt>=0)
    if(ArgV[cpt+1])
        {
	INI_file(ArgV[cpt+1]);
        gamespec=1;
        }
    }

cpt=find("resdir");
if(cpt>=0)
if(ArgV[cpt+1])
	strcpy(resdir,ArgV[cpt+1]);

/*
cpt=find("mus_chan");
if(cpt>=0)
if(ArgV[cpt+1])
  	mus_chan = atoi(ArgV[cpt+1]);

cpt=find("sfx_chan");
if(cpt>=0)
if(ArgV[cpt+1])
        {
  	sfx_chan = atoi(ArgV[cpt+1]);
        if(sfx_chan < 4)
            sfx_chan = 4;
        }
*/

if(VideoMode==-1)
  {
  cpt=find("VideoMode");
  if(cpt>=0)
  if(ArgV[cpt+1])
  	VideoMode=atoi(ArgV[cpt+1]);
  }

cpt=find("console_x");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	conx=atoi(ArgV[cpt+1]);
    }

cpt=find("console_y");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	cony=atoi(ArgV[cpt+1]);
    }

cpt=find("console_w");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	conwid=atoi(ArgV[cpt+1]);
    else
        conwid = 900;
    if(conwid > 80)
        ithe_panic("Usage: -console_w <width>","Width must be in characters, 2 to 80.  It is NOT in pixels.");
    }

cpt=find("console_h");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	conlen=atoi(ArgV[cpt+1])-1;
    else
        conlen = 900;   // Not valid, therefore print the message below.
    if(conlen > 60)
        ithe_panic("Usage: -console_h <height>","Height must be in lines, 2 to 60.  It is NOT in pixels.");
    }

cpt=find("driftlevel");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	driftlevel=atoi(ArgV[cpt+1]);
    }

cpt=find("map_width");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	map_W=atoi(ArgV[cpt+1]);
    }

cpt=find("map_height");
if(cpt>=0)
    {
    if(ArgV[cpt+1])
  	map_H=atoi(ArgV[cpt+1]);
    }

cpt=find("view_w");
if(cpt>=0)
	{
	if(ArgV[cpt+1])
		VSW=atoi(ArgV[cpt+1]);
	}

cpt=find("view_h");
if(cpt>=0)
	{
	if(ArgV[cpt+1])
		VSH=atoi(ArgV[cpt+1]);
	}

cpt=find("loadingpic");
if(cpt>=0 && graflog ==-1)
	{
	if(ArgV[cpt+1])
		strcpy(loadingpic,ArgV[cpt+1]);
	else
		ithe_safepanic("Usage: -loadingpic MYFILE.PCX","The picture can be PCX, BMP or LBM (more later)");
	}

cpt=find("loadingpic_console");
if(cpt>=0)
	{
	if(ArgV[cpt+1])
		logx=atoi(ArgV[cpt+1]);
	if(ArgV[cpt+2])
		logy=atoi(ArgV[cpt+2]);
	if(ArgV[cpt+3])
		loglen=atoi(ArgV[cpt+3]);
	if(logx>128)
		logx=4;
	if(logy+(loglen*8)>480)
		logy=4;
	if(loglen>60)
                loglen=8;
	}

if(found("quickstart"))
    quickstart=1;

if(found("notimer"))
    ire_notimer=1;

if(found("show_imap"))
    show_imap = 1;
cpt=find("imapxy");
if(cpt>=0)
	{
	if(ArgV[cpt+1])
		imapx=atol(ArgV[cpt+1]);
	if(ArgV[cpt+2])
		imapy=atol(ArgV[cpt+2]);
	}

if(found("probe_ivo"))
    probeInvalidObject= 1;

cpt=find("destroy_hp");
if(cpt>=0)
if(ArgV[cpt+1])
  	destroyhp = atoi(ArgV[cpt+1]);

// Set limits for the script editor

cpt = find("max_sprites");
if(cpt>=0)
    spr_alloc = atol(ArgV[cpt+1]);

cpt = find("max_sequences");
if(cpt>=0)
    seq_alloc = atol(ArgV[cpt+1]);

cpt = find("max_vrms");
if(cpt>=0)
    vrm_alloc = atol(ArgV[cpt+1]);

cpt = find("max_characters");
if(cpt>=0)
    chr_alloc = atol(ArgV[cpt+1]);

cpt = find("max_tiles");
if(cpt>=0)
    til_alloc = atol(ArgV[cpt+1]);

cpt = find("max_wavs");
if(cpt>=0)
    wav_alloc = atol(ArgV[cpt+1]);

cpt = find("max_music");
if(cpt<0)
	cpt = find("max_mods");
if(cpt>=0)
    mus_alloc = atol(ArgV[cpt+1]);

cpt = find("max_pe_files");
if(cpt>=0)
    pef_alloc = atol(ArgV[cpt+1]);

// Starting date

cpt = find("start_minute");
if(cpt>=0)
    ig_min = atol(ArgV[cpt+1]);

cpt = find("start_hour");
if(cpt>=0)
    ig_hour = atol(ArgV[cpt+1]);

cpt = find("start_day");
if(cpt>=0)
    ig_day = atol(ArgV[cpt+1]);

cpt = find("start_month");
if(cpt>=0)
    ig_month = atol(ArgV[cpt+1]);

cpt = find("start_year");
if(cpt>=0)
    ig_year = atol(ArgV[cpt+1]);

cpt = find("start_day_of_month");
if(cpt>=0)
    ig_year = atol(ArgV[cpt+1]);

// Calendar definitions

cpt = find("minutes_per_hour");
if(cpt>=0)
    min_hour = atol(ArgV[cpt+1]);

cpt = find("hours_per_day");
if(cpt>=0)
    hour_day = atol(ArgV[cpt+1]);

cpt = find("days_per_week");
if(cpt>=0)
    day_week = atol(ArgV[cpt+1]);

cpt = find("days_per_month");
if(cpt>=0)
    day_month = atol(ArgV[cpt+1]);

cpt = find("months_per_year");
if(cpt>=0)
    month_year = atol(ArgV[cpt+1]);

// More things

cpt=find("file");
if(cpt>=0)
	for(cpt++;cpt<ArgC;cpt++)
		if(ArgV[cpt])
          	{
//        	RAR_Add(ArgV[cpt]);
//			ilog_quiet("Add %s\n",ArgV[cpt]);
			printf("Add %s\n",ArgV[cpt]);
			adddatafile(ArgV[cpt]);
            }

// It is of vital importance that -LIST comes last, or the INI parser will
// fall over, when the globals ArgV and ArgC are clobbered.

cpt=find("list");
if(cpt>=0)
if(ArgV[cpt+1])
    INI_file(ArgV[cpt+1]);

return;
}

short OSCLI_find(short ArgC,char **ArgV,char *param)
{
short ctr;
char tp[]=COOKIE_OscliTmp;

strcpy(tp,"-");
strcat(tp,param);
//ilog_quiet("Look for %s\n",tp);
for(ctr=0;ctr<ArgC;ctr++) // was <=
	if(ArgV[ctr])
		if(stricmp(tp,ArgV[ctr])==0)
			return(ctr);

strcpy(tp,"--");
strcat(tp,param);
for(ctr=0;ctr<ArgC;ctr++)
	if(ArgV[ctr])
		if(stricmp(tp,ArgV[ctr])==0)
			return(ctr);

#if defined(__DJGPP__) || defined(_WIN32) // Only for the DOS families
strcpy(tp,"/");
strcat(tp,param);
for(ctr=0;ctr<ArgC;ctr++)
	if(ArgV[ctr])
		if(stricmp(tp,ArgV[ctr])==0)
			return(ctr);
#endif

return -1;
}

char OSCLI_found(short ArgC,char **ArgV,char *param)
{
if(OSCLI_find(ArgC,ArgV,param) >= 0)
	return 1;
return 0;
}

void INI_file(char *ini)
{
int words,len,ctr;
char linebuf[1024];
char *word[64];
char *line;
FILE *fp;

//printf("Processing: %s\n",ini);


fp=fopen(ini,"r");
if(!fp)
	ithe_safepanic("Could not open listfile:",ini);

	do	{
	line=fgets(linebuf,1023,fp);
	if(line)
		if(line[0]!='#'&&line[0]!=';'&&line[0]!='%')
			{
//			puts(line);
			words=0;
			word[words++]=&line[0];
			len=strlen(line);                     
			for(ctr=0;ctr<len&&words<15;ctr++)
				if(line[ctr]==' '||line[ctr]=='\t')
					{
					line[ctr]=0;
					word[words++]=&line[ctr+1];
					}
				else
					if(line[ctr]<' ')
						line[ctr]=0;
			OSCLI(words,word);
			}
	} while(line);
fclose(fp);
}

