/*
 *      PissEasy compiler Virtual Machine
 *
 *      A lot of this code will need checking/fixing on 64-bit architectures
 *      to prevent 64-bit pointers being truncated by 32-bit pointer math
 */

#define PTR int	// This must match pointer size but support binary ops



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <allegro.h>

#include "../ithelib.h"
#include "../console.h"
#include "../oscli.h"
#include "../vrm_in.hpp"
#include "../object.hpp"
#include "opcodes.h"
#include "pe_api.hpp"

#ifdef __BEOS__
#define _WIN32
#endif


// Game manipulation functions:

#include "../core.hpp"
#include "../gamedata.h"
#include "../init.h"
#include "../media.h"
#include "../object.hpp"
#include "../linklist.hpp"
#include "../sound.h"
#include "../sys.hpp"
#include "../nuspeech.hpp"

extern void SetDarkness(int d);
extern void Call_VRM(int num);
extern char *BestName(OBJECT *o);
extern OBJECT *GameGetObject(int x,int y);
extern TILE *GetTile(int x,int y);
extern void MoveToTop(OBJECT *o);
extern void MoveToFloor(OBJECT *o);
extern void spillcontents(OBJECT *bag,int x,int y);
extern void gen_largemap(void);
extern void RedrawMap(void);
extern int isSolid(int x,int y);
extern void CheckHurt(OBJECT *list);
extern int GetBulk(OBJECT *obj);
extern int getYN(char *q);
extern int get_num(int no);
extern void ResyncEverything(void);
extern void Restart(void);
extern int get_key_debounced();
extern int AddToParty(OBJECT *new_member);
extern void object_sound(char *a, OBJECT *b);
extern int FindPath(OBJECT *start, OBJECT *end, int diagonal);
extern int SumObjects(OBJECT *c,char *o,int q);
extern int TakeQuantity(OBJECT *c,char *o,int q);
extern int MoveQuantity(OBJECT *s,OBJECT *d,char *o,int q);
extern void ResumeSchedule(OBJECT *o);
extern void CheckTime();
extern OBJECT *find_nearest(OBJECT *o, char *type);
extern void set_light(int x, int y, int x2, int y2, int light);


extern char *pe_get_function(int opcode);

extern char user_input[128];
extern "C" char DebugVM;
extern int do_lightning;
extern int irekey;
extern int pevm_err;

// Defines

#define VMOP(x) vmop[PEVM_##x] = PV_##x;
#define OPMASK 15 // Max 16 operators (at present)

// Variables

static int cur_x=0;		// Text cursor X,Y
static int cur_y=0;
static int whilebreak=0;
static int printlogging=0;
static int abortkernel=0;
static char vm_up=0;

char debcurfunc[1024];
char debmsg[1024];

// Functions

static void PV_Invalid();
static void PV_Return();
static void PV_CallVRM();
static void PV_Printstr();
static void PV_Printint();
static void PV_PrintXint();
static void PV_PrintCR();
static void PV_ClearLine();
static void PV_Callfunc();
static void PV_Let_iei();
static void PV_Let_ieipi();
static void PV_Add();
static void PV_Goto();
static void PV_If_iei();
static void PV_If_oics();
static void PV_If_oeo();
static void PV_If_oEo();
static void PV_If_o();
static void PV_If_no();
static void PV_For();
static void PV_Searchin();
static void PV_Listafter();
static void PV_PlayMusic();
static void PV_PlaySound();
static void PV_ObjSound();
static void PV_Create();
static void PV_CreateS();
static void PV_Destroy();
static void PV_SetDarkness();
static void PV_SetFlag();
static void PV_GetFlag();
static void PV_ResetFlag();
static void PV_If_on();
static void PV_If_non();
static void PV_If_tn();
static void PV_If_ntn();
static void PV_MoveObject();
static void PV_XferObject();
static void PV_ShowObject();
static void PV_PrintXYi();
static void PV_PrintXYs();
static void PV_GotoXY();
static void PV_TextColour();
static void PV_SetSequenceN();
static void PV_SetSequenceS();
static void PV_Lightning();
static void PV_Bestname();
static void PV_GetObject();
static void PV_GetSolidObject();
static void PV_GetFirstObject();
static void PV_GetTile();
static void PV_GetObjectBelow();
static void PV_ChangeObject();
static void PV_ChangeObjectS();
static void PV_ReplaceObject();
static void PV_ReplaceObjectS();
static void PV_SetDir();
static void PV_RedrawText();
static void PV_RedrawMap();
static void PV_MoveToPocket();
static void PV_XferToPocket();
static void PV_MoveFromPocket();
static void PV_ForceFromPocket();
static void PV_Spill();
static void PV_SpillXY();
static void PV_FadeOut();
static void PV_FadeIn();
static void PV_MoveToTop();
static void PV_MoveToFloor();
static void PV_If_LOS();
static void PV_CheckHurt();
static void PV_IfSolid();
static void PV_SetPFlag();
static void PV_GetPFlag();
static void PV_SetUFlag();
static void PV_GetUFlag();
static void PV_SetLocal();
static void PV_GetLocal();
static void PV_WeighObject();
static void PV_GetBulk();
static void PV_InputInt();
static void PV_IfInPocket();
static void PV_IfNInPocket();
static void PV_ReSyncEverything();
static void PV_TalkTo1();
static void PV_TalkTo2();
static void PV_Random();
static void PV_GetYN();
static void PV_GetYNN();
static void PV_Restart();
static void PV_ScrTileN();
static void PV_ScrTileS();
static void PV_While1();
static void PV_While2();
static void PV_Break();
static void PV_GetKey();
static void PV_DoAct();
static void PV_DoActS();
static void PV_DoActTo();
static void PV_DoActSTo();
static void PV_StopAct();
static void PV_ResumeAct();
static void PV_GetFuncV();
static void PV_GetFuncP();
static void PV_GetContainer();
static void PV_SetLeader();
static void PV_SetMember();
static void PV_AddMember();
static void PV_DelMember();
static void PV_MoveTowards8();
static void PV_MoveTowards4();
static void PV_WaitFor();
static void PV_WaitForAnimation();
static void PV_If_oonscreen();
static void PV_If_not_oonscreen();
static void PV_TakeQty();
static void PV_MoveQty();
static void PV_CountQty();
static void PV_CopySchedule();
static void PV_AllOnscreen();
static void PV_CheckTime();
static void PV_FindNear();
static void PV_FindTag();
static void PV_SetLight();
static void PV_LastOp();
static void PV_MovePartyToObj();
static void PV_MovePartyFromObj();

static void PV_Printaddr();
static void PV_PrintLog();
static void PV_Dump();
static void PV_NOP();
static void PV_Dofus();

static int OP_Invalid(int a, int b);
static int OP_Add(int a, int b);
static int OP_Sub(int a, int b);
static int OP_Mul(int a, int b);
static int OP_Div(int a, int b);
static int OP_Equal(int a, int b);
static int OP_NotEqual(int a, int b);
static int OP_Less(int a, int b);
static int OP_LessEq(int a, int b);
static int OP_Great(int a, int b);
static int OP_GreatEq(int a, int b);
static int OP_Modulus(int a, int b);

static int *GET_INT();
static int *SAFE_INT();


static void SI_RecursiveLoop(OBJECT *obj, int funccall);

void InitVM();
void CallVM(char *function);
void CallVMnum(int pos);
void DumpVM();
static void VMbug();


static PEVM *AddVM();
static void DelVM();
static void microkernel();
static void Bang(char *errmsg);
static PEVM *vmstack;
static PEVM *curvm;
static char *lastip=NULL;

static void (*vmop[256])(void);
static int (*Operator[16])(int x,int y);
static char *oplist[16];

// Code

inline unsigned char GET_BYTE()
{
return (unsigned char)(*curvm->ip++);
}

#define CHECK_POINTER_CORE(x,f,l) \
if(!x)\
	{\
	Bug("ERROR:  Null pointer in %s at %d\n",f,l);\
	if(DebugVM)\
		{\
		if(curvm)\
			{\
			Bug("        Dumping VM to bootlog.txt...\n");\
			Bug("(ip = 0x%x)\n",curvm->ip-curvm->code);\
			DumpVM();\
			}\
		else\
			{\
			Bug("No running VM!\n");\
			return;\
			}\
		}\
	else\
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");\
	Bug("        Current VM cannot continue.\n");\
	VMbug();\
	Bug("Terminating crashed VM...\n");\
	DelVM();\
	return;\
	}

#define CHECK_POINTER(x); CHECK_POINTER_CORE(x,__FILE__,__LINE__);


inline int GET_DWORD()
{
int x;
x = *(int *)(++curvm->ip); // Skip addressing byte
curvm->ip+=sizeof(int);
return x;
}

inline int *GET_DWORD_PTR()
{
int *x;
x = (int *)(curvm->ip);
curvm->ip+=sizeof(int *);
return x;
}


inline void *GET_PTR()
{
void *x;
x = *(void **)(curvm->ip);
curvm->ip+=sizeof(void *);
return x;
}

inline char *GET_STR(int len)
{
char *x;
x = curvm->ip;
curvm->ip+=len;
return x;
}

inline int *GET_MEMBER_PTR()
{
TILE *o,t;
char **obj;
int *ptr,**pointer;
void *var2;
unsigned char refs,ctr;
char debugchain[2048];
char msg[128];

//obj = (OBJECT **)GET_PTR();
obj = (char **)GET_PTR();
ptr = (int *)*obj;

if(DebugVM)
	sprintf(debugchain,"%x->",obj);

refs = GET_BYTE();

for(ctr=0;ctr<refs;ctr++)
	{
	var2 = (void *)GET_PTR();
	ptr = (int *)((char *)ptr + (PTR)var2);
	pointer = (int **)ptr;
	ptr = (int *)*ptr;
	if(DebugVM)
		{
		sprintf(msg,"%x->",pointer);
		strcat(debugchain,msg);
		}

	if(!ptr)
		{
		Bug("DANGER!  Encountered NULL data in structure lookup!\n");
		if(DebugVM)
			{
			Bug("         Dereference chain:\n");
			Bug("         %sNULL\n",debugchain);
			Bug("         Dumping VM to bootlog.txt...\n");
			DumpVM();
			}
		else
			Bug("         If repeatable, run IRE with -DebugVM for more info.\n");
		Bug("         Current VM cannot continue.\n");
		VMbug();
		Bug("Terminating crashed VM...\n");
		DelVM();
		return NULL;
		}
	}
return (int *)pointer;
}

inline int *GET_ARRAY()
{
int **array;
int *idx;
int size;
char msg[128];

array = (int **)GET_PTR();
//ilog_quiet("array base = %x\n",array);
size = GET_DWORD();
//ilog_quiet("size = %d\n",size);
idx = (int *)GET_INT();
//ilog_quiet("idx = %x\n",*idx);

if(!idx)
	{
	Bug("ERROR:  Array pointer corrupt\n");
	if(DebugVM)
		{
		Bug("        Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("        Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

// Make sure it doesn't exceed array bounds

if(*idx<0 || *idx > size)
	{
	Bug("ERROR:  Array access out of bounds\n");
	Bug("        Array is %d elements long, tried to access %d\n",size,*idx);
	if(DebugVM)
		{
		Bug("        Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("        Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

if(DebugVM)
	{
	ilog_quiet("Before\n");
	ilog_quiet("%x/%x\n",*idx,size);
	ilog_quiet("array = %x\n",array);
	}

//(*array) += *idx; // Now points to appropriate element (it had better, anyway)
array += *idx; // Now points to appropriate element (it had better, anyway)

if(DebugVM)
	{
	ilog_quiet("After\n");
	ilog_quiet("%x/%x\n",*idx,size);
	ilog_quiet("Now array = %x\n",array);
	}

M_chk(array);

return (int *)array;
}


inline int *CHECK_ARRAY()
{
int **array;
int *idx;
int size;
char msg[128];

array = (int **)GET_PTR();
size = GET_DWORD();
idx = (int *)GET_INT();

if(!idx)
	return NULL;

//ilog_quiet("(%d/%d) ",*idx,size);

// Make sure it doesn't exceed array bounds

if(*idx<0 || *idx > size)
	return NULL; // Bad

if(!array)
	return NULL;

//(*array) += *idx; // Now points to appropriate element (it had better, anyway)
return (int *)array;
}

inline int *GET_MEMBER()
{
OBJECT **obj;
int *ptr,**pointer;
void *var2;
unsigned char refs,ctr;
char debugchain[2048];
char msg[128];

obj = (OBJECT **)GET_PTR();
ptr = (int *)*obj;
if(!ptr)
	{
	Bug("DANGER!  Trying to access NULL structure\n");
	if(DebugVM)
		{
		Bug("         Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("         If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("         Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

if(DebugVM)
	sprintf(debugchain,"%x->",obj);

refs = GET_BYTE();
for(ctr=0;ctr<refs;ctr++)
	{
	var2 = (void *)GET_PTR();
	ptr = (int *)((char *)ptr + (PTR)var2);
	pointer = (int **)ptr;
	if(ptr)
		ptr = (int *)*ptr;
	if(DebugVM)
		{
		sprintf(msg,"%x->",pointer);
		strcat(debugchain,msg);
		}

	if(!ptr) // Ignore the last one
	if(ctr<refs-1)
		{
		Bug("DANGER!  Encountered NULL data in structure lookup!\n");
		if(DebugVM)
			{
			Bug("         Dereference chain:\n");
			Bug("         %sNULL\n",debugchain);
			Bug("         Dumping VM to bootlog.txt...\n");
			DumpVM();
			}
		else
			Bug("         If repeatable, run IRE with -DebugVM for more info.\n");
		Bug("         Current VM cannot continue.\n");
		VMbug();
		Bug("Terminating crashed VM...\n");
		DelVM();
		return NULL;
		}
	}
return (int *)pointer;
}

// Traverse the chain and make sure it's accessible.
// Although we want to abort on an error, we MUST traverse the whole thing
// otherwise the VM's EIP will be pointing at at the wrong address and the
// VM will crash.

inline int *CHECK_MEMBER()
{
OBJECT **obj;
int *ptr,**pointer;
void *var2;
unsigned char refs,ctr;
int good;

good=1;
obj = (OBJECT **)GET_PTR();
if(!obj)
	good=0;
if(good)
	{
	ptr = (int *)*obj;
	if(!ptr)
		good=0;
	}

refs = GET_BYTE();
for(ctr=0;ctr<refs;ctr++)
	{
	var2 = (void *)GET_PTR();
	if(good)
		{
		ptr = (int *)((char *)ptr + (PTR)var2);
		pointer = (int **)ptr;
		ptr = (int *)*ptr;
		}
	if(!ptr)
		good=0;
	}
if(good)
	return (int *)pointer;
return NULL;
}

inline int *SKIP_MEMBER()
{
OBJECT **obj;
void *var2;
unsigned char refs,ctr;
obj = (OBJECT **)GET_PTR();
refs = GET_BYTE();
for(ctr=0;ctr<refs;ctr++)
	var2 = (void *)GET_PTR();
return (int *)obj;
}

inline int *GET_ARRAY_MEMBER_PTR()
{
OBJECT **obj;
void *var2;
unsigned char refs,ctr;
int *ptr,**pointer;
int **array;
int *idx;
int size;
char msg[128];
char debugchain[2048];

array = (int **)GET_PTR();
//ilog_quiet("Marray = %x\n",array);
size = GET_DWORD();
//ilog_quiet("Msize = %d\n",size);
idx = (int *)GET_INT();
//ilog_quiet("Midx = %x\n",idx);

//obj = (OBJECT **)GET_PTR();
refs = GET_BYTE();
//ilog_quiet("Refs = %d\n",refs);

if(!idx || !array)
	{
	Bug("ERROR:  Array pointer corrupt\n");
	if(DebugVM)
		{
		Bug("        Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("        Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

// Make sure it doesn't exceed array bounds

if(*idx<0 || *idx > size)
	{
	Bug("ERROR:  Array access out of bounds\n");
	Bug("        Array is %d elements long, tried to access %d\n",size,*idx);
	if(DebugVM)
		{
		Bug("        Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("        Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

if(DebugVM)
	{
	ilog_quiet("Before\n");
	ilog_quiet("%x/%x\n",*idx,size);
	ilog_quiet("array = %x\n",array);
	}

//(*array) += *idx; // Now points to appropriate element (it had better, anyway)
array += *idx; // Now points to appropriate element (it had better, anyway)

// Okay, now do the member lookup
ptr = *array;

//refs = GET_BYTE();
for(ctr=0;ctr<refs;ctr++)
	{
	var2 = (void *)GET_PTR();
	ptr = (int *)((char *)ptr + (PTR)var2);
	pointer = (int **)ptr;
	ptr = (int *)*ptr;
	if(DebugVM)
		{
		sprintf(msg,"%x->",pointer);
		strcat(debugchain,msg);
		}

	if(!ptr) // Ignore the last one
		{
		Bug("DANGER!  Encountered NULL data in structure lookup!\n");
		if(DebugVM)
			{
			Bug("         Dereference chain:\n");
			Bug("         %sNULL\n",debugchain);
			Bug("         Dumping VM to bootlog.txt...\n");
			DumpVM();
			}
		else
			Bug("         If repeatable, run IRE with -DebugVM for more info.\n");
		Bug("         Current VM cannot continue.\n");
		VMbug();
		Bug("Terminating crashed VM...\n");
		DelVM();
		return NULL;
		}
	}


if(DebugVM)
	{
	ilog_quiet("After\n");
	ilog_quiet("%x/%x\n",*idx,size);
	ilog_quiet("array = %x\n",array);
	}

return (int *)pointer;
}

inline int *GET_ARRAY_MEMBER()
{
OBJECT **obj;
void *var2;
unsigned char refs,ctr;
int *ptr,**pointer;
int **array;
int *idx;
int size;
char msg[128];
char debugchain[2048];

array = (int **)GET_PTR();
//ilog_quiet("array = %x\n",array);
size = GET_DWORD();
//ilog_quiet("size = %d\n",size);
idx = (int *)GET_INT();
//ilog_quiet("idx = %x\n",idx);

//obj = (OBJECT **)GET_PTR();
refs = GET_BYTE();
//ilog_quiet("Refs = %d\n",refs);

if(!idx || !array)
	{
	Bug("ERROR:  Array pointer corrupt\n");
	if(DebugVM)
		{
		Bug("        Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("        Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

// Make sure it doesn't exceed array bounds

if(*idx<0 || *idx > size)
	{
	Bug("ERROR:  Array access out of bounds\n");
	Bug("        Array is %d elements long, tried to access %d\n",size,*idx);
	if(DebugVM)
		{
		Bug("        Dumping VM to bootlog.txt...\n");
		DumpVM();
		}
	else
		Bug("        If repeatable, run IRE with -DebugVM for more info.\n");
	Bug("        Current VM cannot continue.\n");
	VMbug();
	Bug("Terminating crashed VM...\n");
	DelVM();
	return NULL;
	}

if(DebugVM)
	{
	ilog_quiet("Before\n");
	ilog_quiet("%x/%x\n",*idx,size);
	ilog_quiet("array = %x\n",array);
	}

//(*array) += *idx; // Now points to appropriate element (it had better, anyway)
array += *idx; // Now points to appropriate element (it had better, anyway)

// Okay, now do the member lookup
ptr = *array;

//refs = GET_BYTE();
for(ctr=0;ctr<refs;ctr++)
	{
	var2 = (void *)GET_PTR();
	ptr = (int *)((char *)ptr + (PTR)var2);
	pointer = (int **)ptr;
	ptr = (int *)*ptr;
	if(DebugVM)
		{
		sprintf(msg,"%x->",pointer);
		strcat(debugchain,msg);
		}

	if(!ptr) // Ignore the last one
	if(ctr<refs-1)
		{
		Bug("DANGER!  Encountered NULL data in structure lookup!\n");
		if(DebugVM)
			{
			Bug("         Dereference chain:\n");
			Bug("         %sNULL\n",debugchain);
			Bug("         Dumping VM to bootlog.txt...\n");
			DumpVM();
			}
		else
			Bug("         If repeatable, run IRE with -DebugVM for more info.\n");
		Bug("         Current VM cannot continue.\n");
		VMbug();
		Bug("Terminating crashed VM...\n");
		DelVM();
		return NULL;
		}
	}


if(DebugVM)
	{
	ilog_quiet("After\n");
	ilog_quiet("%x/%x\n",*idx,size);
	ilog_quiet("array = %x\n",array);
	}

return (int *)pointer;
}

inline int *SKIP_ARRAY_MEMBER()
{
OBJECT **obj;
void *var2;
unsigned char refs,ctr;
int **array;
int *idx;
int size;
char msg[128];

array = (int **)GET_PTR();
size = GET_DWORD();
idx = (int *)GET_INT();

if(!idx)
	return NULL;

//ilog_quiet("(%d/%d) ",*idx,size);

//obj = (OBJECT **)GET_PTR();
refs = GET_BYTE();
//ilog_quiet("[Refs = %d]\n",refs);
for(ctr=0;ctr<refs;ctr++)
	{
	var2 = (void *)GET_PTR();
//	ilog_quiet("<%x> ",var2);
	}

// Make sure it doesn't exceed array bounds

if(*idx<0 || *idx > size)
	return NULL; // Bad

if(!array)
	return NULL;

//(*array) += *idx; // Now points to appropriate element (it had better, anyway)
//return (int *)array;
return (int *)obj;
}



OBJECT **GET_OBJECT()
{
unsigned char flags;
flags = GET_BYTE();
if(flags&ACC_INDIRECT)
	return (OBJECT **)GET_PTR();
if(flags&ACC_MEMBER)
	return (OBJECT **)GET_MEMBER_PTR();
if(flags&ACC_ARRAY)
	return (OBJECT **)GET_ARRAY();
Bug("Fatal: Unknown data method '%x' in GET_OBJECT()\n",flags);
return NULL;
}

OBJECT **CHECK_OBJ()
{
unsigned char flags;
flags = GET_BYTE();
if(flags&ACC_INDIRECT)
	return (OBJECT **)GET_PTR();
if(flags&ACC_MEMBER)
	return (OBJECT **)CHECK_MEMBER();
if(flags&ACC_ARRAY)
	return (OBJECT **)GET_ARRAY();
Bug("Fatal: Unknown data method '%x' in CHECK_OBJ()\n",flags);
return NULL;
}

int *GET_INT()
{
unsigned char flags;
flags = GET_BYTE();
if(!flags)
	return GET_DWORD_PTR();
if(flags == ACC_INDIRECT)
	return (int *)GET_PTR();
if(flags == ACC_MEMBER)
	return GET_MEMBER();
if(flags == ACC_ARRAY)
	return (int *)GET_ARRAY();
if(flags == ACC_ARRAY| ACC_MEMBER)
	return (int *)GET_ARRAY_MEMBER();
Bug("Fatal: Unknown data method '%x' in GET_INT()\n",flags);
return NULL;
}

int *SAFE_INT()
{
unsigned char flags;
flags = GET_BYTE();
if(!flags)
	return GET_DWORD_PTR();
if(flags == ACC_INDIRECT)
	return (int *)GET_PTR();
if(flags == ACC_MEMBER)
	return SKIP_MEMBER();
if(flags == ACC_ARRAY)
	return (int *)CHECK_ARRAY();
if(flags == ACC_ARRAY | ACC_MEMBER)
	return (int *)SKIP_ARRAY_MEMBER();
Bug("Fatal: Unknown data method '%x' in GET_INT()\n",flags);
return NULL;
}

char *GET_STRING()
{
unsigned char len;
unsigned char flags;
char **str;

flags = GET_BYTE();
if(!flags || flags & ACC_IMMSTR)
	{
	len = GET_BYTE();
	return GET_STR(len);
	}
if(flags == ACC_INDIRECT)
	{
	str = (char **)GET_PTR();
	if(!str) return NULL;
	return *str;
	}
if(flags == ACC_MEMBER)
	{
	str = (char **)GET_MEMBER_PTR();
	if(!str) return NULL;
	return *str;
	}
if(flags == ACC_ARRAY|ACC_MEMBER)
	{
	str = (char **)GET_ARRAY_MEMBER_PTR();
	if(!str) return NULL;
	return *str;
	}
Bug("Fatal: Unknown data method '%x' in GET_STRING()\n",flags);
return NULL;
}


/*
 *      VM functions follow
 */


void InitVM()
{
int ctr;
// Set up opcode jump table

for(ctr=0;ctr<256;ctr++)
	vmop[ctr]=PV_Invalid;

// VMOP macro splices strings in preprocessor to make VM maintenance easier
// For example VMOP(Return);
// compiles as: vmop[PEVM_Return] = PV_Return;

VMOP(Return);
VMOP(CallVRM);
VMOP(Printint);
VMOP(Printstr);
VMOP(PrintXint);
VMOP(PrintCR);
VMOP(ClearLine);
VMOP(Callfunc);
VMOP(Let_iei);
VMOP(Let_ieipi);
VMOP(Add);
VMOP(Goto);
VMOP(If_iei);
VMOP(If_oics);
VMOP(If_oeo);
VMOP(If_oEo);
VMOP(If_o);
VMOP(If_no);
VMOP(For);
VMOP(Searchin);
VMOP(Listafter);
VMOP(PlayMusic);
VMOP(PlaySound);
VMOP(ObjSound);
VMOP(Create);
VMOP(CreateS);
VMOP(Destroy);
VMOP(SetFlag);
VMOP(GetFlag);
VMOP(ResetFlag);
VMOP(If_on);
VMOP(If_non);
VMOP(If_tn);
VMOP(If_ntn);
VMOP(MoveObject);
VMOP(XferObject);
VMOP(ShowObject);
VMOP(SetDarkness);
VMOP(GotoXY);
VMOP(PrintXYs);
VMOP(PrintXYi);
VMOP(TextColour);
VMOP(SetSequenceS);
VMOP(SetSequenceN);
VMOP(Lightning);
VMOP(Bestname);
VMOP(GetObject);
VMOP(GetSolidObject);
VMOP(GetFirstObject);
VMOP(GetTile);
VMOP(GetObjectBelow);
VMOP(ChangeObject);
VMOP(ChangeObjectS);
VMOP(ReplaceObject);
VMOP(ReplaceObjectS);
VMOP(SetDir);
VMOP(RedrawMap);
VMOP(RedrawText);
VMOP(MoveToPocket);
VMOP(XferToPocket);
VMOP(MoveFromPocket);
VMOP(ForceFromPocket);
VMOP(Spill);
VMOP(SpillXY);
VMOP(FadeOut);
VMOP(FadeIn);
VMOP(MoveToTop);
VMOP(MoveToFloor);
VMOP(If_LOS);
VMOP(CheckHurt);
VMOP(IfSolid);
VMOP(SetPFlag);
VMOP(GetPFlag);
VMOP(SetUFlag);
VMOP(GetUFlag);
VMOP(SetLocal);
VMOP(GetLocal);
VMOP(WeighObject);
VMOP(GetBulk);
VMOP(InputInt);
VMOP(IfInPocket);
VMOP(IfNInPocket);
VMOP(ReSyncEverything);
VMOP(TalkTo1);
VMOP(TalkTo2);
VMOP(Random);
VMOP(GetYN);
VMOP(GetYNN);
VMOP(Restart);
VMOP(ScrTileN);
VMOP(ScrTileS);
VMOP(While1);
VMOP(While2);
VMOP(Break);
VMOP(GetKey);
VMOP(DoAct);
VMOP(DoActS);
VMOP(DoActTo);
VMOP(DoActSTo);
VMOP(StopAct);
VMOP(ResumeAct);
VMOP(GetFuncV);
VMOP(GetFuncP);
VMOP(GetContainer);
VMOP(SetLeader);
VMOP(SetMember);
VMOP(AddMember);
VMOP(DelMember);
VMOP(MoveTowards8);
VMOP(MoveTowards4);
VMOP(WaitFor);
VMOP(WaitForAnimation);
VMOP(If_oonscreen);
VMOP(If_not_oonscreen);
VMOP(TakeQty);
VMOP(MoveQty);
VMOP(CountQty);
VMOP(CopySchedule);
VMOP(AllOnscreen);
VMOP(CheckTime);
VMOP(FindNear);
VMOP(FindTag);
VMOP(SetLight);
VMOP(Printaddr);
VMOP(PrintLog);
VMOP(MovePartyToObj);
VMOP(MovePartyFromObj);

// These are used internally by the compiler

VMOP(LastOp);
VMOP(Dump);
VMOP(NOP);
VMOP(Dofus);

// Now set up Operator jumptable

for(ctr=0;ctr<16;ctr++)
	Operator[ctr]=OP_Invalid;

Operator[1]=OP_Add;
Operator[2]=OP_Sub;
Operator[3]=OP_Div;
Operator[4]=OP_Mul;
Operator[5]=OP_Equal;
Operator[6]=OP_NotEqual;
Operator[7]=OP_Less;
Operator[8]=OP_LessEq;
Operator[9]=OP_Great;
Operator[10]=OP_GreatEq;
Operator[11]=OP_Modulus;

oplist[0]="<!>";
oplist[1]="+";
oplist[2]="-";
oplist[3]="/";
oplist[4]="*";
oplist[5]="=";
oplist[6]="!=";
oplist[7]="<";
oplist[8]="<=";
oplist[9]=">";
oplist[10]=">=";
oplist[11]="MOD";
oplist[12]="?12?";
oplist[13]="?13?";
oplist[14]="?14?";
oplist[15]="?15?";

vm_up=1;
ilog_quiet("Script VM initialised\n");
}

char *CurrentVM()
{
return debcurfunc;
}

void CallVM(char *function)
{
PEVM *p;
int pos,ctr;

if(!vm_up)
	ithe_panic("Tried to use PEVM before InitVM()","Aiee!");

if(!function)
    {
    Bug("Tried to call NULL function\n");
	VMbug();
    return;
    }

strcpy(debcurfunc,function);

pos = getnum4PE(function);
if(pos == -1)
    return; // Error already reported by getnum4PE

p=AddVM();
p->size = PElist[pos].size;
p->codelen = PElist[pos].codelen;
//ilog_quiet("s-Alloc code (%s)\n",function);
p->code = (char *)M_get(sizeof(char),p->size);
memcpy(p->code,PElist[pos].code,p->size);
//memset(p->reg,0,sizeof(OBJECT *)*4);
p->name = PElist[pos].name;
p->file = PElist[pos].file;
p->ip = p->code;
curvm = p;

DumpVM();

microkernel();
}

void VMbug()
{
PEVM *t;

if(!curvm)
	{
	Bug("Serious..  not in a function.  Error in game loop?\n");
	return;
	}
    
Bug("Error in function %s, in file %s\n",curvm->name,curvm->file);

ilog_quiet("\n");
ilog_quiet("VM Stack trace:\n");
ilog_quiet("Earliest\n");
for(t=vmstack;t;t=t->next)
	ilog_quiet("  %s (%s)\n",t->name,t->file);
ilog_quiet("Latest\n");
}

void CallVMnum(int pos)
{
PEVM *p;
int ctr;

if(!vm_up)
	ithe_panic("Tried to use PEVM before InitVM()","Aiee!");

if(pos < 0 || pos > PEtot)
    {
    Bug("CALL: Invalid function %d (0x%x)\n",pos,pos);
	VMbug();
    return;
    }

strcpy(debcurfunc,PElist[pos].name);

p=AddVM();
p->size = PElist[pos].size;
p->codelen= PElist[pos].codelen;
//ilog_quiet("n-Alloc code (%s)\n",debcurfunc);
p->code = (char *)M_get(sizeof(unsigned char),p->size);
memcpy(p->code,PElist[pos].code,p->size);
p->name = PElist[pos].name;
p->file = PElist[pos].file;
p->ip = p->code;
curvm = p;

DumpVM();

microkernel();
}


void microkernel()
{
unsigned char opcode;
int key;
static int kinst=0;

// Set up a mutex for the kernel
/*
if(kinst)
	return;
kinst=1;
*/

if(!curvm)
    ithe_panic("VM started on an empty stomach",NULL);

do  {
    lastip = curvm->ip;    // Preserve IP (for error messages)
    opcode = *curvm->ip++; // Fetch
	if(DebugVM)
		ilog_quiet("MK: %04x [%02x]\n",((unsigned long)lastip - (unsigned long)curvm->code),opcode);
	vmop[opcode]();        // Execute
	if(abortkernel)
		{
		abortkernel=0;
		return;
		}
    } while(curvm);

// Stop the kernel
//kinst=0;
}


void Bang(char *err)
{
char msg[128];

if(!curvm)
	{
	Bug("All virtual machines shut down.\n");
	return;
	}

VMbug();

ilog_quiet(">>> VIRTUAL MACHINE CRASHED...\n");
ilog_quiet(">>> DISASSEMBLY FOLLOWS...\n");
ilog_quiet(">>> IP = 0x%x\n",(int)lastip-(int)curvm->code);
ilog_quiet("==========================\n");
DebugVM=1;
DumpVM();
ilog_quiet("==========================\n");

sprintf(msg,"In function %s in file %s, check bootlog.txt",curvm->name,curvm->file);
ithe_panic(err,msg);
}

// AddVM - create a new VM at the top of the stack

PEVM *AddVM()
{
PEVM *t;

// Absorb spurious abort codes
abortkernel=0;

// If the subroutine stack is empty, most likely we're starting up the VM.
// create the new VM and return the address

if(!vmstack)
	{
//	ilog_quiet("vmst: Alloc VM\n");
	vmstack = (PEVM *)M_get(1,sizeof(PEVM));
	vmstack->next = NULL;
//	ilog_quiet("Add first VM: now running %x\n",vmstack);
	return vmstack;
	}

// Otherwise, seek to the end of the stack and add a new entry, return it

for(t=vmstack;t;t=t->next)
	if(t->next == NULL)
		{
//		ilog_quiet("vmad: Alloc VM\n");
		t->next = (PEVM *)M_get(1,sizeof(PEVM));
		t->next->next = NULL;
//		ilog_quiet("Add VM: now running %x\n",t->next);
		return t->next;
		}

// Oh no!

ithe_panic("AddVM: Whoops",NULL);
return NULL;
}

// DelVM - pop the top Virtual Machine off the stack

void DelVM()
{
PEVM *t;

// Kill the running VM
abortkernel=1;


if(!curvm)
	{
	Bug("DelVM: No running VM\n");
	return;
	}

// Tried to pop from empty subroutine stack, decide what to do

if(!vmstack)
	if(!curvm)
		Bang("Stack underflow");
	else
		Bang("Internal error: code 'LOVELY ON THE WATER'");

// Ok, seek to end of subroutine stack, erase last VM and pop to last-but-one

for(t=vmstack;t;t=t->next)
	if(t->next == curvm)
		{
		t->next->next = NULL;
//		ilog_quiet("vmde: Free Code\n");
		M_free(t->next->code);
//		ilog_quiet("vmde: Free VM\n");
		M_free(t->next);
		t->next=NULL;

		curvm = t;
//		ilog_quiet("Del VM: now running %x\n",t);
		return;
		}

// Uh, that didn't work, try this instead

if(vmstack == curvm)
	{
	vmstack->next = NULL;
//	ilog_quiet("vmdl: Kill Code\n");
	M_free(vmstack->code);
//	ilog_quiet("NOT KILLING %x\n",vmstack);
	M_free(vmstack);
	vmstack=NULL;
	curvm = NULL;
//	ilog_quiet("Del last VM: now NOT running\n");
	return;
	}

// Oh no!

ithe_panic("DelVM: Holy f---!  Could not delete virtual machine!",NULL);
}

void DumpVM()
{
int ctr,len;
unsigned int proglen;
int data;
char *func;
unsigned char opcode,op;
char *oldip;

if(!DebugVM)
	return;

if(!curvm)
	{
	ilog_quiet("DumpVM: Not running.\n");
	return;
	}

ilog_quiet("(%s)\n",curvm->name);
ilog_quiet("(%x-%x)\n",(unsigned int)curvm->code,(unsigned int)curvm->code+(unsigned int)curvm->size);

oldip = curvm->ip;
curvm->ip = curvm->code;
proglen = (unsigned int)curvm->codelen + (unsigned int)curvm->ip;

for(;(unsigned int)curvm->ip<proglen;)
	{
	if(curvm->ip == lastip)
		ilog_quiet("-> ");
	else
		ilog_quiet("   ");
	ilog_quiet("%04x ",curvm->ip-curvm->code);
	opcode = *curvm->ip++; // Fetch
	func=pe_get_function(opcode);
	ilog_quiet(" [%02x] %s ",opcode,&func[1]); // Skip first byte (no of operands)
	len = func[0];
	for(ctr=0;ctr<len;ctr++)
		{
		M_chk(curvm->ip);
		op = (unsigned char)*curvm->ip;
		switch(op)
			{
			case ACC_IMMEDIATE:
			ilog_quiet("%d ",GET_DWORD());
			break;

			case ACC_INDIRECT:
			ilog_quiet("[%x] ",GET_DWORD());
			break;

			case ACC_MEMBER:
			ilog_quiet("[%x] ",SAFE_INT());
			break;

			case ACC_ARRAY:
			ilog_quiet("[%x] ",SAFE_INT());
			break;

			case ACC_ARRAY | ACC_MEMBER:
			ilog_quiet("[%x] ",SAFE_INT());
			break;

			case ACC_IMMSTR:
			ilog_quiet("%s ",GET_STRING());
			break;

			case ACC_JUMP:
			ilog_quiet(" jmp->%x ",GET_DWORD());
			break;

			case ACC_OPERATOR|1:
			GET_BYTE();
			break;

			default:
			if(op & ACC_OPERATOR)
				{
				ilog_quiet("%s ",oplist[op&OPMASK]);
				GET_BYTE();
				}
			else
				ilog_quiet("??? ");
			};
		}
	if(len == -1)
		{
		ilog_quiet("%d ",GET_PTR());
		ilog_quiet("[%x] ",GET_PTR());
		}
	ilog_quiet("\n");
	}

curvm->ip = oldip;

}

// Helper functions for VM opcodes

void SI_RecursiveLoop(OBJECT *obj, int funccall)
{
OBJECT *temp;

for(temp = obj;temp;temp=temp->next)
    if(temp->flags.on)
        {
		current_object = temp;
		CallVMnum(funccall);
        if(temp->pocket)
            SI_RecursiveLoop(temp->pocket,funccall);
        }
}

/*
 *	Now the opcodes:
 */

void PV_Invalid()
{
char msg[128];
sprintf(msg,"Invalid VM opcode 0x%02x",(*lastip)&0xff);
Bang(msg);
}

void PV_Return()
{
DelVM();
if(DebugVM)
	DumpVM();
/*
if(curvm)
	ilog_quiet("VM now : %s\n",curvm->name);
else
	ilog_quiet("VM now empty\n");
*/
}

void PV_CallVRM()
{
int *vrm;
vrm = GET_INT();
CHECK_POINTER(vrm);
Call_VRM(*vrm);      // Call it
}

void PV_Printstr()
{
char *str;
str = GET_STRING();
//CHECK_POINTER(str);
if(!str)
	str="<null>";
irecon_printf(str); // Print it
if(DebugVM || printlogging)
	ilog_quiet("%s",str);
}

void PV_Printint()
{
int *num;

num = GET_INT();
CHECK_POINTER(num);

irecon_printf("%d",*num);
if(DebugVM || printlogging)
	ilog_quiet("%d",*num);
}


void PV_PrintXint()
{
int *num;
num = GET_INT();
CHECK_POINTER(num);

irecon_printf("%08x",*num);       // Print it
if(DebugVM || printlogging)
	ilog_quiet("%08x",*num);
}

void PV_PrintCR()
{
irecon_newline();
if(DebugVM || printlogging)
	ilog_quiet("\n");
}

void PV_ClearLine()
{
irecon_cls();
}

void PV_Callfunc()
{
int *func;

func = GET_INT();
CHECK_POINTER(func);
if(DebugVM)
	ilog_quiet("calling %d/%d\n",*func,PEtot);
CallVMnum(*func);
}

// LET <int> equal <number>

void PV_Let_iei()
{
int *a,*b;

a = GET_INT();
CHECK_POINTER(a);
b = GET_INT();
CHECK_POINTER(b);

*a = *b;
}

// LET <int> equal <int> [operator] <int>

void PV_Let_ieipi()
{
int *a,*b,*c;
unsigned char op;

a = GET_INT();
CHECK_POINTER(a);
b = GET_INT();
CHECK_POINTER(b);
op = GET_BYTE();
c = GET_INT();
CHECK_POINTER(c);

*a = Operator[op&OPMASK](*b,*c);
}

// ADD <int> [operator] <int>

void PV_Add()
{
int *a,*b;
unsigned char op;

a = GET_INT();
CHECK_POINTER(a);
op = GET_BYTE();
b = GET_INT();
CHECK_POINTER(b);

*a = Operator[op&OPMASK](*a,*b);
}


// Goto

void PV_Goto()
{
int jumpoffset;

jumpoffset = GET_DWORD();
curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
}

// IF <int> = <number> then continue, else jump to offset

void PV_If_iei()
{
int jumpoffset,*a,*b;
unsigned char op;

a = (int *)GET_INT();
CHECK_POINTER(a);
op=GET_BYTE();
//ilog_quiet("Operator %d\n",op&OPMASK);
b = (int *)GET_INT();
CHECK_POINTER(b);
jumpoffset = GET_DWORD();

if(!Operator[op&OPMASK](*a,*b))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF <object> Is Called <string> then continue, else jump to offset
// The string is stored first, not second

void PV_If_oics()
{
int jumpoffset;
char *str;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);
jumpoffset = GET_DWORD();

// If there was an accident, assume it's not true and continue

if(!*obj)
	{
	Bug("If object is called '%s': Object has not been initialised!\n",str);
	VMbug();
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	return; // Leave quickly
	}

// Ok, do the condition

if(stricmp((*obj)->name,str))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF <object> = <object> then continue, else jump to offset

void PV_If_oeo()
{
int jumpoffset;
OBJECT **a,**b,*c1,*c2;

a = CHECK_OBJ();
b = CHECK_OBJ();
jumpoffset = GET_DWORD();

// Just because it's NULL doesn't mean it can't be equal

if(a)
	c1=*a;
else
	c1=(OBJECT *)a;

if(b)
	c2=*b;
else
	c2=(OBJECT *)b;

// Ok, try the condition

if(!(c1 == c2))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF <object> != <object> then continue, else jump to offset

void PV_If_oEo()
{
int jumpoffset;
OBJECT **a,**b,*c1,*c2;

a = CHECK_OBJ();
b = CHECK_OBJ();
jumpoffset = GET_DWORD();

// Just because it's NULL doesn't mean it can't be equal

if(a)
	c1=*a;
else
	c1=(OBJECT *)a;

if(b)
	c2=*b;
else
	c2=(OBJECT *)b;

// Ok, try the condition

if(!(c1 != c2))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF <object> then continue, else jump to offset

void PV_If_o()
{
int jumpoffset;
OBJECT **obj;
int ok;

obj = CHECK_OBJ();
jumpoffset = GET_DWORD();

ok=0;
if(obj)
	if(*obj)
		ok=1;

if(!ok)
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

void PV_If_no()
{
int jumpoffset;
OBJECT **obj;
int ok;

obj = CHECK_OBJ();
jumpoffset = GET_DWORD();

ok=0;
if(obj)
	if(*obj)
		ok=1;

if(ok)
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}


// FOR <integer> = <number> TO <number> (continue, unless false, then jump exit)

void PV_For()
{
int jumpoffset,*control,*start,*end,delta,oob;
char *firsttime;

control = GET_INT();
CHECK_POINTER(control);
start = GET_INT();
CHECK_POINTER(start);
end = GET_INT();
CHECK_POINTER(end);

firsttime = curvm->ip++;      // This is the 'first time' flag
jumpoffset = GET_DWORD();

if(*start<*end)
    delta=1;
else
    delta=-1;

// Kludge to make it go the whole range

if(*firsttime & 1)
    {
    *control-=delta;
    *firsttime = ACC_OPERATOR;
    }

*control+=delta;

// Check for out of bounds

oob=0;
if(delta > 0)
   {
   if(*control > *end)
       oob=1;
   }
else
   {
   if(*control < *end)
       oob=1;
   }

// If we're out of bounds, stop

if(oob)
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    *firsttime=1|ACC_OPERATOR;           // Reset the first-time flag
    }
}

// Search inside an object's pockets calling a function as we go

void PV_Searchin()
{
OBJECT **obj;
OBJECT *old;
int func;

// First get the object we're going to search inside of

obj = GET_OBJECT();
CHECK_POINTER(obj);
func = GET_DWORD();

if(!*obj)
	{
	Bug("Search_inside Obj calling <func>: Obj not initialised!\n");
	VMbug();
	return;
	}

old = current_object;
SI_RecursiveLoop((*obj)->pocket,func);
current_object = old;
}

// List all objects after this one calling a function as we go

void PV_Listafter()
{
OBJECT **obj;
OBJECT *old,*ptr;
int func;

// First get the object we're going to search

obj = GET_OBJECT();
CHECK_POINTER(obj);
func = GET_DWORD();

if(!*obj)
	{
	Bug("List_After Obj calling <func>: Obj not initialised!\n");
	VMbug();
	return;
	}

old = current_object;
for(ptr=*obj;ptr;ptr=ptr->next)
	{
	current_object=ptr;
    CallVMnum(func);
	}
current_object = old;
}

void PV_Create()
{
OBJECT **obj,*temp;
int *var;

obj = GET_OBJECT();
CHECK_POINTER(obj);
var = GET_INT();
CHECK_POINTER(var);

temp = OB_Alloc();
temp->curdir = temp->dir[0];
OB_Init(temp,CHlist[*var].name);
OB_SetDir(temp,1);
TransferToPocket(temp,syspocket);
CreateContents(temp);

*obj = temp;
}

void PV_CreateS()
{
OBJECT **obj,*temp;
char *str;
int type;

obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);

type = getnum4char(str);
if(type == -1)
	{
	Bug("Create obj = %s : Could not find character '%s'\n",str,str);
	VMbug();
	return;
	}

temp = OB_Alloc();
temp->curdir = temp->dir[0];
OB_Init(temp,CHlist[type].name);
OB_SetDir(temp,1);
TransferToPocket(temp,syspocket);
CreateContents(temp);

*obj = temp;
}

void PV_Destroy()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

if(*obj == player)
	{
    Bug("Attempted to destroy the player!\n");
	VMbug();
	return;
	}

if(*obj)
	{
//	ilog_printf("pevm: destroy %s\n",(*obj)->name);
	(*obj)->flags.on = 0;   // Mark Pending-delete
	pending_delete=1;       // Tell the garbage collector it will be needed
	}

*obj = NULL;
}


void PV_PlayMusic()
{
S_PlayMusic(GET_STRING());
}

void PV_PlaySound()
{
S_PlaySample(GET_STRING(),sf_volume);
}

void PV_ObjSound()
{
OBJECT **obj;
char *str;

str = GET_STRING();
CHECK_POINTER(str);
obj = GET_OBJECT();
CHECK_POINTER(obj);

object_sound(str, *obj);
//S_PlaySample(GET_STRING(),sf_volume);
}

void PV_SetDarkness()
{
int *var,dl;
var = GET_INT();
CHECK_POINTER(var);
dl = *var;
if(dl<0)
	dl=0;
if(dl>255)
	dl=255;

SetDarkness(dl);
}

// set_flag <object> <flag> = <i>

void PV_SetFlag()
{
int *var,*flag;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
flag = GET_INT();
CHECK_POINTER(flag);
var = GET_INT();
CHECK_POINTER(var);

set_flag(*obj,*flag,*var);
}

// get_flag <i> = <object> <flag>

void PV_GetFlag()
{
int *var,*flag;
OBJECT **obj;

var = GET_INT();
CHECK_POINTER(var);
obj = GET_OBJECT();
CHECK_POINTER(obj);
flag = GET_INT();
CHECK_POINTER(flag);

*var = get_flag(*obj,*flag);
}

// reset_flag <object> <flag>

void PV_ResetFlag()
{
int *flag;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
flag = GET_INT();
CHECK_POINTER(flag);

default_flag(*obj,*flag);
}

// IF <object> <flag> then continue, else jump to offset

void PV_If_on()
{
int jumpoffset,*flag;
OBJECT **obj;

obj = CHECK_OBJ();	// Allow a NULL object

flag = GET_INT();
CHECK_POINTER(flag);
jumpoffset = GET_DWORD();

// If we don't have a valid object, return false
if(!obj)
	{
	curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	return;
	}

if(!get_flag(*obj,*flag))
	{
	curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	}
}

// IF NOT <object> <flag> then continue, else jump to offset

void PV_If_non()
{
int jumpoffset,*flag;
OBJECT **obj;

obj = CHECK_OBJ();	// Allow a NULL object

flag = GET_INT();
CHECK_POINTER(flag);
jumpoffset = GET_DWORD();

// If it's null, assume it's false (do nothing)
if(!obj)
	return;

if(get_flag(*obj,*flag))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF <tile> <flag> then continue, else jump to offset

void PV_If_tn()
{
int jumpoffset,*flag;
TILE **tile;

tile = (TILE **)GET_OBJECT();
CHECK_POINTER(tile);
flag = GET_INT();
CHECK_POINTER(flag);
jumpoffset = GET_DWORD();

if(!get_tileflag(*tile,*flag))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF NOT <tile> <flag> then continue, else jump to offset

void PV_If_ntn()
{
int jumpoffset,*flag;
TILE **tile;

tile = (TILE **)GET_OBJECT();
CHECK_POINTER(tile);
flag = GET_INT();
CHECK_POINTER(flag);
jumpoffset = GET_DWORD();

if(get_tileflag(*tile,*flag))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// Move <object> <x> <y> and set pe_result_ok accordingly

void PV_MoveObject()
{
int *x,*y;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

// Do it and record the result

if(move_object(*obj,*x,*y))
	pe_result_ok=1;
else
	pe_result_ok=0;
}

// Forcibly move <object> <x> <y>

void PV_XferObject()
{
int *x,*y;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

transfer_object(*obj,*x,*y);
}

// Display <object> onscreen at <x> <y>

void PV_ShowObject()
{
int *x,*y;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

if(!*obj)
	return;
draw_rle_sprite(swapscreen,(*obj)->form->seq[0]->image,*x,*y);
}

void PV_PrintXYs()
{
char *str;
str = GET_STRING();
CHECK_POINTER(str);

irecon_printxy(cur_x,cur_y,str); // Print it
cur_x += text_length(font,str);
}

void PV_PrintXYi()
{
int *num;
char str[128];

num = GET_INT();
CHECK_POINTER(num);

sprintf(str,"%d",*num);
irecon_printxy(cur_x,cur_y,str); // Print it
cur_x += text_length(font,str);
}

void PV_GotoXY()
{
int *x,*y;

x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

cur_x = *x;
cur_y = *y;
}

void PV_TextColour()
{
int *r,*g,*b;

r = GET_INT();
CHECK_POINTER(r);
g = GET_INT();
CHECK_POINTER(g);
b = GET_INT();
CHECK_POINTER(b);

irecon_colour(*r,*g,*b);
}

void PV_SetSequenceS()
{
char *str;
OBJECT **obj;
obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);

OB_SetSeq(*obj,str);
}

void PV_SetSequenceN()
{
OBJECT **obj;
int *n;
obj = GET_OBJECT();
CHECK_POINTER(obj);
n = GET_INT();
CHECK_POINTER(n);

OB_SetSeq(*obj,SQlist[*n].name);
}

void PV_Lightning()
{
do_lightning=GET_DWORD();
}

void PV_Bestname()
{
OBJECT **obj;
char **n;
obj = GET_OBJECT();
CHECK_POINTER(obj);
n = (char **)GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(n);

*n=BestName(*obj);
}

void PV_GetObject()
{
OBJECT **obj;
int *x,*y;
obj = (OBJECT **)GET_INT(); // Hack: get pointer without final NULL check
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

*obj = GameGetObject(*x,*y);
}

void PV_GetSolidObject()
{
OBJECT **obj;
int *x,*y;
obj = (OBJECT **)GET_INT(); // Hack: get pointer without final NULL check
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

*obj = GetSolidObject(*x,*y);
}

void PV_GetFirstObject()
{
OBJECT **obj;
int *x,*y;
obj = (OBJECT **)GET_INT(); // Hack: get pointer without final NULL check
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

*obj = GetObjectBase(*x,*y);
}

void PV_GetTile()
{
TILE **tile;
int *x,*y;
tile = (TILE **)GET_INT(); // Hack: get pointer without final NULL check
CHECK_POINTER(tile);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

*tile = GetTile(*x,*y);
}

void PV_GetObjectBelow()
{
OBJECT **obj1,**obj2;
obj1 = (OBJECT **)GET_INT(); // Hack: get pointer without final NULL check
CHECK_POINTER(obj1);
obj2 = GET_OBJECT();
CHECK_POINTER(obj2);

*obj1 = get_object_below(*obj2);
}

void PV_ChangeObject()
{
OBJECT **obj;
int *var;

STATS ostats;
char resname[32];
char pname[32];
char speech[128];
int tcache;
int partie;     // Must preserve this for dead party members

obj = GET_OBJECT();
CHECK_POINTER(obj);

var = GET_INT();
CHECK_POINTER(var);

//ilog_quiet("<%x>\n",*obj);
if(!*obj)
	{
	if(*var)
		Bug("ChangeObject <null> %s\n",CHlist[*var].name);
	else
		Bug("ChangeObject <null> <null>\n");
	return;
	}

partie = (*obj)->flags.party;
memcpy(&ostats,(*obj)->stats,sizeof(STATS));
strcpy(resname,(*obj)->funcs->resurrect);
strcpy(speech,(*obj)->funcs->talk);
tcache = (*obj)->funcs->tcache;
strcpy(pname,(*obj)->personalname);
OB_Init(*obj,CHlist[*var].name);
memcpy((*obj)->stats,&ostats,sizeof(STATS));
strcpy((*obj)->funcs->resurrect,resname);
strcpy((*obj)->personalname,pname);
strcpy((*obj)->funcs->talk,speech);
(*obj)->flags.party = partie;
(*obj)->funcs->tcache = tcache;
}


void PV_ChangeObjectS()
{
OBJECT **obj;
char *str;
int type;

STATS ostats;
char resname[32];
char pname[32];
char speech[128];
int tcache;
int partie;     // Must preserve this for dead party members

obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);

type = getnum4char(str);
if(type == -1)
	{
	Bug("Change obj %s : Could not find character '%s'\n",str,str);
	VMbug();
	return;
	}

partie = (*obj)->flags.party;
memcpy(&ostats,(*obj)->stats,sizeof(STATS));
strcpy(resname,(*obj)->funcs->resurrect);
strcpy(speech,(*obj)->funcs->talk);
tcache = (*obj)->funcs->tcache;
strcpy(pname,(*obj)->personalname);
OB_Init(*obj,CHlist[type].name);
memcpy((*obj)->stats,&ostats,sizeof(STATS));
strcpy((*obj)->funcs->resurrect,resname);
strcpy((*obj)->personalname,pname);
strcpy((*obj)->funcs->talk,speech);
(*obj)->flags.party = partie;
(*obj)->funcs->tcache = tcache;
}

void PV_ReplaceObject()
{
OBJECT **obj;
int *var,k;

obj = GET_OBJECT();
CHECK_POINTER(obj);
var = GET_INT();
CHECK_POINTER(var);

k=fullrestore;
fullrestore=0;
OB_Init(*obj,CHlist[*var].name);
fullrestore=k;
gen_largemap();     // Recalc solid objects cache
}

void PV_ReplaceObjectS()
{
OBJECT **obj;
int type,k;
char *str;

obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);

type = getnum4char(str);
if(type == -1)
	{
	Bug("Replace obj %s : Could not find character '%s'\n",str,str);
	VMbug();
	return;
	}

k=fullrestore;
fullrestore=0;
OB_Init(*obj,CHlist[type].name);
fullrestore=k;
gen_largemap();     // Recalc solid objects cache
}

void PV_SetDir()
{
OBJECT **obj;
int *n;
obj = GET_OBJECT();
CHECK_POINTER(obj);
n = GET_INT();
CHECK_POINTER(n);
OB_SetDir(*obj,*n);
}

void PV_RedrawText()
{
irecon_update();            // Redraw all text output
Show();                     // And blit the screen again
}

void PV_RedrawMap()
{
RedrawMap();
Show();                     // And blit the screen again
}

// Forcibly move <object> <to> <pocket>

void PV_MoveToPocket()
{
OBJECT **pocket,**obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
pocket = GET_OBJECT();
CHECK_POINTER(pocket);

move_to_pocket(*obj,*pocket);
}

// Forcibly move <object> <to> <pocket>

void PV_XferToPocket()
{
OBJECT **pocket,**obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
pocket = GET_OBJECT();
CHECK_POINTER(pocket);

transfer_to_pocket(*obj,*pocket);
}

// move <object> <from> <pocket> to <x> <y>

void PV_MoveFromPocket()
{
OBJECT **pocket,**obj;
int *x,*y;

obj = GET_OBJECT();
CHECK_POINTER(obj);
pocket = GET_OBJECT();
CHECK_POINTER(pocket);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

MoveFromPocket(*obj,*pocket,*x,*y);
}

// Forcibly move <object> <from> <pocket> to <x> <y>

void PV_ForceFromPocket()
{
OBJECT **pocket,**obj;
int *x,*y;

obj = GET_OBJECT();
CHECK_POINTER(obj);
pocket = GET_OBJECT();
CHECK_POINTER(pocket);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

ForceFromPocket(*obj,*pocket,*x,*y);
}

// Spill contents of <object> at current location

void PV_Spill()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

spillcontents(*obj,(*obj)->x,(*obj)->y);
}

// Spill contents of <object> at <X> <Y>

void PV_SpillXY()
{
OBJECT **obj;
int *x,*y;

obj = GET_OBJECT();
CHECK_POINTER(obj);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

spillcontents(*obj,*x,*y);
}

// Fade Out

void PV_FadeOut()
{
vrmfade_out();
}

// Fade Out

void PV_FadeIn()
{
vrmfade_in();
}

// Move object to top of the heap

void PV_MoveToTop()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

MoveToTop(*obj);
}

// Move object to bottom of the heap

void PV_MoveToFloor()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

MoveToFloor(*obj);
}

// IF line-of-sight between two objects continue, else jump to offset
// The string is stored first, not second

void PV_If_LOS()
{
int jumpoffset;
OBJECT **obj1,**obj2;

obj1 = GET_OBJECT();
CHECK_POINTER(obj1);
obj2 = GET_OBJECT();
CHECK_POINTER(obj2);
jumpoffset = GET_DWORD();

// If there was an accident, assume it's not true and continue

if(!*obj1)
	{
	Bug("If_line_of_sight: Object1 has not been initialised!\n");
	VMbug();
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	return; // Leave quickly
	}
if(!*obj2)
	{
	Bug("If_line_of_sight: Object2 has not been initialised!\n");
	VMbug();
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	return; // Leave quickly
	}

// Ok, do the condition

if(!line_of_sight((*obj1)->x,(*obj1)->y,(*obj2)->x,(*obj2)->y))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// Check if an object has been hurt

void PV_CheckHurt()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

CheckHurt(*obj);
}

// IF a specified square is solid then continue, else jump to offset

void PV_IfSolid()
{
int jumpoffset,*a,*b;
unsigned char op;

a = (int *)GET_INT();
CHECK_POINTER(a);
op=GET_BYTE();
b = (int *)GET_INT();
CHECK_POINTER(b);
jumpoffset = GET_DWORD();

if(!isSolid(*a,*b))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// set_pflag <object> <pflag> = <i>

void PV_SetPFlag()
{
int *var,*flag;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
flag = GET_INT();
CHECK_POINTER(flag);

var = GET_INT();
CHECK_POINTER(var);

set_pflag(*obj,*flag,*var);
}

// get_pflag <i> = <object> <pflag>

void PV_GetPFlag()
{
int *var,*flag;
OBJECT **obj;

var = GET_INT();
CHECK_POINTER(var);
obj = GET_OBJECT();
CHECK_POINTER(obj);
flag = GET_INT();
CHECK_POINTER(flag);

*var = get_pflag(*obj,*flag);
}

// set_user_flag <string> = <i>

void PV_SetUFlag()
{
int *var;
char *str;

str = GET_STRING();
CHECK_POINTER(str);
var = GET_INT();
CHECK_POINTER(var);

Set_tFlag(str,*var);
}

// get_user_flag <i> = <string>

void PV_GetUFlag()
{
int *var;
char *str;

var = GET_INT();
CHECK_POINTER(var);
str = GET_STRING();
CHECK_POINTER(str);

*var = Get_tFlag(str);
}

// set_local <person> <flag> = <state>

void PV_SetLocal()
{
int *var;
char *str;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);
var = GET_INT();
CHECK_POINTER(var);

NPC_set_lFlag(*obj,str,player,*var);
}

// get_local <var> = <person> <flag>

void PV_GetLocal()
{
int *var;
char *str;
OBJECT **obj;

var = GET_INT();
CHECK_POINTER(var);
obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);

*var = NPC_get_lFlag(*obj,str,player);
}

// weigh_object <var> = <object>

void PV_WeighObject()
{
int *var;
OBJECT **obj;

var = GET_INT();
CHECK_POINTER(var);
obj = GET_OBJECT();
CHECK_POINTER(obj);

*var = WeighObject(*obj);
}

// get_bulk <var> = <object>

void PV_GetBulk()
{
int *var;
OBJECT **obj;

var = GET_INT();
CHECK_POINTER(var);
obj = GET_OBJECT();
CHECK_POINTER(obj);

*var = GetBulk(*obj);
}

// input <intvar>

void PV_InputInt()
{
int *var;

var = GET_INT();
CHECK_POINTER(var);

if(!get_num(*var))
    *var = 0;
else
	*var = atoi(user_input);
}

// IF object is in a pocket then continue, else jump to offset

void PV_IfInPocket()
{
int jumpoffset;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

jumpoffset = GET_DWORD();

// If there was an accident, assume it's not true and continue

if(!*obj)
	{
	Bug("If_In_Pocket: Object has not been initialised!\n");
	VMbug();
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	return; // Leave quickly
	}

// Ok, do the condition

if(!InPocket(*obj))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

// IF object is NOT in a pocket then continue, else jump to offset

void PV_IfNInPocket()
{
int jumpoffset;
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);
jumpoffset = GET_DWORD();

// If there was an accident, assume it's not true and continue

if(!*obj)
	{
	Bug("If_In_Pocket: Object has not been initialised!\n");
	VMbug();
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	return; // Leave quickly
	}

// Ok, do the condition

if(InPocket(*obj))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

void PV_ReSyncEverything()
{
ResyncEverything();
}

// Talk to (file, startpage)

void PV_TalkTo2()
{
char *str1,*str2;

str1 = GET_STRING();
CHECK_POINTER(str1);
str2 = GET_STRING();
CHECK_POINTER(str2);

talk_to(str1,str2); // Str1 is null-protected
}

// Talk to (file)

void PV_TalkTo1()
{
char *str1;

str1 = GET_STRING();
CHECK_POINTER(str1);
talk_to(str1,"start"); // This is null-protected
}

// random <var> max <maxnum>

void PV_Random()
{
int *var;
int *min;
int *max,tot;
int minimum;

var = GET_INT();
CHECK_POINTER(var);
min = GET_INT();
CHECK_POINTER(min);
max = GET_INT();
CHECK_POINTER(max);

if(*max < *min)
	{
	*var = 0; // Prevent division by zero
	Bug("Random: maximum is less minimum!\n");
	VMbug();
	return;
	}

minimum = abs(*min);

tot = (*max)+minimum;

if(tot<1)
	{
	*var = 0; // Prevent division by zero
	Bug("Random: maximum and minimum are less than 1\n");
	VMbug();
	}

if(*max > 0)
	*var = (rand()%tot)-minimum;
else
	{
	*var = 0; // Prevent division by zero
	Bug("Random: maximum is less than 1!\n");
	VMbug();
	}
}

void PV_GetYN()
{
int jumpoffset;
char *str;

str = GET_STRING();
CHECK_POINTER(str);
jumpoffset = GET_DWORD();

// Ok, do the condition

if(!getYN(str))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

void PV_GetYNN()
{
int jumpoffset;
char *str;

str = GET_STRING();
CHECK_POINTER(str);
jumpoffset = GET_DWORD();

// Ok, do the condition

if(getYN(str))
    {
    curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
    }
}

void PV_Restart()
{
Restart();
}

// scroll_tile <num> <x> <y>

void PV_ScrTileN()
{
int *num,*x,*y;

num = GET_INT();
CHECK_POINTER(num);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

if(*num<0 || *num>=TItot)
	{
	Bug("No such tile number %d in %s\n",*num,debcurfunc);
	return;
	}

if(*x<0)
	*x=32+(*x);
if(*y<0)
	*y=32+(*y);

TIlist[*num].sdx=*x;
TIlist[*num].sdy=*y;

// If zero, reset the scroll offsets, since the user probably wants it to
// go back to normal

if(*x == 0 && *y == 0)
	{
	TIlist[*num].sx=0;
	TIlist[*num].sy=0;
	}

}

// scroll_tile "name" <x> <y>

void PV_ScrTileS()
{
int ctr,*x,*y;
char *name;

name = GET_STRING();
CHECK_POINTER(name);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

if(*x<0)
	*x=32+(*x);
if(*y<0)
	*y=32+(*y);

for(ctr=0;ctr<TItot;ctr++)
    if(!istricmp(TIlist[ctr].name,name))
		{
		TIlist[ctr].sdx=*x;
		TIlist[ctr].sdy=*y;

		// If zero, reset the scroll offsets, since the user probably wants it to
		// go back to normal

		if(*x == 0 && *y == 0)
			{
			TIlist[ctr].sx=0;
			TIlist[ctr].sy=0;
			}
		return;
		}

Bug("No such tile '%s' in %s\n",name,debcurfunc);
}

// While <exp> is nonzero go around the loop else fall through

void PV_While1()
{
int jumpoffset,*a;

a = (int *)GET_INT();
CHECK_POINTER(a);

jumpoffset = GET_DWORD();

if((!*a) || whilebreak)
	; // Do Nothing
else
	{
	curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	}
whilebreak=0; // If we've done a Break, absorb it
}

// While <x> = <y> go around the loop else fall through

void PV_While2()
{
int jumpoffset,*a,*b;
unsigned char op;

a = (int *)GET_INT();
CHECK_POINTER(a);
op=GET_BYTE();
b = (int *)GET_INT();
CHECK_POINTER(b);

jumpoffset = GET_DWORD();

if(Operator[op&OPMASK](*a,*b) && !whilebreak)
	{
	curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	}
whilebreak=0; // If we've done a Break, absorb it
}

// Break from a While loop or For Loop

void PV_Break()
{
whilebreak=1;
}

// Break from a While loop or For Loop

void PV_GetKey()
{
irekey=get_key_debounced();
}


// Set Object's current activity

void PV_DoAct()
{
OBJECT **obj;
int *act;

obj = GET_OBJECT();
CHECK_POINTER(obj);
act = GET_INT();
CHECK_POINTER(act);

if(!*obj)
	{
	Bug("Do_Activity: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

//ilog_quiet("actnum %x %d NULL\n",*obj,*act);
//ilog_quiet("%s %s\n",(*obj)->name,PElist[*act].name);

ActivityNum(*obj,*act,NULL);
}

// Set Object's current activity by string

void PV_DoActS()
{
OBJECT **obj;
char *act;

obj = GET_OBJECT();
CHECK_POINTER(obj);
act = GET_STRING();
CHECK_POINTER(act);

if(!*obj)
	{
	Bug("Do_Activity: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

//ilog_quiet("actstr %x %d NULL\n",*obj,getnum4PE(act));

ActivityNum(*obj,getnum4PE(act),NULL);
}

// Set Object's current activity (with a target)

void PV_DoActTo()
{
OBJECT **obj,**target;
int *act;

obj = GET_OBJECT();
CHECK_POINTER(obj);
act = GET_INT();
CHECK_POINTER(act);
target = GET_OBJECT();
CHECK_POINTER(target);

if(!*obj)
	{
	Bug("Do_Activity: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

if(!*target)
	{
	Bug("Do_Activity: Target has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

//ilog_quiet("actnum %x %d %s\n",*obj,*act,(*target)->name);

ActivityNum(*obj,*act,*target);
}

// Set Object's current activity by string (with a target)

void PV_DoActSTo()
{
OBJECT **obj,**target;
char *act;

obj = GET_OBJECT();
CHECK_POINTER(obj);
act = GET_STRING();
CHECK_POINTER(act);
target = GET_OBJECT();
CHECK_POINTER(target);

if(!*obj)
	{
	Bug("Do_Activity: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

if(!*target)
	{
	Bug("Do_Activity: Target has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

//ilog_quiet("actstr %x %d %s\n",*obj,getnum4PE(act),(*target)->name);

ActivityNum(*obj,getnum4PE(act),*target);
}

// Halt scheduled activity (and stand there stupidly)

void PV_StopAct()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

if(!*obj)
	{
	Bug("stop_activity: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

//ilog_quiet("stop action for %s\n",(*obj)->name);

Inactive(*obj);
}


// Resume scheduled activity

void PV_ResumeAct()
{
OBJECT **obj;

obj = GET_OBJECT();
CHECK_POINTER(obj);

if(!*obj)
	{
	Bug("resume_activity: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

//ilog_quiet("resume action for %s\n",(*obj)->name);

ResumeSchedule(*obj);
}


// Debugging function: Get VRM name from a number

void PV_GetFuncV()
{
char **str;
int *a;

str = (char **)GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(str);
a = GET_INT();
CHECK_POINTER(a);

if(*a < 1 || *a > COtot)
	*str=NULL;
else
	*str=COlist[*a].name;
}


// Debugging function: Get PEscript name from a number

void PV_GetFuncP()
{
char **str;
int *a;

str = (char **)GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(str);
a = GET_INT();
CHECK_POINTER(a);

if(*a < 1 || *a > PEtot)
	*str=NULL;
else
	*str=PElist[*a].name;
}

// Debugging function: Get PEscript name from a number

void PV_GetContainer()
{
OBJECT **obj1;
OBJECT **obj2;

obj1 = GET_OBJECT();
CHECK_POINTER(obj1);
obj2 = GET_OBJECT();
CHECK_POINTER(obj2);

*obj1=NULL; // Set to Nothing

if(!*obj2)
	return; // Nothing to look for

*obj1 = LookInAllPocketsFor(*obj2);
}

void PV_SetLeader()
{
OBJECT **o;
int *p,*f;

o = GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(o);
p = GET_INT();
CHECK_POINTER(p);
f = GET_INT();
CHECK_POINTER(f);

choose_leader(*o, PElist[*p].name, PElist[*f].name);
//int choose_leader(OBJECT *x, char *player_call, char *follower_call)
}

void PV_SetMember()
{
OBJECT **o;
int *p;

o = GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(o);
p = GET_INT();
CHECK_POINTER(p);

//int choose_member(OBJECT *x, char *player_call)
choose_member(*o, PElist[*p].name);
}

void PV_AddMember()
{
OBJECT **o;

o = GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(o);

AddToParty(*o);
}

void PV_DelMember()
{
OBJECT **o;

o = GET_OBJECT(); // Hack, we need the pointer not just the string
CHECK_POINTER(o);

SubFromParty(*o);
ResumeSchedule(*o);

}

void PV_MoveTowards8()
{
OBJECT **o1,**o2;
char *s1,*s2;

pevm_err=-1; // Assume failure

o1 = GET_OBJECT();
CHECK_POINTER(o1);
o2 = GET_OBJECT();
CHECK_POINTER(o2);

if(!*o1 || !*o2)
	{
	s1="Null";
	s2="Null";

	if(*o1)
		s1=BestName(*o1);
	if(*o2)
		s2=BestName(*o2);
	Bug("FindPath (%s,%s)\n",s1,s2);
	return;
	}

pevm_err=FindPath(*o1,*o2,1);
}

void PV_MoveTowards4()
{
OBJECT **o1,**o2;
char *s1,*s2;

pevm_err=-1; // Assume failure

o1 = GET_OBJECT();
CHECK_POINTER(o1);
o2 = GET_OBJECT();
CHECK_POINTER(o2);

if(!*o1 || !*o2)
	{
	s1="Null";
	s2="Null";

	if(*o1)
		s1=BestName(*o1);
	if(*o2)
		s2=BestName(*o2);
	Bug("FindPath (%s,%s)\n",s1,s2);
	return;
	}

pevm_err=FindPath(*o1,*o2,0);
}

void PV_WaitForAnimation()
{
OBJECT **o;

o = GET_OBJECT();
CHECK_POINTER(o);

if((*o)->form->flags&6)	// Ignore if looped or stepped
	return;				// (Otherwise it will loop forever)

if(InPocket(*o))			// Objects in pockets do not animate
	return;

for(;(*o)->sdir;RedrawMap());
}

void PV_WaitFor()
{
int *ms;
ms = GET_INT();
CHECK_POINTER(ms);
waitfor(*ms);
}

// IF <object> is onscreen then continue, else jump to offset

void PV_If_oonscreen()
{
int jumpoffset;
OBJECT **obj,*a;
int ok;

obj = CHECK_OBJ();
jumpoffset = GET_DWORD();

ok=0;
if(obj)
	if(*obj)
		{
		a=*obj;
		if(a->x >=mapx)
			if(a->x <= mapx2)
				if(a->y >=mapy)
					if(a->y <= mapy2)
						ok=1;    // Yes
		}
    
if(!ok)
	{
	curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	}
}

void PV_If_not_oonscreen()
{
int jumpoffset;
OBJECT **obj,*a;
int ok;

obj = CHECK_OBJ();
jumpoffset = GET_DWORD();

ok=0;
if(obj)
	if(*obj)
		{
		a=*obj;
		if(a->x >=mapx)
			if(a->x <= mapx2)
				if(a->y >=mapy)
					if(a->y <= mapy2)
						ok=1;    // Yes
		}
    
if(ok)
	{
	curvm->ip = curvm->code+jumpoffset;  // Mess with IP for the jump
	}
}

// Take quantity of objects from a container

void PV_TakeQty()
{
OBJECT **obj;
int *amount;
char *str;

amount = GET_INT();
CHECK_POINTER(amount);
str = GET_STRING();
CHECK_POINTER(str);
obj = GET_OBJECT();
CHECK_POINTER(obj);

if(!*obj)
	{
	Bug("Take: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

pe_result_ok=TakeQuantity(*obj,str,*amount);
}

// Move quantity of objects from one container to another

void PV_MoveQty()
{
OBJECT **src,**dest;
int *amount;
char *str;

amount = GET_INT();
CHECK_POINTER(amount);
str = GET_STRING();
CHECK_POINTER(str);
src = GET_OBJECT();
CHECK_POINTER(src);
dest = GET_OBJECT();
CHECK_POINTER(dest);

if(!*src)
	{
	Bug("Move: source object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

if(!*dest)
	{
	Bug("Move: destination object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

pe_result_ok = MoveQuantity(*src,*dest,str,*amount);
}


// Count quantity of objects in a container

void PV_CountQty()
{
OBJECT **obj;
int *amount;
char *str;

amount = GET_INT();
CHECK_POINTER(amount);
str = GET_STRING();
CHECK_POINTER(str);
obj = GET_OBJECT();
CHECK_POINTER(obj);

if(!*obj)
	{
	Bug("Count: Object has not been initialised!\n");
	VMbug();
	return; // Leave quickly
	}

*amount=SumObjects((*obj)->pocket,str,0);
}

void PV_CopySchedule()
{
OBJECT **o1,**o2;
char *s1,*s2;
int ctr;

o1 = GET_OBJECT();
CHECK_POINTER(o1);
o2 = GET_OBJECT();
CHECK_POINTER(o2);

if(!*o1 || !*o2)
	{
	s1="Null";
	s2="Null";

	if(*o1)
		s1=BestName(*o1);
	if(*o2)
		s2=BestName(*o2);
	Bug("CopySchedule(%s -> %s)\n",s1,s2);
	return;
	}

// Make sure the source has a schedule

if(!(*o1)->schedule)
	return;

// Make sure

if(!(*o2)->schedule)
	(*o2)->schedule = (SCHEDULE *)M_get(24,sizeof(SCHEDULE));

memcpy((*o2)->schedule,(*o1)->schedule,sizeof(SCHEDULE)*24);
}

// Call a function for all objects onscreen

void PV_AllOnscreen()
{
OBJECT *old,*cur;
int func,x,y;

// First get the object we're going to search inside of

func = GET_DWORD();

old = current_object;

for(y=mapy;y<mapy2;y++)
	for(x=mapy;x<mapx2;x++)
		{
		current_object = GameGetObject(x,y);
		if(current_object)
			if(current_object->stats->hp > 0)
				if(current_object->stats->intel > 0)
					CallVMnum(func);
		}

current_object = old;
}

void PV_CheckTime()
{
CheckTime();
}

void PV_FindNear()
{
OBJECT **o1;
OBJECT **o2;
char *str;

o1 = GET_OBJECT();
CHECK_POINTER(o1);
str = GET_STRING();
CHECK_POINTER(str);
o2 = GET_OBJECT();
CHECK_POINTER(o2);

*o1 = find_nearest(*o2,str);
}


void PV_FindTag()
{
OBJLIST *t;
OBJECT **obj;
char *str;
int *tag;

obj = GET_OBJECT();
CHECK_POINTER(obj);
str = GET_STRING();
CHECK_POINTER(str);
tag = GET_INT();
CHECK_POINTER(tag);


for(t=MasterList;t;t=t->next)
	if(t->ptr->tag == *tag)
		if(str)
			{
			if(!istricmp(str,t->ptr->name))
				{
				*obj = t->ptr;
				return;
				}
			}
		else
			{
			*obj = t->ptr;
			return;
			}

*obj = NULL;
}

// Set light state for a region between two control objects.

void PV_SetLight()
{
int *tag,*state;
OBJLIST *t;
OBJECT *start,*end;

tag = GET_INT();
CHECK_POINTER(tag);
state = GET_INT();
CHECK_POINTER(state);

start=NULL;
end=NULL;

for(t=MasterList;t;t=t->next)
	if(t->ptr->tag == *tag)
		{
		if(!istricmp("light_start",t->ptr->name))
			start=t->ptr;
		if(!istricmp("light_end",t->ptr->name))
			end=t->ptr;
		}

if(!start)
	{
	Bug("Cannot find LIGHT_START object with tag %d\n",*tag);
	return;
	}

if(!end)
	{
	Bug("Cannot find LIGHT_END object with tag %d\n",*tag);
	return;
	}

set_light(start->x,start->y,end->x,end->y,*state);
}


void PV_Printaddr()
{
int *num;
num = GET_INT();
irecon_printf("%-08x",num);
if(DebugVM || printlogging)
	ilog_quiet("%-08x",num);
}

// Log console output?

void PV_PrintLog()
{
int *p;

p = GET_INT();
CHECK_POINTER(p);

printlogging = *p;

// Special effects for debugging purposes:

switch(printlogging)
	{
	case 1:
	irecon_savecol();
	irecon_colour(200,0,0);
	Bug("Problems in %s\n",debcurfunc);
	break;

	case 0:
	irecon_loadcol();
	irecon_printf("\n");
	break;
	}
}


void PV_MovePartyToObj()
{
int ctr;
OBJECT **o1;

o1 = GET_OBJECT();
CHECK_POINTER(o1);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	if(party[ctr])
		TransferToPocket(party[ctr],*o1);

}

void PV_MovePartyFromObj()
{
int ctr,*x,*y;
OBJECT **o1;

o1 = GET_OBJECT();
CHECK_POINTER(o1);
x = GET_INT();
CHECK_POINTER(x);
y = GET_INT();
CHECK_POINTER(y);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	if(party[ctr])
		ForceFromPocket(party[ctr],*o1,*x,*y);

}

//
// LASTOP: new opcodes go above this line
//


// Print various status items (for debugging)

void PV_LastOp()
{
OBJLIST *t;
int ctr;
int *num;
num = GET_INT();

switch(*num)
	{
	case 0:
	ctr=0;
	for(t=ActiveList;t;t=t->next) ctr++;
	irecon_printf("%d Active objects\n",ctr);
	break;

	case 1:
	ilog_printf("current = %x\n",current_object);
	if(current_object)
		ilog_printf("current = '%s'\n",current_object->name);
	break;

	default:
	Bug("Unknown status request %d\n",*num);
	}

}


// DumpVM

void PV_Dump()
{
int dvm;
dvm = DebugVM;
DebugVM=1;
DumpVM();
DebugVM = dvm;
}

// No Operation

void PV_NOP()
{
}

// DO Fix-Up System, internal system for local variable management

void PV_Dofus()
{
int *data,num,*var,ctr;
PTR ptr,val,dataorg;


//#define DOFUSDUMP // For debugging the VM system
//#define DOFUS_DEBUG

#ifdef DOFUS_DEBUG
ilog_quiet("in %s\n",debcurfunc);
#define DOFUS_LOG(x,y); ilog_quiet(x,y);
#define DOFUS_LOG2(x,y); ilog_quiet(x,y,(y)-(int)(curvm->code));
#else
#define DOFUS_LOG(x,y); ;
#define DOFUS_LOG2(x,y); ;
#endif

#ifdef DOFUSDUMP
//ilog_quiet("%s\n",debcurfunc);
ilog_quiet("Doofus: dump before: ");
for(ctr=0;ctr<curvm->size;ctr++)
     {
     if(!(ctr % 16))
         ilog_quiet("\n%-3x  ",ctr/16);
     ilog_quiet("%02x ",curvm->code[ctr]&255);
     }
ilog_quiet("\n");
//DebugVM=1;
//DumpVM();
#endif

num = (int)GET_PTR(); // Get number of patches

DOFUS_LOG("%x patches\n",num);

// Get the offset of the patch table into the VM's memory space
data = (int *)GET_PTR();

DOFUS_LOG("ptoffset = %x\n",data);

// Turn the patch table offset into an actual pointer
data = (int *)((PTR)curvm->code+(PTR)data);

DOFUS_LOG2("patchptr = %x [%x]\n",data);

// Variable slots are after the patch table, i.e. the patch table's address
// plus the no. of patches times the word size x2 (each patch has two entries)
dataorg = (PTR)data+(PTR)(num*sizeof(PTR)*2); // Variable slots

DOFUS_LOG2("data org = %x [%x]\n---\n",dataorg);

// Patch each of the entries

for(ctr=0;ctr<num;ctr++)
    {
	// Get the first word (variable offset)
	var = (int *)*data;
DOFUS_LOG2("var offset = %x [%x]\n",var);

	// Increment the data pointer in a portable manner
	data = (int *)((PTR)data + (PTR)sizeof(int)); // Increment data ptr
DOFUS_LOG2("dataptr = %x [%x]\n",data);

	// Get the second word (variable slot number) and increment again
	ptr = (PTR)*data;
DOFUS_LOG("slot number = %x\n",ptr);
	data = (int *)((PTR)data + (PTR)sizeof(int)); // Increment data ptr
DOFUS_LOG2("dataptr = %x [%x]\n",data);

	// Make the code offset into a pointer
	var = (int*)((PTR)curvm->ip + (PTR)var); // Now points to patch address
DOFUS_LOG2("patch address = %x [%x]\n",var);

	// Get the value that was there, this is the variable's default value
	// We put that into the var.slot later.  This could be done at build time
	// which may speed things up slightly.
	val = *var;
DOFUS_LOG("def value = %x\n",val);

	// Replace it with the pointer to the variable slot (that the VM needs)
	*var = (dataorg+(ptr<<2));     // Patch it so it's a pointer to variable
DOFUS_LOG2("varslot ptr = %x [%x]\n",*var);

	// Fill the variable slot with the default value
	*(int *)*var = val;            // Write default value into the variable
#ifdef DOFUS_DEBUG
ilog_quiet("round again.. [%d/%d]\n\n",ctr+1,num);
#endif

	}
#ifdef DOFUSDUMP
ilog_quiet("Doofus: dump after\n");
DumpVM();
ilog_quiet("Doofus: signing off\n");
#endif
}


//
//	Math Operators
//

int OP_Invalid(int a, int b)
{
Bug("Unknown operator! Either 'if %d ? %d' or let x = %d ? %d\n",a,b,a,b);
VMbug();
return 0;
}

int OP_Add(int a, int b)
{
return a+b;
}

int OP_Sub(int a, int b)
{
return a-b;
}

int OP_Div(int a, int b)
{
if(b==0)
	return 0; // No, this is not mathematically correct
return a/b;
}

int OP_Mul(int a, int b)
{
return a*b;
}

int OP_Equal(int a, int b)
{
return (a==b);
}

int OP_NotEqual(int a, int b)
{
//ilog_quiet("cmp %x != %x : result %d\n",a,b,(a!=b));
return (a!=b);
}

int OP_Less(int a, int b)
{
return (a<b);
}

int OP_LessEq(int a, int b)
{
return (a<=b);
}

int OP_Great(int a, int b)
{
return (a>b);
}

int OP_GreatEq(int a, int b)
{
return (a>=b);
}

int OP_Modulus(int a, int b)
{
if(b==0)
	return 0; // No, this is not mathematically correct
return a%b;
}

