//
//      Game-To-Script interface.
//      Contains all the system calls that VRMs and PE files can use.
//

#define IRE_SYSTEM_MODULE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#define NO_FORTIFY
#include "ithelib.h"
#include "media.h"
#include "vrm_in.hpp"
#include "sound.h"
#include "sys.hpp"
#include "loadfile.h"
#include "nuspeech.hpp"
#include "linklist.hpp"
#include "gamedata.h"
#include "console.h"
#include "init.h"

#include "seer/seer.h"

#define LIBEXPORT(x) scAdd_External_Symbol(#x,(void*)x);
#define LIBEXVAR(x)  scAdd_External_Symbol(#x,(void*)&x);

#define SOUND_DIST 4 // maximum distance without attenuation
#define SOUND_DROP 8 // 8 volume points every square

void DT();

// Variables

extern "C" char show_vrm_calls;     // From OSCLI
extern long COtot;
extern int do_lightning;
extern char pending_delete;
extern char fullrestore;
extern int sf_volume;
static char *nullstr="<NULL>";

// Master function

void Init_VRMs();

// These functions and variables are exported

extern "C" char *lastvrmcall;
extern "C" char *blametrace;


extern OBJECT *OB_Alloc();
extern void OB_Free(OBJECT *a);
extern void DestroyObject(OBJECT *a);
extern void DeletePending();
extern void OB_Init(OBJECT *a,char *name);
extern void OB_SetDir(OBJECT *objsel,int dir);
extern void OB_SetSeq(OBJECT *objsel,char *name);
void Call_VRM(int num);
void call_vrm(char *name);
extern TILE *GetTile(int x,int y);
extern int WeighObject(OBJECT *obj);
extern int GetBulk(OBJECT *obj);
extern int AddToParty(OBJECT *new_member);
extern void SubFromParty(OBJECT *member);
extern int isSolid(int x,int y);
extern int InPocket(OBJECT *obj);
extern int get_key();
extern int get_key_debounced();
extern void Restart(void);
extern void SetDarkness(int l);
extern void spillcontents(OBJECT *bag,int x,int y);
extern void NPC_Converse(char *file, char *start);
extern void CheckHurt(OBJECT *list);
extern void RedrawMap();
extern char *GetVrmName(int num);
extern int IsTileSolid(int x,int y);
extern int IsTileWater(int x,int y);

extern void Inactive(OBJECT *o);
extern void ActivityName(OBJECT *o,char *activity, OBJECT *target);
extern void ActivityNum(OBJECT *o,int activity, OBJECT *target);

extern OBJECT *GetBestPlace(int x,int y);
extern OBJECT *GetTopObject(int x,int y);
extern OBJECT *GetRawObjectBase(int x,int y);
extern void MoveToPocket(OBJECT *src,OBJECT *dest);
extern void TransferToPocket(OBJECT *src,OBJECT *dest);
extern int MoveFromPocket(OBJECT *obj,OBJECT *container,int x,int y);
extern int ForceFromPocket(OBJECT *obj,OBJECT *container,int x,int y);
extern void MoveToTop(OBJECT *obj);
extern void MoveToFloor(OBJECT *obj);

extern void MoveToMap(int x, int y, OBJECT *object);
extern void TakeObject(int x,int y, OBJECT *container);
extern int MoveObject(OBJECT *object,int x,int y);
extern void TransferObject(OBJECT *object,int x,int y);
extern OBJECT *GetObject(int x,int y);
extern OBJECT *GetSolidObject(int x,int y);

extern OBJECT *GetSolidObject(int x,int y);
extern OBJECT *GameGetObject(int x,int y);
extern OBJECT *GetObjectBase(int x,int y);
extern void CreateContents(OBJECT *o);
extern int TakeQuantity(OBJECT *c,char *o,int q);
extern int MoveQuantity(OBJECT *s,OBJECT *d,char *o,int q);
extern void AddQuantity(OBJECT *c,char *o,int q);
extern int SumObjects(OBJECT *c,char *o,int q);
extern OBJECT *LookInAllPocketsForTag(int tag);
extern OBJECT *LookInAllPocketsFor(OBJECT *obj);
extern OBJECT *GetFirstObject(OBJECT *cont, char *name);
extern int getYN(char *q);
extern void ML_Add(OBJLIST **m,OBJECT *o);
extern void ML_Del(OBJLIST **m,OBJECT *o);
extern void AL_Add(OBJLIST **a,OBJECT *o);
extern void gen_largemap();
extern void setPflag(OBJECT *o,int n,int s);
extern int getPflag(OBJECT *o,int n);
extern int get_num(int no);
extern int get_str();
extern char user_input[128];
extern int FindPath(OBJECT *object, OBJECT *end, int diagonal);
extern int CanRoute(OBJECT *start, OBJECT *end, int diagonal);
extern int IsOnscreen(char *pname);
extern void CheckTime();
extern void ResumeSchedule(OBJECT *o);
extern void ResyncEverything();
extern char *BestName(OBJECT *o);
extern void CallVM(char *function);
extern void CallVMnum(int pos);

static void sysBug(char *msg, ...);
static void check_object(OBJECT *o, char *label);
static void Call_VM(char *c);
static void Call_VM_num(int n);

void set_light(int x, int y, int x2, int y2, int light);


/*
 *    VRM_RegisterExvar()  -  Register a PE global so SEER can use it too
 */

void VRM_RegisterExvar(char *name,void *ptr)
{
scAdd_External_Symbol(name,ptr);
}

/*
 *    Init_VRMs()  -  Initialise the VRM system
 */

void Init_VRMs()
{
ilog_printf("Preinitialise VRMs:\n");
ilog_printf("  VRM Subsystem is: SeeR %s\n",SeeR_VERSION_STR);

// Here is the export table itself

LIBEXPORT(strcmp)
scAdd_External_Symbol("stricmp",istricmp);  // Symbol name != identifier
//LIBEXPORT(stricmp)
LIBEXPORT(strcpy)
LIBEXPORT(malloc)
LIBEXPORT(free)
LIBEXPORT(realloc)
LIBEXPORT(waitfor)

// My console IO functions

LIBEXPORT(print)
//LIBEXPORT(clear)
scAdd_External_Symbol("clear",vrmclear);  // Symbol name != identifier
LIBEXPORT(printxy)
LIBEXPORT(get_number)
LIBEXPORT(Bug)
LIBEXPORT(best_name)

scAdd_External_Symbol("printf",print);  // Symbol name != identifier
scAdd_External_Symbol("Bug",sysBug);    // and again
scAdd_External_Symbol("textcolour",TextColor);
scAdd_External_Symbol("textcolor",TextColor);
scAdd_External_Symbol("move_towards_8",move_towards);

// Object control functions

LIBEXPORT(create_object)
LIBEXPORT(change_object)
LIBEXPORT(replace_object)
LIBEXPORT(remove_object)
LIBEXPORT(delete_pending)
LIBEXPORT(set_object_direction)
LIBEXPORT(set_object_sequence)
LIBEXPORT(set_object_behaviour)
LIBEXPORT(add_goal)
LIBEXPORT(end_goal)
LIBEXPORT(wipe_goals)
LIBEXPORT(kill_goals)
LIBEXPORT(show_object)
LIBEXPORT(get_tile)
LIBEXPORT(is_tile_solid)
LIBEXPORT(is_tile_water)
LIBEXPORT(get_object)
LIBEXPORT(get_solid_object)
LIBEXPORT(get_first_object)
LIBEXPORT(get_object_below)
LIBEXPORT(move_to_pocket)
LIBEXPORT(transfer_to_pocket)
LIBEXPORT(move_from_pocket)
LIBEXPORT(force_from_pocket)
LIBEXPORT(move_to_top)
LIBEXPORT(move_to_floor)
LIBEXPORT(weigh_object)
LIBEXPORT(get_bulk)
LIBEXPORT(is_solid)
LIBEXPORT(spill_contents)
LIBEXPORT(spill_contents_at)
LIBEXPORT(move_object)
LIBEXPORT(transfer_object)
LIBEXPORT(in_pocket)
LIBEXPORT(set_flag)
LIBEXPORT(get_flag)
LIBEXPORT(default_flag)
LIBEXPORT(take_quantity)
LIBEXPORT(move_quantity)
LIBEXPORT(add_quantity)
LIBEXPORT(count_objects)
LIBEXPORT(find_object_with_tag)
LIBEXPORT(find_container)
LIBEXPORT(search_container)
LIBEXPORT(find_nearest)
LIBEXPORT(set_pflag)
LIBEXPORT(get_pflag)
LIBEXPORT(set_user_flag)
LIBEXPORT(get_user_flag)
LIBEXPORT(set_local)
LIBEXPORT(get_local)
LIBEXPORT(move_forward)
LIBEXPORT(move_backward)
LIBEXPORT(turn_l)
LIBEXPORT(turn_r)
LIBEXPORT(wait_for_animation)
LIBEXPORT(move_towards)
LIBEXPORT(move_towards_4)
LIBEXPORT(move_thick)
LIBEXPORT(character_onscreen)
LIBEXPORT(object_onscreen)
LIBEXPORT(count_active_objects)
LIBEXPORT(resume_schedule)
LIBEXPORT(resync_everything)
LIBEXPORT(find_objects_with_tag)

//Miscellaneous functions

LIBEXPORT(call_vrm)
LIBEXPORT(call_vrm_number)
LIBEXPORT(get_vrm_name)
LIBEXPORT(play_song)
LIBEXPORT(play_sound)
LIBEXPORT(object_sound)
LIBEXPORT(stop_song)
LIBEXPORT(start_song)
LIBEXPORT(object_is_called)
LIBEXPORT(redraw)
LIBEXPORT(get_input)
LIBEXPORT(rnd)
LIBEXPORT(restart)
LIBEXPORT(set_darkness)
LIBEXPORT(line_of_sight)
LIBEXPORT(talk_to)
LIBEXPORT(check_hurt)
LIBEXPORT(redraw_map)
LIBEXPORT(get_yn)
scAdd_External_Symbol("lightning",___lightning);
LIBEXPORT(scroll_tile)
LIBEXPORT(scroll_tile_number)
scAdd_External_Symbol("fade_out",vrmfade_out);
scAdd_External_Symbol("fade_in",vrmfade_in);
LIBEXPORT(check_time)
LIBEXPORT(check_object)
scAdd_External_Symbol("callvm",Call_VM);
scAdd_External_Symbol("callvmnum",Call_VM_num);

// Member control

LIBEXPORT(add_to_party)
LIBEXPORT(remove_from_party)
LIBEXPORT(choose_member)
LIBEXPORT(choose_leader)
LIBEXPORT(move_party_to_object)
LIBEXPORT(move_party_from_object)
LIBEXPORT(set_light)

// System variables

LIBEXVAR(syspocket)
LIBEXVAR(show_roof)
LIBEXVAR(player)
LIBEXVAR(victim)
LIBEXVAR(current_tile)
LIBEXVAR(current_object)
LIBEXVAR(person)
LIBEXVAR(storage)
LIBEXVAR(party)
LIBEXVAR(partyname)
LIBEXVAR(new_x)
LIBEXVAR(new_y)
//LIBEXVAR(key)
scAdd_External_Symbol("key",(void *)&irekey);
LIBEXVAR(game_minute)
LIBEXVAR(game_hour)
LIBEXVAR(game_day)
LIBEXVAR(game_month)
LIBEXVAR(game_year)
LIBEXVAR(day_of_week)
LIBEXVAR(days_passed)
LIBEXVAR(Fader)
LIBEXVAR(SoundFixed)
LIBEXVAR(mapx)
LIBEXVAR(mapy)
LIBEXVAR(mapx2)
LIBEXVAR(mapy2)

#ifdef DEBUGSEER
scOpen_Debug("sire.deb");
atexit(DT);
#endif
}

/*
 *    Call_VRM(int n) - call a VRM function by number
 */

void Call_VRM(int v)
{
char *callvrmname;
callvrmname=lastvrmcall;
blametrace=callvrmname;

if(v>COtot || v<0)
       {
       if(lastvrmcall)
           Bug("Invalid VRM number %d called from VRM %s\n",v,lastvrmcall);
       else
			{
           Bug("Invalid VRM number %d called from Engine\n",v);
			}
//       Bug("player = %s [%x], current_object = %s [%x], victim = %s [%x]\n",BestName(player),player,BestName(current_object),current_object,BestName(victim),victim);
       return;
       }

lastvrmcall = COlist[v].name; // This is the VRM we crashed in
COlist[v].count++; // Log number of calls

if(show_vrm_calls)
	ilog_quiet("Entering %s from %s\n",lastvrmcall,callvrmname);

#ifdef CHECK_OBJECT_STRICT
//ilog_quiet("Entering %s from %s\n",lastvrmcall,callvrmname);
if(current_object)
	{
    CHECK_OBJECT(current_object);
	}
if(victim)
	{
    CHECK_OBJECT(victim);
	}
if(player)
	{
    CHECK_OBJECT(player);
	}
#endif

if(!COlist[v].instance)
    {
    COlist[v].instance = scCreate_Instance(COlist[v].vrm,"");
    if(!COlist[v].instance)
        {
        if(scErrorNo != scOK)
            {
            ilog_quiet("SeeR instancing error #%d\n",scErrorNo);
            ilog_quiet("Error: %s\n",scErrorMsg);
            ilog_quiet("Line: %s\n\n",scErrorLine);
            }
        ilog_quiet("Fault with VRM %d: %s\n",v,COlist[v].name);
        ithe_panic("Could not create instance of function",COlist[v].name);
        }
    }
scCall_Instance((scInstance *)COlist[v].instance,scGet_Symbol((scInstance *)COlist[v].instance,"vrm_entrypoint"),0,0);
if(scErrorNo != scOK)
    {
    Bug("SeeR processing error #%d calling %s\n",scErrorNo,COlist[v].name);
    Bug("Error: %s\n",scErrorMsg);
    Bug("Line: %s\n\n",scErrorLine);
    redraw();
    }

if(show_vrm_calls)
	ilog_quiet("Leaving %s\n",lastvrmcall);

lastvrmcall=callvrmname;
blametrace=NULL;
}


///////////////////////////////////

OBJECT *create_object(STRING name, int x, int y)
{
OBJECT *temp;
//boot2("***%s\n",name);
temp = OB_Alloc();
temp->curdir = temp->dir[0];
OB_Init(temp,name);
OB_SetDir(temp,1);
MoveToMap(x,y,temp);
CreateContents(temp);
return(temp);
}

void remove_object(OBJECT *object)
{
if(!object)
    {
    Bug("remove_object: attempt to remove NULL\n");
    redraw();
    return;
    }
CHECK_OBJECT(object);

if(object == player)
    {
    Bug("remove_object: cannot remove the PLAYER\n");
    redraw();
    return;
    }

object->flags.on = 0;   // Mark Pending-delete
pending_delete=1;       // Tell the garbage collector it will be needed
}

void delete_pending()
{
DeletePending();
}

void change_object(OBJECT *obj,char *name)
{
STATS ostats;
char resname[32];
char pname[32];
char speech[128];
int tcache;
int partie;     // Must preserve this for dead party members

CHECK_OBJECT(obj);

if(obj && name)
   {
   partie = obj->flags.party;
   memcpy(&ostats,obj->stats,sizeof(STATS));
   strcpy(resname,obj->funcs->resurrect);
   strcpy(speech,obj->funcs->talk);
   tcache = obj->funcs->tcache;
   strcpy(pname,obj->personalname);
   OB_Init(obj,name);
   memcpy(obj->stats,&ostats,sizeof(STATS));
   strcpy(obj->funcs->resurrect,resname);
   strcpy(obj->personalname,pname);
   strcpy(obj->funcs->talk,speech);
   obj->flags.party = partie;
   obj->funcs->tcache = tcache;
   }
}


void replace_object(OBJECT *obj,STRING name)
{
int k;
if(obj && name)
    {
	CHECK_OBJECT(obj);
    k=fullrestore;
    fullrestore=0;
    OB_Init(obj,name);
    fullrestore=k;
    gen_largemap();     // Recalc solid objects cache
    }
}

void set_object_direction(OBJECT *obj,int dir)
{
if(obj)
	{
	CHECK_OBJECT(obj);
	OB_SetDir(obj,dir);
	}
}

void set_object_sequence(OBJECT *obj,char *seq)
{
if(obj && seq)
	{
	CHECK_OBJECT(obj);
	OB_SetSeq(obj,seq);
	}
}

void print(char *msg, ...)
{
char temp[4096];
va_list ap;

if(!msg)
    {
    Bug("Attemped to print NULL message\n");
    va_end(ap);
    redraw();
    return;
    }

va_start(ap, msg);
vsprintf(temp,msg,ap);          // Temp is now the message
va_end(ap);

irecon_print(temp);
}

void printxy(int x,int y,char *msg, ...)
{
char temp[128];
va_list ap;

va_start(ap, msg);
vsprintf(temp,msg,ap);          // Temp now is the message
irecon_printxy(x,y,temp);
va_end(ap);
}

void TextColor(int r,int g, int b)
{
irecon_colour(r,g,b);
}

void vrmclear()
{
irecon_cls();
}

void show_object(OBJECT *z,int x,int y)
{
if(!z)
   return;
CHECK_OBJECT(z);
draw_rle_sprite(swapscreen,z->form->seq[0]->image,x,y);
}

void call_vrm(char *name)
{
int v;
if(name[0] == 0)
    {
    Bug("Invalid VRM name called in function %s\n",lastvrmcall);
    return;
    }

v = getnum4VRM(name);
if(show_vrm_calls)
   ilog_quiet("chain_vrm: vrm '%s' was called, ID returned was %d\n",name,v);

Call_VRM(v);
}

void call_vrm_number(int num)
{
if(show_vrm_calls)
   ilog_quiet("chain_vrm: vrm %d was called\n",num);
if(num<0)
         Bug("Invalid VRM number called in function %s\n",lastvrmcall);
Call_VRM(num);
}


char *get_vrm_name(int num)
{
return GetVrmName(num);
}


OBJECT *get_first_object(int x,int y)
{
return GetObjectBase(x,y);//GetTopObject(x,y);
}

OBJECT *get_object(int x,int y)
{
return GameGetObject(x,y);
}

OBJECT *get_solid_object(int x,int y)
{
return GetSolidObject(x,y);
}

TILE *get_tile(int x,int y)
{
return GetTile(x,y);
}

void move_to_pocket(OBJECT *src,OBJECT *dest)
{
CHECK_OBJECT(src);
CHECK_OBJECT(dest);
if(src->flags.quantity)
    {
    AddQuantity(dest,src->name,src->stats->quantity);
    src->flags.on = 0;      // Mark Pending-delete
    pending_delete=1;       // Tell the garbage collector it will be needed
    }
else
    MoveToPocket(src,dest);
}

void transfer_to_pocket(OBJECT *src,OBJECT *dest)
{
CHECK_OBJECT(src);
CHECK_OBJECT(dest);
if(src->flags.quantity)
    {
    AddQuantity(dest,src->name,src->stats->quantity);
    src->flags.on = 0;      // Mark Pending-delete
    pending_delete=1;       // Tell the garbage collector it will be needed
    }
else
    TransferToPocket(src,dest);
}

int move_from_pocket(OBJECT *obj,OBJECT *container,int x,int y)
{
CHECK_OBJECT(obj);
CHECK_OBJECT(container);
if(!MoveFromPocket(obj,container,x,y))
    return 0;
return 1;
}


int force_from_pocket(OBJECT *obj,OBJECT *container,int x,int y)
{
CHECK_OBJECT(obj);
CHECK_OBJECT(container);
if(!ForceFromPocket(obj,container,x,y))
    return 0;
return 1;
}


int move_object(OBJECT *obj,int x, int y)
{
CHECK_OBJECT(obj);
if(!MoveObject(obj,x,y))
    return 0;
return 1;
}


void transfer_object(OBJECT *obj,int x, int y)
{
CHECK_OBJECT(obj);
TransferObject(obj,x,y);
}


void spill_contents(OBJECT *bag)
{
CHECK_OBJECT(bag);
spillcontents(bag,bag->x,bag->y);
}

void spill_contents_at(OBJECT *bag,int x, int y)
{
CHECK_OBJECT(bag);
spillcontents(bag,x,y);
}

int weigh_object(OBJECT *obj)
{
CHECK_OBJECT(obj);
return(WeighObject(obj));
}

int get_bulk(OBJECT *obj)
{
CHECK_OBJECT(obj);
return(GetBulk(obj));
}


void play_song(char *a)
{
S_PlayMusic(a);
}


void play_sound(char *a)
{
S_PlaySample(a,sf_volume);
}


void object_sound(char *a, OBJECT *b)
{
int vol,dist,dx,dy,svol;
OBJECT *temp;

svol = sf_volume;

if(SoundFixed)
    {
    S_PlaySample(a,sf_volume);
    return;
    }

if(!b)
    {
    S_PlaySample(a,sf_volume);
    return;
    }

CHECK_OBJECT(b);

if(!b->flags.on)
    {
    S_PlaySample(a,sf_volume);
    return;
    }

if(b->flags.inpocket)
    {
    temp = LookInAllPocketsFor(b);
    if(temp != player)
        {
//    if(temp->flags.person)
//        return;
        b->x = temp->x;
        b->y = temp->y;
        svol = sf_volume >> 2;
        svol = (svol<<1) + svol;       // Vol is now 3/4 sf_volume
        }
    }

vol = svol;

dx = player->x - b->x;
if(dx<0) dx=-dx;
dy = player->y - b->y;
if(dy<0) dy=-dy;

dist=dx;
if(dy>dx) dist=dy;

if(dist>SOUND_DIST)
    {
    dist-=SOUND_DIST;
    vol = svol - (dist*SOUND_DROP);
    if(vol<0) vol=0;
    }

S_PlaySample(a,vol);
}


void stop_song()
{
S_PauseMusic();
}


void start_song()
{
S_UnPauseMusic();
}


int object_is_called(OBJECT *obj,char *str)
{
if(!obj) return 0;
if(!str) return 0;
CHECK_OBJECT(obj);

/*
if(!OB_Check(obj))
	{
	Bug("check_object detected a Fuck-up in object_is_called(\"%s\")\n",str);
	Bug("Detailed report follows:\n");
	CHECK_OBJECT(obj);
	ithe_panic("Fuck-up detected, check logfile",NULL);
	}
*/
if(!obj->name)
	ithe_panic("check_object","Object has NULL name");

if(!istricmp(obj->name,str))
    return 1;
return 0;
}

void redraw()
{
irecon_update();            // Redraw all text output
Show();                     // And blit the screen again
}


void get_input()
{
irekey=get_key_debounced();
}

int get_number(int no)
{
if(!get_num(no))
    return 0;
return atoi(user_input);
}

int is_solid(int x,int y)
{
return isSolid(x,y);
}


int rnd(int max)
{
if(max==0) return 0;    // Prevent division by 0
return(rand() % max);
}

int choose_member(OBJECT *x, char *player_call)
{
int ctr;

if(!x)
    return 0;
//C_printf("choose member: %s\n",x->personalname?x->personalname:x->name);
CHECK_OBJECT(x);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	if(party[ctr])
		Inactive(party[ctr]);

player = x;
//KillGoals(player,21);                   // Wipe all previous movement goals
//AddGoal(player,21,player_call,NULL);    // and install a new one
ActivityName(player,player_call,NULL);
return 1;
}


int choose_leader(OBJECT *x, char *player_call, char *follower_call)
{
int ctr;

if(!x)
	return 0;

CHECK_OBJECT(x);
//C_printf("choose leader: %s\n",x->personalname?x->personalname:x->name);

for(ctr=0;ctr<MAX_MEMBERS;ctr++)
	if(party[ctr])
		{
		Inactive(party[ctr]);
		if(party[ctr] != x)
			ActivityName(party[ctr],follower_call,NULL);
//		ilog_printf("setting '%s' to follow function '%s'\n",party[ctr]->name,follower_call);
		}

player = x;
ActivityName(player,player_call,NULL);
return 1;
}


int add_to_party(OBJECT *new_member)
{
CHECK_OBJECT(new_member);
return(AddToParty(new_member));
}


void remove_from_party(OBJECT *member)
{
CHECK_OBJECT(member);
SubFromParty(member);
ResumeSchedule(member);
}


void set_object_behaviour(OBJECT *obj,char *vrm)
{
Bug("set_object_behaviour() is obsolete, please use add_goal() instead\n");
//CHECK_OBJECT(obj);
//AddGoal(obj,-10,vrm,NULL);
}

void add_goal(OBJECT *o, int priority, char *vrm, OBJECT *target)
{
if(o && vrm)
	{
	CHECK_OBJECT(o);
#ifdef CHECK_OBJECT_STRICT
	if(target)
		{
		CHECK_OBJECT(target);
		}
#endif
//	Bug("%s does %s to %x\n",o->name,vrm,target);
	ActivityName(o,vrm,target);
	}
}

void end_goal(OBJECT *o)
{
if(o)
	{
	CHECK_OBJECT(o);
	Inactive(o);
	}
}

void wipe_goals(OBJECT *o)
{
if(o)
	{
	CHECK_OBJECT(o);
	Inactive(o);
	}
}

void kill_goals(OBJECT *o, int priority)
{
if(o)
	{
	CHECK_OBJECT(o);
	Inactive(o);
	}
}

void restart()
{
Restart();
}

void set_darkness(int l)
{
if(l<0)
	l=0;
if(l>255)
	l=255;
SetDarkness(l);
}

void talk_to(char *speechfile, char *start)
{
char filename[1024];
if(!speechfile)
    return;
if(!speechfile[0])
    return;
if(loadfile(speechfile,filename))
    NPC_Converse(filename,start);
else
    Bug("CONVERSATION FILE '%s' NOT FOUND\n",speechfile);
}

void check_hurt(OBJECT *obj)
{
#ifdef CHECK_OBJECT_STRICT
if(obj)
	{
	CHECK_OBJECT(obj);
	}
#endif
CheckHurt(obj);
}

void redraw_map()
{
RedrawMap();
}

void waitfor(unsigned ms)
{
//rest(ms);
rest_callback(ms,S_PollMusic);
}

// Line of sight routine, using J. Grant's interpretation of Bresenham
// From the ITG32 graphics library (1995)

int line_of_sight(int xa, int ya, int xb, int yb)
{
int t, distance;
int xerr=0, yerr=0, delta_x, delta_y;
int incx, incy;
int xo,yo;

OBJECT *obj;

// Don't count the origin as a solid object
xo = xa;
yo = ya;

// Find the distance to go in each plane.
        delta_x = xb - xa;
        delta_y = yb - ya;

// Find out the direction
	if(delta_x > 0) incx = 1;
	else if(delta_x == 0) incx = 0;
	else incx = -1;

	if(delta_y > 0) incy = 1;
	else if(delta_y == 0) incy = 0;
	else incy = -1;

	// Find which way is were mostly going
	delta_x = abs(delta_x);
	delta_y = abs(delta_y);
	if(delta_x > delta_y) distance=delta_x;
	else distance=delta_y;

	// Draw the god dam line.
	for(t = 0; t <= distance + 1; t++) {
                if(isSolid(xa,ya))
                    if(!(xa == xo && ya == yo))
                        if(!(xa == xb && ya == yb))
			   {

			   obj = GetSolidObject(xa,ya);

			   if(!obj)
			      return 0;
				CHECK_OBJECT(obj);

			   if(!obj->flags.tabletop)

                              return 0;

			   }

		xerr += delta_x;
		yerr += delta_y;
		if(xerr > distance) {
			xerr -= distance;
			xa += incx;
		}
		if(yerr > distance) {
			yerr -= distance;
			ya += incy;
		}
	}
return 1;
}

void move_to_top(OBJECT *object)
{
CHECK_OBJECT(object);
MoveToTop(object);
}

void move_to_floor(OBJECT *object)
{
CHECK_OBJECT(object);
MoveToFloor(object);
}

int in_pocket(OBJECT *obj)
{
CHECK_OBJECT(obj);
return(InPocket(obj));
}

void DT()
{
#ifdef DEBUGSEER
scClose_Debug();
#endif
}

void set_flag(OBJECT *target, int flag, int value)
{
if(!target)
    return;
CHECK_OBJECT(target);

switch(flag)
    {
    case IS_ON:
    if(target == player)
        {
        Bug("set_flag: Attempted to switch off the player\n");
        redraw();
        return;
        }
    target->flags.on=value;
    if(value == 0)
        pending_delete = 1;
    break;

    case CAN_OPEN:
    target->flags.willopen=value;
    break;

    case IS_OVERLAY:
    Bug("Flag OVERLAY is read-only\n");
    redraw();
    break;

    case IS_SOLID:
    target->flags.solid=value;
    break;

    case IS_FRAGILE:
    target->flags.fragile=value;
    break;

    case IS_TRIGGER:
    Bug("Flag TRIGGER is read-only\n");
    redraw();
    break;

    case IS_INVISIBLE:
    target->flags.invisible=value;
    break;

    case IS_PARTY:
    target->flags.party=value;
    break;

    case IS_FIXED:
    target->flags.fixed=value;
    break;

    case IS_CONTAINER:
    target->flags.container=value;
    break;

    case IS_TRANSLUCENT:
    target->flags.translucent=value;
    break;

    case IS_LARGE:
    Bug("Flag LARGE is read-only\n");
    redraw();
    break;

    case IS_SPIKEPROOF:
    target->flags.spikeproof=value;
    break;

    case CAN_WIELD:
    target->flags.wield=value;
    break;

    case DID_STEPUPDATE:
    target->flags.stepupdated=value;
    break;

    case IS_RANGED:
    target->flags.rangeweapon=value;
    break;

    case DOES_BLOCKLIGHT:
    target->flags.blocklight=value;
    break;

    case IS_TABLETOP:
    target->flags.tabletop=value;
    break;

    case DID_INIT:
    target->flags.didinit=value;
    break;

    case DID_UPDATE:
    target->flags.didupdate=value;
    break;

    case IS_PERSON:
    target->flags.person=value;
    break;

    case IN_POCKET:
    target->flags.inpocket=value;
    break;

    case IS_QUANTITY:
    target->flags.quantity=value;
    break;

    case IS_WATER:
    target->flags.watery=value;
    break;

    case IS_SHADOW:
    target->flags.shadow=value;
    break;

    case IS_FEMALE:
    target->stats->npcflags.female=value;
    break;

    case KNOW_NAME:
    target->stats->npcflags.know_name=value;
    break;

    case IS_HERO:
    target->stats->npcflags.is_hero=value;
    break;

    case CANT_EAT:
    target->stats->npcflags.cant_eat=value;
    break;

    case IS_CRITICAL:
    target->stats->npcflags.critical=value;
    break;

    case NOT_CLOSE_DOOR:
    target->stats->npcflags.no_shutdoor=value;
    break;

    case IS_SYMLINK:
    target->stats->npcflags.symlink=value;
    break;

    case IS_BIOLOGICAL:
    target->stats->npcflags.biological=value;
    break;

    case IS_GUARD:
    target->stats->npcflags.guard=value;
    break;

    default:
    Bug("Attempt to set unknown flag %08x\n",flag);
    redraw();
    break;
    }
}

int get_flag(OBJECT *target, int flag)
{
if(!target)
    return 0;
CHECK_OBJECT(target);

switch(flag)
    {
    case IS_ON:
    return (target->flags.on);
    break;

    case CAN_OPEN:
    return (target->flags.willopen);
    break;

    case IS_OVERLAY:
    return (target->flags.overlay);
    break;

    case IS_SOLID:
    return (target->flags.solid);
    break;

    case IS_FRAGILE:
    return (target->flags.fragile);
    break;

    case IS_TRIGGER:
    return (target->flags.trigger);
    break;

    case IS_INVISIBLE:
    return (target->flags.invisible);
    break;

    case IS_PARTY:
    return (target->flags.party);
    break;

    case IS_FIXED:
    return (target->flags.fixed);
    break;

    case IS_CONTAINER:
    return (target->flags.container);
    break;

    case IS_TRANSLUCENT:
    return (target->flags.translucent);
    break;

    case IS_LARGE:
    return (target->flags.large);
    break;

    case IS_SPIKEPROOF:
    return (target->flags.spikeproof);
    break;

    case CAN_WIELD:
    return (target->flags.wield);
    break;

    case DID_STEPUPDATE:
    return (target->flags.stepupdated);
    break;

    case IS_RANGED:
    return (target->flags.rangeweapon);
    break;

    case DOES_BLOCKLIGHT:
    return (target->flags.blocklight);
    break;

    case IS_TABLETOP:
    return (target->flags.tabletop);
    break;

    case DID_INIT:
    return (target->flags.didinit);
    break;

    case DID_UPDATE:
    return (target->flags.didupdate);
    break;

    case IS_PERSON:
    return (target->flags.person);
    break;

    case IN_POCKET:
    return (target->flags.inpocket);
    break;

    case IS_QUANTITY:
    return (target->flags.quantity);
    break;

    case IS_WATER:
    return (target->flags.watery);
    break;

    case IS_SHADOW:
    return (target->flags.shadow);
    break;

    case IS_FEMALE:
    return (target->stats->npcflags.female);
    break;

    case KNOW_NAME:
    return (target->stats->npcflags.know_name);
    break;

    case IS_HERO:
    return (target->stats->npcflags.is_hero);
    break;

    case CANT_EAT:
    return (target->stats->npcflags.cant_eat);
    break;

    case IS_CRITICAL:
    return (target->stats->npcflags.critical);
    break;

    case NOT_CLOSE_DOORS:
    return (target->stats->npcflags.no_shutdoor);
    break;

    case IS_SYMLINK: // Protection:  if there's nothing to link to, fail
	if(target->stats->owner)
	    return (target->stats->npcflags.symlink);
	else
		return 0;
    break;

    case IS_BIOLOGICAL:
    return (target->stats->npcflags.biological);
    break;

    case IS_GUARD:
    return (target->stats->npcflags.guard);
    break;
    }

Bug("Attempt to read unknown flag %08x\n",flag);
redraw();
return 0;
}

void default_flag(OBJECT *target, int flag)
{
int id;
if(!target)
    return;
CHECK_OBJECT(target);

id = getnum4char(target->name);
if(id == -1)
    return;

switch(flag)
    {
    case IS_ON:
    target->flags.on=1;
    break;

    case CAN_OPEN:
    target->flags.willopen=CHlist[id].flags.willopen;
    break;

    case IS_OVERLAY:
    Bug("Flag OVERLAY is read-only\n");
    redraw();
    break;

    case IS_SOLID:
    target->flags.solid=CHlist[id].flags.solid;
    break;

    case IS_FRAGILE:
    target->flags.fragile=CHlist[id].flags.fragile;
    break;

    case IS_TRIGGER:
    Bug("Flag TRIGGER is read-only\n");
    redraw();
    break;

    case IS_INVISIBLE:
    target->flags.invisible=CHlist[id].flags.invisible;
    break;

    case IS_PARTY:
    target->flags.party=CHlist[id].flags.party;
    break;

    case IS_FIXED:
    target->flags.fixed=CHlist[id].flags.fixed;
    break;

    case IS_CONTAINER:
    target->flags.container=CHlist[id].flags.container;
    break;

    case IS_TRANSLUCENT:
    target->flags.translucent=CHlist[id].flags.translucent;
    break;

    case IS_LARGE:
    Bug("Flag LARGE is read-only\n");
    redraw();
    break;

    case IS_SPIKEPROOF:
    target->flags.spikeproof=CHlist[id].flags.spikeproof;
    break;

    case CAN_WIELD:
    target->flags.wield=CHlist[id].flags.wield;
    break;

    case DID_STEPUPDATE:
    target->flags.stepupdated=CHlist[id].flags.stepupdated;
    break;

    case IS_RANGED:
    target->flags.rangeweapon=CHlist[id].flags.rangeweapon;
    break;

    case DOES_BLOCKLIGHT:
    target->flags.blocklight=CHlist[id].flags.blocklight;
    break;

    case IS_TABLETOP:
    target->flags.tabletop=CHlist[id].flags.tabletop;
    break;

    case DID_INIT:
    target->flags.didinit=CHlist[id].flags.didinit;
    break;

    case DID_UPDATE:
    target->flags.didupdate=CHlist[id].flags.didupdate;
    break;

    case IS_PERSON:
    target->flags.person=CHlist[id].flags.person;
    break;

    case IN_POCKET:
    target->flags.inpocket=CHlist[id].flags.inpocket;
    break;

    case IS_QUANTITY:
    target->flags.quantity=CHlist[id].flags.quantity;
    break;

    case IS_WATER:
    target->flags.shadow=CHlist[id].flags.watery;
    break;

    case IS_SHADOW:
    target->flags.shadow=CHlist[id].flags.shadow;
    break;

    case IS_FEMALE:
    target->stats->npcflags.female=CHlist[id].stats->npcflags.female;
    break;

    case KNOW_NAME:
    target->stats->npcflags.know_name=CHlist[id].stats->npcflags.know_name;
    break;

    case IS_HERO:
    target->stats->npcflags.is_hero=CHlist[id].stats->npcflags.is_hero;
    break;

    case CANT_EAT:
    target->stats->npcflags.cant_eat=CHlist[id].stats->npcflags.cant_eat;
    break;

    case IS_CRITICAL:
    target->stats->npcflags.critical=CHlist[id].stats->npcflags.critical;
    break;

    case NOT_CLOSE_DOOR:
    target->stats->npcflags.no_shutdoor=CHlist[id].stats->npcflags.no_shutdoor;
    break;

    case IS_SYMLINK:
    target->stats->npcflags.symlink=CHlist[id].stats->npcflags.symlink;
    break;

    case IS_BIOLOGICAL:
    target->stats->npcflags.biological=CHlist[id].stats->npcflags.biological;
    break;

    case IS_GUARD:
    target->stats->npcflags.guard=CHlist[id].stats->npcflags.guard;
    break;

    default:
    Bug("Attempt to default unknown flag %08x\n",flag);
    redraw();
    break;
    }
}

int get_tileflag(TILE *target, int flag)
{
if(!target)
    return 0;

switch(flag)
    {
    case IS_ON:
    return (target->flags.on);
    break;

    case CAN_OPEN:
    return (target->flags.willopen);
    break;

    case IS_OVERLAY:
    return (target->flags.overlay);
    break;

    case IS_SOLID:
    return (target->flags.solid);
    break;

    case IS_FRAGILE:
    return (target->flags.fragile);
    break;

    case IS_TRIGGER:
    return (target->flags.trigger);
    break;

    case IS_INVISIBLE:
    return (target->flags.invisible);
    break;

    case IS_PARTY:
    return (target->flags.party);
    break;

    case IS_FIXED:
    return (target->flags.fixed);
    break;

    case IS_CONTAINER:
    return (target->flags.container);
    break;

    case IS_TRANSLUCENT:
    return (target->flags.translucent);
    break;

    case IS_LARGE:
    return (target->flags.large);
    break;

    case IS_SPIKEPROOF:
    return (target->flags.spikeproof);
    break;

    case CAN_WIELD:
    return (target->flags.wield);
    break;

    case DID_STEPUPDATE:
    return (target->flags.stepupdated);
    break;

    case IS_RANGED:
    return (target->flags.rangeweapon);
    break;

    case DOES_BLOCKLIGHT:
    return (target->flags.blocklight);
    break;

    case IS_TABLETOP:
    return (target->flags.tabletop);
    break;

    case DID_INIT:
    return (target->flags.didinit);
    break;

    case DID_UPDATE:
    return (target->flags.didupdate);
    break;

    case IS_PERSON:
    return (target->flags.person);
    break;

    case IN_POCKET:
    return (target->flags.inpocket);
    break;

    case IS_QUANTITY:
    return (target->flags.quantity);
    break;

    case IS_WATER:
    return (target->flags.watery);
    break;

    case IS_SHADOW:
    return (target->flags.shadow);
    break;

    case IS_FEMALE:
	Bug("Tried to read flag IS_FEMALE from a Tile: not allowed for tiles\n");
	return 0;
	break;

    case KNOW_NAME:
	Bug("Tried to read flag KNOW_NAME from a Tile: not allowed for tiles\n");
	return 0;
	break;

    case IS_HERO:
	Bug("Tried to read flag IS_HERO from a Tile: not allowed for tiles\n");
	return 0;
	break;

    case CANT_EAT:
	Bug("Tried to read flag CANT_EAT from a Tile: not allowed for tiles\n");
	return 0;
	break;

    case IS_CRITICAL:
	Bug("Tried to read flag IS_CRITICAL from a Tile: not allowed for tiles\n");
	return 0;
	break;

    case NOT_CLOSE_DOORS:
	Bug("Tried to read flag NOT_CLOSE_DOORS from a Tile: not allowed for tiles\n");
    return 0;
    break;

    case IS_SYMLINK:
	Bug("Tried to read flag IS_SYMLINK from a Tile: not allowed for tiles\n");
    return 0;
    break;

    case IS_BIOLOGICAL:
	Bug("Tried to read flag IS_BIOLOGICAL from a Tile: not allowed for tiles\n");
    return 0;
    break;

    case IS_GUARD:
	Bug("Tried to read flag IS_GUARD from a Tile: not allowed for tiles\n");
    return 0;
    break;
    }

Bug("Attempt to read unknown flag %08x\n",flag);
redraw();
return 0;
}


int take_quantity(OBJECT *container, char *objecttype, int quantity)
{
CHECK_OBJECT(container);
return (TakeQuantity(container,objecttype,quantity));
}

void add_quantity(OBJECT *container, char *objecttype, int quantity)
{
CHECK_OBJECT(container);
AddQuantity(container,objecttype,quantity);
}

int move_quantity(OBJECT *src, OBJECT *dest, char *objecttype, int quantity)
{
if(!src)
    {
    Bug("move_quantity with NULL source\n");
    return 0;
    }
CHECK_OBJECT(src);
if(!dest)
    {
    Bug("move_quantity with NULL destination\n");
    return 0;
    }
CHECK_OBJECT(dest);
return (MoveQuantity(src,dest,objecttype,quantity));
}

int count_objects(OBJECT *container, char *objecttype)
{
CHECK_OBJECT(container);
return SumObjects(container->pocket,objecttype,0);
}

OBJECT *find_object_with_tag(int tag,char *name)
{
OBJLIST *t;

for(t=MasterList;t;t=t->next)
    if(t->ptr->tag == tag)
        if(name)
            {
            if(!istricmp(name,t->ptr->name))
                return t->ptr;
            }
        else
            return t->ptr;

return NULL;
}

int get_yn(char *question)
{
return(getYN(question));
}

void set_pflag(OBJECT *a,int flag, int state)
{
if(!a)
	return;
CHECK_OBJECT(a);
if(flag<0 || flag>31)
    {
    Bug("Attempt to set invalid p-flag %d for object %s, in a VRM\n",flag,a->name);
    return;
    }
setPflag(a,flag,state);
}

int get_pflag(OBJECT *a,int flag)
{
if(!a)
    return 0;
CHECK_OBJECT(a);
if(flag<0 || flag>31)
    {
    Bug("Attempt to read invalid p-flag %d for object %s, in a VRM\n",flag,a->name);
    return 0;
    }
return(getPflag(a,flag));
}

void set_user_flag(char *a, int s)
{
Set_tFlag(a,s);
}

int get_user_flag(char *a)
{
return(Get_tFlag(a));
}


void set_local(OBJECT *plyr, OBJECT *obj, char *a, int s)
{
NPC_set_lFlag(obj,a,plyr,s);
}

int get_local(OBJECT *plyr, OBJECT *obj, char *a)
{
return(NPC_get_lFlag(obj,a,plyr));
}

int move_forward(OBJECT *a)
{
if(!a)
    return 0;
CHECK_OBJECT(a);

a->flags.stepupdated=1; // Force animation

if(a->curdir == CHAR_U)
    return MoveObject(a,a->x,a->y-1);
if(a->curdir == CHAR_D)
    return MoveObject(a,a->x,a->y+1);
if(a->curdir == CHAR_L)
    return MoveObject(a,a->x-1,a->y);
if(a->curdir == CHAR_R)
    return MoveObject(a,a->x+1,a->y);
return 0;
}

int move_backward(OBJECT *a)
{
if(!a)
    return 0;
CHECK_OBJECT(a);

a->flags.stepupdated=1; // Force animation

if(a->curdir == CHAR_U)
    return MoveObject(a,a->x,a->y+1);
if(a->curdir == CHAR_D)
    return MoveObject(a,a->x,a->y-1);
if(a->curdir == CHAR_L)
    return MoveObject(a,a->x+1,a->y);
if(a->curdir == CHAR_R)
    return MoveObject(a,a->x-1,a->y);
return 0;
}

int turn_l(OBJECT *a)
{
if(!a)
    return 0;
CHECK_OBJECT(a);

switch(a->curdir)
    {
    case CHAR_U:
    OB_SetDir(a,CHAR_L);
    break;

    case CHAR_L:
    OB_SetDir(a,CHAR_D);
    break;

    case CHAR_D:
    OB_SetDir(a,CHAR_R);
    break;

    case CHAR_R:
    OB_SetDir(a,CHAR_U);
    break;
    }
return a->curdir;
}

int turn_r(OBJECT *a)
{
if(!a)
    return 0;
CHECK_OBJECT(a);

switch(a->curdir)
    {
    case CHAR_U:
    OB_SetDir(a,CHAR_R);
    break;

    case CHAR_R:
    OB_SetDir(a,CHAR_D);
    break;

    case CHAR_D:
    OB_SetDir(a,CHAR_L);
    break;

    case CHAR_L:
    OB_SetDir(a,CHAR_U);
    break;
    }
return a->curdir;
}

void wait_for_animation(OBJECT *obj)
{
if(!obj)
    return;
CHECK_OBJECT(obj);

if(obj->form->flags&6)  // Ignore if looped or stepped
    return;             // (Otherwise it will never come back)

if(InPocket(obj))       // Objects in pockets do not animate
    return;

for(;obj->sdir;RedrawMap());
}

void ___lightning(int ticks)
{
do_lightning=ticks;
}

OBJECT *find_container(OBJECT *obj)
{
if(!obj)
	return NULL;
CHECK_OBJECT(obj);
return LookInAllPocketsFor(obj);
}

int is_tile_solid(int x,int y)
{
return IsTileSolid(x,y);
}

int is_tile_water(int x,int y)
{
return IsTileWater(x,y);
}

void move_party_to_object(OBJECT *o)
{
if(!o)
	return;
CHECK_OBJECT(o);
for(int ctr=0;ctr<MAX_MEMBERS;ctr++)
    if(party[ctr])
        TransferToPocket(party[ctr],o);
}

void move_party_from_object(OBJECT *o, int x, int y)
{
if(!o)
	return;
CHECK_OBJECT(o);
for(int ctr=0;ctr<MAX_MEMBERS;ctr++)
    if(party[ctr])
        ForceFromPocket(party[ctr],o,x,y);
}


int move_towards(OBJECT *object, OBJECT *end)
{
char *s1,*s2;

if(!object || !end)
    {
    s1="Null";
    s2="Null";

    if(object)
        s1=BestName(object);
    if(end)
        s2=BestName(end);
    Bug("FindPath (%s,%s)\n",s1,s2);
    return 0;
    }

CHECK_OBJECT(object);
CHECK_OBJECT(end);

//ilog_quiet("findpath(%s,%s)",object->name,end->name);
if(!stricmp(object->name,"plate_broken"))
	ithe_panic("Oh no not again","plate is posessed");
return FindPath(object,end,1);
}

int move_towards_4(OBJECT *object, OBJECT *end)
{
char *s1,*s2;

if(!object || !end)
    {
    s1="Null";
    s2="Null";

    if(object)
        s1=BestName(object);
    if(end)
        s2=BestName(end);
    Bug("FindPath (%s,%s)\n",s1,s2);
    return 0;
    }

CHECK_OBJECT(object);
CHECK_OBJECT(end);

return FindPath(object,end,0);
}

int move_thick(OBJECT *object, OBJECT *end)
{
char *s1,*s2;

if(!object || !end)
    {
    s1="Null";
    s2="Null";

    if(object)
        s1=object->name;
    if(end)
        s2=end->name;
    Bug("MoveThick (%s,%s)\n",s1,s2);
    return 0;
    }

CHECK_OBJECT(object);
CHECK_OBJECT(end);

object->user->dx=0;
if(object->x < end->x)
    object->user->dx=1;
if(object->x > end->x)
    object->user->dx=-1;
object->user->dy=0;
if(object->y < end->y)
    object->user->dy=1;
if(object->y > end->y)
    object->user->dy=-1;

if(object->user->dy<0)
    OB_SetDir(object,CHAR_U);
if(object->user->dy>0)
    OB_SetDir(object,CHAR_D);
if(object->user->dx<0)
    OB_SetDir(object,CHAR_L);
if(object->user->dx>0)
    OB_SetDir(object,CHAR_R);
MoveObject(object,object->x+object->user->dx,object->y+object->user->dy);
object->flags.stepupdated=1; // Force animation
return 1;
}

int character_onscreen(char *pname)
{
return IsOnscreen(pname);
}

void scroll_things()
{
int ctr,x,y;
for(ctr=0;ctr<TItot;ctr++)
    {
    x = TIlist[ctr].name[1];
    if(x<'A' || x>'Z')
        break;
    x =((x-'A')%3)-1;
    y = TIlist[ctr].name[1];
    if(y<'A' || y>'Z')
        break;
    y =((y-'A')%5)-2;
    TIlist[ctr].sdx=x;
    TIlist[ctr].sdy=y;
    }
}

void scroll_tile(char *name, int x, int y)
{
int ctr;

if(x<0)
    x=32+x;
if(y<0)
    y=32+y;

for(ctr=0;ctr<TItot;ctr++)
    if(!istricmp(TIlist[ctr].name,name))
        {
        TIlist[ctr].sdx=x;
        TIlist[ctr].sdy=y;
        // If zero, reset the scroll offsets, since the user probably wants it to
        // go back to normal

        if(x == 0 && y == 0)
            {
            TIlist[ctr].sx=0;
            TIlist[ctr].sy=0;
            }
        return;
        }
Bug("No such tile '%s'\n",name);
}

void scroll_tile_number(int no, int x, int y)
{
if(no<0 || no>=TItot)
    {
    Bug("No such tile number %d\n",no);
    return;
    }

if(x<0)
    x=32+x;
if(y<0)
    y=32+y;

TIlist[no].sdx=x;
TIlist[no].sdy=y;

// If zero, reset the scroll offsets, since the user probably wants it to
// go back to normal

if(x == 0 && y == 0)
    {
    TIlist[no].sx=0;
    TIlist[no].sy=0;
    }
}

void sysBug(char *msg, ...)
{
char temp[4096];
va_list ap;

if(!msg)
    {
    Bug("Attemped to Bug-print NULL message\n");
    va_end(ap);
    redraw();
    return;
    }

va_start(ap, msg);
vsprintf(temp,msg,ap);          // Temp is now the message
va_end(ap);

Bug(temp);
}

OBJECT *search_container(OBJECT *container,char *name)
{
if(!container)
	return NULL;
CHECK_OBJECT(container);
return GetFirstObject(container->pocket, name);
}

void vrmfade_in()
{
for(Fader=255;Fader>0;Fader-=8)
    {
    RedrawMap();
    rest(1);
	S_PollMusic();
    }
Fader=0;
}

void vrmfade_out()
{
for(Fader=0;Fader<255;Fader+=8)
    {
    RedrawMap();
    rest(1);
	S_PollMusic();
    }
Fader=255;
}

void check_time()
{
CheckTime();
}

OBJECT *find_nearest(OBJECT *o, char *type)
{
#define INF 2000000000
char *s1,*s2;
OBJECT *temp;
int x,y,mix,miy,max,may,steps;

// the three possible cases, best of each
OBJECT *mine_ob=NULL;
int mine_st=INF;
OBJECT *public_ob=NULL;
int public_st=INF;
OBJECT *private_ob=NULL;
int private_st=INF;

if(!o || !type)
    {
    s1="Null";
    s2="Null";

    if(o)
        s1=o->name;
    if(type)
        s2=type;
    Bug("find_nearest(%s,%s)\n",s1,s2);
    return 0;
    }

CHECK_OBJECT(o);

// Now build the search area and make sure it is within the bounds of the map

mix = o->x - 16;
miy = o->y - 16;
if(mix<0)
    mix=0;
if(miy<0)
    miy=0;
max=mix+32;
may=miy+32;

if(max>curmap->w)
    max=curmap->w;
if(may>curmap->h)
    may=curmap->h;

// Ok, start searching

for(y=miy;y<may;y++)
    for(x=mix;x<max;x++)
        for(temp=GetRawObjectBase(x,y);temp;temp=temp->next)
            if(temp->flags.on)
                if(!istricmp(temp->name,type))
                    {
                    steps = CanRoute(temp,o,1);
                    if(steps >= 0)
                        {
                        if(temp->stats->owner == o)    // best case, mine
                            {
                            if(steps<mine_st)
                                {
                                mine_st = steps;
                                mine_ob = temp;
                                }
                            }
                        else
                        if(temp->stats->owner == NULL)    // next best case, public
                            {
                            if(steps<public_st)
                                {
                                public_st = steps;
                                public_ob = temp;
                                }
                            }
                        else                // worst case, someone else's
                            {
                            if(steps<private_st)
                                {
                                private_st = steps;
                                private_ob = temp;
                                }
                            }
                        }
                    }

// Now we should have the best possible case for each of the three types

if(mine_ob)
    return mine_ob;

if(public_ob)
    return public_ob;

return private_ob;
}

/*
 *      return a string with the optimal description of the object
 */

char *best_name(OBJECT *o)
{
if(!o)
	return nullstr;
CHECK_OBJECT(o);
return BestName(o);
}

/*
 *      count_active_objects() - diagnostic data
 */

int count_active_objects()
{
OBJLIST *t;
int ctr=0;

for(t=ActiveList;t;t=t->next) ctr++;
return ctr;
}



int object_onscreen(OBJECT *a)
{
if(!a)
      {
      Bug("Tried to find out if NULL is onscreen\n");
      return 0;
      }
CHECK_OBJECT(a);

// Is it inside the window?

if(a->x >=mapx)
    if(a->x <= mapx2)
        if(a->y >=mapy)
            if(a->y <= mapy2)
                return 1;    // Yes

return 0; // No
}

void resume_schedule(OBJECT *o)
{
if(!o)
	return;
CHECK_OBJECT(o);
ResumeSchedule(o);
}

void resync_everything()
{
ResyncEverything();
}

OBJECT *get_object_below(OBJECT *o)
{
OBJECT *temp;
if(!o)
    {
    Bug("get_object_below(NULL);\n");
    return NULL;
    }
CHECK_OBJECT(o);

if(o->flags.inpocket)   // Don't bother
    return NULL;

temp = GetObjectBase(o->x,o->y);
if(!temp)
    {
    Bug("get_object_below found nothing: map corrupt?\n");
    return NULL;
    }

if(temp == o)   // Nothing below it
    return NULL;

for(;temp->next;temp=temp->next)
    if(temp->next == o)
        return temp;

//Bug("Ooh bugger!\n");
return NULL;
}

void check_object(OBJECT *target, char *label)
{
if(!target)
	return;

if(!OB_Check(target))
	{
	Bug("check_object detected a Fuck-up in VRM-space (\"%s\")\n",label);
	ithe_panic("Fuck-up detected, check logfile",NULL);
	}
}

void set_light(int x, int y, int x2, int y2, int light)
{
int ctr,w;
if(x<0 || y<0 || x2>curmap->w || y2>curmap->h)
	{
	Bug("Tried to Set Light (%d,%d-%d,%d) outside map (%d,%d-%d,%d)\n",x,y,x2,y2,0,0,curmap->w,curmap->h);
	return;
	}
w=(x2-x)+1;
if(w<0)
	{
	Bug("Bogus light range (%d,%d to %d,%d)\n",x,y,x2,y2);
	return;
	}

for(ctr=y;ctr<=y2;ctr++)
	memset(&curmap->lightst[(ctr*curmap->w)+x],light,w);
}

void find_objects_with_tag(int tag,char *name, int *num, OBJECT **obj)
{
OBJLIST *t;
int ctr;

if(!obj)
	return;
if(!num)
	return;
if(*num<1)
	return;

ctr=0;
for(t=MasterList;t;t=t->next)
	if(t->ptr->tag == tag)
		if(name)
			if(!istricmp(name,t->ptr->name))
				if(ctr < *num)
					obj[ctr++]=t->ptr;
*num=ctr;
return;
}

void Call_VM(char *c)
{
CallVM(c);
}

void Call_VM_num(int n)
{
CallVMnum(n);
}

