//
//      VRM_IN.HPP - Internal version of vrm.hpp
//

#ifndef IRE_NATIVE_EXECUTION_MODULE_HEADER_FILE
#define IRE_NATIVE_EXECUTION_MODULE_HEADER_FILE

#include "core.hpp"

typedef char *STRING;


#include "seer/seer.h"

#define TYPE "C"
#ifndef NULL
  #define NULL 0L
#endif


/////// Virtual Runnable Module subroutines ////////

#define AND &&
#define OR ||
#define NOT !

extern TYPE void print(char *msg, ...);
extern TYPE void printF(char *msg, ...);
extern TYPE void vrmclear();
extern TYPE void printxy(int x,int y,char *msg, ...);
extern TYPE void TextColor(int r,int g, int b);

extern TYPE OBJECT *create_object(STRING name, int x, int y);
extern TYPE void delete_pending(void);
extern TYPE void remove_object(OBJECT *obj);
extern TYPE void change_object(OBJECT *obj, STRING name);
extern TYPE void replace_object(OBJECT *obj, STRING name);
extern TYPE void set_object_direction(OBJECT *obj, int dir);
extern TYPE void set_object_sequence(OBJECT *obj, STRING name);
extern TYPE void set_object_behaviour(OBJECT *obj, STRING name);
extern TYPE void add_goal(OBJECT *o, int priority, char *vrm, OBJECT *target);
extern TYPE void end_goal(OBJECT *o);
extern TYPE void wipe_goals(OBJECT *o);
extern TYPE void kill_goals(OBJECT *o,int p);
#define set_object_behavior set_object_behaviour
extern TYPE void spill_contents(OBJECT *bag);
extern TYPE void spill_contents_at(OBJECT *bag,int x,int y);
extern TYPE int get_flag(OBJECT *target, int flag);
extern TYPE int get_tileflag(TILE *target, int flag);
extern TYPE void set_flag(OBJECT *target, int flag, int value);
extern TYPE void default_flag(OBJECT *target, int flag);
extern TYPE void add_quantity(OBJECT *container, char *objecttype, int quantity);
extern TYPE int take_quantity(OBJECT *container, char *objecttype, int quantity);
extern TYPE int move_quantity(OBJECT *src, OBJECT *dest, char *objecttype, int quantity);
extern TYPE int count_objects(OBJECT *container, char *objecttype);
extern TYPE OBJECT *find_object_with_tag(int a, char *name);
extern TYPE OBJECT *find_container(OBJECT *ex);
extern TYPE int get_pflag(OBJECT *target, int flag);
extern TYPE void set_pflag(OBJECT *target, int flag, int state);
extern TYPE int get_user_flag(char *flag);
extern TYPE void set_user_flag(char *flag, int state);
extern TYPE int get_local(OBJECT *pl,OBJECT *ch,char *flag);
extern TYPE void set_local(OBJECT *pl,OBJECT *ch,char *flag, int state);
extern TYPE void show_object(OBJECT *z,int x,int y);
extern TYPE void call_vrm(char *name);
extern TYPE void call_vrm_number(int name);
extern TYPE char *get_vrm_name(int num);
extern TYPE void play_song(char *song);
extern TYPE void play_sound(char *sound);
extern TYPE void object_sound(char *sound, OBJECT *o);
extern TYPE void start_song();
extern TYPE void stop_song();
extern TYPE int object_is_called(OBJECT *obj,char *str);
extern TYPE int rnd(int max);
extern TYPE void waitfor(unsigned millisec);
extern TYPE int get_number(int no);

extern TYPE TILE *get_tile(int x,int y);
extern TYPE OBJECT *get_object(int x,int y);
extern TYPE OBJECT *get_solid_object(int x,int y);
extern TYPE OBJECT *get_first_object(int x,int y);
extern TYPE OBJECT *get_object_below(OBJECT *o);
extern TYPE OBJECT *search_container(OBJECT *pocket, char *name);
extern TYPE int count_active_objects();
extern TYPE void find_objects_with_tag(int a, char *name, int *num, OBJECT **list);


extern TYPE void move_to_pocket(OBJECT *src,OBJECT *dest);
extern TYPE void transfer_to_pocket(OBJECT *src,OBJECT *dest);
extern TYPE int force_from_pocket(OBJECT *obj,OBJECT *container,int x,int y);
extern TYPE int move_from_pocket(OBJECT *obj,OBJECT *container,int x,int y);
extern TYPE void move_to_top(OBJECT *obj);
extern TYPE void move_to_floor(OBJECT *obj);
extern TYPE int weigh_object(OBJECT *obj);
extern TYPE int get_bulk(OBJECT *obj);
extern TYPE int is_tile_solid(int x,int y);
extern TYPE int is_tile_water(int x,int y);
extern TYPE int is_solid(int x,int y);
extern TYPE int in_pocket(OBJECT *o);
extern TYPE int move_object(OBJECT *src,int x,int y);
extern TYPE void transfer_object(OBJECT *src,int x,int y);
extern TYPE int line_of_sight(int xa, int ya, int xb, int yb);
extern TYPE int move_forward(OBJECT *a);
extern TYPE int move_backward(OBJECT *a);
extern TYPE int turn_l(OBJECT *a);
extern TYPE int turn_r(OBJECT *a);
extern TYPE void wait_for_animation(OBJECT *a);
extern TYPE int move_towards(OBJECT *object, OBJECT *end);
extern TYPE int move_towards_4(OBJECT *object, OBJECT *end);
extern TYPE int move_towards_8(OBJECT *object, OBJECT *end);
extern TYPE int move_thick(OBJECT *object, OBJECT *end);

extern TYPE int add_to_party(OBJECT *new_member);
extern TYPE void remove_from_party(OBJECT *member);
extern TYPE int choose_member(OBJECT *member, char *player);
extern TYPE int choose_leader(OBJECT *leader, char *player, char *others);
extern TYPE void move_party_from_object(OBJECT *o, int x, int y);
extern TYPE void move_party_to_object(OBJECT *o);

extern TYPE void set_darkness(int level);
extern TYPE void redraw();
extern TYPE void restart();
extern TYPE void get_input();
extern TYPE void talk_to(char *speechfile,char *startpage);
extern TYPE void check_hurt(OBJECT *list);
extern TYPE void redraw_map();
extern TYPE int get_yn(char *question);
extern TYPE void ___lightning(int ticks);
extern TYPE void scroll_tile(char *name, int x, int y);
extern TYPE void scroll_tile_number(int no, int x, int y);
extern TYPE void vrmfade_in();
extern TYPE void vrmfade_out();
extern TYPE void check_time();
extern TYPE OBJECT *find_nearest(OBJECT *o, char *type);
extern TYPE char *best_name(OBJECT *o);
extern TYPE int character_onscreen(char *pname);
extern TYPE int object_onscreen(OBJECT *a);
extern TYPE void resume_schedule(OBJECT *o);
extern TYPE void resync_everything();
extern TYPE void check_object(OBJECT *o, char *label);
extern TYPE void set_light(int x, int y, int x2, int y2, int light);

/*
extern OBJECT *syspocket;
extern OBJECT *player;
extern OBJECT *victim;
extern OBJECT *current_object;
extern TILE   *current_tile;
extern STRING message;
extern int    *storage;
extern OBJECT **objectstore;
extern "C" int show_roof;
extern int    irekey;
extern int    new_x,new_y;
extern int    mapx,mapy;
extern int    mapx2,mapy2;
*/

#define hurt_object(x,y) {x->stats->hp-=y; check_hurt(x);}

#define UP 0
#define DOWN 1
#define LEFT 2
#define RIGHT 3

#define CHAR_U 0
#define CHAR_D 1
#define CHAR_L 2
#define CHAR_R 3

#define IS_ON           0x00000001
#define CAN_OPEN        0x00000002
#define IS_OVERLAY      0x00000004
#define IS_SOLID        0x00000008
#define IS_FRAGILE      0x00000010
#define IS_TRIGGER      0x00000020
#define IS_INVISIBLE    0x00000040
#define IS_PARTY        0x00000080
#define IS_FIXED        0x00000100
#define IS_CONTAINER    0x00000200
#define IS_TRANSLUCENT  0x00000400
#define IS_LARGE        0x00000800
#define IS_SPIKEPROOF   0x00001000
#define CAN_WIELD       0x00002000
#define DID_STEPUPDATE  0x00004000
#define IS_RANGED       0x00008000
#define DOES_BLOCKLIGHT 0x00010000
#define IS_TABLETOP     0x00020000
#define DID_INIT        0x00040000
#define DID_UPDATE      0x00080000
#define IS_PERSON       0x00100000
#define IN_POCKET       0x00200000
#define IS_QUANTITY     0x00400000
#define IS_BOAT         0x00800000
#define IS_WATER        0x00800000
#define IS_SHADOW       0x01000000

#define IS_FEMALE       0x10000001
#define KNOW_NAME       0x10000002
#define IS_HERO         0x10000004
#define CANT_EAT        0x10000008
#define CANT_DRINK      0x10000008      // Synonymous with CANT_EAT
#define IS_CRITICAL     0x10000010
#define NOT_CLOSE_DOOR  0x10000020
#define NOT_CLOSE_DOORS 0x10000020
#define IS_SYMLINK      0x10000040
#define IS_BIOLOGICAL   0x10000080
#define IS_GUARD        0x10000100


#define KEY_A                 1
#define KEY_B                 2
#define KEY_C                 3
#define KEY_D                 4
#define KEY_E                 5
#define KEY_F                 6
#define KEY_G                 7
#define KEY_H                 8
#define KEY_I                 9
#define KEY_J                 10
#define KEY_K                 11
#define KEY_L                 12
#define KEY_M                 13
#define KEY_N                 14
#define KEY_O                 15
#define KEY_P                 16
#define KEY_Q                 17
#define KEY_R                 18
#define KEY_S                 19
#define KEY_T                 20
#define KEY_U                 21
#define KEY_V                 22
#define KEY_W                 23
#define KEY_X                 24
#define KEY_Y                 25
#define KEY_Z                 26
#define KEY_0                 27
#define KEY_1                 28
#define KEY_2                 29
#define KEY_3                 30
#define KEY_4                 31
#define KEY_5                 32
#define KEY_6                 33
#define KEY_7                 34
#define KEY_8                 35
#define KEY_9                 36
#define KEY_0_PAD             37
#define KEY_1_PAD             38
#define KEY_2_PAD             39
#define KEY_3_PAD             40
#define KEY_4_PAD             41
#define KEY_5_PAD             42
#define KEY_6_PAD             43
#define KEY_7_PAD             44
#define KEY_8_PAD             45
#define KEY_9_PAD             46
#define KEY_F1                47
#define KEY_F2                48
#define KEY_F3                49
#define KEY_F4                50
#define KEY_F5                51
#define KEY_F6                52
#define KEY_F7                53
#define KEY_F8                54
#define KEY_F9                55
#define KEY_F10               56
#define KEY_F11               57
#define KEY_F12               58
#define KEY_ESC               59
#define KEY_TILDE             60
#define KEY_MINUS             61
#define KEY_EQUALS            62
#define KEY_BACKSPACE         63
#define KEY_TAB               64
#define KEY_OPENBRACE         65
#define KEY_CLOSEBRACE        66
#define KEY_ENTER             67
#define KEY_COLON             68
#define KEY_QUOTE             69
#define KEY_BACKSLASH         70
#define KEY_BACKSLASH2        71
#define KEY_COMMA             72
#define KEY_STOP              73
#define KEY_DOT               73
#define KEY_SLASH             74
#define KEY_SPACE             75
#define KEY_INSERT            76
#define KEY_DEL               77
#define KEY_HOME              78
#define KEY_END               79
#define KEY_PGUP              80
#define KEY_PGDN              81
#define KEY_LEFT              82
#define KEY_RIGHT             83
#define KEY_UP                84
#define KEY_DOWN              85
#define KEY_SLASH_PAD         86
#define KEY_ASTERISK          87
#define KEY_MINUS_PAD         88
#define KEY_PLUS_PAD          89
#define KEY_DEL_PAD           90
#define KEY_ENTER_PAD         91
#define KEY_PRTSCR            92
#define KEY_PAUSE             93
#define KEY_YEN               94
#define KEY_YEN2              95

#define KEY_MODIFIERS         96

#define KEY_LSHIFT            96
#define KEY_RSHIFT            97
#define KEY_LCONTROL          98
#define KEY_RCONTROL          99
#define KEY_ALT               100
#define KEY_ALTGR             101
#define KEY_LWIN              102
#define KEY_RWIN              103
#define KEY_MENU              104
#define KEY_SCRLOCK           105
#define KEY_NUMLOCK           106
#define KEY_CAPSLOCK          107

#define MAX_MEMBERS 128

#endif

